/* The CSGEditorWindow class represents the window for editing CSGObjects. */

/* Copyright (C) 1999-2002 by Peter Eastman

   This program is free software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the Free Software
   Foundation; either version 2 of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but WITHOUT ANY 
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
   PARTICULAR PURPOSE.  See the GNU General Public License for more details. */

package artofillusion;

import artofillusion.*;
import artofillusion.math.*;
import artofillusion.object.*;
import artofillusion.texture.*;
import artofillusion.ui.*;
import java.awt.*;
import java.awt.event.*;

public class CSGEditorWindow extends Frame implements EditingWindow, KeyListener, ActionListener, ItemListener
{
  EditingWindow parentWindow;
  SceneViewer theView;
  EditingTool defaultTool, currentTool;
  Label helpText;
  CSGObject oldObject, theObject;
  UndoRecord undo;
  ToolPalette tools;
  MenuBar menubar;
  Menu editMenu, objectMenu;
  MenuItem undoItem, objectMenuItem[];
  CheckboxMenuItem displayItem[];
  Scene theScene;
  Callback onClose;

  public CSGEditorWindow(EditingWindow parent, String title, CSGObject obj, Callback onClose)
  {
    super(title);
    parentWindow = parent;
    oldObject = obj;
    theObject = (CSGObject) obj.duplicate();
    this.onClose = onClose;
    theScene = new Scene();
    theScene.addObject(obj.getObject1().object.duplicate(), obj.getObject1().coords.duplicate(), obj.getObject1().name, null);
    theScene.addObject(obj.getObject2().object.duplicate(), obj.getObject2().coords.duplicate(), obj.getObject2().name, null);
    add(helpText = new Label(), "South");
    Panel p1 = new Panel(), p2, p3, p4;
    add(p1, "Center");
    p1.setLayout(new BorderLayout());
    p1.add(p2 = new Panel(), "North");
    p2.setLayout(new BorderLayout());
    p2.add(p3 = new Panel(), "West");
    p1.add(theView = new SceneViewer(theScene, p3, this), "Center");
    p2.add(p3 = new Panel(), "East");
    p3.add(Translate.button("ok", this));
    p3.add(Translate.button("cancel", this));
    p4 = new Panel();
    add(p4, "West");
    p4.setLayout(new BorderLayout());
    p4.add(tools = new ToolPalette(1, 5), "North");
    EditingTool metaTool, altTool;
    tools.addTool(defaultTool = new MoveObjectTool(this));
    tools.addTool(new RotateObjectTool(this));
    tools.addTool(new ScaleObjectTool(this));
    tools.addTool(metaTool = new MoveViewTool(this));
    tools.addTool(altTool = new RotateViewTool(this));
    tools.selectTool(defaultTool);
    theView.setMetaTool(metaTool);
    theView.setAltTool(altTool);
    helpText.setBackground(ModellingApp.APP_BACKGROUND_COLOR);
    p2.setBackground(ModellingApp.APP_BACKGROUND_COLOR);
    p3.setBackground(ModellingApp.APP_BACKGROUND_COLOR);
    p4.setBackground(ModellingApp.APP_BACKGROUND_COLOR);
    menubar = new MenuBar();
    setMenuBar(menubar);
    createEditMenu();
    createObjectMenu();
    recursivelyAddListeners(this);
    pack();
    Dimension d1 = Toolkit.getDefaultToolkit().getScreenSize(), d2;
    d2 = new Dimension((d1.width*2)/3, (d1.height*2)/3);
    setSize(d2);
    setLocation((d1.width-d2.width)/2, (d1.height-d2.height)/2);
    tools.requestFocus();
    updateMenus();
    show();
  }

  void createEditMenu()
  {
    editMenu = Translate.menu("edit");
    menubar.add(editMenu);
    editMenu.add(undoItem = Translate.menuItem("undo", this));
    undoItem.setEnabled(false);
    editMenu.addSeparator();
    editMenu.add(Translate.menuItem("selectAll", this));
    editMenu.add(Translate.menuItem("properties", this));
  }

  void createObjectMenu()
  {
    Menu displayMenu;
    objectMenu = Translate.menu("object");
    menubar.add(objectMenu);
    objectMenuItem = new MenuItem [5];
    objectMenu.add(objectMenuItem[0] = Translate.menuItem("editObject", this));
    objectMenu.add(objectMenuItem[1] = Translate.menuItem("objectLayout", this));
    objectMenu.add(objectMenuItem[2] = Translate.menuItem("transformObject", this));
    objectMenu.add(objectMenuItem[3] = Translate.menuItem("alignObjects", this));
    objectMenu.add(Translate.menuItem("centerObjects", this));
    objectMenu.add(objectMenuItem[4] = Translate.menuItem("convertToTriangle", this));
    objectMenu.addSeparator();
    if (ModellingApp.getPreferences().getObjectPreviewRenderer() != null)
      objectMenu.add(Translate.menuItem("renderPreview", this));
    objectMenu.add(displayMenu = Translate.menu("displayMode"));
    displayItem = new CheckboxMenuItem [3];
    displayMenu.add(displayItem[0] = Translate.checkboxMenuItem("wireframeDisplay", this, theView.getRenderMode() == ViewerCanvas.RENDER_WIREFRAME));
    displayMenu.add(displayItem[1] = Translate.checkboxMenuItem("shadedDisplay", this, theView.getRenderMode() == ViewerCanvas.RENDER_FLAT));
    displayMenu.add(displayItem[2] = Translate.checkboxMenuItem("smoothDisplay", this, theView.getRenderMode() == ViewerCanvas.RENDER_SMOOTH));
  }

  /* We need to be notified of key events which take place anywhere in the window.  To make
     sure this happens, we need to add the window as a KeyListener to every component
     contained in it. */
  
  void recursivelyAddListeners(Component c)
  {
    Component children[];
    
    c.addKeyListener(this);
    if (c instanceof Container)
      {
	children = ((Container) c).getComponents();
	for (int i = 0; i < children.length; i++)
	  recursivelyAddListeners(children[i]);
      }
  }

  /* EditingWindow methods. */

  public boolean confirmClose()
  {
    return true;
  }

  public void setTool(EditingTool tool)
  {
    theView.setTool(tool);
    currentTool = tool;
  }

  public void setHelpText(String text)
  {
    helpText.setText(text);
  }
  
  public Frame getFrame()
  {
    return this;
  }

  public void updateImage()
  {
    theView.updateImage();
    theView.repaint();
  }
  
  public void updateMenus()
  {
    int selected[] = theScene.getSelection();
    objectMenuItem[0].setEnabled(selected.length == 1);
    objectMenuItem[1].setEnabled(selected.length > 0);
    objectMenuItem[2].setEnabled(selected.length > 0);
    objectMenuItem[3].setEnabled(selected.length > 0);
    objectMenuItem[4].setEnabled(selected.length == 1);
  }

  public void setUndoRecord(UndoRecord command)
  {
    undo = command;
    undoItem.setEnabled(true);
    if (undo.isRedo())
      undoItem.setLabel(Translate.text("menu.redo"));
    else
      undoItem.setLabel(Translate.text("menu.undo"));
    updateMenus();
  }
  
  public Scene getScene()
  {
    return theScene;
  }
  
  public ViewerCanvas getView()
  {
  	return theView;
  }

  /* KeyListener methods. */
  
  public void keyPressed(KeyEvent e)
  {
    int code = e.getKeyCode();
    
    if (e.isShiftDown() && (code == KeyEvent.VK_LEFT || code == KeyEvent.VK_RIGHT || code == KeyEvent.VK_UP || code == KeyEvent.VK_DOWN))
      tools.keyPressed(e);
    else
      currentTool.keyPressed(e, theView);
  }
  
  public void keyReleased(KeyEvent e)
  {
  }
  
  public void keyTyped(KeyEvent e)
  {
  }

  public void actionPerformed(ActionEvent e)
  {
    String command = e.getActionCommand();

    setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
    if (command.equals("ok"))
      {
	updateFromScene();
	oldObject.copyObject(theObject);
	oldObject = theObject = null;
	theScene = null;
	dispose();
	onClose.execute();
	parentWindow.updateImage();
	parentWindow.updateMenus();
      }
    else if (command.equals("cancel"))
      {
	oldObject = theObject = null;
	theScene = null;
	dispose();
      }
    else if (command.equals("undo"))
      undoCommand();
    else if (command.equals("selectAll"))
      selectAllCommand();
    else if (command.equals("properties"))
      propertiesCommand();
    else if (command.equals("editObject"))
      editObjectCommand();
    else if (command.equals("objectLayout"))
      objectLayoutCommand();
    else if (command.equals("transformObject"))
      transformObjectCommand();
    else if (command.equals("alignObjects"))
      alignObjectsCommand();
    else if (command.equals("centerObjects"))
      centerObjectsCommand();
    else if (command.equals("convertToTriangle"))
      convertToTriangleCommand();
    else if (command.equals("renderPreview"))
      renderPreviewCommand();
    setCursor(Cursor.getDefaultCursor());
  }

  public void itemStateChanged(ItemEvent e)
  {
    Object source = e.getSource();
    int i;
    
    for (i = 0; i < displayItem.length; i++)
      displayItem[i].setState(source == displayItem[i]);
    if (source == displayItem[0])
      theView.setRenderMode(ViewerCanvas.RENDER_WIREFRAME);
    else if (source == displayItem[1])
      theView.setRenderMode(ViewerCanvas.RENDER_FLAT);
    else if (source == displayItem[2])
      theView.setRenderMode(ViewerCanvas.RENDER_SMOOTH);
    updateImage();
  }
  
  /* Update the object based on the scene being used for editing it. */
  
  private void updateFromScene()
  {
    theObject.setComponentObjects(theScene.getObject(0), theScene.getObject(1));
  }

  void undoCommand()
  {
    if (undo == null)
      return;
    setUndoRecord(undo.execute());
    updateImage();
    updateMenus();
  }
  
  void selectAllCommand()
  {
    theScene.setSelection(new int [] {0, 1});
    updateImage();
    updateMenus();
  }
  
  void propertiesCommand()
  {
    updateFromScene();
    CSGDialog dial = new CSGDialog(this, theObject);
  }
  
  void editObjectCommand()
  {
    int sel[] = theScene.getSelection();
    Object3D obj;
    
    if (sel.length != 1)
      return;
    obj = theScene.getObject(sel[0]).object;
    if (obj.editable())
      {
	setUndoRecord(new UndoRecord(this, false, UndoRecord.COPY_OBJECT, new Object [] {obj, obj.duplicate()}));
	obj.edit(this, theScene.getObject(sel[0]), null);  /*******************/
      }
  }
  
  void objectLayoutCommand()
  {
    int i, sel[] = theScene.getSelection();
    TransformDialog dlg;
    Object3D obj[] = new Object3D [sel.length];
    CoordinateSystem coords[] = new CoordinateSystem [sel.length];
    Vec3 orig, size;
    double angles[], values[];
	
    if (sel.length == 0)
      return;
    setUndoRecord(new UndoRecord(this, false));
    for (i = 0; i < sel.length; i++)
      {
	obj[i] = theScene.getObject(sel[i]).object;
	coords[i] = theScene.getObject(sel[i]).coords;
	undo.addCommand(UndoRecord.COPY_OBJECT, new Object [] {obj[i], obj[i].duplicate()});
	undo.addCommand(UndoRecord.COPY_COORDS, new Object [] {coords[i], coords[i].duplicate()});
      }
    if (sel.length == 1)
      {
	orig = coords[0].getOrigin();
	angles = coords[0].getRotationAngles();
	size = theScene.getObject(sel[0]).getBounds().getSize();
	dlg = new TransformDialog(this, "Layout for '"+theScene.getObject(sel[0]).name+"'", 
	    new double [] {orig.x, orig.y, orig.z, angles[0], angles[1], angles[2], 
	    size.x, size.y, size.z}, false, false);
	values = dlg.getValues();
	if (!Double.isNaN(values[0]))
	  orig.x = values[0];
	if (!Double.isNaN(values[1]))
	  orig.y = values[1];
	if (!Double.isNaN(values[2]))
	  orig.z = values[2];
	if (!Double.isNaN(values[3]))
	  angles[0] = values[3];
	if (!Double.isNaN(values[4]))
	  angles[1] = values[4];
	if (!Double.isNaN(values[5]))
	  angles[2] = values[5];
	if (!Double.isNaN(values[6]))
	  size.x = values[6];
	if (!Double.isNaN(values[7]))
	  size.y = values[7];
	if (!Double.isNaN(values[8]))
	  size.z = values[8];
	coords[0].setOrigin(orig);
	coords[0].setOrientation(angles[0], angles[1], angles[2]);
	obj[0].setSize(size.x, size.y, size.z);
      }
    else
      {
	dlg = new TransformDialog(this, "Layout for Selected Objects", false, false);
	values = dlg.getValues();
	for (i = 0; i < sel.length; i++)
	  {
	    orig = coords[i].getOrigin();
	    angles = coords[i].getRotationAngles();
	    size = theScene.getObject(sel[i]).getBounds().getSize();
	    if (!Double.isNaN(values[0]))
	      orig.x = values[0];
	    if (!Double.isNaN(values[1]))
	      orig.y = values[1];
	    if (!Double.isNaN(values[2]))
	      orig.z = values[2];
	    if (!Double.isNaN(values[3]))
	      angles[0] = values[3];
	    if (!Double.isNaN(values[4]))
	      angles[1] = values[4];
	    if (!Double.isNaN(values[5]))
	      angles[2] = values[5];
	    if (!Double.isNaN(values[6]))
	      size.x = values[6];
	    if (!Double.isNaN(values[7]))
	      size.y = values[7];
	    if (!Double.isNaN(values[8]))
	      size.z = values[8];
	    coords[i].setOrigin(orig);
	    coords[i].setOrientation(angles[0], angles[1], angles[2]);
	    obj[i].setSize(size.x, size.y, size.z);
	  }
      }
    updateImage();
  }

  void transformObjectCommand()
  {
    int i, sel[] = theScene.getSelection();
    TransformDialog dlg;
    Object3D obj;
    CoordinateSystem coords;
    Vec3 orig, size;
    double values[];
    Mat4 m;

    if (sel.length == 0)
      return;
    if (sel.length == 1)
      dlg = new TransformDialog(this, "Transform '"+theScene.getObject(sel[0]).name+"'", 
		new double [] {0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 1.0, 1.0}, true, true);
    else
      dlg = new TransformDialog(this, "Transform Selected Objects", 
		new double [] {0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 1.0, 1.0}, true, true);
    values = dlg.getValues();
    setUndoRecord(new UndoRecord(this, false));
    for (i = 0; i < sel.length; i++)
      {
	obj = theScene.getObject(sel[i]).object;
	coords = theScene.getObject(sel[i]).coords;
	undo.addCommand(UndoRecord.COPY_OBJECT, new Object [] {obj, obj.duplicate()});
	undo.addCommand(UndoRecord.COPY_COORDS, new Object [] {coords, coords.duplicate()});
	orig = coords.getOrigin();
	size = obj.getBounds().getSize();
	if (!Double.isNaN(values[0]))
          orig.x += values[0];
	if (!Double.isNaN(values[1]))
	  orig.y += values[1];
	if (!Double.isNaN(values[2]))
	  orig.z += values[2];
	m = Mat4.identity();
	if (!Double.isNaN(values[3]))
	  m = m.times(Mat4.xrotation(values[3]*Math.PI/180.0));
	if (!Double.isNaN(values[4]))
	  m = m.times(Mat4.yrotation(values[4]*Math.PI/180.0));
	if (!Double.isNaN(values[5]))
	  m = m.times(Mat4.zrotation(values[5]*Math.PI/180.0));
	if (!Double.isNaN(values[6]))
	  size.x *= values[6];
	if (!Double.isNaN(values[7]))
	  size.y *= values[7];
	if (!Double.isNaN(values[8]))
	  size.z *= values[8];
	coords.setOrigin(orig);
	coords.transformAxes(m);
	obj.setSize(size.x, size.y, size.z);
      }
    updateImage();
  }

  void alignObjectsCommand()
  {
    int i, sel[] = theScene.getSelection();
    ComponentsDialog dlg;
    Object3D obj;
    CoordinateSystem coords;
    Vec3 alignTo, orig, center;
    double values[];
    Choice xchoice, ychoice, zchoice;
    Panel px = new Panel(), py = new Panel(), pz = new Panel();
    ValueField vfx, vfy, vfz;
    BoundingBox bounds;

    if (sel.length == 0)
      return;
    px.add(xchoice = new Choice());
    xchoice.add("Do not align");
    xchoice.add("Right");
    xchoice.add("Center");
    xchoice.add("Left");
    xchoice.add("Origin");
    px.add(new Label("align to"));
    px.add(vfx = new ValueField(Double.NaN, ValueField.NONE, 5));
    py.add(ychoice = new Choice());
    ychoice.add("Do not align");
    ychoice.add("Top");
    ychoice.add("Center");
    ychoice.add("Bottom");
    ychoice.add("Origin");
    py.add(new Label("align to"));
    py.add(vfy = new ValueField(Double.NaN, ValueField.NONE, 5));
    pz.add(zchoice = new Choice());
    zchoice.add("Do not align");
    zchoice.add("Front");
    zchoice.add("Center");
    zchoice.add("Back");
    zchoice.add("Origin");
    pz.add(new Label("align to"));
    pz.add(vfz = new ValueField(Double.NaN, ValueField.NONE, 5));
    dlg = new ComponentsDialog(this, "Align to which point:", 
		new Component [] {px, py, pz}, new String [] {"X", "Y", "Z"});
    if (!dlg.clickedOk())
      return;
    setUndoRecord(new UndoRecord(this, false));
    
    // Determine the position to align the objects to.
    
    alignTo = new Vec3();
    for (i = 0; i < sel.length; i++)
      {
	obj = theScene.getObject(sel[i]).object;
	coords = theScene.getObject(sel[i]).coords;
	bounds = obj.getBounds();
	bounds = bounds.transformAndOutset(coords.fromLocal());
	center = bounds.getCenter();
	orig = coords.getOrigin();
	if (!Double.isNaN(vfx.getValue()))
	  alignTo.x += vfx.getValue();
	else if (xchoice.getSelectedIndex() == 1)
	  alignTo.x += bounds.maxx;
	else if (xchoice.getSelectedIndex() == 2)
	  alignTo.x += center.x;
	else if (xchoice.getSelectedIndex() == 3)
	  alignTo.x += bounds.minx;
	else if (xchoice.getSelectedIndex() == 4)
	  alignTo.x += orig.x;
	if (!Double.isNaN(vfy.getValue()))
	  alignTo.y += vfy.getValue();
	else if (ychoice.getSelectedIndex() == 1)
	  alignTo.y += bounds.maxy;
	else if (ychoice.getSelectedIndex() == 2)
	  alignTo.y += center.y;
	else if (ychoice.getSelectedIndex() == 3)
	  alignTo.y += bounds.miny;
	else if (ychoice.getSelectedIndex() == 4)
	  alignTo.y += orig.y;
	if (!Double.isNaN(vfz.getValue()))
	  alignTo.z += vfz.getValue();
	else if (zchoice.getSelectedIndex() == 1)
	  alignTo.z += bounds.maxz;
	else if (zchoice.getSelectedIndex() == 2)
	  alignTo.z += center.z;
	else if (zchoice.getSelectedIndex() == 3)
	  alignTo.z += bounds.minz;
	else if (zchoice.getSelectedIndex() == 4)
	  alignTo.z += orig.z;
      }
    alignTo.scale(1.0/sel.length);
    
    // Now transform all of the objects.
    
    for (i = 0; i < sel.length; i++)
      {
	obj = theScene.getObject(sel[i]).object;
	coords = theScene.getObject(sel[i]).coords;
	bounds = obj.getBounds();
	bounds = bounds.transformAndOutset(coords.fromLocal());
	center = bounds.getCenter();
	orig = coords.getOrigin();
	undo.addCommand(UndoRecord.COPY_COORDS, new Object [] {coords, coords.duplicate()});
	if (xchoice.getSelectedIndex() == 1)
	  orig.x += alignTo.x-bounds.maxx;
	else if (xchoice.getSelectedIndex() == 2)
	  orig.x += alignTo.x-center.x;
	else if (xchoice.getSelectedIndex() == 3)
	  orig.x += alignTo.x-bounds.minx;
	else if (xchoice.getSelectedIndex() == 4)
	  orig.x += alignTo.x-orig.x;
	if (ychoice.getSelectedIndex() == 1)
	  orig.y += alignTo.y-bounds.maxy;
	else if (ychoice.getSelectedIndex() == 2)
	  orig.y += alignTo.y-center.y;
	else if (ychoice.getSelectedIndex() == 3)
	  orig.y += alignTo.y-bounds.miny;
	else if (ychoice.getSelectedIndex() == 4)
	  orig.y += alignTo.y-orig.y;
	if (zchoice.getSelectedIndex() == 1)
	  orig.z += alignTo.z-bounds.maxz;
	else if (zchoice.getSelectedIndex() == 2)
	  orig.z += alignTo.z-center.z;
	else if (zchoice.getSelectedIndex() == 3)
	  orig.z += alignTo.z-bounds.minz;
	else if (zchoice.getSelectedIndex() == 4)
	  orig.z += alignTo.z-orig.z;
	coords.setOrigin(orig);
      }
    updateImage();
  }  

  void centerObjectsCommand()
  {
    BoundingBox bounds = null;
    
    // Determine the bounding box for all objects.
    
    for (int i = 0; i < theScene.getNumObjects(); i++)
      {
	ObjectInfo info = theScene.getObject(i);
	BoundingBox b = info.getBounds().transformAndOutset(info.coords.fromLocal());
	if (bounds == null)
	  bounds = b;
	else
	  bounds = bounds.merge(b);
      }
    Vec3 center = bounds.getCenter();
    setUndoRecord(new UndoRecord(this, false));
    
    // Center the objects.
    
    for (int i = 0; i < theScene.getNumObjects(); i++)
      {
	ObjectInfo info = theScene.getObject(i);
	undo.addCommand(UndoRecord.COPY_COORDS, new Object [] {info.coords, info.coords.duplicate()});
	info.coords.setOrigin(info.coords.getOrigin().minus(center));
      }
    updateImage();
  }  

  void convertToTriangleCommand()
  {
    int sel[] = theScene.getSelection();
    double tol;
    Object3D obj, mesh;
    ObjectInfo info;
    Texture tex;
    TextureMapping map;

    if (sel.length != 1)
      return;
    info = theScene.getObject(sel[0]);
    obj = info.object;
    if (obj.canConvertToTriangleMesh() == Object3D.CANT_CONVERT)
      return;
    setUndoRecord(new UndoRecord(this, false, UndoRecord.COPY_OBJECT_INFO, new Object [] {info, info.duplicate()}));
    if (obj.canConvertToTriangleMesh() == Object3D.EXACTLY)
      mesh = obj.convertToTriangleMesh(0.0);
    else
      {
	EntriesDialog dlg = new EntriesDialog(this, "Specify tolerance for triangle mesh:",
		new String[] {"Max Error"}, new String [] {"0.1"});
	String val[] = dlg.getEntries();
	if (val == null)
	  return;
	try
	  {
	    tol = Double.valueOf(val[0]).doubleValue();
	  }
	catch (NumberFormatException ex)
	  {
	    new MessageDialog(this, "Illegal entry for error tolerance.");
	    return;
	  }
	if (tol < 0.0)
	  {
	    new MessageDialog(this, "Error tolerance cannot be negative.");
	    return;
	  }
	mesh = obj.convertToTriangleMesh(tol);
      }
    if (mesh == null)
      {
	new MessageDialog(this, "Unable to triangulate object.");
	return;
      }
    mesh.setTexture(obj.getTexture());
    mesh.setTextureMapping(obj.getTextureMapping());
    mesh.setMaterial(obj.getMaterial());
    mesh.setMaterialMapping(obj.getMaterialMapping());
    theScene.getObject(sel[0]).object = mesh;
    theView.updateImage();
    updateMenus();
  }

  public void renderPreviewCommand()
  {
    Scene sc = new Scene();
    Camera theCamera = theView.getCamera();
    Component parent = this;
    double dist = theCamera.getDistToScreen();
    Renderer rend = ModellingApp.getPreferences().getObjectPreviewRenderer();

    if (rend == null)
      return;
    updateFromScene();
    sc.addObject(new DirectionalLight(new RGBColor(1.0f, 1.0f, 1.0f), 0.8f), theCamera.getCameraCoordinates(), "", null);
    sc.addObject(theObject, new CoordinateSystem(new Vec3(), Vec3.vz(), Vec3.vy()), "", null);
    while (!(parent instanceof Frame))
      parent = parent.getParent();
    theView.adjustCamera(true);
    rend.configurePreview();
    ObjectInfo cameraInfo = new ObjectInfo(new SceneCamera(), theCamera.getCameraCoordinates(), "");
    new RenderingDialog((Frame) parent, rend, sc, theCamera, cameraInfo);
  }
}