/* Copyright (C) 2003 by Peter Eastman

   This program is free software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the Free Software
   Foundation; either version 2 of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but WITHOUT ANY 
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
   PARTICULAR PURPOSE.  See the GNU General Public License for more details. */

package artofillusion;

import artofillusion.image.*;
import artofillusion.image.filter.*;
import artofillusion.math.*;
import artofillusion.object.*;
import artofillusion.ui.*;
import java.awt.*;
import java.awt.event.*;
import java.util.Vector;

/** This is dialog in which the user can edit the list of filters attached to a camera. */

public class CameraFilterDialog extends Dialog implements ActionListener, ItemListener, RenderListener
{
  private SceneCamera theCamera;
  private Vector filters;
  private ImageFilter oldFilters[];
  private Class filterClasses[];
  private List allFiltersList, cameraFiltersList;
  private Button addButton, deleteButton, upButton, downButton, previewButton, okButton, cancelButton;
  private ScrollPane editorPane;
  private Canvas preview;
  private ComplexImage unfilteredImage;
  private Image displayImage;
  private boolean doneRendering, doneFiltering;
  
  private static final int PREVIEW_WIDTH = 200;
  private static final int PREVIEW_HEIGHT = 150;
  
  public CameraFilterDialog(EditingWindow parent, SceneCamera camera, CoordinateSystem cameraCoords)
  {
    super(parent.getFrame(), "Filters", true);
    theCamera = camera;
    filters = new Vector();
    oldFilters = theCamera.getImageFilters();
    for (int i = 0; i < oldFilters.length; i++)
      {
        filters.addElement(oldFilters[i]);
        oldFilters[i] = oldFilters[i].duplicate();
      }
    
    // Layout the major sections of the window.
    
    Panel listsPanel = new Panel(), previewPanel = new Panel(), okPanel = new Panel();
    editorPane = new ScrollPane();
    setLayout(new GridBagLayout());
    GridBagConstraints gc = new GridBagConstraints();
    gc.gridx = 0;
    gc.fill = GridBagConstraints.BOTH;
    gc.weightx = gc.weighty = 1.0;
    add(listsPanel, gc);
    add(editorPane, gc);
    gc.gridx = 1;
    gc.gridheight = 2;
    gc.weightx = gc.weighty = 0.0;
    add(previewPanel, gc);
    gc.gridx = 0;
    gc.gridy = 2;
    gc.gridheight = 1;
    gc.gridwidth = 2;
    add(okPanel, gc);
    
    // Layout the Lists and the buttons between them.
    
    Panel buttonsPanel = new Panel();
    listsPanel.setLayout(new GridBagLayout());
    gc = new GridBagConstraints();
    gc.gridy = 0;
    gc.fill = GridBagConstraints.BOTH;
    gc.weightx = gc.weighty = 1.0;
    listsPanel.add(allFiltersList = new List(), gc);
    gc.fill = GridBagConstraints.NONE;
    gc.weightx = 0.0;
    listsPanel.add(buttonsPanel, gc);
    gc.fill = GridBagConstraints.BOTH;
    gc.weightx = 1.0;
    listsPanel.add(cameraFiltersList = new List(), gc);
    buttonsPanel.setLayout(new GridLayout(4, 1));
    buttonsPanel.add(addButton = Translate.button("add", " >>", this));
    buttonsPanel.add(deleteButton = Translate.button("delete", this));
    buttonsPanel.add(upButton = Translate.button("moveUp", this));
    buttonsPanel.add(downButton = Translate.button("moveDown", this));
    allFiltersList.addItemListener(this);
    cameraFiltersList.addItemListener(this);
    
    // Fill in the Lists.
    
    filterClasses = ModellingApp.getImageFilters();
    for (int i = 0; i < filterClasses.length; i++)
      {
        try
          {
            allFiltersList.add(((ImageFilter) filterClasses[i].newInstance()).getName());
          }
        catch (Exception ex)
          {
            ex.printStackTrace();
          }
      }
    rebuildFilterList();
    if (cameraFiltersList.getItemCount() > 0)
      cameraFiltersList.select(0);
    
    // Create the preview panel.
    
    preview = new Canvas() {
      public void paint(Graphics g)
      {
        paintPreview(g);
      }
      
      public Dimension getPreferredSize()
      {
        return new Dimension(PREVIEW_WIDTH, PREVIEW_HEIGHT);
      }
    };
    previewPanel.setLayout(new BorderLayout());
    previewPanel.add(preview, BorderLayout.CENTER);
    Panel p = new Panel();
    p.add(previewButton = new Button("Update Preview"));
    previewButton.addActionListener(this);
    previewButton.setEnabled(false);
    previewPanel.add(p, BorderLayout.SOUTH);
    
    // Create the OK and Cancel buttons.
    
    okPanel.add(okButton = Translate.button("ok", this));
    okPanel.add(cancelButton = Translate.button("cancel", this));
    updateComponents();
    
    // Begin rendering the preview.
    
    Camera cam = theCamera.createCamera(PREVIEW_WIDTH, PREVIEW_HEIGHT, cameraCoords);
    SceneCamera sc = new SceneCamera() {
      public int getComponentsForFilters()
      {
        return -1; // Force the renderer to generate all possible components.
      }
    };
    sc.copyObject(theCamera);
    Renderer renderer = ModellingApp.getPreferences().getTexturePreviewRenderer();
    renderer.configurePreview();
    renderer.renderScene(parent.getScene(), cam, this, sc);
    
    // Display the window.

    pack();
    show();
  }
  
  public void actionPerformed(ActionEvent ev)
  {
    Object src = ev.getSource();
    if (src == addButton)
      doAdd();
    else if (src == deleteButton)
      doDelete();
    else if (src == upButton)
      doMoveUp();
    else if (src == downButton)
      doMoveDown();
    else if (src == previewButton)
      applyFilters();
    else if (src == okButton)
      doOk();
    else if (src == cancelButton)
      doCancel();
  }
  
  public void itemStateChanged(ItemEvent ev)
  {
    updateComponents();
  }
  
  /** Update the states of various components in the window. */
  
  private void updateComponents()
  {
    addButton.setEnabled(allFiltersList.getSelectedIndex() > -1);
    int selection = cameraFiltersList.getSelectedIndex();
    deleteButton.setEnabled(selection > -1);
    upButton.setEnabled(selection > 0);
    downButton.setEnabled(selection > -1 && selection < filters.size()-1);
    if (selection > -1)
      {
        editorPane.add(((ImageFilter) filters.elementAt(cameraFiltersList.getSelectedIndex())).getConfigPanel());
        editorPane.validate();
      }
    else
      editorPane.removeAll();
  }
  
  /** Rebuild the list of filters. */
  
  private void rebuildFilterList()
  {
    cameraFiltersList.removeAll();
    for (int i = 0; i < filters.size(); i++)
      cameraFiltersList.add(((ImageFilter) filters.elementAt(i)).getName());
  }
  
  /** Save changes and close the window. */
  
  private void doOk()
  {
    ImageFilter filt[] = new ImageFilter [filters.size()];
    filters.copyInto(filt);
    theCamera.setImageFilters(filt);
    dispose();
  }
  
  /** Cancel changes and close the window. */
  
  private void doCancel()
  {
    ImageFilter filt[] = theCamera.getImageFilters();
    for (int i = 0; i < filt.length; i++)
      filt[i].copy(oldFilters[i]);
    dispose();
  }
  
  /** Add a filter to the list. */
  
  private void doAdd()
  {
    int sel = allFiltersList.getSelectedIndex();
    if (sel == -1)
      return;
    try
      {
        filters.addElement(filterClasses[sel].newInstance());
        rebuildFilterList();
        cameraFiltersList.select(filters.size()-1);
      }
    catch (Exception ex)
      {
        ex.printStackTrace();
      }
    updateComponents();
  }
  
  /** Delete a filter from the list. */
  
  private void doDelete()
  {
    int sel = cameraFiltersList.getSelectedIndex();
    if (sel == -1)
      return;
    filters.removeElementAt(sel);
    rebuildFilterList();
    updateComponents();
  }
  
  /** Move the selected filter upward. */
  
  private void doMoveUp()
  {
    int sel = cameraFiltersList.getSelectedIndex();
    if (sel < 1)
      return;
    ImageFilter filt = (ImageFilter) filters.elementAt(sel);
    filters.removeElementAt(sel);
    filters.insertElementAt(filt, sel-1);
    rebuildFilterList();
    cameraFiltersList.select(sel-1);
    updateComponents();
  }
  
  /** Move the selected filter downward. */
  
  private void doMoveDown()
  {
    int sel = cameraFiltersList.getSelectedIndex();
    if (sel == -1 || sel == filters.size()-1)
      return;
    ImageFilter filt = (ImageFilter) filters.elementAt(sel);
    filters.removeElementAt(sel);
    filters.insertElementAt(filt, sel+1);
    rebuildFilterList();
    cameraFiltersList.select(sel+1);
    updateComponents();
  }
  
  /** Repaint the preview canvas. */
  
  private void paintPreview(Graphics g)
  {
    if (displayImage != null)
      g.drawImage(displayImage, 0, 0, null);
    String message = null;
    if (!doneRendering)
      message = "Rendering Preview...";
    else if (!doneFiltering)
      message = "Applying Filter...";
    if (message != null)
      {
        FontMetrics fm = g.getFontMetrics();
        int width = fm.stringWidth(message);
        g.setColor(Color.red);
        g.drawString(message, (PREVIEW_WIDTH-width)/2, PREVIEW_HEIGHT/2);
      }
  }
  
  /** Apply the filters to the image and redisplay it. */
  
  private void applyFilters()
  {
    doneFiltering = false;
    preview.repaint();
    new Thread() {
      public void run()
      {
        ComplexImage img = unfilteredImage.duplicate();
        for (int i = 0; i < filters.size(); i++)
          ((ImageFilter) filters.elementAt(i)).filterImage(img);
        if (filters.size() > 0)
          img.rebuildImage();
        displayImage = img.getImage();
        doneFiltering = true;
        preview.repaint();
      }
    }.start();
  }

  /** The renderer may call this method periodically during rendering, to notify the listener that more of the
      image is complete. */
  
  public void imageUpdated(Image image)
  {
    displayImage = image;
    preview.repaint();
  }
  
  /** The renderer may call this method periodically during rendering, to give the listener text descriptions
      of the current status of rendering. */
  
  public void statusChanged(String status)
  {
  }

  /** This method will be called when rendering is complete. */
  
  public void imageComplete(ComplexImage image)
  {
    unfilteredImage = image;
    displayImage = image.getImage();
    doneRendering = true;
    previewButton.setEnabled(true);
    applyFilters();
  }
  
  /** This method will be called if rendering is canceled. */
  
  public void renderingCanceled()
  {
  }
}