
/* Copyright (C) 1999-2003 by Peter Eastman

   This program is free software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the Free Software
   Foundation; either version 2 of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
   PARTICULAR PURPOSE.  See the GNU General Public License for more details. */

package artofillusion;

import artofillusion.animation.*;
import artofillusion.animation.distortion.*;
import artofillusion.image.*;
import artofillusion.material.*;
import artofillusion.math.*;
import artofillusion.object.*;
import artofillusion.script.*;
import artofillusion.texture.*;
import artofillusion.ui.*;
import bsh.*;
import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.text.*;
import java.util.Vector;

/** The LayoutWindow class represents the main window for creating and laying out scenes. */

public class LayoutWindow extends Frame implements EditingWindow, ItemListener, KeyListener,
                                                   ActionListener, PopupMenuManager, MenuItemFactory
{
  SceneViewer theView[];
  Panel viewPanel[], viewsContainer;
  DividedPanel div1, div2;
  ScrollPane itemTreeScroller;
  Score theScore;
  ToolPalette tools;
  EditingTool defaultTool, currentTool;
  Panel toolsPanel;
  Label helpText, timeLabel, frameLabel;
  TreeList itemTree;
  Scene theScene;
  MenuBar menubar;
  PopupMenu popupMenu;
  UndoRecord undo;
  int numViewsShown, currentView;
  boolean modified, scoreHidden, listHidden;

  /** Create a new LayoutWindow for editing a Scene.  Usually, you will not use this constructor directly.
      Instead, call ModellingApp.newWindow(Scene s). */

  public LayoutWindow(Scene s)
  {
    super(s.getName() == null ? "Untitled" : s.getName());
    theScene = s;
    helpText = new Label();
    createItemList();

    // Create the four SceneViewer panels.

    theView = new SceneViewer [4];
    viewPanel = new Panel [4];
    Panel p;
    MouseListener listen = new MouseAdapter() {
    public void mousePressed(MouseEvent ev)
      {
	for (int i = 0; i < theView.length; i++)
	  if (currentView != i && ev.getSource() == theView[i])
            {
              theView[currentView].setDrawFocus(false);
              theView[i].setDrawFocus(true);
              currentView = i;
              updateMenus();
              updateImage();
            }
      }
    };
    for (int i = 0; i < 4; i++)
      {
	viewPanel[i] = new Panel();
	viewPanel[i].setLayout(new BorderLayout());
	viewPanel[i].add(p = new Panel(), "North");
	viewPanel[i].add(theView[i] = new SceneViewer(theScene, p, this), "Center");
	theView[i].setGrid(theScene.gridSpacing, theScene.gridSubdivisions, theScene.showGrid, theScene.snapToGrid);
	theView[i].addMouseListener(listen);
      }
    theView[1].selectOrientation(2);
    theView[2].selectOrientation(4);
    theView[3].selectOrientation(6);
    theView[3].setPerspective(true);
    theView[currentView].setDrawFocus(true);
    viewsContainer = new Panel();
    viewsContainer.setLayout(new GridBagLayout());
    div1 = new DividedPanel(viewsContainer, itemTreeScroller, 1.0f, 0.0f, false);
    div1.setLiveValidate(false, true);
    p = new Panel();
    p.setLayout(new BorderLayout());
    p.add(div1, BorderLayout.CENTER);
    p.add(helpText, BorderLayout.SOUTH);
    theScore = new Score(this);
    div2 = new DividedPanel(p, theScore, 3.0f, 1.0f, true);
    div2.setLiveValidate(false, true);
    add(div2, BorderLayout.CENTER);
    div2.setBackground(ModellingApp.APP_BACKGROUND_COLOR);
    itemTreeScroller.setBackground(Color.white);

    numViewsShown = 1;

    // Build the tool palette.

    tools = new ToolPalette(2, 7);
    EditingTool metaTool, altTool;
    tools.addTool(defaultTool = new MoveObjectTool(this));
    tools.addTool(new RotateObjectTool(this));
    tools.addTool(new ScaleObjectTool(this));
    tools.addTool(new CreateCubeTool(this));
    tools.addTool(new CreateSphereTool(this));
    tools.addTool(new CreateCylinderTool(this));
    tools.addTool(new CreateSplineMeshTool(this));
    tools.addTool(new CreatePolygonTool(this));
    tools.addTool(new CreateCameraTool(this));
    tools.addTool(new CreateLightTool(this));
    tools.addTool(new CreateCurveTool(this, Curve.INTERPOLATING));
    tools.addTool(new CreateCurveTool(this, Curve.APPROXIMATING));
    tools.addTool(metaTool = new MoveViewTool(this));
    tools.addTool(altTool = new RotateSceneViewTool(this, true));
    tools.selectTool(defaultTool);
    for (int i = 0; i < theView.length; i++)
      {
        theView[i].setMetaTool(metaTool);
        theView[i].setAltTool(altTool);
      }

    // Fill in the left hand panel.

    toolsPanel = new Panel();
    toolsPanel.setLayout(new GridBagLayout());
    GridBagConstraints gc = new GridBagConstraints();
    gc.gridx = 0;
    gc.anchor = gc.NORTH;
    toolsPanel.add(tools, gc);
    gc.fill = gc.HORIZONTAL;
    gc.weighty = 1.0;
    gc.anchor = gc.SOUTH;
    toolsPanel.add(new ShortLabel(Translate.text("Time")+':', Label.CENTER), gc);
    gc.weighty = 0.0;
    toolsPanel.add(timeLabel = new ShortLabel("0.0", Label.CENTER), gc);
    toolsPanel.add(new ShortLabel(Translate.text("Frame")+':', Label.CENTER), gc);
    toolsPanel.add(frameLabel = new ShortLabel("0", Label.CENTER), gc);

    // Build the menubar.
    rebuildMenuBar();

//    menubar = new MenuBar();
//    setMenuBar(menubar);
//    createFileMenu();
//    createEditMenu();
//    createObjectMenu();
//    createToolsMenu();
//    createAnimationMenu();
//    createSceneMenu();
    createPopupMenu();

    toggleViewsCommand();
    recursivelyAddListeners(this);
    addWindowListener(new WindowAdapter() {
      public void windowActivated(WindowEvent e)
	{
	  updateMenus();
	}
      public void windowClosing(WindowEvent e)
	{
	  ModellingApp.closeWindow(LayoutWindow.this);
	} });
    itemTree.setPopupMenuManager(this);
    if (ModellingApp.defaultFont != null)
      setFont(ModellingApp.defaultFont);
    pack();
    div2.setHidden(false, scoreHidden = true);
    Dimension screenDim = Toolkit.getDefaultToolkit().getScreenSize();
    setSize(screenDim.width, screenDim.height-20);
    tools.requestFocus();
    updateMenus();
  }

  /** Loads the (default or cutomized) menu structure, builds and sets
    the menubar. */

  public void rebuildMenuBar()
  {
    MenuStructure ms = new MenuStructure();
    ms.loadStructure("MainWindow", false);
    menubar = ms.createMenuBar(this);
    setMenuBar(menubar);
  }


  /** Create the TreeList containing all the objects in the scene. */

  private void createItemList()
  {
    itemTree = new TreeList(this);
    itemTree.setPreferredSize(130, 100);
    itemTree.addTreeListener(new TreeAdapter() {
      public void elementsMoved(TreeEvent e)
      {
	theScore.rebuildList();
      }

      public void elementDoubleClicked(TreeEvent e)
      {
	editObjectCommand();
      } });
    itemTree.setUpdateEnabled(false);
    for (int i = 0; i < theScene.getNumObjects(); i++)
      {
	ObjectInfo info = theScene.getObject(i);
	if (info.parent == null)
	  itemTree.addElement(new ObjectTreeElement(info, itemTree));
      }
    itemTree.setUpdateEnabled(true);
    ScrollPane sp = new ScrollPane();
    sp.add(itemTree);
    sp.getVAdjustable().setUnitIncrement(10);
    itemTree.addItemListener(this);
    new AutoScroller(itemTree, sp, 0, 10);
    itemTreeScroller = sp;
  }

  /** Rebuild the TreeList of objects, attempting as much as possible to preserve its
      current state. */

  public void rebuildItemList()
  {
    boolean expanded[] = new boolean [theScene.getNumObjects()], selected[] = new boolean [theScene.getNumObjects()];

    for (int i = 0; i < theScene.getNumObjects(); i++)
      {
	ObjectInfo info = theScene.getObject(i);
	TreeElement el = itemTree.findElement(info);
	if (el == null)
	  continue;
	expanded[i] = el.isExpanded();
	selected[i] = el.isSelected();
      }
    itemTree.setUpdateEnabled(false);
    itemTree.removeAllElements();
    for (int i = 0; i < theScene.getNumObjects(); i++)
      {
	ObjectInfo info = theScene.getObject(i);
	if (info.parent == null)
	  itemTree.addElement(new ObjectTreeElement(info, itemTree));
      }
    for (int i = 0; i < theScene.getNumObjects(); i++)
      {
	ObjectInfo info = theScene.getObject(i);
	TreeElement el = itemTree.findElement(info);
	if (el == null)
	  continue;
	el.setExpanded(expanded[i]);
	el.setSelected(selected[i]);
      }
    itemTree.setUpdateEnabled(true);
    theScore.rebuildList();
  }

  /** Create the list of tool scripts in the "All Tool Scripts" menu.*/

  public Menu createScriptsMenu()
  {
    Menu scriptMenu = Translate.menu("allToolScripts");
    addScriptsToMenu(scriptMenu, new File(ModellingApp.TOOL_SCRIPT_DIRECTORY));
    return scriptMenu;
  }


  /** Create the list of tools in the "All Tools" menu. */

  public Menu createToolsMenu()
  {
    Menu toolsMenu = Translate.menu("allTools");
    MenuItem item;
    ModellingTool tools[] = ModellingApp.getModellingTools();

    for (int i = 0; i < tools.length; i++)
    {
      toolsMenu.add(item = new MenuItem(tools[i].getName()));
      item.setActionCommand("modellingTool/"+tools[i].getClass().getName());
      item.setName("modellingTool/"+tools[i].getClass().getName());
      item.addActionListener(this);
    }
    return toolsMenu;
  }

  public Menu createImportMenu()
  {
    Menu importMenu = Translate.menu("import");
    MenuItem item;
    Translator trans[] = ModellingApp.getTranslators();

    for (int i = 0; i < trans.length; i++)
      if (trans[i].canImport())
      {
        importMenu.add(item = new MenuItem(trans[i].getName()));
        item.setActionCommand("import/"+trans[i].getClass().getName());
        item.setName("import/"+trans[i].getClass().getName());
        item.addActionListener(this);
      }
    return importMenu;
  }

  public Menu createExportMenu()
  {
    Menu exportMenu = Translate.menu("export");
    MenuItem item;
    Translator trans[] = ModellingApp.getTranslators();

    for (int i = 0; i < trans.length; i++)
      if (trans[i].canExport())
      {
        exportMenu.add(item = new MenuItem(trans[i].getName()));
        item.setActionCommand("export/"+trans[i].getClass().getName());
        item.setName("export/"+trans[i].getClass().getName());
        item.addActionListener(this);
      }
    return exportMenu;
  }

  private void addScriptsToMenu(Menu menu, File dir)
  {
    String files[] = dir.list();
    if (files == null)
      return;
    for (int i = 0; i < files.length; i++)
      {
        File f = new File(dir, files[i]);
        if (f.isDirectory())
          {
            Menu m = new Menu(files[i]);
            menu.add(m);
            addScriptsToMenu(m, f);
          }
        else if (files[i].endsWith(".bsh") && files[i].length() > 4)
          {
            MenuItem item = new MenuItem(files[i].substring(0, files[i].length()-4));
            item.setActionCommand("beanshell/"+f.getAbsolutePath());
            item.setName("beanshell/"+f.getAbsolutePath());
            item.addActionListener(this);

            menu.add(item);
          }
      }
  }
  
  /** Create a list of all scripts. Used for menu customization.
    Internal recursive version. */
  private void createScriptList(Vector list, File dir)
  {
    String files[] = dir.list();
    if (files == null)
      return;
    for (int i = 0; i < files.length; i++)
      {
        File f = new File(dir, files[i]);
        if (f.isDirectory())
          createScriptList(list, f);
        else if (files[i].endsWith(".bsh") && files[i].length() > 4)
          {
            list.add("beanshell/"+f.getAbsolutePath()+" - "+
              files[i].substring(0, files[i].length()-4));
          }
      }
  }
  
  /** Get a list of all scripts. Used for menu customization.
    The returned strings have the form "<internal name> - <menu label text>"
    */
  
  public String[] getScriptList()
  {
    Vector list = new Vector();
    
    createScriptList(list, new File(ModellingApp.TOOL_SCRIPT_DIRECTORY));
    
    return (String[])list.toArray(new String[0]);
  }

  /** Returns the default label for a menu item. */
  
  public static String getDefaultMenuLabel(String internalname)
  {
    if (internalname.startsWith("modellingTool/"))
    {
      String classname = internalname.substring(14);

      ModellingTool tools[] = ModellingApp.getModellingTools();

      for (int i = 0; i < tools.length; i++)
      {
        if (tools[i].getClass().getName().equals(classname))
          return tools[i].getName();
      }
    }
    else if (internalname.startsWith("import/"))
    {
      String classname = internalname.substring(7);

      Translator trans[] = ModellingApp.getTranslators();

      for (int i = 0; i < trans.length; i++)
      {
        if (trans[i].canImport() &&
            trans[i].getClass().getName().equals(classname))
          return trans[i].getName();
      }
    }
    else if (internalname.startsWith("export/"))
    {
      String classname = internalname.substring(7);

      Translator trans[] = ModellingApp.getTranslators();


      for (int i = 0; i < trans.length; i++)
      {
        if (trans[i].canExport() &&
            trans[i].getClass().getName().equals(classname))
          return trans[i].getName();
      }
      return classname+"???";
    }
    else      
      return Translate.text("menu."+internalname);

    return "";   //???
  }
  
  
  
  private static final String[] MENUITEMS_CHECKBOXITEMS=new String[]
  {"wireframeDisplay", "shadedDisplay", "smoothDisplay"};

  /** Creates a MenuItem which is either a normal MenuItem,
    a CheckboxMenuItem or a whole submenu, depending on
    {@link #isMenuItemCheckbox}.
    An appropriate listener is registered for this automatically.

    @param name Internal name of the item (as recognized by action listener)
    @param label Label of the item (or null for default label)
    @param shortcut Shortcut of the item (or null for default)

    @return The menu item or null if creation is not allowed.
  */

  public MenuItem createMenuItem(String name, MenuShortcut shortcut,
      String label)
  {
    MenuItem result = null;

    if (name.equals("allToolScripts"))
      result = createScriptsMenu();
    else if (name.equals("allTools"))
      result = createToolsMenu();
    else if (name.equals("import"))
      result = createImportMenu();
    else if (name.equals("export"))
      result = createExportMenu();
    else if (name.startsWith("modellingTool/") ||
             name.startsWith("import/") ||
             name.startsWith("export/") )
    {
      result = new MenuItem(getDefaultMenuLabel(name));
      result.addActionListener(this);
      result.setActionCommand(name);
      result.setName(name);
    }
    else if (Utilities.findIndexEqual(MENUITEMS_CHECKBOXITEMS, name) > -1)
      result = Translate.checkboxMenuItem(name, this);
    else
      result = Translate.menuItem(name, this);

    if (label != null)
      result.setLabel(label);

    if (shortcut != null)
      result.setShortcut(shortcut);

    return result;
  }

  /** Create the popup menu. */

  private void createPopupMenu()
  {
    popupMenu = new PopupMenu();
    popupMenu.add(Translate.menuItem("editObject", this, null));
    popupMenu.add(Translate.menuItem("objectLayout", this, null));
    popupMenu.add(Translate.menuItem("setTexture", this, null));
    popupMenu.add(Translate.menuItem("setMaterial", this, null));
    popupMenu.add(Translate.menuItem("renameObject", this, null));
    popupMenu.add(Translate.menuItem("convertToTriangle", this, null));
    popupMenu.add(Translate.menuItem("selectChildren", this, null));
    popupMenu.addSeparator();
    popupMenu.add(Translate.menuItem("hideSelection", this, null));
    popupMenu.add(Translate.menuItem("showSelection", this, null));
    popupMenu.addSeparator();
    popupMenu.add(Translate.menuItem("cut", this, null));
    popupMenu.add(Translate.menuItem("copy", this, null));
    popupMenu.add(Translate.menuItem("clear", this, null));
    add(popupMenu);
  }
//  private void createPopupMenu()
//  {
//    popupMenu = new PopupMenu();
//    popupMenuItem = new MenuItem [12];
//    popupMenu.add(popupMenuItem[0] = Translate.menuItem("editObject", this, null));
//    popupMenu.add(popupMenuItem[1] = Translate.menuItem("objectLayout", this, null));
//    popupMenu.add(popupMenuItem[2] = Translate.menuItem("setTexture", this, null));
//    popupMenu.add(popupMenuItem[3] = Translate.menuItem("setMaterial", this, null));
//    popupMenu.add(popupMenuItem[4] = Translate.menuItem("renameObject", this, null));
//    popupMenu.add(popupMenuItem[5] = Translate.menuItem("convertToTriangle", this, null));
//    popupMenu.add(popupMenuItem[6] = Translate.menuItem("selectChildren", this, null));
//    popupMenu.addSeparator();
//    popupMenu.add(popupMenuItem[7] = Translate.menuItem("hideSelection", this, null));
//    popupMenu.add(popupMenuItem[8] = Translate.menuItem("showSelection", this, null));
//    popupMenu.addSeparator();
//    popupMenu.add(popupMenuItem[9] = Translate.menuItem("cut", this, null));
//    popupMenu.add(popupMenuItem[10] = Translate.menuItem("copy", this, null));
//    popupMenu.add(popupMenuItem[11] = Translate.menuItem("clear", this, null));
//    add(popupMenu);
//  }

  /** Display the popup menu. */

  public void showPopupMenu(Component c, int x, int y)
  {
    updateMenus();
    popupMenu.show(c, x, y);
  }

  /** We need to be notified of key events which take place anywhere in the window.  To make
      sure this happens, we need to add the window as a KeyListener to every component
      contained in it. */

  private void recursivelyAddListeners(Component c)
  {
    Component children[];
    if (!(c instanceof TextComponent))   // TODO(MB) Check
      c.addKeyListener(this);
    if (c instanceof Container)
      {
	children = ((Container) c).getComponents();
	for (int i = 0; i < children.length; i++)
	  recursivelyAddListeners(children[i]);
      }
  }

  /** Set the wait cursor on everything in this window. */

  public void setWaitCursor()
  {
    recursivelySetCursor(this, Cursor.getDefaultCursor(), Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
  }

  /** Remove the wait cursor from everything in this window. */

  public void clearWaitCursor()
  {
    recursivelySetCursor(this, Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR), Cursor.getDefaultCursor());
  }

  private void recursivelySetCursor(Component c, Cursor oldcursor, Cursor newcursor)
  {
    Cursor curs = c.getCursor();
    if (curs == oldcursor)
      c.setCursor(newcursor);
    if (c instanceof Container)
      {
	Container cont = (Container) c;
	for (int i = 0; i < cont.getComponentCount(); i++)
	  recursivelySetCursor(cont.getComponent(i), oldcursor, newcursor);
      }
  }

  public Dimension getMinimumSize()
  {
    return new Dimension(100, 100);
  }

  /* EditingWindow methods. */

  /** This method is called to close the window.  If the Scene has been modified, it first
      gives the user a chance to save the Scene, or to cancel.  If the user cancels it, the
      method returns false.  Otherwise, it closes the window and returns true. */

  public boolean confirmClose()
  {
    if (modified)
      {
	String name = theScene.getName();
	if (name == null)
	  name = "Untitled";
	MessageDialog dlg = new MessageDialog(this, Translate.text("checkSaveChanges", name),
		new String [] {"Save", "Don't Save", "Cancel"});
	if (dlg.getChoice() == 0)
	  {
	    saveCommand();
	    if (modified)
	      return false;
	  }
	if (dlg.getChoice() == 2)
	  return false;
      }
    dispose();
    return true;
  }

  /** Set the selected EditingTool for this window. */

  public void setTool(EditingTool tool)
  {
    for (int i = 0; i < theView.length; i++)
      theView[i].setTool(tool);
    currentTool = tool;
  }

  /** Set the help text displayed at the bottom of the window. */

  public void setHelpText(String text)
  {
    helpText.setText(text);
  }

  /** Get the Frame corresponding to this window.  (Because LayoutWindow is a Frame,
      it simply returns itself.) */

  public Frame getFrame()
  {
    return this;
  }

  /** Update the images displayed in all of the viewport. */

  public void updateImage()
  {
    if (numViewsShown == 1)
      {
	theView[currentView].copyOrientationFromCamera();
	theView[currentView].updateImage();
	theView[currentView].repaint();
      }
    else
      for (int i = 0; i < numViewsShown; i++)
        {
          theView[i].copyOrientationFromCamera();
          theView[i].updateImage();
          theView[i].repaint(10l);
        }
  }

  /** Needed for the MenuUpdateHelper class. */
  private static String[] MENUITEMS_TRACKS = new String[] {
    "xyzOneTrack", "xyzThreeTracks", "proceduralTrack",
    "xyzOneTrack", "xyzThreeTracks", "quaternionTrack",
    "proceduralTrack", "poseTrack", "bendDistortion",
    "customDistortion", "scaleDistortion", "shatterDistortion",
    "twistDistortion", "constraintTrack", "visibilityTrack",
    "textureTrack"
  };

  /** Helper class. On construction, it stores a few values needed to
    decide the state of each menu item when updateMenuItem is called.
  */
  private class MenuUpdateHelper
  {
    Object sel[] = itemTree.getSelectedObjects();
    int numSelObjects = sel.length;
    Track selTrack[] = theScore.getSelectedTracks();
    int numSelTracks = selTrack.length;
    int numSelKeyframes = theScore.getSelectedKeyframes().length;
    boolean canConvert, canSetMaterial, canSetTexture;
    boolean curve, noncurve, enable, disable, hasChildren, hasParent;
    ObjectInfo info;
    Object3D obj;

    public MenuUpdateHelper()
    {
      int i;
      canConvert = canSetMaterial = canSetTexture = (numSelObjects > 0);
      curve = noncurve = enable = disable = hasChildren = hasParent = false;

      for (i = 0; i < numSelObjects; i++)
      {
        info = (ObjectInfo) sel[i];
        obj = info.object;
        if (obj instanceof Curve)
          curve = true;
        else
          noncurve = true;
        if (obj.canConvertToTriangleMesh() == Object3D.CANT_CONVERT)
          canConvert = false;
        if (!obj.canSetTexture())
          canSetTexture = false;
        if (!obj.canSetMaterial())
          canSetMaterial = false;
        if (info.children.length > 0)
          hasChildren = true;
        if (info.parent != null)
          hasParent = true;
      }
      for (i = 0; i < numSelTracks; i++)
      {
        if (selTrack[i].isEnabled())
          disable = true;
        else
          enable = true;
      }
    }

    /** Updates a single menu item
    (setting enabled/disabled, (un)checked ...). */
    protected void updateMenuItem(MenuItem item)
    {
      String itemname = item.getName();
      CheckboxMenuItem cbitem = null;
      if (item instanceof CheckboxMenuItem)
        cbitem = (CheckboxMenuItem)item;

      if ("save".equals(itemname))     // File menu
        item.setEnabled(modified);

      else if ("undo".equals(itemname))   // Edit menu
      {
        item.setEnabled(undo != null);
        if (undo != null)
        {
          if (undo.isRedo())
            item.setLabel(Translate.text("menu.redo"));
          else
            item.setLabel(Translate.text("menu.undo"));
        }
      }
      else if ("cut".equals(itemname))
        item.setEnabled(numSelObjects > 0);
      else if ("copy".equals(itemname))
        item.setEnabled(numSelObjects > 0);
      else if ("paste".equals(itemname))
        item.setEnabled(ModellingApp.getClipboardSize() > 0);
      else if ("clear".equals(itemname))
        item.setEnabled(numSelObjects > 0);
      else if ("selectChildren".equals(itemname))
        item.setEnabled(hasChildren);
      else if ("duplicate".equals(itemname))
        item.setEnabled(numSelObjects > 0);
      else if ("sever".equals(itemname))
        item.setEnabled(numSelObjects > 0);

      else if ("editObject".equals(itemname))  // Object menu
        item.setEnabled(numSelObjects == 1 && obj.editable());
      else if ("objectLayout".equals(itemname))
        item.setEnabled(numSelObjects > 0);
      else if ("transformObject".equals(itemname))
        item.setEnabled(numSelObjects > 0);
      else if ("alignObjects".equals(itemname))
        item.setEnabled((numSelObjects > 0) && numSelObjects > 0);
      else if ("setTexture".equals(itemname))
        item.setEnabled((numSelObjects > 0) && canSetTexture);
      else if ("setMaterial".equals(itemname))
        item.setEnabled((numSelObjects > 0) && canSetMaterial);
      else if ("renameObject".equals(itemname))
        item.setEnabled(numSelObjects == 1);
      else if ("convertToTriangle".equals(itemname))
        item.setEnabled(canConvert && numSelObjects >= 1);
      else if ("hideSelection".equals(itemname))
        item.setEnabled(numSelObjects > 0);
      else if ("showSelection".equals(itemname))
        item.setEnabled(numSelObjects > 0);

      else if ("showScore".equals(itemname))  // Animation menu
        item.setLabel(Translate.text(scoreHidden ? "menu.showScore" : "menu.hideScore"));
      else if ("editKeyframe".equals(itemname))
        item.setEnabled(numSelKeyframes == 1);
      else if ("deleteSelectedKeyframes".equals(itemname))
        item.setEnabled(numSelKeyframes > 0);
      else if ("keyframe".equals(itemname))
        item.setEnabled(numSelTracks > 0);
      else if ("keyframeModified".equals(itemname))
        item.setEnabled(numSelObjects > 0);
      else if ("editTrack".equals(itemname))
        item.setEnabled(numSelTracks == 1);
      else if ("duplicateTracks".equals(itemname))
        item.setEnabled(numSelTracks > 0);
      else if ("deleteTracks".equals(itemname))
        item.setEnabled(numSelTracks > 0);
      else if ("selectAllTracks".equals(itemname))
        item.setEnabled(numSelObjects > 0);
      else if ("enableTracks".equals(itemname))
        item.setEnabled(enable);
      else if ("disableTracks".equals(itemname))
        item.setEnabled(disable);
      else if ("pathFromCurve".equals(itemname))
        item.setEnabled(curve && noncurve);
      else if ("bindToParent".equals(itemname))
        item.setEnabled(hasParent);

      else if (Utilities.findIndexEqual(MENUITEMS_TRACKS, itemname) > -1)
        item.setEnabled(numSelObjects > 0); // Track submenu(s)

      else if ("fourViews".equals(itemname))
      {
        if (numViewsShown == 4)
          item.setLabel(Translate.text("menu.fourViews"));
        else
          item.setLabel(Translate.text("menu.oneView"));
      }
      else if ("hideObjectList".equals(itemname))
        item.setLabel(Translate.text(listHidden ? "menu.showObjectList" : "menu.hideObjectList"));
      else if ("showTemplate".equals(itemname))
        item.setLabel(Translate.text(theView[0].getTemplateShown() ? "menu.showTemplate" : "menu.hideTemplate"));

      else if ("wireframeDisplay".equals(itemname))
        cbitem.setState(theView[currentView].getRenderMode() == ViewerCanvas.RENDER_WIREFRAME);
      else if ("shadedDisplay".equals(itemname))
        cbitem.setState(theView[currentView].getRenderMode() == ViewerCanvas.RENDER_FLAT);
      else if ("smoothDisplay".equals(itemname))
        cbitem.setState(theView[currentView].getRenderMode() == ViewerCanvas.RENDER_SMOOTH);

      else if ("showTemplate".equals(itemname))
      {
        if (theView[0].getTemplateImage() == null)
          item.setEnabled(false);
        else
        {
          item.setEnabled(true);
          boolean wasShown = theView[0].getTemplateShown();
          item.setLabel(Translate.text(wasShown ? "menu.showTemplate" :
            "menu.hideTemplate"));
        }
        item.setEnabled(theView[0].getTemplateImage() != null);
      }
      else if ("frameSelection".equals(itemname))
        item.setEnabled(numSelObjects > 0);
    }
  }


  public void updateMenus()
  {
    MenuItem[] mis;
    MenuUpdateHelper helper = new MenuUpdateHelper();

    mis = Utilities.getMenuItemList(getMenuBar()); // TODO(MB) Use cache
    for(int i = 0; i < mis.length; ++i)
      helper.updateMenuItem(mis[i]);
    mis = Utilities.getMenuItemList(popupMenu);
    for(int i = 0; i < mis.length; ++i)
      helper.updateMenuItem(mis[i]);
  }


  /** Set the UndoRecord which will be executed if the user chooses Undo from the Edit menu. */

  public void setUndoRecord(UndoRecord command)
  {
    undo = command;
    modified = true;
    updateMenus();
  }

  /** Set whether the scene has been modified since it was last saved. */

  public void setModified()
  {
    modified = true;
  }

  /** Add a new object to the scene.  If undo is not null,
      appropriate commands will be added to it to undo this operation. */

  public void addObject(Object3D obj, CoordinateSystem coords, String name, UndoRecord undo)
  {
    addObject(new ObjectInfo(obj, coords, name), undo);
  }

  /** Add a new object to the scene.  If undo is not null,
      appropriate commands will be added to it to undo this operation. */

  public void addObject(ObjectInfo info, UndoRecord undo)
  {
    theScene.addObject(info, undo);
    itemTree.addElement(new ObjectTreeElement(info, itemTree));
    for (int i = 0; i < theView.length ; i++)
      theView[i].rebuildCameraList();
    theScore.rebuildList();
  }

  /** Add a new object to the scene.  If undo is not null,
      appropriate commands will be added to it to undo this operation. */

  public void addObject(ObjectInfo info, int index, UndoRecord undo)
  {
    theScene.addObject(info, index, undo);
    itemTree.addElement(new ObjectTreeElement(info, itemTree), index);
    for (int i = 0; i < theView.length ; i++)
      theView[i].rebuildCameraList();
    theScore.rebuildList();
  }

  /** Remove an object from the scene.  If undo is not null,
      appropriate commands will be added to it to undo this operation. */

  public void removeObject(int which, UndoRecord undo)
  {
    itemTree.removeObject(theScene.getObject(which));
    theScene.removeObject(which, undo);
    for (int i = 0; i < theView.length ; i++)
      theView[i].rebuildCameraList();
    theScore.rebuildList();
  }

  /** Set the name of an object in the scene. */

  public void setObjectName(int which, String name)
  {
    ObjectInfo info = theScene.getObject(which);

    setUndoRecord(new UndoRecord(this, false, UndoRecord.RENAME_OBJECT, new Object [] {new Integer(which), info.name}));
    info.name = name;
    itemTree.repaint();
    for (int i = 0; i < theView.length ; i++)
      theView[i].rebuildCameraList();
    theScore.rebuildList();
  }

  /** Set the time which is currently being displayed. */

  public void setTime(double time)
  {
    theScene.setTime(time);
    theScore.setTime(time);
    NumberFormat nf = NumberFormat.getNumberInstance();
    nf.setMaximumFractionDigits(3);
    timeLabel.setText(nf.format(time));
    frameLabel.setText(Integer.toString((int) Math.round(time*theScene.getFramesPerSecond())));
    theScore.repaint();
    itemTree.repaint();
    updateImage();
  }

  /** Get the Scene associated with this window. */

  public Scene getScene()
  {
    return theScene;
  }

  /** Get the ViewerCanvas which currently has focus. */

  public ViewerCanvas getView()
  {
    return theView[currentView];
  }

  /** Get the Score for this window. */

  public Score getScore()
  {
    return theScore;
  }

  /** Set whether the object list should be displayed. */

  public void setObjectListVisible(boolean visible)
  {
    listHidden = !visible;
    div1.setHidden(false, listHidden);
    updateMenus();
  }

  /** Set whether the score should be displayed. */

  public void setScoreVisible(boolean visible)
  {
    scoreHidden = !visible;
    div2.setHidden(false, scoreHidden);
    updateMenus();
  }

  /** Set whether the window is split into four views. */

  public void setSplitView(boolean split)
  {
    if ((numViewsShown == 1) == split)
      toggleViewsCommand();
  }

  /* From ItemListener. */

  public void itemStateChanged(ItemEvent e)
  {
    Object source = e.getSource();
    int i;

    if (source instanceof CheckboxMenuItem)
      {
        String command = ((CheckboxMenuItem)source).getName();

        if ("wireframeDisplay".equals(command))
          theView[currentView].setRenderMode(ViewerCanvas.RENDER_WIREFRAME);
        else if ("shadedDisplay".equals(command))
          theView[currentView].setRenderMode(ViewerCanvas.RENDER_FLAT);
        else if ("smoothDisplay".equals(command))
          theView[currentView].setRenderMode(ViewerCanvas.RENDER_SMOOTH);
        updateMenus();
	updateImage();
      }
    else
      {
	Object sel[] = itemTree.getSelectedObjects();
	int which[] = new int [sel.length];

	for (i = 0; i < sel.length; i++)
	  which[i] = theScene.indexOf((ObjectInfo) sel[i]);
	setSelection(which);
	if (currentTool.whichClicks() == EditingTool.ALL_CLICKS)
	  tools.selectTool(defaultTool);
	else
	  updateImage();
      }
  }

  /* KeyListener methods. */

  public void keyPressed(KeyEvent e)
  {
    int code = e.getKeyCode();

    if (code == KeyEvent.VK_BACK_SPACE || code == KeyEvent.VK_DELETE)
      clearCommand();
    else if (e.isShiftDown() && (code == KeyEvent.VK_LEFT || code == KeyEvent.VK_RIGHT || code == KeyEvent.VK_UP || code == KeyEvent.VK_DOWN))
      tools.keyPressed(e);
    else
      currentTool.keyPressed(e, theView[currentView]);
  }

  public void keyReleased(KeyEvent e)
  {
  }

  public void keyTyped(KeyEvent e)
  {
  }

  /** Set a single object in the scene to be selected. */

  public void setSelection(int which)
  {
    itemTree.setUpdateEnabled(false);
    clearSelection();
    theScene.setSelection(which);
    itemTree.setSelected(theScene.getObject(which), true);
    itemTree.setUpdateEnabled(true);
    theScore.rebuildList();
    updateMenus();
  }

  /** Set the list of objects in the scene which should be selected. */

  public void setSelection(int which[])
  {
    itemTree.setUpdateEnabled(false);
    clearSelection();
    theScene.setSelection(which);
    for (int i = 0; i < which.length; i++)
      itemTree.setSelected(theScene.getObject(which[i]), true);
    itemTree.setUpdateEnabled(true);
    theScore.rebuildList();
    updateMenus();
  }

  /** Set an object to be selected. */

  public void addToSelection(int which)
  {
    theScene.addToSelection(which);
    itemTree.setSelected(theScene.getObject(which), true);
    theScore.rebuildList();
    updateMenus();
  }

  /** Deselect all objects. */

  public void clearSelection()
  {
    theScene.clearSelection();
    itemTree.deselectAll();
    theScore.rebuildList();
    updateMenus();
  }

  /** Deselect a single object. */

  public void removeFromSelection(int which)
  {
    theScene.removeFromSelection(which);
    itemTree.setSelected(theScene.getObject(which), false);
    theScore.rebuildList();
    updateMenus();
  }

  public void actionPerformed(ActionEvent e)
  {
    String command = e.getActionCommand();
    Object src = e.getSource();
    MenuContainer menu = (src instanceof MenuItem ? ((MenuItem) src).getParent() : null);

    setWaitCursor();
    if (command.equals("new"))
      ModellingApp.newWindow();
    else if (command.equals("open"))
      ModellingApp.openScene(this);
    else if (command.equals("close"))
      ModellingApp.closeWindow(this);
    else if (command.equals("save"))
      saveCommand();
    else if (command.equals("saveas"))
      saveAsCommand();
    else if (command.equals("quit"))
      ModellingApp.quit();
    else if (command.startsWith("import/"))
      importCommand(command.substring(7));
    else if (command.startsWith("export/"))
      exportCommand(command.substring(7));

    else if (command.equals("undo"))
      undoCommand();
    else if (command.equals("cut"))
      cutCommand();
    else if (command.equals("copy"))
      copyCommand();
    else if (command.equals("paste"))
      pasteCommand();
    else if (command.equals("clear"))
      clearCommand();
    else if (command.equals("selectChildren"))
      {
        setSelection(theScene.getSelectionWithChildren());
        updateImage();
      }
    else if (command.equals("selectAll"))
      selectAllCommand();
    else if (command.equals("duplicate"))
      duplicateCommand();
    else if (command.equals("sever"))
      severCommand();
    else if (command.equals("preferences"))
      new PreferencesWindow(this);
    else if (command.equals("editMenu"))
    {
      new MenuEditDialog(this, this);
      rebuildMenuBar();
    }
    else if (command.equals("editObject"))
      editObjectCommand();
    else if (command.equals("objectLayout"))
      objectLayoutCommand();
    else if (command.equals("transformObject"))
      transformObjectCommand();
    else if (command.equals("alignObjects"))
      alignObjectsCommand();
    else if (command.equals("setTexture"))
      setTextureCommand();
    else if (command.equals("setMaterial"))
      setMaterialCommand();
    else if (command.equals("renameObject"))
      renameObjectCommand();
    else if (command.equals("convertToTriangle"))
      convertToTriangleCommand();
    else if (command.equals("hideSelection"))
      setObjectVisibility(false, true);
    else if (command.equals("showSelection"))
      setObjectVisibility(true, true);
    else if (command.equals("showAll"))
      setObjectVisibility(true, false);

    else if (Utilities.findIndexEqual(CREATABLEOBJECTS, command) > -1)
      createObjectCommand(command);
    else if (command.startsWith("modellingTool/"))
      modellingToolCommand(command.substring(14));
    else if (command.equals("createScriptObject"))
      createScriptObjectCommand();
    else if (command.equals("editScript"))
      new ExecuteScriptWindow(this);
    else if (command.equals("showScore"))
      setScoreVisible(scoreHidden);
    else if (command.equals("previewAnimation"))
      new AnimationPreviewer(this);
    else if (command.equals("forwardFrame"))
      {
        double t = theScene.getTime() + 1.0/theScene.getFramesPerSecond();
        setTime(t);
      }
    else if (command.equals("backFrame"))
      {
        double t = theScene.getTime() - 1.0/theScene.getFramesPerSecond();
        setTime(t);
      }
    else if (command.equals("jumpToTime"))
      jumpToTimeCommand();
    else if (command.equals("editKeyframe"))
      theScore.editSelectedKeyframe();
    else if (command.equals("deleteSelectedKeyframes"))
      theScore.deleteSelectedKeyframes();
    else if (command.equals("editTrack"))
      theScore.editSelectedTrack();
    else if (command.equals("keyframe"))
      theScore.keyframeSelectedTracks();
    else if (command.equals("keyframeModified"))
      theScore.keyframeModifiedTracks();
    else if (command.equals("duplicateTracks"))
      theScore.duplicateSelectedTracks();
    else if (command.equals("deleteTracks"))
      theScore.deleteSelectedTracks();
    else if (command.equals("selectAllTracks"))
      theScore.selectAllTracks();
    else if (command.equals("enableTracks"))
      theScore.setTracksEnabled(true);
    else if (command.equals("disableTracks"))
      theScore.setTracksEnabled(false);
    else if (command.equals("pathFromCurve"))
      new PathFromCurveDialog(this, itemTree.getSelectedObjects());
    else if (command.equals("bindToParent"))
      bindToParentCommand();
    else if (command.equals("moveKeyframes"))
      new EditKeyframesDialog(this, EditKeyframesDialog.MOVE);
    else if (command.equals("copyKeyframes"))
      new EditKeyframesDialog(this, EditKeyframesDialog.COPY);
    else if (command.equals("rescaleKeyframes"))
      new EditKeyframesDialog(this, EditKeyframesDialog.RESCALE);
    else if (command.equals("loopKeyframes"))
      new EditKeyframesDialog(this, EditKeyframesDialog.LOOP);
    else if (command.equals("deleteKeyframes"))
      new EditKeyframesDialog(this, EditKeyframesDialog.DELETE);
    else if (command.equals("poseTrack"))
      theScore.addTrack(itemTree.getSelectedObjects(), PoseTrack.class, null, true);
    else if (command.equals("constraintTrack"))
      theScore.addTrack(itemTree.getSelectedObjects(), ConstraintTrack.class, null, true);
    else if (command.equals("visibilityTrack"))
      theScore.addTrack(itemTree.getSelectedObjects(), VisibilityTrack.class, null, true);
    else if (command.equals("textureTrack"))
      theScore.addTrack(itemTree.getSelectedObjects(), TextureTrack.class, null, true);
    else if (command.equals("xyzOneTrack"))
      theScore.addTrack(itemTree.getSelectedObjects(), PositionTrack.class, null, true);
    else if (command.equals("xyzThreeTracks"))
      {
        theScore.addTrack(itemTree.getSelectedObjects(), PositionTrack.class, new Object [] {"Z Position", Boolean.FALSE, Boolean.FALSE, Boolean.TRUE}, true);
        theScore.addTrack(itemTree.getSelectedObjects(), PositionTrack.class, new Object [] {"Y Position", Boolean.FALSE, Boolean.TRUE, Boolean.FALSE}, false);
        theScore.addTrack(itemTree.getSelectedObjects(), PositionTrack.class, new Object [] {"X Position", Boolean.TRUE, Boolean.FALSE, Boolean.FALSE}, false);
      }
    else if (command.equals("proceduralTrack"))
      theScore.addTrack(itemTree.getSelectedObjects(), ProceduralPositionTrack.class, null, true);
    else if (command.equals("xyzOneTrack"))
      theScore.addTrack(itemTree.getSelectedObjects(), RotationTrack.class, new Object [] {"Rotation", Boolean.FALSE, Boolean.TRUE, Boolean.TRUE, Boolean.TRUE}, true);
    else if (command.equals("xyzThreeTracks"))
      {
        theScore.addTrack(itemTree.getSelectedObjects(), RotationTrack.class, new Object [] {"Z Rotation", Boolean.FALSE, Boolean.FALSE, Boolean.FALSE, Boolean.TRUE}, true);
        theScore.addTrack(itemTree.getSelectedObjects(), RotationTrack.class, new Object [] {"Y Rotation", Boolean.FALSE, Boolean.FALSE, Boolean.TRUE, Boolean.FALSE}, false);
        theScore.addTrack(itemTree.getSelectedObjects(), RotationTrack.class, new Object [] {"X Rotation", Boolean.FALSE, Boolean.TRUE, Boolean.FALSE, Boolean.FALSE}, false);
      }
    else if (command.equals("quaternionTrack"))
      theScore.addTrack(itemTree.getSelectedObjects(), RotationTrack.class, new Object [] {"Rotation", Boolean.TRUE, Boolean.TRUE, Boolean.TRUE, Boolean.TRUE}, true);
    else if (command.equals("proceduralTrack"))
      theScore.addTrack(itemTree.getSelectedObjects(), ProceduralRotationTrack.class, null, true);
    else if (command.equals("bendDistortion"))
      theScore.addTrack(itemTree.getSelectedObjects(), BendTrack.class, null, true);
    else if (command.equals("customDistortion"))
      theScore.addTrack(itemTree.getSelectedObjects(), CustomDistortionTrack.class, null, true);
    else if (command.equals("IKTrack"))
      theScore.addTrack(itemTree.getSelectedObjects(), IKTrack.class, null, true);      
    else if (command.equals("scaleDistortion"))
      theScore.addTrack(itemTree.getSelectedObjects(), ScaleTrack.class, null, true);
    else if (command.equals("shatterDistortion"))
      theScore.addTrack(itemTree.getSelectedObjects(), ShatterTrack.class, null, true);
    else if (command.equals("twistDistortion"))
      theScore.addTrack(itemTree.getSelectedObjects(), TwistTrack.class, null, true);
    else if (command.equals("renderScene"))
      renderCommand();
    else if (command.equals("renderImmediately"))
      renderImmediatelyCommand();
    else if (command.equals("fourViews"))
      toggleViewsCommand();
    else if (command.equals("hideObjectList"))
      setObjectListVisible(listHidden);
    else if (command.equals("showTemplate"))
      {
        boolean wasShown = theView[0].getTemplateShown();
        for (int i = 0; i < theView.length; i++)
          theView[i].setShowTemplate(!wasShown);
        updateMenus();
        updateImage();
      }
    else if (command.equals("setTemplate"))
      setTemplateCommand();
    else if (command.equals("grid"))
      setGridCommand();
    else if (command.equals("frameSelection"))
      frameWithCameraCommand(true);
    else if (command.equals("frameScene"))
      frameWithCameraCommand(false);
    else if (command.equals("textures"))
      texturesCommand();
    else if (command.equals("materials"))
      materialsCommand();
    else if (command.equals("images"))
      new ImagesDialog(this, theScene, null);
    else if (command.equals("environment"))
      environmentCommand();
    else if (command.startsWith("beanshell/"))
      executeScript(new File(command.substring(10)));
    clearWaitCursor();
  }

  void importCommand(String transclass)
  {
    Translator trans[] = ModellingApp.getTranslators();

    for (int i = 0; i < trans.length; i++)
      if (trans[i].canImport() &&
          trans[i].getClass().getName().equals(transclass))
	{
	  trans[i].importFile(this);
	  return;
	}
  }

  void exportCommand(String transclass)
  {
    Translator trans[] = ModellingApp.getTranslators();

    for (int i = 0; i < trans.length; i++)
      if (trans[i].canExport() &&
          trans[i].getClass().getName().equals(transclass))
	{
	  trans[i].exportFile(this, theScene);
	  return;
	}
  }

  void modellingToolCommand(String toolclass)
  {
    ModellingTool tools[] = ModellingApp.getModellingTools();
    for(int i = 0; i < tools.length; ++i)
      if (tools[i].getClass().getName().equals(toolclass))
        tools[i].commandSelected(this);
  }

  void saveCommand()
  {
    if (theScene.getName() == null)
      saveAsCommand();
    else
      modified = !ModellingApp.saveScene(theScene, this);
  }

  void saveAsCommand()
  {
    FileDialog fd = new FileDialog(this, "Save Scene", FileDialog.SAVE);
    if (theScene.getName() == null)
      fd.setFile("Untitled.aoi");
    else
      fd.setFile(theScene.getName());
    if (theScene.getDirectory() != null)
      fd.setDirectory(theScene.getDirectory());
    else if (ModellingApp.currentDirectory != null)
      fd.setDirectory(ModellingApp.currentDirectory);
    fd.setCursor(Cursor.getDefaultCursor());
    fd.show();
    if (fd.getFile() == null)
      return;
    theScene.setName(fd.getFile());
    theScene.setDirectory(fd.getDirectory());
    setTitle(fd.getFile());
    modified = !ModellingApp.saveScene(theScene, this);
  }

  void undoCommand()
  {
    if (undo == null)
      return;
    setUndoRecord(undo.execute());
    rebuildItemList();
    updateImage();
    updateMenus();
  }

  void cutCommand()
  {
    copyCommand();
    clearCommand();
  }

  void copyCommand()
  {
    int sel[] = theScene.getSelectionWithChildren();
    if (sel.length == 0)
      return;
    ObjectInfo copy[] = new ObjectInfo [sel.length];
    for (int i = 0; i < sel.length; i++)
      copy[i] = theScene.getObject(sel[i]);
    copy = ObjectInfo.duplicateAll(copy);
    ModellingApp.copyToClipboard(copy, theScene);
    updateMenus();
  }

  void pasteCommand()
  {
    int which[] = new int [ModellingApp.getClipboardSize()], num = theScene.getNumObjects();
    for (int i = 0; i < which.length; i++)
      which[i] = num+i;
    ModellingApp.pasteClipboard(this);
    setSelection(which);
    rebuildItemList();
    updateImage();
  }

  void clearCommand()
  {
    Object sel[] = itemTree.getSelectedObjects();
    boolean any;
    int i, j;

    if (sel.length == 0)
      return;
    theScene.clearSelection();
    setUndoRecord(new UndoRecord(this, false));

    // First remove any selected objects.

    for (i = sel.length-1; i >= 0; i--)
      {
	ObjectInfo info = (ObjectInfo) sel[i];
	int index = theScene.indexOf(info);
	removeObject(index, undo);
      }

    // Now remove any objects whose parents were just deleted.

    do
    {
      any = false;
      for (i = 0; i < theScene.getNumObjects(); i++)
	{
	  ObjectInfo info = theScene.getObject(i);
	  if (info.parent != null && theScene.indexOf(info.parent) == -1)
	    {
	      removeObject(i, undo);
	      i--;
	      any = true;
	    }
	}
    } while (any == true);
    updateMenus();
    updateImage();
  }

  void selectAllCommand()
  {
    int i, which[] = new int [theScene.getNumObjects()];

    for (i = 0; i < which.length; i++)
      which[i] = i;
    setSelection(which);
    updateImage();
  }

  void duplicateCommand()
  {
    Object sel[] = itemTree.getSelectedObjects();
    int i, which[] = new int [sel.length], num = theScene.getNumObjects();

    setUndoRecord(new UndoRecord(this, false));
    for (i = 0; i < sel.length; i++)
      {
	addObject(((ObjectInfo) sel[i]).duplicate(), undo);
	which[i] = num + i;
      }
    setSelection(which);
    updateImage();
  }

  void severCommand()
  {
    Object sel[] = itemTree.getSelectedObjects();
    ObjectInfo info;
    int i;

    setUndoRecord(new UndoRecord(this, false));
    for (i = 0; i < sel.length; i++)
      {
	info = (ObjectInfo) sel[i];
	undo.addCommand(UndoRecord.COPY_OBJECT_INFO, new Object [] {info, info.duplicate()});
	info.object = info.object.duplicate();
      }
  }

  void editObjectCommand()
  {
    int sel[] = theScene.getSelection();
    final Object3D obj;

    if (sel.length != 1)
      return;
    obj = theScene.getObject(sel[0]).object;
    if (obj.editable())
      {
	setUndoRecord(new UndoRecord(this, false, UndoRecord.COPY_OBJECT, new Object [] {obj, obj.duplicate()}));  // TODO(MB) ?
	obj.edit(this, theScene.getObject(sel[0]), new Callback() {     // TODO(MB) Remove the callback some day
	  public void execute()
	  {
	    theScene.objectModified(obj);
	    updateImage();
	    updateMenus();
	  }
	} );
      }
  }

  void objectLayoutCommand()
  {
    int i, sel[] = theScene.getSelection();
    TransformDialog dlg;
    ObjectInfo obj[] = new ObjectInfo [sel.length];
    Vec3 orig, size;
    double angles[], values[];

    if (sel.length == 0)
      return;
    setUndoRecord(new UndoRecord(this, false));
    for (i = 0; i < sel.length; i++)
      {
	obj[i] = theScene.getObject(sel[i]);
	undo.addCommand(UndoRecord.COPY_OBJECT, new Object [] {obj[i].object, obj[i].object.duplicate()});
	undo.addCommand(UndoRecord.COPY_COORDS, new Object [] {obj[i].coords, obj[i].coords.duplicate()});
      }
    if (sel.length == 1)
      {
	orig = obj[0].coords.getOrigin();
	angles = obj[0].coords.getRotationAngles();
	size = obj[0].object.getBounds().getSize();
	dlg = new TransformDialog(this, "Layout for '"+theScene.getObject(sel[0]).name+"'",
	    new double [] {orig.x, orig.y, orig.z, angles[0], angles[1], angles[2],
	    size.x, size.y, size.z}, false, false);
        if (!dlg.clickedOk())
          return;
	values = dlg.getValues();
	if (!Double.isNaN(values[0]))
	  orig.x = values[0];
	if (!Double.isNaN(values[1]))
	  orig.y = values[1];
	if (!Double.isNaN(values[2]))
	  orig.z = values[2];
	if (!Double.isNaN(values[3]))
	  angles[0] = values[3];
	if (!Double.isNaN(values[4]))
	  angles[1] = values[4];
	if (!Double.isNaN(values[5]))
	  angles[2] = values[5];
	if (!Double.isNaN(values[6]))
	  size.x = values[6];
	if (!Double.isNaN(values[7]))
	  size.y = values[7];
	if (!Double.isNaN(values[8]))
	  size.z = values[8];
	obj[0].coords.setOrigin(orig);
	obj[0].coords.setOrientation(angles[0], angles[1], angles[2]);
	obj[0].object.setSize(size.x, size.y, size.z);
	theScene.objectModified(obj[0].object);
        obj[0].object.sceneChanged(obj[0], theScene);
      }
    else
      {
	dlg = new TransformDialog(this, "Layout for Selected Objects", false, false);
        if (!dlg.clickedOk())
          return;
	values = dlg.getValues();
	for (i = 0; i < sel.length; i++)
	  {
	    orig = obj[i].coords.getOrigin();
	    angles = obj[i].coords.getRotationAngles();
	    size = obj[i].object.getBounds().getSize();
	    if (!Double.isNaN(values[0]))
	      orig.x = values[0];
	    if (!Double.isNaN(values[1]))
	      orig.y = values[1];
	    if (!Double.isNaN(values[2]))
	      orig.z = values[2];
	    if (!Double.isNaN(values[3]))
	      angles[0] = values[3];
	    if (!Double.isNaN(values[4]))
	      angles[1] = values[4];
	    if (!Double.isNaN(values[5]))
	      angles[2] = values[5];
	    if (!Double.isNaN(values[6]))
	      size.x = values[6];
	    if (!Double.isNaN(values[7]))
	      size.y = values[7];
	    if (!Double.isNaN(values[8]))
	      size.z = values[8];
	    obj[i].coords.setOrigin(orig);
	    obj[i].coords.setOrientation(angles[0], angles[1], angles[2]);
	    obj[i].object.setSize(size.x, size.y, size.z);
	  }
        for (i = 0; i < sel.length; i++)
          {
            theScene.objectModified(obj[i].object);
            obj[i].object.sceneChanged(obj[i], theScene);
          }
      }
    updateImage();
  }

  void transformObjectCommand()
  {
    int i, sel[] = theScene.getSelection();
    TransformDialog dlg;
    ObjectInfo info;
    Object3D obj;
    CoordinateSystem coords;
    Vec3 orig, size, center;
    double values[];
    Mat4 m;

    if (sel.length == 0)
      return;
    if (sel.length == 1)
      dlg = new TransformDialog(this, "Transform '"+theScene.getObject(sel[0]).name+"'",
		new double [] {0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 1.0, 1.0}, true, true);
    else
      dlg = new TransformDialog(this, "Transform Selected Objects",
		new double [] {0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 1.0, 1.0}, true, true);
    if (!dlg.clickedOk())
      return;
    values = dlg.getValues();

    // Find the center of all selected objects.

    BoundingBox bounds = null;
    for (i = 0; i < sel.length; i++)
      {
	info = theScene.getObject(sel[i]);
	if (bounds == null)
	  bounds = info.getBounds().transformAndOutset(info.coords.fromLocal());
	else
	  bounds = bounds.merge(info.getBounds().transformAndOutset(info.coords.fromLocal()));
      }
    center = bounds.getCenter();
    if (dlg.applyToChildren())
      sel = theScene.getSelectionWithChildren();

    // Determine the rotation matrix.

    m = Mat4.identity();
    if (!Double.isNaN(values[3]))
      m = m.times(Mat4.xrotation(values[3]*Math.PI/180.0));
    if (!Double.isNaN(values[4]))
      m = m.times(Mat4.yrotation(values[4]*Math.PI/180.0));
    if (!Double.isNaN(values[5]))
      m = m.times(Mat4.zrotation(values[5]*Math.PI/180.0));
    setUndoRecord(new UndoRecord(this, false));
    for (i = 0; i < sel.length; i++)
      {
	info = theScene.getObject(sel[i]);
	obj = info.object;
	coords = info.coords;
	undo.addCommand(UndoRecord.COPY_OBJECT, new Object [] {obj, obj.duplicate()});
	undo.addCommand(UndoRecord.COPY_COORDS, new Object [] {coords, coords.duplicate()});
	orig = coords.getOrigin();
	size = obj.getBounds().getSize();
	if (!Double.isNaN(values[0]))
          orig.x += values[0];
	if (!Double.isNaN(values[1]))
	  orig.y += values[1];
	if (!Double.isNaN(values[2]))
	  orig.z += values[2];
	if (!Double.isNaN(values[6]))
	  size.x *= values[6];
	if (!Double.isNaN(values[7]))
	  size.y *= values[7];
	if (!Double.isNaN(values[8]))
	  size.z *= values[8];
	if (dlg.useSelectionCenter())
	  {
	    Vec3 neworig = orig.minus(center);
	    if (!Double.isNaN(values[6]))
	      neworig.x *= values[6];
	    if (!Double.isNaN(values[7]))
	      neworig.y *= values[7];
	    if (!Double.isNaN(values[8]))
	      neworig.z *= values[8];
	    coords.setOrigin(neworig);
	    coords.transformCoordinates(m);
	    coords.setOrigin(coords.getOrigin().plus(center));
	  }
	else
	  {
	    coords.setOrigin(orig);
	    coords.transformAxes(m);
	  }
	obj.setSize(size.x, size.y, size.z);
      }
    for (i = 0; i < sel.length; i++)
      {
        info = theScene.getObject(sel[i]);
        theScene.objectModified(info.object);
        info.object.sceneChanged(info, theScene);
      }
    updateImage();
  }

  void alignObjectsCommand()
  {
    int i, sel[] = theScene.getSelection();
    ComponentsDialog dlg;
    ObjectInfo info;
    Object3D obj;
    CoordinateSystem coords;
    Vec3 alignTo, orig, center;
    double values[];
    Choice xchoice, ychoice, zchoice;
    Panel px = new Panel(), py = new Panel(), pz = new Panel();
    ValueField vfx, vfy, vfz;
    BoundingBox bounds;

    if (sel.length == 0)
      return;
    px.add(xchoice = new Choice());
    xchoice.add("Do not align");
    xchoice.add("Right");
    xchoice.add("Center");
    xchoice.add("Left");
    xchoice.add("Origin");
    px.add(new Label("align to"));
    px.add(vfx = new ValueField(Double.NaN, ValueField.NONE, 5));
    py.add(ychoice = new Choice());
    ychoice.add("Do not align");
    ychoice.add("Top");
    ychoice.add("Center");
    ychoice.add("Bottom");
    ychoice.add("Origin");
    py.add(new Label("align to"));
    py.add(vfy = new ValueField(Double.NaN, ValueField.NONE, 5));
    pz.add(zchoice = new Choice());
    zchoice.add("Do not align");
    zchoice.add("Front");
    zchoice.add("Center");
    zchoice.add("Back");
    zchoice.add("Origin");
    pz.add(new Label("align to"));
    pz.add(vfz = new ValueField(Double.NaN, ValueField.NONE, 5));
    dlg = new ComponentsDialog(this, "Select how to align the objects:",
		new Component [] {px, py, pz}, new String [] {"X", "Y", "Z"});
    if (!dlg.clickedOk())
      return;
    setUndoRecord(new UndoRecord(this, false));

    // Determine the position to align the objects to.

    alignTo = new Vec3();
    for (i = 0; i < sel.length; i++)
      {
	info = theScene.getObject(sel[i]);
	coords = info.coords;
	bounds = info.getBounds();
	bounds = bounds.transformAndOutset(coords.fromLocal());
	center = bounds.getCenter();
	orig = coords.getOrigin();
	if (!Double.isNaN(vfx.getValue()))
	  alignTo.x += vfx.getValue();
	else if (xchoice.getSelectedIndex() == 1)
	  alignTo.x += bounds.maxx;
	else if (xchoice.getSelectedIndex() == 2)
	  alignTo.x += center.x;
	else if (xchoice.getSelectedIndex() == 3)
	  alignTo.x += bounds.minx;
	else if (xchoice.getSelectedIndex() == 4)
	  alignTo.x += orig.x;
	if (!Double.isNaN(vfy.getValue()))
	  alignTo.y += vfy.getValue();
	else if (ychoice.getSelectedIndex() == 1)
	  alignTo.y += bounds.maxy;
	else if (ychoice.getSelectedIndex() == 2)
	  alignTo.y += center.y;
	else if (ychoice.getSelectedIndex() == 3)
	  alignTo.y += bounds.miny;
	else if (ychoice.getSelectedIndex() == 4)
	  alignTo.y += orig.y;
	if (!Double.isNaN(vfz.getValue()))
	  alignTo.z += vfz.getValue();
	else if (zchoice.getSelectedIndex() == 1)
	  alignTo.z += bounds.maxz;
	else if (zchoice.getSelectedIndex() == 2)
	  alignTo.z += center.z;
	else if (zchoice.getSelectedIndex() == 3)
	  alignTo.z += bounds.minz;
	else if (zchoice.getSelectedIndex() == 4)
	  alignTo.z += orig.z;
      }
    alignTo.scale(1.0/sel.length);

    // Now transform all of the objects.

    for (i = 0; i < sel.length; i++)
      {
	info = theScene.getObject(sel[i]);
	coords = info.coords;
	bounds = info.getBounds();
	bounds = bounds.transformAndOutset(coords.fromLocal());
	center = bounds.getCenter();
	orig = coords.getOrigin();
	undo.addCommand(UndoRecord.COPY_COORDS, new Object [] {coords, coords.duplicate()});
	if (xchoice.getSelectedIndex() == 1)
	  orig.x += alignTo.x-bounds.maxx;
	else if (xchoice.getSelectedIndex() == 2)
	  orig.x += alignTo.x-center.x;
	else if (xchoice.getSelectedIndex() == 3)
	  orig.x += alignTo.x-bounds.minx;
	else if (xchoice.getSelectedIndex() == 4)
	  orig.x += alignTo.x-orig.x;
	if (ychoice.getSelectedIndex() == 1)
	  orig.y += alignTo.y-bounds.maxy;
	else if (ychoice.getSelectedIndex() == 2)
	  orig.y += alignTo.y-center.y;
	else if (ychoice.getSelectedIndex() == 3)
	  orig.y += alignTo.y-bounds.miny;
	else if (ychoice.getSelectedIndex() == 4)
	  orig.y += alignTo.y-orig.y;
	if (zchoice.getSelectedIndex() == 1)
	  orig.z += alignTo.z-bounds.maxz;
	else if (zchoice.getSelectedIndex() == 2)
	  orig.z += alignTo.z-center.z;
	else if (zchoice.getSelectedIndex() == 3)
	  orig.z += alignTo.z-bounds.minz;
	else if (zchoice.getSelectedIndex() == 4)
	  orig.z += alignTo.z-orig.z;
	coords.setOrigin(orig);
      }
    for (i = 0; i < sel.length; i++)
      {
	info = theScene.getObject(sel[i]);
        info.object.sceneChanged(info, theScene);
      }
    updateImage();
  }

  void setTextureCommand()
  {
    int sel[] = theScene.getSelection(), i, count = 0;
    ObjectInfo obj[];

    for (i = 0; i < sel.length; i++)
      if (theScene.getObject(sel[i]).object.canSetTexture())
	count++;
    if (count == 0)
      return;
    obj = new ObjectInfo [count];
    for (i = 0; i < sel.length; i++)
      if (theScene.getObject(sel[i]).object.canSetTexture())
	obj[i] = theScene.getObject(sel[i]);
    new ObjectTextureDialog(this, theScene, obj);
    for (i = 0; i < sel.length; i++)
      theScene.objectModified(theScene.getObject(sel[i]).object);
    modified = true;
    updateImage();
  }

  void setMaterialCommand()
  {
    int sel[] = theScene.getSelection(), i, count = 0;
    ObjectInfo obj[], info;

    for (i = 0; i < sel.length; i++)
      {
	info = theScene.getObject(sel[i]);
	if (info.object.canSetMaterial())
	  count++;
      }
    if (count == 0)
      return;
    obj = new ObjectInfo [count];
    for (i = 0; i < sel.length; i++)
      {
	info = theScene.getObject(sel[i]);
	if (info.object.canSetMaterial())
	  obj[i] = info;
      }
    new ObjectMaterialDialog(this, theScene, obj);
    modified = true;
    updateImage();
  }

  void renameObjectCommand()
  {
    int sel[] = theScene.getSelection();
    double tol;
    ObjectInfo info;

    if (sel.length != 1)
      return;
    info = theScene.getObject(sel[0]);
    EntriesDialog dlg = new EntriesDialog(this, "Enter new name for selected object:",
		new String[] {Translate.text("Name")}, new String [] {info.name});
    String val[] = dlg.getEntries();
    if (val == null)
      return;
    setObjectName(sel[0], val[0]);
  }

  void convertToTriangleCommand()
  {
    int sel[] = theScene.getSelection();
    int i, seli;
    double tol;
    Object3D obj;
    TriangleMesh mesh;
    ObjectInfo info = null;
    Texture tex;
    TextureMapping map;
    int convertmode = Object3D.EXACTLY;
    boolean confirmed = false, poseRemConfirmed = false;

    if (sel.length == 0)
      return;

    for(seli = 0; seli < sel.length; ++seli)
    {
      info = theScene.getObject(sel[seli]);
      obj = info.object;
      if (obj.canConvertToTriangleMesh() == Object3D.APPROXIMATELY)
        convertmode = Object3D.APPROXIMATELY;
      else if (obj.canConvertToTriangleMesh() == Object3D.CANT_CONVERT)
      {
        convertmode = Object3D.CANT_CONVERT;
        break;
      }

      if (sel.length == 1)   // TODO(MB)
      {
        // If the object has a Pose track, all Pose keyframes will need to be deleted.
        for (i = 0; i < info.tracks.length; i++)
          if (info.tracks[i] instanceof PoseTrack)
            {
              if (!poseRemConfirmed && !info.tracks[i].isNullTrack())
                {
                  MessageDialog dlg = new MessageDialog(this, Translate.text("convertLosesPosesWarning", info.name),
                    new String [] {Translate.text("button.ok"), Translate.text("button.cancel")});
                  if (dlg.getChoice() == 1)
                    return;
                  confirmed = true;
                  poseRemConfirmed = true;
                }
              if (info.tracks[i].getTimecourse() != null)
                info.tracks[i].getTimecourse().removeAllTimepoints();
              info.pose = null;
            }
      }
    }

    if (confirmed)
      theScore.repaintAll();

    if (convertmode == Object3D.EXACTLY)
      {
	if (!confirmed)
	  {
	    MessageDialog dlg = new MessageDialog(this, Translate.text("confirmConvert", info.name),
	      new String [] {Translate.text("button.ok"), Translate.text("button.cancel")});
	    if (dlg.getChoice() == 1)
	      return;
	  }

        tol = 0.0;
      }
    else
      {
	EntriesDialog dlg = new EntriesDialog(this, "Specify tolerance for triangle mesh:",
		new String[] {"Max Error"}, new String [] {"0.1"});
	String val[] = dlg.getEntries();
	if (val == null)
	  return;
	try
	  {
	    tol = Double.valueOf(val[0]).doubleValue();
	  }
	catch (NumberFormatException ex)
	  {
	    new MessageDialog(this, Translate.text("illegalTolerance"));
	    return;
	  }
	if (tol < 0.0)
	  {
	    new MessageDialog(this, Translate.text("negativeTolerance"));
	    return;
	  }
      }


    if (sel.length == 1)
    {
      info = theScene.getObject(sel[0]);
      obj = info.object;

      mesh = obj.convertToTriangleMesh(tol);

      mesh.setMaterial(obj.getMaterial());
      mesh.setMaterialMapping(obj.getMaterialMapping());
      tex = obj.getTexture();
      map = obj.getTextureMapping();
      if (tex == null && obj.getMaterial() == null)
        {
          tex = theScene.getDefaultTexture();
          map = tex.getDefaultMapping();
        }
      mesh.setTexture(tex);
      mesh.setTextureMapping(map);

      if (poseRemConfirmed)
        theScene.replaceObject(obj, mesh.getPosableObject(), undo);
      theScene.replaceObject(obj, mesh, undo);
    }
    else
    {
      info = theScene.getObject(sel[0]);
      obj = info.object.duplicate();
      Mat4 trans = info.getCoordinates().fromLocal();
      TriangleMesh mesh2;

      mesh = obj.convertToTriangleMesh(tol);
      mesh.transformVertices(trans);

      for(seli = 1; seli < sel.length; ++seli)
      {
        info = theScene.getObject(sel[seli]);
        obj = info.object.duplicate();
        trans = info.getCoordinates().fromLocal();
        mesh2 = obj.convertToTriangleMesh(tol);
        mesh2.transformVertices(trans);

        mesh.addGeometry(mesh2);
      }

      mesh.refreshVertexData();
      // Recenter mesh
      Vec3 center = mesh.getBounds().getCenter();
      trans = Mat4.translation(-center.x, -center.y, -center.z);
      mesh.transformVertices(trans);

      tex = theScene.getDefaultTexture();
      mesh.setTexture(tex);
      mesh.setTextureMapping(tex.getDefaultMapping());

      info = new ObjectInfo(mesh,
          new CoordinateSystem(center, Vec3.vz(), Vec3.vy()), "Combined Mesh");
      info.addTrack(new PositionTrack(info), 0);
      info.addTrack(new RotationTrack(info), 1);

      addObject(info, undo);
    }

    updateImage();
    updateMenus();
  }

  void setObjectVisibility(boolean visible, boolean selectionOnly)
  {
    int i, sel[];
    ObjectInfo info;

    setUndoRecord(new UndoRecord(this, false));
    if (selectionOnly)
      {
	sel = theScene.getSelection();
	for (i = 0; i < sel.length; i++)
	  {
	    info = theScene.getObject(sel[i]);
	    undo.addCommand(UndoRecord.COPY_OBJECT_INFO, new Object [] {info, info.duplicate()});
	    info.visible = visible;
	  }
      }
    else
      for (i = 0; i < theScene.getNumObjects(); i++)
	{
	  info = theScene.getObject(i);
	  undo.addCommand(UndoRecord.COPY_OBJECT_INFO, new Object [] {info, info.duplicate()});
	  info.visible = visible;
	}
    updateImage();
    itemTree.repaint();
  }
  
  /** List of strings recognized by {@link #createObjectCommand}. Needed
    by {@link #actionPerformed}. */
  private static final String[] CREATABLEOBJECTS = new String[]
      {
        "cube", "sphere", "cylinder", "cone", "pointLight", "directionalLight",
        "spotLight", "camera", "null"
      };
      
  void createObjectCommand(String type)
  {
    Object3D obj;
    String name;

    if ("cube".equals(type))
      {
	obj = new Cube(1.0, 1.0, 1.0);
	name = "Cube "+(CreateCubeTool.counter++);
      }
    else if ("sphere".equals(type))
      {
	obj = new Sphere(0.5, 0.5, 0.5);
	name = "Sphere "+(CreateSphereTool.counter++);
      }
    else if ("cylinder".equals(type))
      {
	obj = new Cylinder(1.0, 0.5, 0.5, 1.0);
	name = "Cylinder "+(CreateCylinderTool.counter++);
      }
    else if ("cone".equals(type))
      {
	obj = new Cylinder(1.0, 0.5, 0.5, 0.0);
	name = "Cone "+(CreateCylinderTool.counter++);
      }
    else if ("pointLight".equals(type))
      {
	obj = new PointLight(new RGBColor(1.0f, 1.0f, 1.0f), 1.0f, 0.1);
	name = "Light "+(CreateLightTool.counter++);
      }
    else if ("directionalLight".equals(type))
      {
	obj = new DirectionalLight(new RGBColor(1.0f, 1.0f, 1.0f), 1.0f);
	name = "Light "+(CreateLightTool.counter++);
      }
    else if ("spotLight".equals(type))
      {
	obj = new SpotLight(new RGBColor(1.0f, 1.0f, 1.0f), 1.0f, 20.0, 0.0, 0.1);
	name = "Light "+(CreateLightTool.counter++);
      }
    else if ("camera".equals(type))
      {
	obj = new SceneCamera();
	name = "Camera "+(CreateCameraTool.counter++);
      }
    else
      {
	obj = new NullObject();
	name = "Null";
      }
    obj.setTexture(theScene.getDefaultTexture());
    obj.setTextureMapping(obj.getTexture().getDefaultMapping());
    CoordinateSystem coords = new CoordinateSystem(new Vec3(), Vec3.vz(), Vec3.vy());
    ObjectInfo info = new ObjectInfo(obj, coords, name);
    Vec3 orig = coords.getOrigin();
    double angles[] = coords.getRotationAngles();
    Vec3 size = info.getBounds().getSize();
    TransformDialog dlg = new TransformDialog(this, "Layout for '"+name+"'",
	new double [] {orig.x, orig.y, orig.z, angles[0], angles[1], angles[2],
	size.x, size.y, size.z}, false, false);
    if (!dlg.clickedOk())
      return;
    double values[] = dlg.getValues();
    if (!Double.isNaN(values[0]))
      orig.x = values[0];
    if (!Double.isNaN(values[1]))
      orig.y = values[1];
    if (!Double.isNaN(values[2]))
      orig.z = values[2];
    if (!Double.isNaN(values[3]))
      angles[0] = values[3];
    if (!Double.isNaN(values[4]))
      angles[1] = values[4];
    if (!Double.isNaN(values[5]))
      angles[2] = values[5];
    if (!Double.isNaN(values[6]))
      size.x = values[6];
    if (!Double.isNaN(values[7]))
      size.y = values[7];
    if (!Double.isNaN(values[8]))
      size.z = values[8];
    coords.setOrigin(orig);
    coords.setOrientation(angles[0], angles[1], angles[2]);
    obj.setSize(size.x, size.y, size.z);
    info.addTrack(new PositionTrack(info), 0);
    info.addTrack(new RotationTrack(info), 1);
    setUndoRecord(new UndoRecord(this, false));
    addObject(info, undo);
    setSelection(theScene.getNumObjects()-1);
    updateImage();
  }

  void createScriptObjectCommand()
  {
    // Prompt the user to select a name and, optionally, a predefined script.

    TextField nameField = new TextField("Script");
    Choice scriptChoice = new Choice();
    scriptChoice.add("New Script");
    String files[] = new File(ModellingApp.OBJECT_SCRIPT_DIRECTORY).list();
    if (files != null)
      for (int i = 0; i < files.length; i++)
        if (files[i].endsWith(".bsh") && files[i].length() > 4)
          scriptChoice.add(files[i].substring(0, files[i].length()-4));
    ComponentsDialog dlg = new ComponentsDialog(this, "New Scripted Objects",
      new Component [] {nameField, scriptChoice}, new String [] {Translate.text("Name"), "Script"});
    if (!dlg.clickedOk())
      return;

    // If they are using a predefined script, load it.

    String scriptText = "";
    if (scriptChoice.getSelectedIndex() > 0)
      {
        try
          {
            String scriptName = scriptChoice.getSelectedItem()+".bsh";
            File f = new File(ModellingApp.OBJECT_SCRIPT_DIRECTORY, scriptName);
            scriptText = ModellingApp.loadFile(f);
          }
        catch (IOException ex)
          {
            new MessageDialog(this, "Error Reading Script", new String []
              {Translate.text("errorReadingScript"), ex.getMessage()});
            return;
          }
      }
    ScriptedObject obj = new ScriptedObject("");
    ObjectInfo info = new ObjectInfo(obj, new CoordinateSystem(), nameField.getText());
    setUndoRecord(new UndoRecord(this, false));
    addObject(info, undo);
    setSelection(theScene.getNumObjects()-1);
    updateImage();
    obj.setScript(scriptText);
    obj.edit(this, info, null);
  }

  void jumpToTimeCommand()
  {
    ValueField timeField = new ValueField(theScene.getTime(), ValueField.NONE);
    ComponentsDialog dlg = new ComponentsDialog(this, "Jump To Time",
      new Component [] {timeField}, new String [] {Translate.text("Time")});

    if (!dlg.clickedOk())
      return;
    double t = timeField.getValue();
    double fps = theScene.getFramesPerSecond();
    t = Math.round(t*fps)/(double) fps;
    setTime(t);
  }

  void bindToParentCommand()
  {
    MessageDialog dlg = new MessageDialog(this, Translate.text("confirmBindParent"),
        new String [] {Translate.text("button.ok"), Translate.text("button.cancel")});
    if (dlg.getChoice() == 1)
      return;
    int sel[] = theScene.getSelection();

    setUndoRecord(new UndoRecord(this, false));
    for (int i = 0; i < sel.length; i++)
      {
        ObjectInfo info = theScene.getObject(sel[i]);
        if (info.parent == null)
          continue;
        Skeleton s = info.parent.getSkeleton();
        ObjectRef relObj = new ObjectRef(info.parent);
        if (s != null)
          {
            double nearest = Double.MAX_VALUE;
            Joint jt[] = s.getJoints();
            Vec3 pos = info.coords.getOrigin();
            for (int j = 0; j < jt.length; j++)
              {
                ObjectRef r = new ObjectRef(info.parent, jt[j]);
                double dist = r.getCoords().getOrigin().distance2(pos);
                if (dist < nearest)
                  {
                    relObj = r;
                    nearest = dist;
                  }
              }
          }
        undo.addCommand(UndoRecord.COPY_OBJECT_INFO, new Object [] {info, info.duplicate()});
        PositionTrack pt = new PositionTrack(info);
        pt.setCoordsObject(relObj);
        info.addTrack(pt, 0);
        pt.setKeyframe(theScene.getTime(), theScene);
        RotationTrack rt = new RotationTrack(info);
        rt.setCoordsObject(relObj);
        info.addTrack(rt, 1);
        rt.setKeyframe(theScene.getTime(), theScene);
      }
    theScore.rebuildList();
    theScore.repaint();
  }

  void renderCommand()
  {
    new RenderSetupDialog(this, theScene);
  }

  void renderImmediatelyCommand()
  {
    RenderSetupDialog.renderImmediately(this, theScene);
  }

  void toggleViewsCommand()
  {
    GridBagConstraints gc = new GridBagConstraints();

    viewsContainer.removeAll();
    gc.gridx = gc.gridy = 0;
    gc.gridheight = 2;
    gc.anchor = GridBagConstraints.NORTH;
    gc.fill = GridBagConstraints.BOTH;
    viewsContainer.add(toolsPanel, gc);
    gc.weightx = gc.weighty = 1;
    if (numViewsShown == 4)
      {
       gc.gridx = 1;
	viewsContainer.add(viewPanel[currentView], gc);
	numViewsShown = 1;
      }
    else
      {
       gc.gridheight = 1;
	gc.gridx = 1;
	viewsContainer.add(viewPanel[0], gc);
	gc.gridx = 2;
	viewsContainer.add(viewPanel[1], gc);
	gc.gridy = 1;
	viewsContainer.add(viewPanel[3], gc);
	gc.gridx = 1;
	viewsContainer.add(viewPanel[2], gc);
	numViewsShown = 4;
      }
    viewsContainer.validate();
    updateMenus();
    updateImage();
  }

  void setTemplateCommand()
  {
    FileDialog dlg = new FileDialog(this, "Select an image file", FileDialog.LOAD);
    dlg.show();
    if (dlg.getFile() == null)
      return;
    try
      {
        theView[0].setTemplateImage(new File(dlg.getDirectory(), dlg.getFile()));
      }
    catch (InterruptedException ex)
      {
        new MessageDialog(this, Translate.text("errorLoadingImage"));
      }
    for (int i = 0; i < theView.length; i++)
      theView[i].setShowTemplate(true);
    updateMenus();
    updateImage();
  }

  void setGridCommand()
  {
    ComponentsDialog dlg;
    ValueField spaceField, divField;
    Checkbox showBox, snapBox;
    int i;

    spaceField = new ValueField(theScene.gridSpacing, ValueField.POSITIVE);
    divField = new ValueField(theScene.gridSubdivisions, ValueField.POSITIVE+ValueField.INTEGER);
    showBox = new Checkbox("Show Grid", theScene.showGrid);
    snapBox = new Checkbox("Snap To Grid", theScene.snapToGrid);
    dlg = new ComponentsDialog(this, "Select grid options:",
		new Component [] {spaceField, divField, showBox, snapBox},
		new String [] {"Grid Spacing", "Snap-To Subdivisions", "", ""});
    if (!dlg.clickedOk())
      return;
    theScene.gridSpacing = spaceField.getValue();
    theScene.gridSubdivisions = (int) divField.getValue();
    theScene.showGrid = showBox.getState();
    theScene.snapToGrid = snapBox.getState();
    for (i = 0; i < theView.length; i++)
      theView[i].setGrid(theScene.gridSpacing, theScene.gridSubdivisions, theScene.showGrid, theScene.snapToGrid);
    updateImage();
  }

  void frameWithCameraCommand(boolean selectionOnly)
  {
    int sel[] = theScene.getSelectionWithChildren();
    BoundingBox bb = null;

    if (selectionOnly)
      for (int i = 0; i < sel.length; i++)
        {
          ObjectInfo info = theScene.getObject(sel[i]);
          BoundingBox bounds = info.getBounds().transformAndOutset(info.coords.fromLocal());
          if (bb == null)
            bb = bounds;
          else
            bb = bb.merge(bounds);
        }
    else
      for (int i = 0; i < theScene.getNumObjects(); i++)
        {
          ObjectInfo info = theScene.getObject(i);
          BoundingBox bounds = info.getBounds().transformAndOutset(info.coords.fromLocal());
          if (bb == null)
            bb = bounds;
          else
            bb = bb.merge(bounds);
        }
    if (bb == null)
      return;
    if (numViewsShown == 1)
      theView[currentView].frameBox(bb);
    else
      for (int i = 0; i < theView.length; i++)
        theView[i].frameBox(bb);
    updateImage();
  }

  void texturesCommand()
  {
    theScene.showTexturesDialog(this);
  }

  void materialsCommand()
  {
    theScene.showMaterialsDialog(this);
  }

  void environmentCommand()
  {
    final RGBColor ambColor = theScene.getAmbientColor(), envColor = theScene.getEnvironmentColor(), fogColor = theScene.getFogColor();
    final RGBColor oldAmbColor = ambColor.duplicate(), oldEnvColor = envColor.duplicate(), oldFogColor = fogColor.duplicate();
    final Panel ambPatch = ambColor.getSample(50, 30), envPatch = envColor.getSample(50, 30), fogPatch = fogColor.getSample(50, 30);
    final Checkbox fogBox = new Checkbox("Environment Fog", theScene.getFogState());
    final ValueField fogField = new ValueField(theScene.getFogDistance(), ValueField.POSITIVE);
    final Panel envPanel = new Panel();
    final Choice envChoice = new Choice();
    final Button envButton = new Button("Set:");
    final Label envLabel = new Label();
    final CardLayout envLayout = new CardLayout();
    final Sphere envSphere = new Sphere(1.0, 1.0, 1.0);
    final ObjectInfo envInfo = new ObjectInfo(envSphere, new CoordinateSystem(), "Environment");
    final ComponentsDialog dlg;

    envChoice.add("Solid Color");
    envChoice.add("Texture - Diffuse");
    envChoice.add("Texture - Emissive");
    envChoice.select(theScene.getEnvironmentMode());
    envPanel.setLayout(envLayout);
    Panel p = new Panel();
    p.add(envButton);
    p.add(envLabel);
    envPanel.add(envPatch, "solid");
    envPanel.add(p, "texture");
    if (theScene.getEnvironmentMode() == Scene.ENVIRON_SOLID)
      envLayout.first(envPanel);
    else
      envLayout.last(envPanel);
    envInfo.setTexture(theScene.getEnvironmentTexture(), theScene.getEnvironmentMapping());
    envLabel.setText(envSphere.getTexture().getName());
    envChoice.addItemListener(new ItemListener()
      {
        public void itemStateChanged(ItemEvent e)
        {
	  if (envChoice.getSelectedIndex() == Scene.ENVIRON_SOLID)
	    envLayout.show(envPanel, "solid");
	  else
	    envLayout.show(envPanel, "texture");
        }
      });
    final Callback envTextureCallback = new Callback() {
      public void execute()
      {
        envLabel.setText(envSphere.getTexture().getName());
        envButton.invalidate();
        envLabel.invalidate();
        envPanel.validate();
      }
    };
    envButton.addActionListener(new ActionListener()
      {
        public void actionPerformed(ActionEvent e)
        {
          setCursor(envPanel.getParent(), Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
          ObjectTextureDialog otd = new ObjectTextureDialog(LayoutWindow.this, theScene, new ObjectInfo [] {envInfo});
          otd.setCallback(envTextureCallback);
          setCursor(envPanel.getParent(), Cursor.getDefaultCursor());
        }

        private void setCursor(Component comp, Cursor cur)
        {
          comp.setCursor(cur);
          if (!(comp instanceof Container))
            return;
          Component children[] = ((Container) comp).getComponents();
          for (int i = 0; i < children.length; i++)
            setCursor(children[i], cur);
        }
      });
    ambPatch.addMouseListener(new MouseAdapter()
      {
        public void mouseClicked(MouseEvent e)
        {
          new ColorChooser(LayoutWindow.this, "Ambient Color", ambColor);
          ambPatch.setBackground(ambColor.getColor());
          ambPatch.repaint();
        }
      });
    envPatch.addMouseListener(new MouseAdapter()
      {
        public void mouseClicked(MouseEvent e)
        {
          new ColorChooser(LayoutWindow.this, "Environment Color", envColor);
          envPatch.setBackground(envColor.getColor());
          envPatch.repaint();
        }
      });
    fogPatch.addMouseListener(new MouseAdapter()
      {
        public void mouseClicked(MouseEvent e)
        {
          new ColorChooser(LayoutWindow.this, "Fog Color", fogColor);
          fogPatch.setBackground(fogColor.getColor());
          fogPatch.repaint();
        }
      });
    Callback okCallback = new Callback() {
      public void execute()
      {
        theScene.setFog(fogBox.getState(), fogField.getValue());
        theScene.setEnvironmentMode(envChoice.getSelectedIndex());
        theScene.setEnvironmentTexture(envSphere.getTexture());
        theScene.setEnvironmentMapping(envSphere.getTextureMapping());
        modified = true;
      }
    };
    Callback cancelCallback = new Callback() {
      public void execute()
      {
        ambColor.copy(oldAmbColor);
        envColor.copy(oldEnvColor);
        fogColor.copy(oldFogColor);
      }
    };
    dlg = new ComponentsDialog(LayoutWindow.this, "Environment Properties:",
	new Component [] {ambPatch, envChoice, envPanel, fogBox, fogPatch, fogField},
	new String [] {"Ambient Color", "Environment", "", "", "Fog Color", "Fog Distance"}, okCallback, cancelCallback);
  }

  /** Execute the tool script contained in a file, passing a reference to this window in its "window" variable. */

  public void executeScript(File f)
  {
    // Read the script from the file.

    String scriptText = null;
    try
      {
        scriptText = ModellingApp.loadFile(f);
      }
    catch (IOException ex)
      {
        new MessageDialog(this, "Error Reading Script", new String []
          {Translate.text("errorReadingScript"), ex.getMessage()});
        return;
      }
    try
      {
        ToolScript script = ScriptRunner.parseToolScript(scriptText);
        script.execute(this);
      }
    catch (Exception e)
      {
        ScriptRunner.displayError(e, this, 1);
      }
    updateImage();
  }
}