/* This class implements the dialog box which is used to choose material mappings for objects. 
   It presents a list of all mappings which can be used with the current object and material,
   and allows the user to select one. */

/* Copyright (C) 2000-2002 by Peter Eastman

   This program is free software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the Free Software
   Foundation; either version 2 of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but WITHOUT ANY 
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
   PARTICULAR PURPOSE.  See the GNU General Public License for more details. */

package artofillusion;

import artofillusion.material.*;
import artofillusion.object.*;
import java.awt.*;
import java.awt.event.*;
import java.lang.reflect.*;
import java.util.Vector;

public class MaterialMappingDialog extends Dialog implements ActionListener, ItemListener
{
  Frame fr;
  Object3D obj;
  Vector mappings;
  Choice mapChoice;
  MaterialPreviewer preview;
  MaterialMapping map, oldMapping;
  Panel editingPanel;

  /* Create a dialog for editing the material mapping for a particular object. */

  public MaterialMappingDialog(Frame parent, Object3D obj)
  {
    super(parent, "Material Mapping", true);
    
    fr = parent;
    this.obj = obj;
    map = obj.getMaterialMapping();
    oldMapping = map.duplicate();
    
    // Make a list of all material mappings which can be used for this object and material.
    
    mappings = new Vector();
    Class allMappings[] = ModellingApp.getMaterialMappings();
    for (int i = 0; i < allMappings.length; i++)
      {
	try
	  {
	    Method mtd = allMappings[i].getMethod("legalMapping", new Class [] {Object3D.class, Material.class});
	    Material mat = obj.getMaterial();
	    Boolean result = (Boolean) mtd.invoke(null, new Object [] {obj, mat});
	    if (result.booleanValue())
	      mappings.addElement(allMappings[i]);
	  }
	catch (Exception ex)
	{
	}
      }

    // Add the various components to the dialog.
    
    GridBagConstraints gc = new GridBagConstraints();
    setLayout(new GridBagLayout());
    gc.gridwidth = 2;
    add(preview = new MaterialPreviewer(obj.getTexture(), obj.getMaterial(), obj.duplicate(), 160, 160), gc);
    preview.setMaterialMapping(obj.getMaterialMapping());
    gc.gridwidth = 1;
    gc.gridy = 1;
    add(new Label("Mapping:"), gc);
    add(mapChoice = new Choice(), gc);
    Method mtd;
    for (int i = 0; i < mappings.size(); i++)
      {
	try
	  {
	    mtd = ((Class) mappings.elementAt(i)).getMethod("getName", null);
	    mapChoice.add((String) mtd.invoke(null, null));
	    if (mappings.elementAt(i) == map.getClass())
	      mapChoice.select(i);
	  }
	catch (IllegalAccessException ex)
	  {
	  }
	catch (IllegalArgumentException ex)
	  {
	  }
	catch (java.lang.reflect.InvocationTargetException ex)
	  {
	  }
	catch (NoSuchMethodException ex)
	  {
	  }
	catch (SecurityException ex)
	  {
	  }
      }
    mapChoice.addItemListener(this);
    gc.gridwidth = 2;
    gc.gridy = 2;
    add(editingPanel = map.getEditingPanel(obj, preview), gc);

    // Add the buttons at the bottom.

    Panel p = new Panel();
    Button b;

    gc.gridwidth = 2;
    gc.gridy = 3;
    add(p, gc);
    p.add(b = new Button("Ok"));
    b.setActionCommand("ok");
    b.addActionListener(this);
    p.add(b = new Button("Cancel"));
    b.setActionCommand("cancel");
    b.addActionListener(this);

    // Show the dialog.

    pack();
    setResizable(false);
    ModellingApp.centerWindow(this);
    show();
  }
  
  public void actionPerformed(ActionEvent e)
  {
    String command = e.getActionCommand();
    String finalStrings[] = new String [9];
    int i;

    if (command.equals("cancel"))
      setMapping(oldMapping);
    dispose();
  }

  public void itemStateChanged(ItemEvent e)
  {
    try
      {
	Class cls = (Class) mappings.elementAt(mapChoice.getSelectedIndex());
	if (cls == map.getClass())
	  return;
	Constructor con = cls.getConstructor(new Class [] {Material.class});
	Material mat =  obj.getMaterial();
	setMapping((MaterialMapping) con.newInstance(new Object [] {mat}));
	remove(editingPanel);
	GridBagConstraints gc = new GridBagConstraints();
	gc.gridwidth = 2;
	gc.gridy = 2;
	add(editingPanel = map.getEditingPanel(obj, preview), gc);
	pack();
	preview.render();
      }
    catch (IllegalAccessException ex)
      {
      }
    catch (InstantiationException ex)
      {
      }
    catch (InvocationTargetException ex)
      {
      }
    catch (NoSuchMethodException ex)
      {
      }
  }
  
  /* Set the mapping for the object being edited. */
  
  private void setMapping(MaterialMapping newmap)
  {
    map = newmap;
    obj.setMaterialMapping(newmap);
    preview.setMaterialMapping(newmap);
    preview.render();
  }
}