/* The MeshEditorWindow class represents the window for editing Mesh objects.  This is an
   abstract class, with subclasses for various types of objects. */

/* Copyright (C) 1999-2002 by Peter Eastman

   This program is free software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the Free Software
   Foundation; either version 2 of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
   PARTICULAR PURPOSE.  See the GNU General Public License for more details. */

package artofillusion;

import artofillusion.animation.*;
import artofillusion.math.*;
import artofillusion.object.*;
import artofillusion.texture.*;
import artofillusion.ui.*;
import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.util.Vector;

public abstract class MeshEditorWindow extends Frame implements
    ActionListener, EditingWindow, ItemListener, KeyListener, MenuItemFactory
{
  EditingWindow parentWindow;
  MeshViewer theView[];
  protected Panel viewPanel[], viewsContainer;
  int numViewsShown, currentView;
  MeshSelectionHolder baseSelHolder;
  SkeletonControlHolder baseSkelControlHolder;

  ToolPalette tools;
  Panel toolsPanel;
  EditingTool defaultTool, currentTool, metaTool, altTool;

  Label helpText;
  Object3D oldObject;
  MenuBar menubar;
  Menu viewMenu;
  // MenuItem undoItem, templateItem;
  // CheckboxMenuItem displayItem[], coordsItem[], showItem[];
  UndoRecord undo;
  /** List of all MenuItem and CheckboxMenuItem used for the editor window. */
  protected MenuItem[] menuitemsCache;
  protected Callback onClose;
  protected ObjectInfo objInConstructor;  // TODO(MB) Remove this

  /** 'Or' it to bit mask, if topology changing is allowed. Constant for some derived classes. */
  public static final int OPTION_ALLOWTOPOLOGY = 1;

  /** 'Or' it to bit mask, if window should allow to store/discard changes with "Ok" or "Cancel".
    Constant for some derived classes. */
  public static final int OPTION_ALLOWSOKCANCEL = 2;

  static int meshTension = 2;
  static int tensionDistance = 0;
  static int lastNumViewsShown = 1;
  static final double[] tensionArray = {5.0, 3.0, 2.0, 1.0, 0.5};

  public MeshEditorWindow(EditingWindow parent, String title, ObjectInfo obj,
    Callback onClose)
  {
    super(title);
    objInConstructor = obj;
    parentWindow = parent;
//    oldMesh = (Mesh) obj;
    oldObject = obj.getObject3D().duplicate();
    obj.getObject3D().prepareForEditing();

    if (ModellingApp.defaultFont != null)
      setFont(ModellingApp.defaultFont);
    addWindowListener(new WindowAdapter() {
      public void windowClosing(WindowEvent e)
      {
        MessageDialog dlg = new MessageDialog(MeshEditorWindow.this,
            "Save changes made in this window?",
            new String [] {"Save Changes", "Discard Changes", Translate.text("button.cancel")});
        if (dlg.getChoice() == 0)
          actionPerformed(new ActionEvent(this, ActionEvent.ACTION_PERFORMED, "ok"));
        if (dlg.getChoice() == 1)
          actionPerformed(new ActionEvent(this, ActionEvent.ACTION_PERFORMED, "cancel"));
      }
    } );

    // Initialize the four possible viewers

    numViewsShown = 4;
    currentView = 0;
    theView = new MeshViewer [4];
    viewPanel = new Panel [4];
    Panel p;
    MouseListener listen = new MouseAdapter() {
    public void mousePressed(MouseEvent ev)
      {
	for (int i = 0; i < theView.length; i++)
	  if (currentView != i && ev.getSource() == theView[i])
            {
              theView[currentView].setDrawFocus(false);
              theView[i].setDrawFocus(true);
              currentView = i;
              updateMenus();
              updateImage();
            }
      }
    };

    // Initialize skel. control holder
    
    baseSkelControlHolder = new SkeletonControlHolder(obj);
    for (int i = 0; i < 4; i++)
      {
	viewPanel[i] = new Panel();
	viewPanel[i].setLayout(new BorderLayout());
	viewPanel[i].add(p = new Panel(), "North");
	viewPanel[i].add(theView[i] = createMeshViewer(obj, p), "Center");
        theView[i].setMetaTool(getMetaTool());
        theView[i].setAltTool(getAltTool());
        theView[i].setSkeletonControlHolder(baseSkelControlHolder);

//	theView[i].setGrid(theScene.gridSpacing, theScene.gridSubdivisions, theScene.showGrid, theScene.snapToGrid);
	theView[i].addMouseListener(listen);
      }
    theView[1].selectOrientation(2);
    theView[2].selectOrientation(4);
    theView[3].selectOrientation(6);
    theView[3].setPerspective(true);
    theView[currentView].setDrawFocus(true);
    viewsContainer = new Panel();
    viewsContainer.setLayout(new GridBagLayout());

    // Lay out the editor window
    ObjectInfo info = obj;
    this.onClose = onClose;
    add(helpText = new Label(), "South");
    Panel p1 = new Panel(), p2, p3;
    add(p1, "Center");
    p1.setLayout(new BorderLayout());
    p1.add(p2 = new Panel(), "North");
    p2.setLayout(new BorderLayout());
    p2.add(p3 = new Panel(), "West");
    p1.add(viewsContainer, "Center");
    p2.add(p3 = new Panel(), "East");
    p3.add(Translate.button("ok", this));
    p3.add(Translate.button("cancel", this));
    initBaseSelectionHolder(obj);    
    initWindowMenus(obj);
    toolsPanel = createToolbarPanel();
    add(toolsPanel, "West");

    helpText.setBackground(ModellingApp.APP_BACKGROUND_COLOR);
    p2.setBackground(ModellingApp.APP_BACKGROUND_COLOR);
    p3.setBackground(ModellingApp.APP_BACKGROUND_COLOR);
    toolsPanel.setBackground(ModellingApp.APP_BACKGROUND_COLOR);


    if (lastNumViewsShown == 1)
    {
      toggleViewsCommand();
    }
    else if (lastNumViewsShown == 4)   // TODO(MB) Better solution
    {
      toggleViewsCommand();
      toggleViewsCommand();
    }
  }

  /* We need to be notified of key events which take place anywhere in the window.  To make
     sure this happens, we need to add the window as a KeyListener to every component
     contained in it. */

  protected void recursivelyAddListeners(Component c)
  {
    Component children[];

    if (!(c instanceof TextComponent))
      c.addKeyListener(this);
    if (c instanceof Container)
      {
	children = ((Container) c).getComponents();
	for (int i = 0; i < children.length; i++)
	  recursivelyAddListeners(children[i]);
      }
  }

  /** Tests if menu item item should be shown in the menus for
    this object editor. */
  protected boolean isMenuItemAllowed(MenuItem item)
  {
    return true;
  }

  /** List of names of menu items which should be CheckboxMenuItem.
  Used by #isMenuItemNameCheckbox.*/

  private static final String[] MENUITEMS_CHECKBOXITEMS=new String[]
    { "wireframeDisplay", "shadedDisplay", "smoothDisplay",
      "transparentDisplay", "coordinateSystem", "localCoords",
      "sceneCoords", "controlMesh", "surface", "skeleton",
      "entireScene", "autosyncDragging"
    };


  /** Tests if menu item item should be a checkbox item. */
  protected boolean isMenuItemNameCheckbox(String itemname)
  {
    return (Utilities.findIndexEqual(MENUITEMS_CHECKBOXITEMS, itemname) > -1);
  }

  /** Creates a MenuItem which is either a normal MenuItem,
    a CheckboxMenuItem or a whole submenu, depending on
    {@link #isMenuItemCheckbox}.
    A appropriate listener is registered for this automatically
    If {@link #isMenuItemAllowed} returns false on it, createMenuItem
    returns null.
  */

  protected MenuItem createMenuItem(String name)
  {  // TODO(MB) Remove
    return createMenuItem(name, null, null);
  }

  /** Creates a MenuItem which is either a normal MenuItem,
    a CheckboxMenuItem or a whole submenu, depending on
    {@link #isMenuItemCheckbox}.
    A appropriate listener is registered for this automatically.

    @param name Internal name of the item (as recognized by action listener)
    @param shortcut Shortcut of the item (or null for default)
    @param label Label of the item (or null for default label)

    @return The menu item or null if creation is not allowed.
  */

  public MenuItem createMenuItem(String name, MenuShortcut shortcut, String label)
  {
    MenuItem result;

    if (isMenuItemNameCheckbox(name))
      result = Translate.checkboxMenuItem(name, this);
    else
      result = Translate.menuItem(name, this);

    if (!isMenuItemAllowed(result))
      return null;

    if (label != null)
      result.setLabel(label);

    if (shortcut != null)
      result.setShortcut(shortcut);

    return result;
  }

  /** The itemnames array
    is a list of names for the menu items. The special name "separator"
    creates a separating line between groups of menu items.
    <P>The method tests automatically if the item is allowed and if
    it should be a checkbox item. Multiple subsequent separators and
    separators at the end of the menu are automatically removed.
  */

  protected Menu createMenu(Object[] itemnames)
  {
    boolean separator = false;
    MenuItem mi;
    Menu result = Translate.menu((String)itemnames[0]);

    for(int i = 1; i < itemnames.length; ++i)
    {
      if ("separator".equals(itemnames[i]))
      {
        separator = true;  // Actual creation of separators is delayed
        continue;
      }
      mi = null;
      if (itemnames[i] instanceof String)
        mi = createMenuItem((String)itemnames[i]);
      else if (itemnames[i] instanceof Object[])
	mi = createMenu((Object[])itemnames[i]);

      if (mi == null)
	continue;

      if (separator)
      {
        result.addSeparator();
        separator = false;
      }

      if (mi instanceof Menu)   //Little hack because Menu.add will otherwise not work
        result.add((Menu)mi);
      else
        result.add(mi);
    }

    return result;
  }

  protected void createViewMenu()
  {
    viewMenu = createMenu(new Object[]{"view", // Menu title
      new Object[]{"displayMode",              // Title of submenu
        "wireframeDisplay",
        "shadedDisplay",
        "smoothDisplay",
        "transparentDisplay" },
      new Object[]{"show",
        "controlMesh",
        "surface",
        "skeleton",
        "entireScene" },
      new Object[]{"coordinateSystem",
        "localCoords",
        "sceneCoords" },
      "setViewSettingsForAllViews",
      "separator",

      "fourViews",
      "autosyncDragging",
      "separator",
      
      "grid",
      "createNewViewNewSelection",
      "showTemplate",
      "setTemplate" });
    menubar.add(viewMenu);
  }

  public void toggleViewsCommand()
  {
    GridBagConstraints gc = new GridBagConstraints();

    viewsContainer.removeAll();
    gc.gridx = gc.gridy = 0;
    gc.gridheight = 2;
    gc.anchor = GridBagConstraints.NORTH;
    gc.fill = GridBagConstraints.BOTH;
    viewsContainer.add(toolsPanel, gc);
    gc.weightx = gc.weighty = 1;
    if (numViewsShown == 4)
      {
        for(int i = 0; i < theView.length; ++i)
        {
          if (i != currentView)
            theView[i].setVisible(false);
        }
        gc.gridx = 1;
	viewsContainer.add(viewPanel[currentView], gc);
	numViewsShown = 1;
      }
    else
      {
       gc.gridheight = 1;
	gc.gridx = 1;
	viewsContainer.add(viewPanel[0], gc);
	gc.gridx = 2;
	viewsContainer.add(viewPanel[1], gc);
	gc.gridy = 1;
	viewsContainer.add(viewPanel[3], gc);
	gc.gridx = 1;
	viewsContainer.add(viewPanel[2], gc);
	numViewsShown = 4;
        for(int i = 0; i < theView.length; ++i)
          theView[i].setVisible(true);

      }
    viewsContainer.validate();
    updateMenus();
    updateImage();
    lastNumViewsShown = numViewsShown;
  }


  protected Menu createShowMenu()  // TODO(MB) Remove method
  {
    Menu menu = Translate.menu("show");
    menu.add(Translate.checkboxMenuItem("controlMesh", this, theView[0].getMeshVisible()));
    menu.add(Translate.checkboxMenuItem("surface", this, theView[0].getSurfaceVisible()));
    menu.add(Translate.checkboxMenuItem("skeleton", this, theView[0].getSkeletonVisible()));
    menu.add(Translate.checkboxMenuItem("entireScene", this, theView[0].getSceneVisible()));
    return menu;
  }

  /** Updates a single menu item
    (setting enabled/disabled, (un)checked ...). */

  protected void updateMenuItem(MenuItem item)
  {
    String itemname = item.getName();
    MeshViewer currView = getCurrentView();
    CheckboxMenuItem cbitem = null;
    if (item instanceof CheckboxMenuItem)
      cbitem = (CheckboxMenuItem)item;

    if ("undo".equals(itemname))
    {
      item.setEnabled(undo != null);
      if (undo != null)
      {
        if (undo.isRedo())
          item.setLabel(Translate.text("menu.redo"));
        else
          item.setLabel(Translate.text("menu.undo"));
      }
    }
    else if ("localCoords".equals(itemname))
      cbitem.setState(!currView.getUseWorldCoords());
    else if ("sceneCoords".equals(itemname))
      cbitem.setState(currView.getUseWorldCoords());
    else if ("wireframeDisplay".equals(itemname))
      cbitem.setState(currView.getRenderMode() == ViewerCanvas.RENDER_WIREFRAME);
    else if ("shadedDisplay".equals(itemname))
      cbitem.setState(currView.getRenderMode() == ViewerCanvas.RENDER_FLAT);
    else if ("smoothDisplay".equals(itemname))
      cbitem.setState(currView.getRenderMode() == ViewerCanvas.RENDER_SMOOTH);
    else if ("transparentDisplay".equals(itemname))
      cbitem.setState(currView.getRenderMode() == ViewerCanvas.RENDER_TRANSPARENT);
    else if ("controlMesh".equals(itemname))
      cbitem.setState(currView.getMeshVisible());
    else if ("surface".equals(itemname))
      cbitem.setState(currView.getSurfaceVisible());
    else if ("skeleton".equals(itemname))
      cbitem.setState(currView.getSkeletonVisible());
    else if ("entireScene".equals(itemname))
      cbitem.setState(currView.getSceneVisible());
    else if ("autosyncDragging".equals(itemname))
      cbitem.setState(currView.getDraggingAutosync());
    else if ("showTemplate".equals(itemname))
    {
      if (currView.getTemplateImage() == null)
        item.setEnabled(false);
      else
      {
        item.setEnabled(true);
        boolean wasShown = currView.getTemplateShown();
        item.setLabel(Translate.text(wasShown ? "menu.showTemplate" :
          "menu.hideTemplate"));
      }
    }
  }

  public void updateMenus()
  {
    if (menuitemsCache == null)
      menuitemsCache = Utilities.getMenuItemList(getMenuBar());
    for(int i = 0; i < menuitemsCache.length; ++i)
      updateMenuItem(menuitemsCache[i]);
  }

  public MeshViewer getCurrentView()
  {
    return theView[currentView];
  }

  /** Create a MeshViewer which is appropriate for a particular subclass.
    E. g. a TubeViewer for a TubeEditorWindow. */

  public abstract MeshViewer createMeshViewer(ObjectInfo obj, Panel p);

  public abstract void initBaseSelectionHolder(ObjectInfo obj);

  public abstract void initWindowMenus(ObjectInfo obj);

  public EditingTool getMetaTool()
  {
    if (metaTool == null)
      metaTool = new MoveViewTool(this);

    return metaTool;
  }

  public EditingTool getAltTool()
  {
    if (altTool == null)
      altTool = new RotateViewTool(this);

    return altTool;
  }

  /** Create the toolbar panel on the left of an object editor. */

  public abstract Panel createToolbarPanel();

  /* EditingWindow methods. */

  public boolean confirmClose()
  {
    return true;
  }

  public void setHelpText(String text)
  {
    helpText.setText(text);
  }

  /** Sets the scene for all views. */
  public void setSceneForViews(Scene sc, ObjectInfo thisObject)
  {
    for(int i=0; i < theView.length; ++i)
      theView[i].setScene(sc, thisObject);
  }

  /** Sets the autosync dragging mode for all views. */
  public void setDraggingAutosyncForViews(boolean autosyncDragging)
  {
    for(int i=0; i < theView.length; ++i)
      theView[i].setDraggingAutosync(autosyncDragging);
  }

  public Frame getFrame()
  {
    return this;
  }

  public void updateImage()
  {

    if (numViewsShown == 1)
      {
	theView[currentView].updateImage();
	theView[currentView].repaint();
      }
    else
      for (int i = 0; i < numViewsShown; i++)
        {
          theView[i].updateImage();
          theView[i].repaint(10l);
        }
  }

  public void setUndoRecord(UndoRecord command)
  {
    undo = command;
    updateMenus();
  }

  /** Returns true iff it "knows" a menu command and calls it.
   If the method is overridden by subclasses of MeshEditorWindow
   the super method should be called. */   // TODO(MB) Something is bad here
  protected boolean handleMenuAction(String command)
  {
    if (callCommandMethod(command+"Command"))
      return true;
    else if (command.equals("ok"))
      commitChanges(onClose);
    else if (command.equals("cancel"))
      rollbackChanges();

    return false;
  }

  /** Tries to call a method with no parameters and void return value.
    @param methodname name of the method to call
    @return true iff the method was found and called
  */
  protected boolean callCommandMethod(String methodname)
  {
    try
    {
      //System.out.println("MeshEditorWindow.callCommandMethod "+methodname);
      getClass().getMethod(methodname, null).invoke(this, new Object[0]);
      return true;
    }
    catch(NoSuchMethodException nme)
    {
      return false;
    }
    catch(java.lang.reflect.InvocationTargetException ite)
    {
      ite.getTargetException().printStackTrace();
      return false;
    }
    catch(Exception e)  // TODO(MB) Refine exception handling
    {
      e.printStackTrace();
      return false;
    }
  }

  /** Creates a new view, containing the same Object3D with a
   new VertexSelection. */
  protected abstract MeshEditorWindow createNewViewNewSelection();

  /** Creates a new view, containing the same Object3D and the
   same selection */
  protected MeshEditorWindow createNewViewSameSelection()
  {
    return null;
  } // TODO(MB) Implement

  public Scene getScene()
  {
    return getCurrentView().getScene();
  }

  /* KeyListener Methods. */

  public void keyPressed(KeyEvent e)
  {
    int code = e.getKeyCode();

    if (e.isShiftDown() && (code == KeyEvent.VK_LEFT || code == KeyEvent.VK_RIGHT || code == KeyEvent.VK_UP || code == KeyEvent.VK_DOWN))
      tools.keyPressed(e);
    else
    currentTool.keyPressed(e, getCurrentView());
  }

  public void keyReleased(KeyEvent e)
  {
  }

  public void keyTyped(KeyEvent e)
  {
  }

  /** Convenience method. */
  protected void object3DChangedDuringEditor()
  {
    ((MeshViewer) getCurrentView()).getObject().object.informChanged(this, ModelEvent.CHANGEDUR_OBJECTEDITOR);
  }


  public void undoCommand()
  {
    if (undo == null)
      return;
    setUndoRecord(undo.execute());

    object3DChangedDuringEditor();
  }

  /** Called if "Ok" was pressed. */
  public void commitChanges(Callback onClose)
  {
    Object3D theMesh = getCurrentView().getObject().getObject3D();
    if (oldObject.getMaterial() != null)
      {
	if (!theMesh.isClosed())
	  {
	    setCursor(Cursor.getDefaultCursor());
	    MessageDialog dlg = new MessageDialog(this, new String [] {
	      "You have modified this object so that it is no longer a closed surface.",
	      "This means that it can no longer have a material assigned to it.  Are",
	      "you sure you want to continue?"}, new String [] {Translate.text("button.ok"), Translate.text("button.cancel")});
	    if (dlg.getChoice() == 1)
	      return;
	    theMesh.setMaterial(null);
	    theMesh.setMaterialMapping(null);
	    theMesh.pushUndo();  // TODO(MB) ???
	  }
	else
	  {
	    // TODO(MB) Check if these lines can be removed
	    // theMesh.setMaterial(oldObjectInfo.getObject3D().getMaterial());
	    // theMesh.setMaterialMapping(oldObjectInfo.getObject3D().getMaterialMapping());
	  }
      }
    theMesh.unprepareAfterEditing();
    dispose();
    onClose.execute();
    theMesh.informChanged(this, ModelEvent.CHANGEDUR_OBJECTEDITORCOMMIT);
    parentWindow.updateImage();
    parentWindow.updateMenus();
    for(int i = 0; i < theView.length; ++i)
      theView[i].dispose();
  }

  /** Call if "Cancel" was pressed. */
  public void rollbackChanges()
  {
    Object3D theMesh = getCurrentView().getObject().getObject3D();
    theMesh.copyObject(oldObject);
    theMesh.informChanged(this, ModelEvent.CHANGEDUR_OBJECTEDITORROLLBACK);
    theMesh.unprepareAfterEditing();
    dispose();
    for(int i = 0; i < theView.length; ++i)
      theView[i].dispose();
  }


  /** Get the MeshViewer for this window. */

  public ViewerCanvas getView()
  {
    return getCurrentView();   // TODO(MB)  Right?
  }

  public void itemStateChanged(ItemEvent e)
  {
    Object source = e.getSource();
    CheckboxMenuItem sourceitem = null;
    MeshViewer currView = getCurrentView();
    String itemname = "";

    if (source instanceof MenuItem)
    {
      sourceitem = (CheckboxMenuItem)source;
      itemname = sourceitem.getName();
    }

    if ("localCoords".equals(itemname) || "sceneCoords".equals(itemname))
      currView.setUseWorldCoords("sceneCoords".equals(itemname));
    else if ("controlMesh".equals(itemname))
      currView.setMeshVisible(sourceitem.getState());
    else if ("surface".equals(itemname))
      currView.setSurfaceVisible(sourceitem.getState());
    else if ("skeleton".equals(itemname))
      currView.setSkeletonVisible(sourceitem.getState());
    else if ("entireScene".equals(itemname))
      currView.setSceneVisible(sourceitem.getState());
    else if ("autosyncDragging".equals(itemname))
      setDraggingAutosyncForViews(sourceitem.getState());
    else if ("wireframeDisplay".equals(itemname))
      currView.setRenderMode(ViewerCanvas.RENDER_WIREFRAME);
    else if ("shadedDisplay".equals(itemname))
      currView.setRenderMode(ViewerCanvas.RENDER_FLAT);
    else if ("smoothDisplay".equals(itemname))
      currView.setRenderMode(ViewerCanvas.RENDER_SMOOTH);
    else if ("transparentDisplay".equals(itemname))
      currView.setRenderMode(ViewerCanvas.RENDER_TRANSPARENT);
    

    updateMenus();
    updateImage();
  }

  /* Allow the user to enter new coordinates for one or more vertices. */

  public void setPointsCommand()
  {
    int i, j = 0, num = 0;
    final Mesh theMesh = (Mesh) getCurrentView().getObject().object;
    Mesh oldMesh = (Mesh) theMesh.duplicate();
    Skeleton s = theMesh.getSkeleton();
    Joint jt[] = null;
    final int selected[] = getCurrentView().getSelectionDistance();
    final CoordinateSystem coords = getCurrentView().thisObjectInScene.coords;
    final MeshVertex vert[] = theMesh.getVertices();
    final Vec3 points[] = new Vec3 [vert.length];
    final ValueField xField, yField, zField;
    ValueSlider weightSlider = null;
    Choice jointChoice = null;
    double weight = -1.0;
    int joint = -2;
    String title;
    ComponentsDialog dlg;

    for (i = 0; i < selected.length; i++)
      if (selected[i] == 0)
	{
	  num++;
	  j = i;
	  if (weight == -1.0)
	    weight = vert[i].ikWeight;
	  else if (vert[i].ikWeight != weight)
	    weight = Double.NaN;
	  if (joint == -2)
	    joint = vert[i].ikJoint;
	  else if (vert[i].ikJoint != joint)
	    joint = -3;
	}
    if (num == 0)
      return;
    if (num == 1)
      {
	Vec3 pos = vert[j].r;
        if (getCurrentView().getUseWorldCoords() && coords != null)
          pos = coords.fromLocal().times(pos);
        xField = new ValueField(pos.x, ValueField.NONE, 5);
	yField = new ValueField(pos.y, ValueField.NONE, 5);
	zField = new ValueField(pos.z, ValueField.NONE, 5);
	title = "Enter coordinates for selected vertex:";
      }
    else
      {
	xField = new ValueField(Double.NaN, ValueField.NONE, 5);
	yField = new ValueField(Double.NaN, ValueField.NONE, 5);
	zField = new ValueField(Double.NaN, ValueField.NONE, 5);
	title = "Enter coordinates for selected vertices:";
      }
    TextListener tl = new TextListener() {
      public void textValueChanged(TextEvent ev)
      {
        for (int i = 0; i < selected.length; i++)
          {
            points[i] = vert[i].r;
            if (selected[i] == 0)
              {
                if (getCurrentView().getUseWorldCoords() && coords != null)
                  coords.fromLocal().transform(points[i]);
                if (!Double.isNaN(xField.getValue()))
                  points[i].x = xField.getValue();
                if (!Double.isNaN(yField.getValue()))
                  points[i].y = yField.getValue();
                if (!Double.isNaN(zField.getValue()))
                  points[i].z = zField.getValue();
                if (getCurrentView().getUseWorldCoords() && coords != null)
                  coords.toLocal().transform(points[i]);
              }
          }
        theMesh.setVertices(points);
        for(int i = 0; i < theView.length; ++i)
          theView[i].setMesh(theMesh);
        updateImage();
        // ??? theView.repaint();
      }
    };
    xField.addTextListener(tl);
    yField.addTextListener(tl);
    zField.addTextListener(tl);
    if (s == null)
      dlg = new ComponentsDialog(this, title, new Component [] {xField, yField, zField},
        new String [] {"X", "Y", "Z"});
    else
      {
        weightSlider = new ValueSlider(0.0, 1.0, 100, weight);
        jointChoice = new Choice();
        jointChoice.add("None");
        jt = s.getJoints();
        for (i = 0; i < jt.length; i++)
          jointChoice.add(jt[i].name);
        if (joint == -3)
          jointChoice.add("");
        jointChoice.select(0);
        for (i = 0; i < jt.length; i++)
          if (jt[i].id == joint)
            jointChoice.select(i+1);
        if (joint == -3)
          jointChoice.select(jt.length+1);
        dlg = new ComponentsDialog(this, title, new Component [] {xField, yField, zField, jointChoice, weightSlider},
          new String [] {"X", "Y", "Z", "IK Bone", "IK Weight"});
      }
    if (dlg.clickedOk())
      {
        for (i = 0; i < selected.length; i++)
          if (selected[i] == 0)
            {
              if (weightSlider != null && !Double.isNaN(weightSlider.getValue()))
                vert[i].ikWeight = weightSlider.getValue();
              if (jointChoice != null)
                {
                  if (jointChoice.getSelectedIndex() == 0)
                    vert[i].ikJoint = -1;
                  else if (jointChoice.getSelectedIndex() <= jt.length)
                    vert[i].ikJoint = jt[jointChoice.getSelectedIndex()-1].id;
                }
            }
        setUndoRecord(new UndoRecord(this, false, UndoRecord.COPY_OBJECT, new Object [] {theMesh, oldMesh}));
        object3DChangedDuringEditor();
      }
    else
      {
        theMesh.copyObject((Object3D) oldMesh);
        for(i = 0; i < theView.length; ++i)
          theView[i].setMesh(theMesh);
        updateImage();
        // ??? theView.repaint();
      }
  }

  /* Allow the user to transform one or more vertices. */

  public void transformPointsCommand()
  {
    int i, j;
    Mesh theMesh = (Mesh) getCurrentView().getObject().object;
    int selected[] = getCurrentView().getSelectionDistance();
    MeshVertex vert[] = theMesh.getVertices();
    Vec3 points[] = new Vec3 [vert.length], center;
    double val[] = new double [3];
    TransformPointsDialog dlg;
    CoordinateSystem coords = getCurrentView().thisObjectInScene.coords;
    Mat4 m;

    for (i = 0; i < selected.length && selected[i] == -1; i++);
    if (i == selected.length)
      return;
    dlg = new TransformPointsDialog(this);
    val = dlg.getValues();
    if (val == null)
      return;
    setUndoRecord(new UndoRecord(this, false, UndoRecord.COPY_OBJECT, new Object [] {theMesh, theMesh.duplicate()}));
    m = Mat4.translation(val[0], val[1], val[2]);
    m = m.times(Mat4.xrotation(val[3]*Math.PI/180.0));
    m = m.times(Mat4.yrotation(val[4]*Math.PI/180.0));
    m = m.times(Mat4.zrotation(val[5]*Math.PI/180.0));
    m = m.times(Mat4.scale(val[6], val[7], val[8]));
    if (getCurrentView().getUseWorldCoords() && coords != null)
      m = coords.toLocal().times(m).times(coords.fromLocal());
    if (dlg.useSelectionCenter())
      {
	center = new Vec3();
	for (i = j = 0; i < selected.length; i++)
	  if (selected[i] == 0)
	    {
	      center.add(vert[i].r);
	      j++;
	    }
	center.scale(1.0/j);
	m = Mat4.translation(center.x, center.y, center.z).times(m).times(Mat4.translation(-center.x, -center.y, -center.z));
      }
    for (i = 0; i < selected.length; i++)
      {
	points[i] = vert[i].r;
	if (selected[i] == 0)
	  points[i] = m.times(points[i]);
      }
    theMesh.setVertices(points);
    object3DChangedDuringEditor();
  }

  /* Displace selected vertices by a random amount. */

  public void randomizeCommand()
  {
    int i;
    Mesh theMesh = (Mesh) getCurrentView().getObject().object;
    int selected[] = getCurrentView().getSelectionDistance();
    CoordinateSystem coords = getCurrentView().thisObjectInScene.coords;
    MeshVertex vert[] = theMesh.getVertices();
    Vec3 points[] = new Vec3 [vert.length];
    ValueField xfield, yfield, zfield;
    ComponentsDialog dlg;

    for (i = 0; i < selected.length && selected[i] == -1; i++);
    if (i == selected.length)
      return;
    xfield = new ValueField(0.0, ValueField.NONE);
    yfield = new ValueField(0.0, ValueField.NONE);
    zfield = new ValueField(0.0, ValueField.NONE);
    dlg = new ComponentsDialog(this, "Maximum random displacement:", new Component[]
    		{xfield, yfield, zfield}, new String[] {"X", "Y", "Z"});
    if (!dlg.clickedOk())
      return;
    setUndoRecord(new UndoRecord(this, false, UndoRecord.COPY_OBJECT, new Object [] {theMesh, theMesh.duplicate()}));
    for (i = 0; i < selected.length; i++)
      {
	points[i] = vert[i].r;
	if (selected[i] == 0)
	  {
            if (getCurrentView().getUseWorldCoords() && coords != null)
              coords.fromLocal().transform(points[i]);
	    points[i].x += (1.0-2.0*Math.random())*xfield.getValue();
	    points[i].y += (1.0-2.0*Math.random())*yfield.getValue();
	    points[i].z += (1.0-2.0*Math.random())*zfield.getValue();
            if (getCurrentView().getUseWorldCoords() && coords != null)
              coords.toLocal().transform(points[i]);
	  }
      }
    theMesh.setVertices(points);
    object3DChangedDuringEditor();
  }

  /** Center the mesh. */

  public void centerCommand()
  {
    Mesh theMesh = (Mesh) getCurrentView().getObject().object;
    MeshVertex vert[] = theMesh.getVertices();
    Skeleton skeleton = theMesh.getSkeleton();
    Joint joint[] = skeleton.getJoints();
    CoordinateSystem coords = getCurrentView().thisObjectInScene.coords;
    Vec3 center = theMesh.getBounds().getCenter(), points[] = new Vec3 [vert.length];

    setUndoRecord(new UndoRecord(this, false, UndoRecord.COPY_OBJECT, new Object [] {theMesh, theMesh.duplicate()}));
    if (getCurrentView().getUseWorldCoords() && coords != null)
    {
      coords.fromLocal().transform(center);
      coords.toLocal().transformDirection(center);
    }
    for (int i = 0; i < vert.length; i++)
      points[i] = vert[i].r.minus(center);
    theMesh.setVertices(points);
    for (int i = 0; i < joint.length; i++)
      joint[i].coords.setOrigin(joint[i].coords.getOrigin().minus(center));
    object3DChangedDuringEditor();
  }

  /* Allow the user to set the mesh tension. */

  public void setTensionCommand()
  {
    ComponentsDialog dlg;
    ValueField distanceField;
    Choice tensionChoice;

    distanceField = new ValueField((double) tensionDistance, ValueField.NONNEGATIVE+ValueField.INTEGER);
    tensionChoice = new Choice();
    tensionChoice.add("Very low");
    tensionChoice.add("Low");
    tensionChoice.add("Normal");
    tensionChoice.add("High");
    tensionChoice.add("Very High");
    tensionChoice.select(meshTension);
    dlg = new ComponentsDialog(this, "Set tension parameters:",
		new Component [] {distanceField, tensionChoice},
		new String [] {"Maximum Distance", "Tension"});
    if (!dlg.clickedOk())
      return;
    tensionDistance = (int) distanceField.getValue();
    meshTension = tensionChoice.getSelectedIndex();
  }

  public void createNewViewNewSelectionCommand()
  {
    createNewViewNewSelection().show();
  }

  public void createNewViewSameSelectionCommand()
  {
    createNewViewSameSelection().show();
  }

  public static double getMeshTension()
  {
    return tensionArray[meshTension];
  }

  public static int getTensionDistance()
  {
    return tensionDistance;
  }

  /* Allow the user to set the texture parameters for selected vertices. */

  public void setParametersCommand()
  {
    Mesh theMesh = (Mesh) getCurrentView().getObject().object;
    final MeshVertex vert[] = theMesh.getVertices();
    ObjectInfo info = getCurrentView().getObject();
    TextureParameter param[] = info.texParam;
    int i, j, k, paramIndex[] = null;
    final int selected[] = getCurrentView().getSelectionDistance();
    double value[];
    ValueField field[];
    ValueSlider slider[];
    String label[];
    PanelDialog dlg;

    for (j = 0; j < selected.length && selected[j] != 0; j++);
    if (j == selected.length)
      return;
    if (param != null)
      {
        // Find the list of per-vertex parameters.

        int num = 0;
        for (i = 0; i < param.length; i++)
          if (info.texParamPerVertex[i])
            num++;
        paramIndex = new int [num];
        for (i = 0, k = 0; k < param.length; k++)
          if (info.texParamPerVertex[k])
            paramIndex[i++] = k;
      }
    if (paramIndex == null || paramIndex.length == 0)
      {
        new MessageDialog(this, "This object's texture has no per-vertex parameters.");
        return;
      }
    value = new double [paramIndex.length];
    for (i = 0; i < paramIndex.length; i++)
      value[i] = vert[j].param[paramIndex[i]];
    for (; j < selected.length; j++)
      for (i = 0; i < paramIndex.length; i++)
        if (selected[j] == 0 && vert[j].param[paramIndex[i]] != value[i])
          value[i] = Double.NaN;

    // Define an inner class used for resetting parameters that represent texture coordinates.

    class ResetButton extends Button implements ActionListener
    {
      int xindex = -1, yindex = -1, zindex = -1;
      ValueField xfield, yfield, zfield;

      public ResetButton()
      {
	super("Reset");
	addActionListener(this);
      }

      public void addParam(int index, int type, ValueField field)
      {
	if (type == TextureParameter.X_COORDINATE)
	  {
	    xindex = index;
	    xfield = field;
	  }
	else if (type == TextureParameter.Y_COORDINATE)
	  {
	    yindex = index;
	    yfield = field;
	  }
	else if (type == TextureParameter.Z_COORDINATE)
	  {
	    zindex = index;
	    zfield = field;
	  }
      }

      public void actionPerformed(ActionEvent ev)
      {
	MessageDialog dl = new MessageDialog(MeshEditorWindow.this, new String []
	  {"Reset texture mapping coordinates for the selected vertices",
	  "to be equal to the current vertex positions?"}, new String [] {Translate.text("button.ok"), Translate.text("button.cancel")});
	if (dl.getChoice() == 1)
	  return;
	double xval = Double.NaN, yval = Double.NaN, zval = Double.NaN;
	for (int ind = 0; ind < selected.length; ind++)
	  if (selected[ind] == 0)
	    {
	      // Reset the texture coordinates for this vertex.

	      if (xindex > -1)
		{
		  vert[ind].param[xindex] = vert[ind].r.x;
		  if (Double.isNaN(xval))
		    {
		      xval = vert[ind].r.x;
		      xfield.setValue(xval);
		    }
		  else if (xval != vert[ind].r.x)
		    xfield.setValue(Double.NaN);
		}
	      if (yindex > -1)
		{
		  vert[ind].param[yindex] = vert[ind].r.y;
		  if (Double.isNaN(yval))
		    {
		      yval = vert[ind].r.y;
		      yfield.setValue(yval);
		    }
		  else if (yval != vert[ind].r.y)
		    yfield.setValue(Double.NaN);
		}
	      if (zindex > -1)
		{
		  vert[ind].param[zindex] = vert[ind].r.z;
		  if (Double.isNaN(zval))
		    {
		      zval = vert[ind].r.z;
		      zfield.setValue(zval);
		    }
		  else if (zval != vert[ind].r.z)
		    zfield.setValue(Double.NaN);
		}
	    }
      }
    }

    // Build the panel for editing the values.

    field = new ValueField [paramIndex.length];
    slider = new ValueSlider [paramIndex.length];
    Panel p = new Panel(), coordsPanel = null;
    ResetButton reset = null;
    GridBagConstraints gc = new GridBagConstraints();
    Object lastOwner = null;
    int line = 0;
    p.setLayout(new GridBagLayout());
    if (info.object.getTexture() instanceof LayeredTexture)
      {
	// This is a layered texture, so we want to group the parameters by layer.

	LayeredMapping map = (LayeredMapping) info.object.getTextureMapping();
	Texture layer[] = map.getLayers();
	for (k = 0; k < layer.length; k++)
	  {
            coordsPanel = null;
	    gc.gridx = 0;
	    gc.gridy = line++;
	    gc.anchor = gc.WEST;
	    gc.gridwidth = gc.REMAINDER;
	    gc.insets.left = 0;
	    p.add(new Label("Layer "+(k+1)+": "+layer[k].getName()), gc);
	    TextureParameter layerParam[] = map.getLayerParameters(k);
	    boolean any = false;
	    for (i = 0; i < paramIndex.length; i++)
	      {
		// Determine whether this parameter is actually part of this layer.

		int m;
		final TextureParameter pm = param[paramIndex[i]];
		for (m = 0; m < layerParam.length; m++)
		  if (layerParam[m].equals(pm))
		    break;
		if (m == layerParam.length)
		  continue;
		any = true;

		// It is, so add it.

		Component toAdd;
		if (pm.minVal == -Double.MAX_VALUE || pm.maxVal == Double.MAX_VALUE)
		  {
		    toAdd = field[i] = new ValueField(value[i], ValueField.NONE, 5);
		    field[i].setValueChecker(new ValueChecker() {
		      public boolean isValid(double v)
		      {
			return (v >= pm.minVal && v <= pm.maxVal);
		      }
		    } );
		  }
		else
		  toAdd = slider[i] = new ValueSlider(pm.minVal, pm.maxVal, 100, value[i]);
		if (pm.type == TextureParameter.NORMAL_PARAMETER)
		  {
		    gc.gridy = line++;
		    gc.gridx = 0;
		    gc.gridwidth = 1;
		    gc.insets.left = 10;
		    p.add(new Label(pm.name), gc);
		    gc.gridx = 1;
		    gc.gridwidth = gc.REMAINDER;
		    gc.insets.left = 5;
		    p.add(toAdd, gc);
		    if (coordsPanel != null)
		      coordsPanel.add(reset);
		    coordsPanel = null;
		    reset = null;
		  }
		else if (coordsPanel == null)
		  {
		    coordsPanel = new Panel();
		    gc.gridy = line++;
		    gc.gridx = 0;
		    gc.insets.left = 10;
		    p.add(new Label("Texture Mapping Coordinates:"), gc);
		    gc.gridy = line++;
		    gc.gridx = 0;
		    gc.insets.left = 20;
		    p.add(coordsPanel, gc);
		    coordsPanel.add(new Label(pm.name));
		    coordsPanel.add(toAdd);
		    reset = new ResetButton();
		    reset.addParam(paramIndex[i], pm.type, field[i]);
		  }
		else
		  {
		    coordsPanel.add(new Label(pm.name));
		    coordsPanel.add(toAdd);
		    reset.addParam(paramIndex[i], pm.type, field[i]);
		  }
	      }
	    if (coordsPanel != null)
	      coordsPanel.add(reset);
	    if (!any)
	      {
		gc.gridy = line++;
		gc.gridx = 0;
		gc.gridwidth = gc.REMAINDER;
		gc.insets.left = 10;
		p.add(new Label("(No per-vertex parameters)"), gc);
	      }
	  }
      }
    else
      {
	// This is a simple texture, so just list off all the parameters.

	gc.gridy = line++;
	gc.anchor = gc.WEST;
	gc.gridwidth = gc.REMAINDER;
	p.add(new Label("Texture: "+info.object.getTexture().getName()), gc);
	for (i = 0; i < paramIndex.length; i++)
	  {
	    final TextureParameter pm = param[paramIndex[i]];
	    Component toAdd;
	    if (pm.minVal == -Double.MAX_VALUE || pm.maxVal == Double.MAX_VALUE)
	      {
		toAdd = field[i] = new ValueField(value[i], ValueField.NONE, 5);
		field[i].setValueChecker(new ValueChecker() {
		  public boolean isValid(double v)
		  {
		    return (v >= pm.minVal && v <= pm.maxVal);
		  }
		} );
	      }
	    else
	      toAdd = slider[i] = new ValueSlider(pm.minVal, pm.maxVal, 100, value[i]);
	    if (pm.type == TextureParameter.NORMAL_PARAMETER)
	      {
		gc.gridy = line++;
		gc.gridx = 0;
		gc.gridwidth = 1;
		gc.insets.left = 10;
		p.add(new Label(pm.name), gc);
		gc.gridx = 1;
		gc.gridwidth = gc.REMAINDER;
		gc.insets.left = 5;
		p.add(toAdd, gc);
		if (coordsPanel != null)
		  coordsPanel.add(reset);
		coordsPanel = null;
		coordsPanel = null;
	      }
	    else if (coordsPanel == null)
	      {
		coordsPanel = new Panel();
		gc.gridy = line++;
		gc.gridx = 0;
		gc.insets.left = 10;
		p.add(new Label("Texture Mapping Coordinates:"), gc);
		gc.gridy = line++;
		gc.gridx = 0;
		gc.insets.left = 20;
		p.add(coordsPanel, gc);
		coordsPanel.add(new Label(pm.name));
		coordsPanel.add(toAdd);
		reset = new ResetButton();
		reset.addParam(paramIndex[i], pm.type, field[i]);
	      }
	    else
	      {
		coordsPanel.add(new Label(pm.name));
		coordsPanel.add(toAdd);
		reset.addParam(paramIndex[i], pm.type, field[i]);
	      }
	  }
	if (coordsPanel != null)
	  coordsPanel.add(reset);
      }
    dlg = new PanelDialog(this, "Texture parameters for selected points:", p);
    if (!dlg.clickedOk())
      return;
    setUndoRecord(new UndoRecord(this, false, UndoRecord.COPY_OBJECT, new Object [] {theMesh, theMesh.duplicate()}));
    for (i = 0; i < selected.length; i++)
      {
	if (selected[i] == 0)
	  for (j = 0; j < slider.length; j++)
	    {
	      if (field[j] != null && !Double.isNaN(field[j].getValue()))
		vert[i].param[paramIndex[j]] = field[j].getValue();
	      if (slider[j] != null && !Double.isNaN(slider[j].getValue()))
		vert[i].param[paramIndex[j]] = slider[j].getValue();
	    }
      }
  }

  /** Allow the user to set the template image. */

  public void setTemplateCommand()
  {
    FileDialog dlg = new FileDialog(this, "Select an image file", FileDialog.LOAD);
    dlg.show();
    if (dlg.getFile() == null)
      return;
    try
      {
        getCurrentView().setTemplateImage(new File(dlg.getDirectory(), dlg.getFile()));
      }
    catch (InterruptedException ex)
      {
        new MessageDialog(this, "An error occurred while loading the image file.");
      }
    getCurrentView().setShowTemplate(true);
    updateMenus();
    updateImage();
  }

  /* Delete the select joint from the skeleton. */

  public void deleteJointCommand()
  {
    Mesh theMesh = (Mesh) getCurrentView().getObject().object;
    Skeleton s = theMesh.getSkeleton();

    if (s == null)
      return;
    Joint j = s.getJoint(getCurrentView().getSelectedJoint());
    if (j == null)
      return;
    MessageDialog dlg = new MessageDialog(this, "Delete the bone '"+j.name+
      (j.children.length == 0 ? "'?" : "' and all of its children?"),
      new String [] {"Yes", "No"});
    if (dlg.getChoice() == 1)
      return;
    setUndoRecord(new UndoRecord(this, false, UndoRecord.COPY_OBJECT, new Object [] {theMesh, theMesh.duplicate()}));
    s.deleteJoint(getCurrentView().getSelectedJoint());
    getCurrentView().setSelectedJoint(j.parent == null ? -1 : j.parent.id);
    object3DChangedDuringEditor();  // TODO(MB) ?
    updateMenus();
  }

  /* Allow the user to edit the selected joint. */

  public void editJointCommand()
  {
    Mesh theMesh = (Mesh) getCurrentView().getObject().object;
    Skeleton s = theMesh.getSkeleton();

    if (s == null)
      return;
    Joint j = s.getJoint(getCurrentView().getSelectedJoint());
    if (j == null)
      return;
    new JointEditorDialog(this, getCurrentView(), j.id);
    updateImage();
    updateMenus();
  }

  /* Present a window for binding the selected vertices to the skeleton. */

  public void bindSkeletonCommand()
  {
    Mesh theMesh = (Mesh) getCurrentView().getObject().object;
    Skeleton s = theMesh.getSkeleton();
    if (s == null)
      return;

    // Find the selected vertices.

    int i, j, selected[] = getCurrentView().getSelectionDistance();
    for (j = 0; j < selected.length && selected[j] != 0; j++);
    if (j == selected.length)
      return;

    // Prompt the user.

    ValueSlider blendSlider = new ValueSlider(0.0, 1.0, 100, 0.5);
    ComponentsDialog dlg = new ComponentsDialog(this, "Bind selected points to skeleton?",
      new Component [] {blendSlider}, new String [] {"IK Weight Blending"});
    if (!dlg.clickedOk())
      return;
    setUndoRecord(new UndoRecord(this, false, UndoRecord.COPY_OBJECT, new Object [] {theMesh, theMesh.duplicate()}));
    double blend = blendSlider.getValue();

    // Find the position and axis vectors for each joint.

    Joint joint[] = s.getJoints();
    Vec3 pos[] = new Vec3 [joint.length], axis[] = new Vec3 [joint.length];
    for (i = 0; i < joint.length; i++)
      {
	pos[i] = joint[i].coords.getOrigin();
	if (joint[i].parent == null || joint[i].length.pos == 0.0)
	  continue;
	axis[i] = joint[i].coords.getZDirection();
	axis[i] = axis[i].times(1.0/axis[i].length());
      }

    // Loop over vertices and decide which joint to bind each one to.

    MeshVertex vert[] = theMesh.getVertices();
    double dist[] = new double [joint.length];
    for (i = 0; i < selected.length; i++)
      {
	if (selected[i] != 0)
	  continue;
	int nearest = -1;

	// Find which bone it is nearest to.

	for (j = 0; j < joint.length; j++)
	  {
	    dist[j] = distToBone(vert[i].r, joint[j], axis[j]);
	    if (nearest == -1 || dist[j] < dist[nearest])
	      nearest = j;
	  }
	if (nearest == -1)
	  continue;

	// Find the secondary bone.

	int second = -1;
	if (joint[nearest].parent != null)
	  second = s.findJointIndex(joint[nearest].parent.id);
	for (j = 0; j < joint[nearest].children.length; j++)
	  {
	    int k = s.findJointIndex(joint[nearest].children[j].id);
	    if (k != -1 && (second == -1 || dist[k] < dist[second]))
	      second = k;
	  }

	// Select the binding parameters.

	if (second == -1)
	  {
	    vert[i].ikJoint = joint[nearest].id;
	    vert[i].ikWeight = 1.0;
	  }
	else if (joint[nearest].parent != null && joint[second].id == joint[nearest].parent.id)
	  {
	    vert[i].ikJoint = joint[nearest].id;
	    double ratio = dist[nearest]/dist[second];
	    if (ratio <= 1.0-blend)
	      vert[i].ikWeight = 1.0;
	    else
	      vert[i].ikWeight = 0.5+0.5*(1.0-ratio)/blend;
	  }
	else
	  {
	    double ratio = dist[nearest]/dist[second];
	    if (ratio <= 1.0-blend)
	      {
		vert[i].ikJoint = joint[nearest].id;
		vert[i].ikWeight = 1.0;
	      }
	    else
	      {
		vert[i].ikJoint = joint[second].id;
		vert[i].ikWeight = 0.5-0.5*(1.0-ratio)/blend;
	      }
	  }
	vert[i].ikWeight = 0.001*Math.round(vert[i].ikWeight*1000.0);
      }
    object3DChangedDuringEditor();
  }

  /* Calculate the distance between a vertex and a bone. */

  private double distToBone(Vec3 v, Joint j, Vec3 axis)
  {
    Vec3 end = j.coords.getOrigin();
    if (axis == null)
      return end.distance(v);
    Vec3 base = j.parent.coords.getOrigin();
    Vec3 diff = v.minus(base);
    double dot = diff.dot(axis);
    if (dot < 0.0)
      return base.distance(v);
    if (dot > j.length.pos)
      return end.distance(v);
    diff.subtract(axis.times(dot));
    return diff.length();
  }

  /** Set the grid options for the current window. */

  public void setGridCommand()
  {
    ComponentsDialog dlg;
    ValueField spaceField, divField;
    Checkbox showBox, snapBox;

    spaceField = new ValueField(getCurrentView().gridSpacing, ValueField.POSITIVE);
    divField = new ValueField(getCurrentView().gridSubdivisions, ValueField.POSITIVE+ValueField.INTEGER);
    showBox = new Checkbox("Show Grid", getCurrentView().showGrid);
    snapBox = new Checkbox("Snap To Grid", getCurrentView().snapToGrid);
    dlg = new ComponentsDialog(this, "Select grid options:",
		new Component [] {spaceField, divField, showBox, snapBox},
		new String [] {"Grid Spacing", "Snap-To Subdivisions", "", ""});
    if (!dlg.clickedOk())
      return;
    for(int i = 0; i < theView.length; ++i)
    {
      theView[i].gridSpacing = spaceField.getValue();
      theView[i].gridSubdivisions = (int) divField.getValue();
      theView[i].showGrid = showBox.getState();
      theView[i].snapToGrid = snapBox.getState();
      theView[i].setGrid(spaceField.getValue(), (int) divField.getValue(), showBox.getState(), snapBox.getState());
    }
    updateImage();
  }
  
  /** Propagate the settings of the currentviwe to all other views. */
  public void setViewSettingsForAllViewsCommand()
  {
    MeshViewer cv = getCurrentView();
    for(int i = 0; i < theView.length; ++i)
    {
      if (cv == theView[i])
        continue;
      
      theView[i].copyViewSettings(cv);
      theView[i].updateImage();
      theView[i].repaint();
    }
  }    

  /* Given a list of deltas which will be added to the selected vertices, calculate the
     corresponding deltas for the unselected vertices according to the mesh tension. */

  public abstract void adjustDeltas(Vec3 delta[]);
}