/* MoveObjectTool is an EditingTool used for moving objects in a scene. */

/* Copyright (C) 1999-2002 by Peter Eastman

   This program is free software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the Free Software
   Foundation; either version 2 of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
   PARTICULAR PURPOSE.  See the GNU General Public License for more details. */

package artofillusion;

import artofillusion.math.*;
import artofillusion.object.*;
import artofillusion.ui.*;
import java.awt.*;
import java.awt.event.*;
import java.util.Vector;

public class MoveObjectTool extends EditingTool
{
  static Image icon, selectedIcon;
  Point clickPoint;
  Vec3 objectPos[];
  Vector toMove;
  ObjectInfo clickedObject;
  boolean dragged, applyToChildren = true;
  EventProcessor process;

  public MoveObjectTool(EditingWindow fr)
  {
    super(fr);
    icon = loadImage("move.gif");
    selectedIcon = loadImage("selected/move.gif");
  }

  public void activate()
  {
    super.activate();
    theWindow.setHelpText(Translate.text("moveObjectTool.helpText"));
  }

  public int whichClicks()
  {
    return OBJECT_CLICKS;
  }

  public Image getIcon()
  {
    return icon;
  }

  public Image getSelectedIcon()
  {
    return selectedIcon;
  }

  public void mousePressedOnObject(MouseEvent e, ViewerCanvas view, int obj)
  {
    Scene theScene = theWindow.getScene();
    int i, sel[];

    toMove = new Vector();
    clickedObject = theScene.getObject(obj);
    if (applyToChildren)
      sel = theScene.getSelectionWithChildren();
    else
      sel = theScene.getSelection();
    for (i = 0; i < sel.length; i++)
      toMove.addElement(theScene.getObject(sel[i]));
    objectPos = new Vec3 [toMove.size()];
    for (i = 0; i < objectPos.length; i++)
      {
        ObjectInfo info = (ObjectInfo) toMove.elementAt(i);
        objectPos[i] = info.coords.getOrigin();
      }
    clickPoint = e.getPoint();
    dragged = false;
    process = new EventProcessor();
  }

  public void mouseDragged(final MouseEvent e, final ViewerCanvas view)
  {
    // Since this can be a somewhat CPU intensive operation, we perform it on a
    // separate thread to avoid blocking the AWT even thread.

    System.out.println("Move mouseDragged");
    Callback cb = new Callback() {
      public void execute()
      {
        Scene theScene = theWindow.getScene();
        Camera cam = view.getCamera();
        Point dragPoint = e.getPoint();
        CoordinateSystem c;
        int i, dx, dy;
        Vec3 v;

        if (!dragged)
          {
            UndoRecord undo;
            theWindow.setUndoRecord(undo = new UndoRecord(theWindow, false));
            for (i = 0; i < toMove.size(); i++)
              {
                ObjectInfo info = (ObjectInfo) toMove.elementAt(i);
                c = info.coords;
                undo.addCommand(UndoRecord.COPY_COORDS, new Object [] {c, c.duplicate()});
              }
            dragged = true;
          }
        dx = dragPoint.x - clickPoint.x;
        dy = dragPoint.y - clickPoint.y;
        if (e.isShiftDown() && !e.isControlDown())
          {
            if (Math.abs(dx) > Math.abs(dy))
              dy = 0;
            else
              dx = 0;
          }
        if (e.isControlDown())
          v = cam.getCameraCoordinates().getZDirection().times(-dy*0.01);
        else
          v = cam.findDragVector(clickedObject.coords.getOrigin(), dx, dy);
        for (i = 0; i < toMove.size(); i++)
          {
            ObjectInfo info = (ObjectInfo) toMove.elementAt(i);
            c = info.coords;
            c.setOrigin(objectPos[i].plus(v));
          }
        theWindow.updateImage();
        theWindow.setHelpText(Translate.text("moveObjectTool.dragText",
          Math.round(v.x*1e5)/1e5+", "+Math.round(v.y*1e5)/1e5+", "+Math.round(v.z*1e5)/1e5));
      }
    };
    process.addEvent(cb);
  }

  public void mouseReleased(MouseEvent e, ViewerCanvas view)
  {
    System.out.println("Move mouseReleased");
    if (process != null)
      process.stopProcessing();
    process = null;
    for (int i = 0; i < toMove.size(); i++)
      {
        ObjectInfo info = (ObjectInfo) toMove.elementAt(i);
        info.object.sceneChanged(info, theWindow.getScene());
      }
    theWindow.setHelpText(Translate.text("moveObjectTool.helpText"));
    toMove = null;
    objectPos = null;
    theWindow.updateImage();
  }

  public void keyPressed(KeyEvent e, ViewerCanvas view)
  {
    Scene theScene = theWindow.getScene();
    Camera cam = view.getCamera();
    Vector objects;
    CoordinateSystem c;
    UndoRecord undo;
    int i, dx, dy, sel[];
    Vec3 v;
    int key = e.getKeyCode();

    // Pressing an arrow key is equivalent to dragging the first selected object by one pixel.

    if (key == KeyEvent.VK_UP)
      {
	dx = 0;
	dy = -1;
      }
    else if (key == KeyEvent.VK_DOWN)
      {
	dx = 0;
	dy = 1;
      }
    else if (key == KeyEvent.VK_LEFT)
      {
	dx = -1;
	dy = 0;
      }
    else if (key == KeyEvent.VK_RIGHT)
      {
	dx = 1;
	dy = 0;
      }
    else
      return;
    if (applyToChildren)
      sel = theScene.getSelectionWithChildren();
    else
      sel = theScene.getSelection();
    if (sel.length == 0)
      return;  // No objects are selected.
    if (e.isControlDown())
      v = cam.getCameraCoordinates().getZDirection().times(-dy*0.01);
    else
      v = cam.findDragVector(theScene.getObject(sel[0]).coords.getOrigin(), dx, dy);
    theWindow.setUndoRecord(undo = new UndoRecord(theWindow, false));
    toMove = new Vector();
    for (i = 0; i < sel.length; i++)
      toMove.addElement(theScene.getObject(sel[i]));
    for (i = 0; i < toMove.size(); i++)
      {
	c = ((ObjectInfo) toMove.elementAt(i)).coords;
	undo.addCommand(UndoRecord.COPY_COORDS, new Object [] {c, c.duplicate()});
	c.setOrigin(c.getOrigin().plus(v));
      }
    for (i = 0; i < toMove.size(); i++)
      {
        ObjectInfo info = (ObjectInfo) toMove.elementAt(i);
        info.object.sceneChanged(info, theWindow.getScene());
      }
    theWindow.updateImage();
  }

  /* Allow the user to set options. */

  public void iconDoubleClicked()
  {
    ComponentsDialog dlg;
    Checkbox childrenBox = new Checkbox("Apply to unselected children of selected objects", applyToChildren);

    dlg = new ComponentsDialog(theFrame, "Options for Move Tool:",
		new Component [] {childrenBox}, new String [] {""});
    if (!dlg.clickedOk())
      return;
    applyToChildren = childrenBox.getState();
  }
}