/* Copyright (C) 1999-2003 by Peter Eastman

   This program is free software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the Free Software
   Foundation; either version 2 of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but WITHOUT ANY 
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
   PARTICULAR PURPOSE.  See the GNU General Public License for more details. */

package artofillusion;

import artofillusion.material.*;
import artofillusion.object.*;
import artofillusion.ui.*;
import java.awt.*;
import java.awt.event.*;
import java.util.Vector;

/** This class implements the dialog box which is used to choose materials for objects. 
    It presents a list of all available materials from which the user can select one.
    If only one object is being editing, it also allows the user to edit the texture mapping
    for that object. */

public class ObjectMaterialDialog extends Dialog implements ActionListener, ItemListener, ListChangeListener
{
  Frame fr;
  Scene sc;
  ObjectInfo obj[];
  Object3D firstObj;
  List matList;
  Button mapButton;
  MaterialPreviewer preview;
  Material oldMaterial;
  MaterialMapping oldMapping;

  public ObjectMaterialDialog(Frame parent, Scene theScene, ObjectInfo objects[])
  {
    super(parent, "Object Material", false);
    
    fr = parent;
    sc = theScene;
    obj = objects;
    firstObj = obj[0].object;
    oldMaterial = firstObj.getMaterial();
    if (oldMaterial == null)
      oldMapping = null;
    else
      oldMapping = firstObj.getMaterialMapping().duplicate();

    // Add the various components to the dialog.
    
    GridBagConstraints gc = new GridBagConstraints();
    setLayout(new GridBagLayout());
    gc.gridwidth = 2;
    if (obj.length == 1)
      add(new Label("Choose material for '"+obj[0].name+"':"), gc);
    else
      add(new Label("Choose material for selected objects:"), gc);
    gc.gridwidth = 1;
    gc.gridx = 0;
    gc.fill = GridBagConstraints.BOTH;
    gc.weighty = 1.0;
    Panel listPanel = new Panel();
    add(listPanel, gc);
    listPanel.setLayout(new BorderLayout());
    listPanel.add(matList = new List(), BorderLayout.CENTER);
    buildList();
    matList.addItemListener(this);
    matList.addActionListener(this);
    Panel p = new Panel();
    listPanel.add(p, BorderLayout.SOUTH);
    Button b;
    p.add(b = new Button("New Material..."));
    b.setActionCommand("new");
    b.addActionListener(this);
    p.add(b = new Button("Materials..."));
    b.setActionCommand("materials");
    b.addActionListener(this);
    gc.fill = GridBagConstraints.NONE;
    gc.weighty = 0.0;
    gc.gridx = 1;
    add(preview = new MaterialPreviewer(firstObj.getTexture(), oldMaterial, 160, 160), gc);
    preview.setTexture(firstObj.getTexture(), firstObj.getTextureMapping());
    preview.setMaterialMapping(oldMapping);
    preview.render();

    // Add the buttons at the bottom.

    p = new Panel();
    gc.gridwidth = 2;
    gc.gridheight = 1;
    gc.gridx = 0;
    add(p, gc);
    p.add(mapButton = new Button("Edit Mapping"));
    mapButton.setActionCommand("map");
    mapButton.addActionListener(this);
    mapButton.setEnabled(!(oldMaterial instanceof UniformMaterial) && oldMaterial != null);
    p.add(Translate.button("ok", this));
    p.add(Translate.button("cancel", this));

    // Show the dialog.

    pack();
    setResizable(false);
    ModellingApp.centerWindow(this);
    sc.addMaterialListener(this);
    show();
  }

  public void dispose()
  {
    sc.removeMaterialListener(this);
    super.dispose();
  }
  
  void buildList()
  {
    matList.removeAll();
    matList.add("None");
    matList.select(0);
    for (int i = 0; i < sc.getNumMaterials(); i++)
      {
        matList.add(sc.getMaterial(i).getName());
        if (firstObj.getMaterial() == sc.getMaterial(i))
          matList.select(i+1);
      }
  }
  
  public void actionPerformed(ActionEvent e)
  {
    String command = e.getActionCommand();
    String finalStrings[] = new String [9];
    int i;

    if (e.getSource() == matList)
      {
        int which = matList.getSelectedIndex()-1;
        if (which > -1)
          {
            Material mat = sc.getMaterial(which);
            mat.edit(fr, sc);
            preview.render();
          }
        return;
      }
    else if (command.equals("cancel"))
      {
        firstObj.setMaterial(oldMaterial);
        firstObj.setMaterialMapping(oldMapping);
      }
    else if (command.equals("map"))
      {
        new MaterialMappingDialog(fr, firstObj);
        preview.setMaterialMapping(firstObj.getMaterialMapping());
        preview.render();
        return;
      }
    else if (command.equals("new"))
      {
        MaterialsDialog.showNewMaterialWindow(this, sc);
	return;
      }
    else if (command.equals("materials"))
      {
        sc.showMaterialsDialog(fr);
        buildList();
        preview.render();
        return;
      }
    else
      {
        for (i = 1; i < obj.length; i++)
          {
            obj[i].object.setMaterial(firstObj.getMaterial());
            obj[i].object.setMaterialMapping(firstObj.getMaterialMapping().duplicate());
          }
      }
    dispose();
  }

  public void itemStateChanged(ItemEvent e)
  {
    if (matList.getSelectedIndex() < 0)
      matList.select(0);
    if (matList.getSelectedIndex() == 0)
      {
        mapButton.setEnabled(false);
        firstObj.setMaterial(null);
        firstObj.setMaterialMapping(null);
        preview.setMaterial(null);
        preview.setMaterialMapping(null);
        preview.render();
        return;
      }
    Material mat = sc.getMaterial(matList.getSelectedIndex()-1);
    firstObj.setMaterial(mat);
    if (mat == oldMaterial)
      firstObj.setMaterialMapping(oldMapping.duplicate());
    else
      firstObj.setMaterialMapping(mat.getDefaultMapping());
    mapButton.setEnabled(!(mat instanceof UniformMaterial));
    preview.setMaterial(mat);
    preview.setMaterialMapping(firstObj.getMaterialMapping());
    preview.render();
  }
  
  /* ListChangeListener methods. */
  
  public void itemAdded(int index, Object obj)
  {
    Material mat = (Material) obj;
    
    matList.add(mat.getName(), index+1);
  }
  
  public void itemRemoved(int index, Object obj)
  {
    Material mat = (Material) obj;
    
    matList.remove(index+1);
    if (firstObj.getMaterial() == mat)
      {
        firstObj.setMaterial(null);
        firstObj.setMaterialMapping(null);
        preview.setMaterial(firstObj.getMaterial());
        preview.setMaterialMapping(firstObj.getMaterialMapping());
        preview.render();
        mapButton.setEnabled(false);
      }
  }
  
  public void itemChanged(int index, Object obj)
  {
    Material mat = (Material) obj;
    
    matList.replaceItem(mat.getName(), index+1);
  }
}