/* Copyright (C) 1999-2003 by Peter Eastman

   This program is free software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the Free Software
   Foundation; either version 2 of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but WITHOUT ANY 
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
   PARTICULAR PURPOSE.  See the GNU General Public License for more details. */

package artofillusion;

import artofillusion.animation.*;
import artofillusion.object.*;
import artofillusion.texture.*;
import artofillusion.ui.*;
import java.awt.*;
import java.awt.event.*;
import java.util.Vector;

/** This class implements the dialog box which is used to choose textures for objects. 
    It presents a list of all available textures from which the user can select one.
    If only one object is being editing, it also allows the user to edit the texture mapping
    for that object. */

public class ObjectTextureDialog extends Dialog implements ActionListener, ItemListener, ListChangeListener
{
  Frame fr;
  Scene sc;
  ObjectInfo obj[], editObj;
  List texList, layerList;
  Button mapButton, addLayerButton, deleteLayerButton, moveUpButton, moveDownButton;
  Button newTextureButton, editTexturesButton;
  Panel contentPanel, listPanel;
  MaterialPreviewer preview;
  Choice typeChoice, blendChoice;
  Checkbox paramPerVertexBox[];
  ValueField paramValueField[];
  ValueSlider paramValueSlider[];
  int fieldParamIndex[];
  ScrollPane paramsScroller;
  Panel paramsPanel;
  Label paramsLabel;
  Callback callback;
  Texture oldTexture;
  TextureMapping oldMapping;
  LayeredTexture layeredTex;
  LayeredMapping layeredMap;

  public ObjectTextureDialog(Frame parent, Scene theScene, ObjectInfo objects[])
  {
    super(parent, "Object Texture", false);
    
    fr = parent;
    sc = theScene;
    obj = objects;
    editObj = obj[0].duplicate();
    editObj.object = editObj.object.duplicate();
    oldTexture = editObj.object.getTexture();
    oldMapping = editObj.object.getTextureMapping();
    if (oldTexture instanceof LayeredTexture)
      {
	layeredMap = (LayeredMapping) oldMapping;
	layeredTex = (LayeredTexture) oldTexture;;
      }
    else
      {
	layeredTex = new LayeredTexture();
	layeredMap = (LayeredMapping) layeredTex.getDefaultMapping();
      }

    // Add the various components to the dialog.
    
    Panel p = new Panel();
    GridBagConstraints gc = new GridBagConstraints();
    p.setLayout(new GridBagLayout());
    gc.gridwidth = 4;
    if (obj.length == 1)
      p.add(new Label("Choose texture for '"+obj[0].name+"':"), gc);
    else
      p.add(new Label("Choose texture for selected objects:"), gc);
    gc.gridwidth = 1;
    typeChoice = new Choice();
    typeChoice.add("Simple Texture");
    typeChoice.add("Layered Texture");
    typeChoice.select((oldTexture instanceof LayeredTexture) ? 1 : 0);
    typeChoice.addItemListener(this);
    if (obj.length == 1)
      {
	gc.gridx = 0;
	p.add(new Label("Type "), gc);
	gc.gridx = 1;
	p.add(typeChoice, gc);
      }
    add(p, BorderLayout.NORTH);
    add(contentPanel = new Panel(), BorderLayout.CENTER);
    contentPanel.setLayout(new GridBagLayout());

    // Create the components that can be put into the content panel.

    texList = new List();
    buildList();
    texList.addItemListener(this);
    texList.addActionListener(this);
    layerList = new List() {
      public Dimension getPreferredSize()
      {
        return texList.getPreferredSize();
      }
    };
    layerList.addItemListener(this);
    for (int i = 0; i < layeredMap.getLayers().length; i++)
      layerList.add((layeredMap.getLayers())[i].getName());
    newTextureButton = new Button("New Texture...");
    editTexturesButton = new Button("Textures...");
    addLayerButton = Translate.button("add", " >>", this);
    deleteLayerButton = Translate.button("delete", this);
    moveUpButton = Translate.button("moveUp", this);
    moveDownButton = Translate.button("moveDown", this);
    newTextureButton.addActionListener(this);
    editTexturesButton.addActionListener(this);
    newTextureButton.setActionCommand("new");
    editTexturesButton.setActionCommand("textures");
    blendChoice = new Choice();
    blendChoice.add("Blend");
    blendChoice.add("Overlay");
    blendChoice.add("Overlay (bumps add)");
    blendChoice.addItemListener(this);
    listPanel = new Panel();
    listPanel.setLayout(new BorderLayout());
    listPanel.add(texList, BorderLayout.CENTER);
    listPanel.add(p = new Panel(), BorderLayout.SOUTH);
    p.add(newTextureButton);
    p.add(editTexturesButton);
   
    // Add the material previewer.
    
    if (oldTexture instanceof LayeredTexture)
      {
	preview = new MaterialPreviewer(layeredTex, editObj.object.getMaterial(), 160, 160);
	preview.setTexture(layeredTex, layeredMap);
      }
    else
      {
	preview = new MaterialPreviewer(oldTexture, editObj.object.getMaterial(), 160, 160);
	preview.setTexture(oldTexture, oldMapping);
      }
    preview.setMaterialMapping(editObj.object.getMaterialMapping());
    preview.getObject().texParamValue = editObj.texParamValue;

    // Add the buttons at the bottom.

    p = new Panel();
    Button b;

    add(p, BorderLayout.SOUTH);
    p.add(mapButton = new Button("Edit Mapping..."));
    mapButton.setActionCommand("map");
    mapButton.addActionListener(this);
    p.add(Translate.button("ok", this));
    p.add(Translate.button("cancel", this));

    // Add the parameters panel.
    
    paramsScroller = new ScrollPane() {
      public Dimension getPreferredSize()
      {
        Dimension dim = paramsPanel.getPreferredSize();
        dim.height = 80;
        return dim;
      }
    };
    paramsPanel = new Panel();
    paramsScroller.add(paramsPanel);
    buildParamList();

    // Show the dialog.

    if (oldTexture instanceof LayeredTexture)
      layoutLayered();
    else
      layoutSimple();
    pack();
    setResizable(false);
    ModellingApp.centerWindow(this);
    updateComponents();
    sc.addTextureListener(this);
    show();
  }

  public void dispose()
  {
    sc.removeTextureListener(this);
    super.dispose();
  }
  
  /** Set a callback to be executed when the window is closed. */
  
  public void setCallback(Callback cb)
  {
    callback = cb;
  }
  
  /** Layout the content panel for a simple texture. */
  
  private void layoutSimple()
  {
    GridBagConstraints gc = new GridBagConstraints();
    
    remove(contentPanel);
    contentPanel = new Panel();
    contentPanel.setLayout(new GridBagLayout());
    gc.fill = GridBagConstraints.BOTH;
    gc.weightx = gc.weighty = 1.0;
    gc.gridy = 0;
    contentPanel.add(listPanel, gc);
    gc.weightx = 0;
    contentPanel.add(preview, gc);
    gc.gridy = 1;
    gc.gridwidth = 2;
    contentPanel.add(paramsLabel = new Label("Texture Parameters"), gc);
    gc.gridy = 2;
    contentPanel.add(paramsScroller, gc);
    add(contentPanel, BorderLayout.CENTER);
  }
  
  /** Layout the content panel for a layered texture. */
  
  private void layoutLayered()
  {
    GridBagConstraints gc = new GridBagConstraints();
    
    remove(contentPanel);
    contentPanel = new Panel();
    contentPanel.setLayout(new GridBagLayout());
    gc.fill = GridBagConstraints.BOTH;
    gc.weightx = 1.0;
    gc.weighty = 1.0;
    gc.gridheight = 5;
    contentPanel.add(listPanel, gc);
    gc.fill = GridBagConstraints.HORIZONTAL;
    gc.weightx = 0.0;
    gc.gridheight = 1;
    gc.gridx = 1;
    contentPanel.add(addLayerButton, gc);
    contentPanel.add(deleteLayerButton, gc);
    contentPanel.add(moveUpButton, gc);
    contentPanel.add(moveDownButton, gc);
    gc.fill = GridBagConstraints.BOTH;
    gc.weightx = 1.0;
    gc.gridheight = 5;
    gc.gridx = 2;
    contentPanel.add(layerList, gc);
    gc.gridheight = 1;
    gc.gridx = 3;
    contentPanel.add(new Label("Blending mode:"), gc);
    contentPanel.add(blendChoice, gc);
    gc.gridwidth = 3;
    gc.gridx = 0;
    gc.gridy = 5;
    contentPanel.add(paramsLabel = new Label("Texture Parameters"), gc);
    gc.gridy = 6;
    contentPanel.add(paramsScroller, gc);
    gc.gridx = 3;
    gc.insets.left = 5;
    contentPanel.add(preview, gc);
    add(contentPanel, BorderLayout.CENTER);
  }
  
  /** Build the list of all available textures. */
  
  private void buildList()
  {
    texList.removeAll();
    for (int i = 0; i < sc.getNumTextures(); i++)
      {
	texList.add((sc.getTexture(i)).getName());
	if (editObj.object.getTexture() == sc.getTexture(i))
	  texList.select(i);
      }
  }
  
  /** Build the list of texture parameters. */
  
  private void buildParamList()
  {
    // Find a list of all parameters, both for the entire texture and for the selected layer.
    
    TextureParameter params[];
    if (editObj.object.getTexture() instanceof LayeredTexture)
      {
	int index = layerList.getSelectedIndex();
	if (index == -1)
	  params = new TextureParameter [0];
	else
	  params = ((LayeredMapping) editObj.object.getTextureMapping()).getLayerParameters(index);
      }
    else
      params = editObj.texParam;
    paramPerVertexBox = new Checkbox [params.length];
    paramValueField = new ValueField [params.length];
    paramValueSlider = new ValueSlider [params.length];
    fieldParamIndex = new int [params.length];
    paramsPanel.removeAll();
    paramsPanel.setLayout(new GridBagLayout());
    GridBagConstraints gc = new GridBagConstraints();
    gc.anchor = gc.NORTHWEST;
    gc.fill = gc.HORIZONTAL;
    for (int i = 0; i < params.length; i++)
      {
	if (params[i].type != TextureParameter.NORMAL_PARAMETER)
	  continue;
	gc.gridy = i;
	gc.insets.right = 10;
        gc.weightx = 0.0;
	paramsPanel.add(new Label(params[i].name), gc);
	double val = params[i].defaultVal;
	boolean perVertex = false;
	int j;
	for (j = 0; j < editObj.texParam.length; j++)
	  if (params[i].equals(editObj.texParam[j]))
	    {
	      val = editObj.texParamValue[j];
	      perVertex = editObj.texParamPerVertex[j];
	      break;
	    }
	final int whichParam = j, whichField = i;
	fieldParamIndex[i] = j;
        gc.weightx = 0.0;
	if (Double.isInfinite(params[i].minVal) || Double.isInfinite(params[i].maxVal))
	  {
	    final TextureParameter p = params[i];
	    paramValueField[i] = new ValueField(val, ValueField.NONE, 5);
	    paramValueField[i].setEnabled(!perVertex);
	    paramValueField[i].setValueChecker(new ValueChecker() {
	      public boolean isValid(double v)
	      {
		return (v >= p.minVal && v <= p.maxVal);
	      }
	    } );
	    paramValueField[i].addTextListener(new TextListener() {
	      public void textValueChanged(TextEvent ev)
	      {
		if (whichParam >= editObj.texParamValue.length)
		  return;
		editObj.texParamValue[whichParam] = preview.getObject().texParamValue[whichParam] = paramValueField[whichField].getValue();
		preview.render();
	      }
	    } );
	    paramsPanel.add(paramValueField[i], gc);
	  }
	else
	  {
	    paramValueSlider[i] = new ValueSlider(params[i].minVal, params[i].maxVal, 100, val);
	    paramValueSlider[i].setEnabled(!perVertex);
	    paramValueSlider[i].addAdjustmentListener(new AdjustmentListener() {
	      public void adjustmentValueChanged(AdjustmentEvent ev)
	      {
		if (whichParam >= editObj.texParamValue.length)
		  return;
		editObj.texParamValue[whichParam] = preview.getObject().texParamValue[whichParam] = paramValueSlider[whichField].getValue();
		preview.render();
	      }
	    } );
	    paramsPanel.add(paramValueSlider[i], gc);
	  }
	gc.insets.left = 10;
        gc.weightx = 0.0;
	paramsPanel.add(paramPerVertexBox[i] = new Checkbox("Per Vertex", perVertex), gc);
	paramPerVertexBox[i].addItemListener(this);
      }
    paramsPanel.invalidate();
    paramsScroller.invalidate();
    paramsPanel.validate();
    paramsScroller.validate();
  }

  public void actionPerformed(ActionEvent e)
  {
    String command = e.getActionCommand();
    boolean layered = (typeChoice.getSelectedIndex() == 1);
    int index = layerList.getSelectedIndex();
    Texture tex = null;
    int i;

    if (e.getSource() == texList)
      {
	tex = sc.getTexture(texList.getSelectedIndex());
        tex.edit(fr, sc);
        preview.render();
        return;
      }
    else if (command.equals("map"))
      {
	new TextureMappingDialog(fr, editObj, index);
	editObj.setTexture(editObj.object.getTexture(), editObj.object.getTextureMapping());
	preview.setTexture(editObj.object.getTexture(), editObj.object.getTextureMapping());
	preview.render();
	return;
      }
    else if (command.equals("new"))
      {
        TexturesDialog.showNewTextureWindow(this, sc);
	return;
      }
    else if (command.equals("textures"))
      {
	sc.showTexturesDialog(fr);
	buildList();
	preview.render();
	return;
      }
    else if (command.equals("add"))
      {
        preview.cancelRendering();
	tex = sc.getTexture(texList.getSelectedIndex());
	layeredMap.addLayer(tex);
	layerList.add(tex.getName(), 0);
        layerList.select(0);
	resetParameters();
	updateComponents();
	return;
      }
    else if (command.equals("delete"))
      {
        preview.cancelRendering();
	layeredMap.deleteLayer(index);
	layerList.remove(index);
	resetParameters();
	updateComponents();
	return;
      }
    else if (command.equals("moveUp"))
      {
        preview.cancelRendering();
	String label = layerList.getItem(index);
	layeredMap.moveLayer(index, index-1);
	layerList.remove(index);
	layerList.add(label, index-1);
	layerList.select(index-1);
	resetParameters();
	updateComponents();
	return;
      }
    else if (command.equals("moveDown"))
      {
        preview.cancelRendering();
	String label = layerList.getItem(index);
	layeredMap.moveLayer(index, index+1);
	layerList.remove(index);
	layerList.addItem(label, index+1);
	layerList.select(index+1);
	resetParameters();
	updateComponents();
	return;
      }
    else if (command.equals("cancel"))
      {
	dispose();
        if (callback != null)
          callback.execute();
	return;
      }
    else
      {
	for (i = 0; i < obj.length; i++)
	  {
	    if (editObj.object.getTexture() instanceof LayeredTexture)
	      {
		LayeredMapping m = (LayeredMapping) editObj.object.getTextureMapping().duplicate();
		obj[i].setTexture(new LayeredTexture(m), m);
	      }
	    else
	      obj[i].setTexture(editObj.object.getTexture(), editObj.object.getTextureMapping().duplicate());
	    System.arraycopy(editObj.texParamValue, 0, obj[i].texParamValue, 0, editObj.texParamValue.length);
	    System.arraycopy(editObj.texParamPerVertex, 0, obj[i].texParamPerVertex, 0, editObj.texParamPerVertex.length);
	  }
        obj[0].object.copyObject(editObj.object);
      }
    if (fr instanceof LayoutWindow)
      {
        ((LayoutWindow) fr).updateImage();
        ((LayoutWindow) fr).getScore().tracksModified(false);
      }
    if (callback != null)
      callback.execute();
    dispose();
  }

  public void itemStateChanged(ItemEvent e)
  {
    Object src = e.getSource();
    boolean layered = (typeChoice.getSelectedIndex() == 1);
    boolean anyselection;
    Texture tex;
    
    for (int i = 0; i < paramPerVertexBox.length; i++)
      if (src == paramPerVertexBox[i])
	{
	  // Toggle whether the parameter is set globally or per-vertex.
	  
	  boolean state = paramPerVertexBox[i].getState();
	  int index = fieldParamIndex[i];
	  if (paramValueField[i] != null)
	    paramValueField[i].setEnabled(!state);
	  if (paramValueSlider[i] != null)
	    paramValueSlider[i].setEnabled(!state);
	  editObj.texParamPerVertex[index] = state;
	  preview.render();
	  return;
	}
    if (src == typeChoice)
      {
	// Switch between simple and layered texture modes.
	
	if (!layered)
	  {
	    texList.select(0);
	    updateComponents();
	    layoutSimple();
	    pack();
	    ModellingApp.centerWindow(this);
	    resetParameters();
	  }
	else
	  {
	    preview.cancelRendering();
	    editObj.setTexture(layeredTex, layeredMap);
	    preview.setTexture(layeredTex, layeredMap);
	    updateComponents();
	    layoutLayered();
	    pack();
	    ModellingApp.centerWindow(this);
	    resetParameters();
	    return;
	  }
      }
    if (src == blendChoice)
      {
	layeredMap.setLayerMode(layerList.getSelectedIndex(), blendChoice.getSelectedIndex());
	preview.render();
	return;
      }
    if (layered)
      anyselection = (layerList.getSelectedIndex() > -1);
    else
      anyselection = (texList.getSelectedIndex() > -1);
    if (!anyselection)
      {
	updateComponents();
	return;
      }
    if (layered)
      tex = layeredMap.getLayer(layerList.getSelectedIndex());
    else
      tex = sc.getTexture(texList.getSelectedIndex());
    if (!layered)
      {
	if (tex == oldTexture)
	  editObj.setTexture(tex, oldMapping.duplicate());
	else
	  editObj.setTexture(tex, tex.getDefaultMapping());
	preview.setTexture(tex, editObj.object.getTextureMapping());
      }
    preview.render();
    updateComponents();
  }
  
  /* Utility routine.  This should be called whenever the list of texture parameters
     might have changed.  It ensures that all parameter lists are correctly updated,
     then re-renders the preview. */
  
  private void resetParameters()
  {
    preview.cancelRendering();
    editObj.setTexture(editObj.object.getTexture(), editObj.object.getTextureMapping());
    preview.getObject().setTexture(preview.getObject().object.getTexture(), preview.getObject().object.getTextureMapping());
    preview.render();
  }

  // Update the status of various components.
  
  private void updateComponents()
  {
    boolean layered = (typeChoice.getSelectedIndex() == 1);
    boolean anyselection = (texList.getSelectedIndex() > -1);
    Texture tex;

    buildParamList();
    if (layered)
      {
	addLayerButton.setEnabled(anyselection);
	int index = layerList.getSelectedIndex();
	if (index == -1)
	  {
	    mapButton.setEnabled(false);
	    deleteLayerButton.setEnabled(false);
	    moveUpButton.setEnabled(false);
	    moveDownButton.setEnabled(false);
	    blendChoice.setEnabled(false);
	    return;
	  }
	tex = layeredMap.getLayer(index);
        mapButton.setEnabled(true);
	deleteLayerButton.setEnabled(true);
	moveUpButton.setEnabled(index > 0);
	moveDownButton.setEnabled(index < layeredMap.getLayers().length-1);
	blendChoice.setEnabled(true);
	blendChoice.select(layeredMap.getLayerMode(index));
      }
    else if (anyselection)
      {
	mapButton.setEnabled(true);
	tex = sc.getTexture(texList.getSelectedIndex());
      }
    else
      mapButton.setEnabled(false);
  }
  
  /* ListChangeListener methods. */
  
  public void itemAdded(int index, Object obj)
  {
    Texture tex = (Texture) obj;
    
    texList.add(tex.getName(), index);
  }
  
  public void itemRemoved(int index, Object obj)
  {
    Texture tex = (Texture) obj;
    
    texList.remove(index);
    if (editObj.object.getTextureMapping() instanceof LayeredMapping)
      {
	Texture layers[] = layeredMap.getLayers();
	for (int i = layers.length-1; i >= 0; i--)
	  if (layers[i] == tex)
	    {
	      layeredMap.deleteLayer(i);
	      layerList.remove(i);
	    }
	preview.render();
	updateComponents();
      }
    else if (editObj.object.getTexture() == tex)
      {
	editObj.setTexture(sc.getDefaultTexture(), sc.getDefaultTexture().getDefaultMapping());
	preview.setTexture(editObj.object.getTexture(), editObj.object.getTextureMapping());
	preview.render();
	updateComponents();
      }
  }
  
  public void itemChanged(int index, Object obj)
  {
    Texture tex = (Texture) obj;
    
    texList.replaceItem(tex.getName(), index);
  }
}