/* Copyright (C) 1999-2003 by Peter Eastman

   This program is free software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the Free Software
   Foundation; either version 2 of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
   PARTICULAR PURPOSE.  See the GNU General Public License for more details. */

package artofillusion;

import artofillusion.image.*;
import artofillusion.object.*;
import artofillusion.ui.*;
import java.awt.*;
import java.awt.event.*;
import java.util.Vector;

/** This class implements the dialog box in which the user can select a renderer, and 
    specify options on how a scene should be rendered. */

public class RenderSetupDialog extends Dialog implements ActionListener, ItemListener
{
  Renderer renderers[];
  ObjectInfo cameras[];
  Scene theScene;
  Choice rendChoice, camChoice;
  Checkbox singleBox, movieBox;
  CheckboxGroup movieGroup;
  ValueField widthField, heightField, startField, endField, fpsField, subimagesField;
  Panel configPanel;

  static Renderer currentRenderer;
  static int currentCamera = 0, width = 400, height = 300;
  static int fps = 30, subimages = 1;
  static double startTime = 0.0, endTime = 1.0;
  static boolean movie;

  public RenderSetupDialog(Frame parent, Scene theScene)
  {
    this(parent, theScene, false);
    }
  
  public RenderSetupDialog(Frame parent, Scene theScene, boolean immediate)
  {
    super(parent, "Rendering Options", true);
    this.theScene = theScene;
    renderers = ModellingApp.getRenderers();

    // Find all the cameras in the scene.

    ObjectInfo obj;
    int i, count;

    for (i = 0, count = 0; i < theScene.getNumObjects(); i++)
      {
	obj = theScene.getObject(i);
	if (obj.object instanceof SceneCamera)
	  count++;
      }
    if (count == 0)
      {
	new MessageDialog((Frame) getParent(), "A scene must contain a camera to be rendered.");
	return;
      }
    if (count <= currentCamera)
      currentCamera = 0;
    cameras = new ObjectInfo [count];
    for (i = 0, count = 0; i < theScene.getNumObjects(); i++)
      {
	obj = theScene.getObject(i);
	if (obj.object instanceof SceneCamera)
	  cameras[count++] = obj;
      }
      
    if (immediate && currentRenderer != null)
    {
      startRendering();
      return;
    }

    if (currentRenderer == null)
      currentRenderer = ModellingApp.getPreferences().getDefaultRenderer();

    layoutDialog();
    pack();
    setResizable(false);
    ModellingApp.centerWindow(this);
    widthField.requestFocus();
    show();
  }

  void layoutDialog()
  {
    int i;
    Button b;
    Panel p = new Panel();
    ObjectInfo obj;
    GridBagConstraints c = new GridBagConstraints();

    // Create the first panel, which general options.

    p.setLayout(new GridBagLayout());
    c.gridx = 0;
    c.anchor = GridBagConstraints.EAST;
    c.fill = GridBagConstraints.HORIZONTAL;
    c.insets = new Insets(0, 0, 0, 5);
    p.add(new Label(Translate.text("Width")+":"), c);
    p.add(new Label(Translate.text("Height")+":"), c);
    p.add(new Label("Render:"), c);
    p.add(new Label(Translate.text("StartTime")+":"), c);
    p.add(new Label(Translate.text("EndTime")+":"), c);
    c.gridx = 1;
    p.add(widthField = new ValueField((double) width, ValueField.POSITIVE+ValueField.INTEGER), c);
    p.add(heightField = new ValueField((double) height, ValueField.POSITIVE+ValueField.INTEGER), c);
    widthField.addActionListener(this);
    heightField.addActionListener(this);
    movieGroup = new CheckboxGroup();
    p.add(singleBox = new Checkbox("Single Image", !movie, movieGroup), c);
    singleBox.addItemListener(this);
    p.add(startField = new ValueField(startTime, ValueField.NONE), c);
    p.add(endField = new ValueField(endTime, ValueField.NONE), c);
    c.gridx = 2;
    p.add(new Label(Translate.text("Renderer")+":"), c);
    p.add(new Label(Translate.text("Camera")+":"), c);
    p.add(movieBox = new Checkbox("Movie", movie, movieGroup), c);
    movieBox.addItemListener(this);
    p.add(new Label(Translate.text("FramesPerSec")+":"), c);
    p.add(new Label(Translate.text("ImagesPerFrame")+":"), c);
    c.gridx = 3;
    p.add(rendChoice = new Choice(), c);
    for (i = 0; i < renderers.length; i++)
      rendChoice.add(renderers[i].getName());
    rendChoice.select(currentRenderer.getName());
    rendChoice.addItemListener(this);
    p.add(camChoice = new Choice(), c);
    for (i = 0; i < cameras.length; i++)
      camChoice.add(cameras[i].name);
    camChoice.select(currentCamera);
    p.add(new Label(), c);
    p.add(fpsField = new ValueField(fps, ValueField.POSITIVE+ValueField.INTEGER), c);
    p.add(subimagesField = new ValueField(subimages, ValueField.POSITIVE+ValueField.INTEGER), c);
    enableMovieComponents();
    add(p, "North");

    // Add the panel containing renderer-specific options.

    add(currentRenderer.getConfigPanel(), "Center");

    // Add the buttons at the bottom.

    p = new Panel();
    p.add(Translate.button("render", this));
    p.add(Translate.button("cancel", this));
    add(p, "South");
    enableMovieComponents();
  }

  public void actionPerformed(ActionEvent e)
  {
    String command = e.getActionCommand();
    int i;
    double t;

    if (command.equals("cancel"))
      {
	dispose();
        remove(currentRenderer.getConfigPanel());
      }
    else
      {
	width = (int) widthField.getValue();
	height = (int) heightField.getValue();
	movie = movieBox.getState();
	startTime = startField.getValue();
	endTime = endField.getValue();
	fps = (int) fpsField.getValue();
	subimages = (int) subimagesField.getValue();
	currentCamera = camChoice.getSelectedIndex();
	if (currentRenderer.recordConfiguration())
	  {
	    Camera cam = new Camera();
	    SceneCamera sc = (SceneCamera) cameras[currentCamera].object;
	    cam.setCameraCoordinates(cameras[currentCamera].coords.duplicate());
	    cam.setSize(width, height);
	    cam.setDistToScreen((height/200.0)/Math.tan(sc.getFieldOfView()*Math.PI/360.0));
	    if (movie)
	      {
		ImageSaver saver = new ImageSaver((Frame) getParent(), width, height, fps);
		if (!saver.clickedOk())
		  return;
		dispose();
                remove(currentRenderer.getConfigPanel());
		new RenderingDialog((Frame) getParent(), currentRenderer, theScene, 
		  cam, cameras[currentCamera], startTime, endTime, fps, subimages, saver);
	      }
	    else
	      {
		dispose();
                remove(currentRenderer.getConfigPanel());
		new RenderingDialog((Frame) getParent(), currentRenderer, theScene, 
		  cam, cameras[currentCamera]);
	      }
	  }
      }
  }
 
  /** Begins rendering, assumes that the configuration is OK. */
  public void startRendering()   // TODO(MB) Remove method
  {
    Camera cam = new Camera();
    System.out.println("cameras "+cameras);
    System.out.println("cameras[currentCamera] "+cameras[currentCamera]);
    SceneCamera sc = (SceneCamera) cameras[currentCamera].object;
    cam.setCameraCoordinates(cameras[currentCamera].coords.duplicate());
    cam.setSize(width, height);
    cam.setDistToScreen((height/200.0)/Math.tan(sc.getFieldOfView()*Math.PI/360.0));
    if (movie)
    {
	ImageSaver saver = new ImageSaver((Frame) getParent(), width, height, fps);
	if (!saver.clickedOk())
	  return;
	dispose();
	new RenderingDialog((Frame) getParent(), currentRenderer, theScene,
	  cam, cameras[currentCamera], startTime, endTime, fps, subimages, saver);
    }
    else
    {
	dispose();
	new RenderingDialog((Frame) getParent(), currentRenderer, theScene,
	  cam, cameras[currentCamera]);
    }
  }

  
  /** Render a still image based on the current settings. */
  
  public static void renderImmediately(Frame parent, Scene theScene)
  {
    // Find the camera to render from.
    
    Vector cameras = new Vector();
    for (int i = 0; i < theScene.getNumObjects(); i++)
      if (theScene.getObject(i).object instanceof SceneCamera)
        cameras.addElement(theScene.getObject(i));
    if (cameras.size() == 0)
      {
	new MessageDialog(parent, "A scene must contain a camera to be rendered.");
	return;
      }
    if (cameras.size() <= currentCamera)
      currentCamera = 0;
    
    // Render the image.
    
    if (currentRenderer == null)
      currentRenderer = ModellingApp.getPreferences().getDefaultRenderer();
    currentRenderer.getConfigPanel();
    currentRenderer.recordConfiguration();
    Camera cam = new Camera();
    ObjectInfo cameraInfo = (ObjectInfo) cameras.elementAt(currentCamera);
    SceneCamera sc = (SceneCamera) cameraInfo.object;
    cam.setCameraCoordinates(cameraInfo.coords.duplicate());
    cam.setSize(width, height);
    cam.setDistToScreen((height/200.0)/Math.tan(sc.getFieldOfView()*Math.PI/360.0));
    new RenderingDialog(parent, currentRenderer, theScene, cam, cameraInfo);
  }
    
  public void itemStateChanged(ItemEvent e)
  {
    Object src = e.getSource();

    if (src == singleBox || src == movieBox)
      {
        enableMovieComponents();
        return;
      }
    remove(currentRenderer.getConfigPanel());
    currentRenderer = renderers[rendChoice.getSelectedIndex()];
    add(currentRenderer.getConfigPanel(), "Center");
    pack();
  }

  private void enableMovieComponents()
  {
    boolean enable = movieBox.getState();

    startField.setEnabled(enable);
    endField.setEnabled(enable);
    fpsField.setEnabled(enable);
    subimagesField.setEnabled(enable);
  }
}