/* Copyright (C) 1999-2003 by Peter Eastman

   This program is free software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the Free Software
   Foundation; either version 2 of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but WITHOUT ANY 
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
   PARTICULAR PURPOSE.  See the GNU General Public License for more details. */

package artofillusion;

import artofillusion.image.*;
import artofillusion.object.*;
import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;
import java.io.*;
import jpegencoder.*;

/** This class implements the dialog box in which the user can watch a scene being rendered. */

public class RenderingDialog extends Dialog implements ActionListener, RenderListener
{
  Canvas canvas;
  Image previewImage;
  ComplexImage theImage;
  Renderer renderer;
  Scene theScene;
  ObjectInfo sceneCamera;
  Camera theCamera;
  double start, end, originalTime;
  ImageSaver imgsaver;
  ImageAverager imgaverager;
  Button closeButton, saveButton;
  Label label1, label2;
  Frame parent;
  int w, h, fps, subimages, currentFrame, currentSubimage, totalFrames;
  long startTime;
  boolean done;

  /** Render a single frame. */

  public RenderingDialog(Frame parent, Renderer rend, Scene sc, Camera cam, ObjectInfo sceneCamera)
  {
    super(parent, true);
    this.parent = parent;
    this.sceneCamera = sceneCamera;
    renderer = rend;
    theScene = sc;
    layoutDialog(parent, cam);
    startTime = System.currentTimeMillis();
    SceneCamera scm = (SceneCamera) sceneCamera.object;
    rend.renderScene(sc, cam, this, scm);
    show();
  }
  
  /** Render an animation. */
  
  public RenderingDialog(Frame parent, Renderer rend, Scene sc, Camera cam, ObjectInfo sceneCamera,
    double start, double end, int fps, int subimages, ImageSaver imgsaver)
  {
    super(parent, true);
    this.parent = parent;
    renderer = rend;
    theScene = sc;
    theCamera = cam;
    this.sceneCamera = sceneCamera;
    this.start = start;
    this.end = end;
    this.fps = fps;
    this.subimages = subimages;
    this.imgsaver = imgsaver;
    originalTime = theScene.getTime();
    totalFrames = (int) Math.ceil((end-start)*fps);
    if (totalFrames <= 0)
      totalFrames = 1;
    if (subimages > 1)
      imgaverager = new ImageAverager(theCamera.getSize().width, theCamera.getSize().height);
    layoutDialog(parent, cam);
    startTime = System.currentTimeMillis();
    sc.setTime(start);
    theCamera.setCameraCoordinates(sceneCamera.coords.duplicate());
    SceneCamera scm = (SceneCamera) sceneCamera.object;
    theCamera.setDistToScreen((h/200.0)/Math.tan(scm.getFieldOfView()*Math.PI/360.0));
    rend.renderScene(sc, theCamera, this, scm);
    show();
  }

  private void layoutDialog(Frame parent, Camera cam)
  {
    Dimension dim = cam.getSize();
    w = dim.width;
    h = dim.height;
    setFont(parent.getFont());
    GridBagConstraints c = new GridBagConstraints();
    setLayout(new GridBagLayout());
    c.anchor = GridBagConstraints.WEST;
    c.weightx = 1;
    c.gridx = 0;
    c.fill = GridBagConstraints.HORIZONTAL;
    add(label1 = new Label(), c);
    add(label2 = new Label(), c);
    c.weightx = 0;
    c.gridx = 1;
    add(closeButton = new Button("Cancel"), c);
    add(saveButton = new Button("Save"), c);
    saveButton.setVisible(false);
    closeButton.setActionCommand("close");
    closeButton.addActionListener(this);
    saveButton.setActionCommand("save");
    saveButton.addActionListener(this);
    c.gridx = 0;
    c.gridwidth = 2;
    ScrollPane sp = new ScrollPane();
    c.weighty = 1.0;
    c.fill = c.BOTH;
    add(sp, c);
    Dimension screendim = Toolkit.getDefaultToolkit().getScreenSize();
    int sw = Math.min(w+sp.getVScrollbarWidth(), screendim.width-50);
    int sh = Math.min(h+sp.getHScrollbarHeight(), screendim.height-100);
    sp.setSize(sw, sh);
    canvas = new Canvas() {
      public void paint(Graphics g)
	{
	  if (previewImage != null)
	    g.drawImage(previewImage, 0, 0, this);
	}

      public Dimension getPreferredSize()
	{
	  return new Dimension(w, h);
	} };
    sp.add(canvas);
    pack();
    ModellingApp.centerWindow(this);
  }

  public void actionPerformed(ActionEvent e)
  {
    String command = e.getActionCommand();

    if (command.equals("close"))
      {
	done = true;
	renderer.cancelRendering(theScene);
	if (imgsaver != null)
          {
            theScene.setTime(originalTime);
            imgsaver.lastMovieImage(); // Ken: soft abort; file should be readable.
          }
	dispose();
      }
    else if (command.equals("save"))
      {
	setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
	ImageSaver saver = new ImageSaver(parent);
	if (saver.clickedOk())
	  saver.saveImage(theImage);
	setCursor(Cursor.getDefaultCursor());
      }
  }
  
  /** Update the label of how much time has elapsed. */
  
  private void updateTimeLabel()
  {
    int sec = (int) ((System.currentTimeMillis()-startTime)/1000);
    int hour = sec/3600;
    sec -= 3600*hour;
    int min = sec/60;
    sec -= 60*min;
    if (hour == 0)
      label2.setText("Elapsed Time: "+min+":"+(sec<10?"0":"")+sec);
    else
      label2.setText("Elapsed Time: "+hour+":"+min+":"+(sec<10?"0":"")+sec);
  }
  
  /** Called when more pixels are available for the current image. */
  
  public void imageUpdated(Image image)
  {
    previewImage = image;
    canvas.repaint();
    updateTimeLabel();
  }
  
  /** Called when the status changes. */
  
  public void statusChanged(String status)
  {
    StringBuffer buf = new StringBuffer();
    if (imgsaver != null)
      {
        buf.append("Frame ");
        buf.append(Integer.toString(currentFrame+1));
        buf.append(" of ");
        buf.append(Integer.toString(totalFrames));
        if (subimages > 1)
          {
            buf.append(", Subimage ");
            buf.append(Integer.toString(currentSubimage+1));
          }
        buf.append(": ");
      }
    buf.append(status);
    buf.append("...");
    label1.setText(buf.toString());
    updateTimeLabel();
  }

  
  /** Called when rendering is finished. */
  
  public void imageComplete(ComplexImage image)
  {
    SceneCamera sc = (SceneCamera) sceneCamera.object;
    if (sc.getImageFilters().length > 0)
      {
        statusChanged("Applying Filters");
        sc.applyImageFilters(image);
      }
    theImage = image;
    previewImage = theImage.getImage();
    canvas.repaint();
    if (imgsaver != null)
      nextFrame();
    if (currentFrame == totalFrames)
      {
        done = true;
        if (imgsaver != null)
          imgsaver.lastMovieImage(); // forwards to instance of MovieEncoder
        label1.setText("Done Rendering.");
        closeButton.setLabel("Close");
        saveButton.setVisible(imgsaver == null);
      }
    pack();
    updateTimeLabel();
  }
  
  /** Called when rendering is cancelled. */
  
  public void renderingCanceled()
  {
    dispose();
  }
  
  /** Save the image which has just finished rendering, and begin the next one. */
  
  private void nextFrame()
  {
    if (done)
      return;
    statusChanged("Saving");
    if (imgaverager != null)
      {
	imgaverager.addImage(theImage);
	currentSubimage++;
	if (currentSubimage == subimages)
	  {
	    imgsaver.saveImage(imgaverager.getAverageImage());
	    imgaverager.clear();
	    currentSubimage = 0;
	    currentFrame++;
	    if (currentFrame == totalFrames)
	      return;
	  }
      }
    else
      {
	imgsaver.saveImage(theImage);
	currentFrame++;
      }
    theScene.setTime(start+(currentFrame*subimages+currentSubimage)/(double) (fps*subimages));
    theCamera.setCameraCoordinates(sceneCamera.coords.duplicate());
    SceneCamera scm = (SceneCamera) sceneCamera.object;
    theCamera.setDistToScreen((h/200.0)/Math.tan(scm.getFieldOfView()*Math.PI/360.0));
    renderer.renderScene(theScene, theCamera, this, scm);
    statusChanged("Rendering");
  }
}