/* A RenderingMesh represents an object to be rendered to the screen.  It is described by
   an array of RenderingTriangles, and arrays describing the positions, normals, and
   parameter values of the vertices.  If any normal vector is null, then each triangle
   uses its own normal vector at the corresponding vertex. */

/* Copyright (C) 1999-2002 by Peter Eastman

   This program is free software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the Free Software
   Foundation; either version 2 of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
   PARTICULAR PURPOSE.  See the GNU General Public License for more details. */

package artofillusion;

import artofillusion.math.*;
import artofillusion.texture.*;

public class RenderingMesh
{
  public Vec3 vert[], norm[], faceNorm[];
  public double param[][], paramTemp[];
  public RenderingTriangle triangle[];

  /* Construct a rendering mesh. */

  public RenderingMesh(Vec3 vert[], Vec3 norm[], RenderingTriangle triangle[])
  {
    int i, j;
    Vec3 v;
    double length;

    this.vert = vert;
    this.norm = norm;
    this.triangle = triangle;
    faceNorm = new Vec3 [triangle.length];

    // Compute the normals for each face.

    for (i = 0; i < triangle.length; i++)
      {
	faceNorm[i] = vert[triangle[i].v2].minus(vert[triangle[i].v1]).cross(vert[triangle[i].v3].minus(vert[triangle[i].v1]));
	length = faceNorm[i].length();
	if (length > 0.0)
	  faceNorm[i].scale(1.0/length);
	triangle[i].setMesh(this);
      }

    // If any triangles refer to null normal vectors, replace them by the triangle normals.

    for (i = 0, j = 0; i < triangle.length; i++)
      if (norm[triangle[i].n1] == null || norm[triangle[i].n2] == null || norm[triangle[i].n3] == null)
        j++;
    if (j == 0)
      return;
    Vec3 newNorm [] = new Vec3 [norm.length+j];
    for (i = 0; i < norm.length; i++)
      newNorm[i] = norm[i];
    for (i = 0, j = norm.length; i < triangle.length; i++)
      if (norm[triangle[i].n1] == null || norm[triangle[i].n2] == null || norm[triangle[i].n3] == null)
	{
	  newNorm[j] = faceNorm[i];
	  if (norm[triangle[i].n1] == null)
	    triangle[i].n1 = j;
	  if (norm[triangle[i].n2] == null)
	    triangle[i].n2 = j;
	  if (norm[triangle[i].n3] == null)
	    triangle[i].n3 = j;
	  j++;
	}
    this.norm = newNorm;
  }

  /* Set the texture parameters for each vertex. */

  public void setParameters(double param[][], TextureMapping mapping)
  {
    this.param = param;
    paramTemp = new double [param[0].length];
    for (int i = 0; i < triangle.length; i++)
      mapping.setParameters(triangle[i], param[triangle[i].v1], param[triangle[i].v2], param[triangle[i].v3], this);
  }

  /* Set the texture parameters for all vertices to be the same. */

  public void setParameters(double p[], TextureMapping mapping)
  {
    param = new double [vert.length][];
    for (int i = 0; i < param.length; i++)
      param[i] = p;
    setParameters(param, mapping);
  }
}