/* ReshapeMeshTool is an EditingTool used for moving the vertices of TriangleMesh objects. */

/* Copyright (C) 1999-2002 by Peter Eastman

   This program is free software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the Free Software
   Foundation; either version 2 of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but WITHOUT ANY 
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
   PARTICULAR PURPOSE.  See the GNU General Public License for more details. */

package artofillusion;

import artofillusion.math.*;
import artofillusion.object.*;
import artofillusion.ui.*;
import artofillusion.object.Mesh.*;
import java.awt.*;
import java.awt.event.*;
import java.util.Vector;

public class ReshapeMeshTool extends EditingTool
{
  boolean dragged;
  static Image icon, selectedIcon;
  Point clickPoint;
  Vec3 clickPos;
  
  public ReshapeMeshTool(EditingWindow fr)
  {
    super(fr);
    icon = loadImage("movePoints.gif");
    selectedIcon = loadImage("selected/movePoints.gif");
  }

  public void activate()
  {
    super.activate();
    theWindow.setHelpText(Translate.text("reshapeMeshTool.helpText"));
  }

  public int whichClicks()
  {
    return HANDLE_CLICKS;
  }

  public Image getIcon()
  {
    return icon;
  }

  public Image getSelectedIcon()
  {
    return selectedIcon;
  }

  public void mousePressedOnHandle(MouseEvent e, ViewerCanvas view, int obj, int handle)
  {
    Mesh mesh = (Mesh) ((MeshViewer) view).getObject().object;
    MeshVertex v[] = mesh.getVertices();
    
    clickPoint = e.getPoint();
    clickPos = v[handle].r;
    dragged = false;
  }
  
  public void mouseDragged(MouseEvent e, ViewerCanvas view)
  {
    Mesh mesh = (Mesh) ((MeshViewer) view).getObject().object;
    MeshVertex vert[] = mesh.getVertices();
    Graphics g = view.getGraphics();
    Camera cam = view.getCamera();
    Point dragPoint = e.getPoint();
    Vec3 v[], drag;
    Vec2 p;
    int i, dx, dy, size = MeshViewer.HANDLE_SIZE;
    int selectDist[] = ((MeshViewer) view).getSelectionDistance();

    if (dragged)
      view.drawImage(g);
    dragged = true;
    dx = dragPoint.x - clickPoint.x;
    dy = dragPoint.y - clickPoint.y;
    if (e.isShiftDown())
      {
	if (Math.abs(dx) > Math.abs(dy))
	  dy = 0;
	else
	  dx = 0;
      }
    v = findDraggedPositions(clickPos, vert, dx, dy, (MeshViewer) view, e.isControlDown());
    g.setColor(Color.gray);
    ((MeshViewer) view).drawDraggedSelection(g, cam, v, true);
    if (e.isControlDown())
      drag = view.getCamera().getCameraCoordinates().getZDirection().times(-dy*0.01);
    else
      drag = view.getCamera().findDragVector(clickPos, dx, dy);
    theWindow.setHelpText(Translate.text("reshapeMeshTool.dragText",
        Math.round(drag.x*1e5)/1e5+", "+Math.round(drag.y*1e5)/1e5+", "+Math.round(drag.z*1e5)/1e5));
    g.dispose();
  }

  public void mouseReleased(MouseEvent e, ViewerCanvas view)
  {
    Object3D meshobj = ((MeshViewer) view).getObject().object;
    Mesh mesh = (Mesh) meshobj;
    MeshVertex vert[] = mesh.getVertices();
    Camera cam = view.getCamera();
    Point dragPoint = e.getPoint();
    int i, dx, dy;
    Vec3 v[];

    dx = dragPoint.x - clickPoint.x;
    dy = dragPoint.y - clickPoint.y;
    if (e.isShiftDown())
      {
	if (Math.abs(dx) > Math.abs(dy))
	  dy = 0;
	else
	  dx = 0;
      }
    if (dx != 0 || dy != 0)
      {
	theWindow.setUndoRecord(new UndoRecord(theWindow, false, UndoRecord.COPY_OBJECT, new Object [] {mesh, mesh.duplicate()}));
	v = findDraggedPositions(clickPos, vert, dx, dy, (MeshViewer) view, e.isControlDown());
	mesh.setVertices(v);
      }
    meshobj.informChanged(this, ModelEvent.CHANGEDUR_OBJECTEDITOR);
    theWindow.setHelpText(Translate.text("reshapeMeshTool.helpText"));
  }

  public void keyPressed(KeyEvent e, ViewerCanvas view)
  {
    Object3D meshobj = ((MeshViewer) view).getObject().object;
    Mesh mesh = (Mesh) meshobj;
    MeshVertex vert[] = mesh.getVertices();
    Camera cam = view.getCamera();
    int i, dx, dy, selectDist[] = ((MeshViewer) view).getSelectionDistance();
    int key = e.getKeyCode();
    Vec3 v[];
    
    // Pressing an arrow key is equivalent to dragging the first selected point by one pixel.
 
    if (key == KeyEvent.VK_UP)
      {
	dx = 0;
	dy = -1;
      }
    else if (key == KeyEvent.VK_DOWN)
      {
	dx = 0;
	dy = 1;
      }
    else if (key == KeyEvent.VK_LEFT)
      {
	dx = -1;
	dy = 0;
      }
    else if (key == KeyEvent.VK_RIGHT)
      {
	dx = 1;
	dy = 0;
      }
    else
      return;
    for (i = 0; i < vert.length && selectDist[i] != 0; i++);
    if (i == vert.length)
      return;
    theWindow.setUndoRecord(new UndoRecord(theWindow, false, UndoRecord.COPY_OBJECT, new Object [] {mesh, mesh.duplicate()}));
    v = findDraggedPositions(vert[i].r, vert, dx, dy, (MeshViewer) view, e.isControlDown());
    mesh.setVertices(v);
    meshobj.informChanged(this, ModelEvent.CHANGEDUR_OBJECTEDITOR);
  }
  
  public static Vec3 [] findDraggedPositions(Vec3 pos, MeshVertex vert[], int dx, int dy, MeshViewer view, boolean controlDown)
  {
    return findDraggedPositions(pos, vert, dx, dy, view, controlDown,
        view.getSelectionDistance());
  }
  
  /** This version takes a separate selectionDistance. Needed for
    CreateVertexTool.*/
  public static Vec3 [] findDraggedPositions(Vec3 pos, MeshVertex vert[],
      int dx, int dy, MeshViewer view, boolean controlDown, int[] selectDist)
  {
    int maxDistance = TriMeshEditorWindow.getTensionDistance(); // TODO(MB) ?!?
    double tension = TriMeshEditorWindow.getMeshTension();
    Vec3 drag[] = new Vec3 [maxDistance+1], v[] = new Vec3 [vert.length];
    int i;
    
    if (controlDown)
      drag[0] = view.getCamera().getCameraCoordinates().getZDirection().times(-dy*0.01);
    else
      drag[0] = view.getCamera().findDragVector(pos, dx, dy);
    for (i = 1; i <= maxDistance; i++)
      drag[i] = drag[0].times(Math.pow((maxDistance-i+1.0)/(maxDistance+1.0), tension));
    for (i = 0; i < vert.length; i++)
      {
	if (selectDist[i] > -1)
	  v[i] = vert[i].r.plus(drag[selectDist[i]]);
	else
	  v[i] = vert[i].r;
      }
    return v;
  }
}