/* Copyright (C) 1999-2000 by Peter Eastman

   This program is free software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the Free Software
   Foundation; either version 2 of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
   PARTICULAR PURPOSE.  See the GNU General Public License for more details. */

package artofillusion;

import java.util.*;


/** Represents a stack of states for undo or redo operation. An UndoManager
  normally has exactly two StateStacks, one undo stack and one redo stack.

  @author Michael Butscher
  @see UndoManager
 */

public class StateStack
{
  /** Vector of ObjectState elements. */
  protected Vector data;
  protected ModelEvent modelevent;

  public StateStack()
  {
    data = new Vector();
    modelevent = new ModelEvent();
  }

  /** Get the object i from the stack where i == 0 means top of stack.
   i == 1 means one below top and so on.

   @throws StateStackException if stack is empty or not deep enough*/
  public ObjectState peek(int i) throws StateStackException
  {
    if (i >= 0 && size() > i)
      return (ObjectState)data.get(size() - 1 - i);

    throw new StateStackException();
  }

  public ObjectState peek() throws StateStackException
  {
    return peek(0);
  }

  /** Return the object at the bottom of this stack. */
  public ObjectState peekBottom() throws StateStackException
  {
    return peek(size() - 1);
  }

  public void remove(int i) throws StateStackException
  {
    if (i >= 0 && size() > i)
    {
      data.remove(size() - 1 - i);
      return;
    }
    
    throw new StateStackException();
  }

  public int size()
  {
    return data.size();
  }

  public boolean isEmpty()
  {
    return size() == 0;
  }

  public void push(ObjectState state, Object cause)
  {
    data.add(state);
    getModelEvent().createClone(this, cause).put("stack push").processSend();
  }

  public ObjectState pop(Object cause) throws StateStackException
  {
    getModelEvent().createClone(this, cause).put("stack pop").processSend();
    ObjectState result = peek(0);
    remove(0);
    return result;
  }

  /** Remove all elements from the stack. */
  public void clear()
  {
    data = new Vector();
  }

  public ModelEvent getModelEvent()
  {
    if (modelevent == null)
      modelevent=new ModelEvent();

    return modelevent;
  }

}