/* This class implements the dialog box which is used to choose texture mappings for objects. 
   It presents a list of all mappings which can be used with the current object and material,
   and allows the user to select one. */

/* Copyright (C) 2000,2002 by Peter Eastman

   This program is free software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the Free Software
   Foundation; either version 2 of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but WITHOUT ANY 
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
   PARTICULAR PURPOSE.  See the GNU General Public License for more details. */

package artofillusion;

import artofillusion.animation.*;
import artofillusion.math.*;
import artofillusion.object.*;
import artofillusion.texture.*;
import java.awt.*;
import java.awt.event.*;
import java.lang.reflect.*;
import java.util.Vector;

public class TextureMappingDialog extends Dialog implements ActionListener, ItemListener
{
  Frame fr;
  ObjectInfo obj;
  Vector mappings;
  Choice mapChoice;
  MaterialPreviewer preview;
  TextureMapping map, oldMapping;
  Panel editingPanel;
  boolean layered;
  int layer;

  /** Create a dialog for editing the texture mapping for a particular object.  If the object
      has a layered texture, then layer is the layer number to edit.  Otherwise, layer is
      ignored. */

  public TextureMappingDialog(Frame parent, ObjectInfo obj, int layer)
  {
    super(parent, "Texture Mapping", true);
    
    fr = parent;
    this.obj = obj;
    this.layer = layer;
    map = obj.object.getTextureMapping();
    layered = (map instanceof LayeredMapping);
    if (layered)
      map = ((LayeredMapping) map).getLayerMapping(layer);
    oldMapping = map.duplicate();
    
    // Make a list of all texture mappings which can be used for this object and texture.
    
    mappings = new Vector();
    Class allMappings[] = ModellingApp.getTextureMappings();
    for (int i = 0; i < allMappings.length; i++)
      {
	try
	  {
	    Method mtd = allMappings[i].getMethod("legalMapping", new Class [] {Object3D.class, Texture.class});
	    Texture tex = layered ? ((LayeredMapping) obj.object.getTextureMapping()).getLayer(layer) 
		: obj.object.getTexture();
	    Boolean result = (Boolean) mtd.invoke(null, new Object [] {obj.object, tex});
	    if (result.booleanValue())
	      mappings.addElement(allMappings[i]);
	  }
	catch (Exception ex)
	{
	}
      }

    // Add the various components to the dialog.
    
    GridBagConstraints gc = new GridBagConstraints();
    setLayout(new GridBagLayout());
    gc.gridwidth = 2;
    Object3D previewObj = (obj.object instanceof Actor ? ((Actor) obj.object).getObject().duplicate() : obj.object.duplicate());
    add(preview = new MaterialPreviewer(obj.object.getTexture(), obj.object.getMaterial(), previewObj, 161, 160), gc);
    setPreviewMapping(map);
    gc.gridwidth = 1;
    gc.gridy = 1;
    gc.insets.right = 5;
    add(new Label("Mapping:"), gc);
    gc.insets.right = 0;
    gc.anchor = gc.WEST;
    add(mapChoice = new Choice(), gc);
    gc.anchor = gc.CENTER;
    Method mtd;
    for (int i = 0; i < mappings.size(); i++)
      {
	try
	  {
	    mtd = ((Class) mappings.elementAt(i)).getMethod("getName", null);
	    mapChoice.add((String) mtd.invoke(null, null));
	    if (mappings.elementAt(i) == map.getClass())
	      mapChoice.select(i);
	  }
	catch (IllegalAccessException ex)
	  {
	  }
	catch (IllegalArgumentException ex)
	  {
	  }
	catch (java.lang.reflect.InvocationTargetException ex)
	  {
	  }
	catch (NoSuchMethodException ex)
	  {
	  }
	catch (SecurityException ex)
	  {
	  }
      }
    mapChoice.addItemListener(this);
    gc.gridwidth = 2;
    gc.gridy = 2;
    add(editingPanel = map.getEditingPanel(obj, preview), gc);

    // Add the buttons at the bottom.

    Panel p = new Panel();
    Button b;

    gc.gridwidth = 2;
    gc.gridy = 3;
    add(p, gc);
    p.add(b = new Button("Ok"));
    b.setActionCommand("ok");
    b.addActionListener(this);
    p.add(b = new Button("Cancel"));
    b.setActionCommand("cancel");
    b.addActionListener(this);

    // Show the dialog.

    pack();
    setResizable(false);
    ModellingApp.centerWindow(this);
    show();
  }
  
  public void actionPerformed(ActionEvent e)
  {
    String command = e.getActionCommand();
    String finalStrings[] = new String [9];
    int i;

    if (command.equals("cancel"))
      setMapping(oldMapping);
    dispose();
  }

  public void itemStateChanged(ItemEvent e)
  {
    try
      {
	Class cls = (Class) mappings.elementAt(mapChoice.getSelectedIndex());
	if (cls == map.getClass())
	  return;
	Constructor con = cls.getConstructor(new Class [] {Texture.class});
	    Texture tex = layered ? ((LayeredMapping) obj.object.getTextureMapping()).getLayer(layer) 
		: obj.object.getTexture();
	setMapping((TextureMapping) con.newInstance(new Object [] {tex}));
	remove(editingPanel);
	GridBagConstraints gc = new GridBagConstraints();
	gc.gridwidth = 2;
	gc.gridy = 2;
	add(editingPanel = map.getEditingPanel(obj, preview), gc);
	pack();
	preview.render();
      }
    catch (IllegalAccessException ex)
      {
      }
    catch (InstantiationException ex)
      {
      }
    catch (InvocationTargetException ex)
      {
      }
    catch (NoSuchMethodException ex)
      {
      }
  }
  
  /** Set the mapping for the object being edited. */
  
  private void setMapping(TextureMapping newmap)
  {
    map = newmap;
    Vec2 uv[] = null;
    if (newmap instanceof UVMapping)
      {
        // Record the current texture coordinates of each vertex.
        
        Mapping2D oldmap;
        if (layered)
          oldmap = (Mapping2D) ((LayeredMapping) obj.object.getTextureMapping()).getLayerMapping(layer);
        else
          oldmap = (Mapping2D) obj.object.getTextureMapping();
        Mesh m;
        if (obj.object instanceof Actor)
          m = (Mesh) ((Actor) obj.object).getObject();
        else
          m = (Mesh) obj.object;
        uv = oldmap.findTextureCoordinates(m, obj.object.getTextureMapping().getParameters());
      }
    if (layered)
      {
        ((LayeredMapping) obj.object.getTextureMapping()).setLayerMapping(layer, newmap);
	obj.object.setTextureMapping(obj.object.getTextureMapping());
      }
    else
      {
	obj.object.setTextureMapping(newmap);
	setPreviewMapping(newmap);
      }
    if (uv != null)
      {
        // Set the texture coordinates at each vertex.
        
        if (obj.object instanceof Actor)
          ((UVMapping) newmap).setTextureCoordinates((Actor) obj.object, obj.object.getTextureMapping().getParameters(), uv);
        else
          ((UVMapping) newmap).setTextureCoordinates((Mesh) obj.object, obj.object.getTextureMapping().getParameters(), uv);
	setPreviewMapping(newmap);
      }
    preview.render();
  }
  
  /** Set the texture mapping for the preview and, if necessary, copy over
      texture coordinates. */
  
  public void setPreviewMapping(TextureMapping newmap)
  {
    preview.setTexture(obj.object.getTexture(), newmap);
    Mesh mesh;
    if (obj.object instanceof Actor && ((Actor) obj.object).getObject() instanceof Mesh)
      mesh = (Mesh) ((Actor) obj.object).getObject();
    else if (obj.object instanceof Mesh)
      mesh = (Mesh) obj.object;
    else
      return;
    Mesh previewMesh = (Mesh) preview.getObject().object;
    MeshVertex vert[] = mesh.getVertices();
    MeshVertex previewVert[] = previewMesh.getVertices();
    for (int i = 0; i < vert.length; i++)
      if (vert[i].param != null && previewVert[i].param != null && vert[i].param.length == previewVert[i].param.length)
        for (int j = 0; j < vert[i].param.length; j++)
          previewVert[i].param[j] = vert[i].param[j];
    preview.render();
  }
}