/* TexturesDialog is a dialog box for editing the list of Textures used in a scene. */

/* Copyright (C) 1999-2002 by Peter Eastman

   This program is free software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the Free Software
   Foundation; either version 2 of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but WITHOUT ANY 
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
   PARTICULAR PURPOSE.  See the GNU General Public License for more details. */

package artofillusion;

import artofillusion.image.*;
import artofillusion.texture.*;
import artofillusion.ui.*;
import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.util.Vector;
import java.util.zip.*;

public class TexturesDialog extends Dialog implements ActionListener, ItemListener, ListChangeListener
{
  Scene theScene;
  Frame parent;
  List texList;
  Button b[];

  public TexturesDialog(Frame fr, Scene sc)
  {
    super(fr, "Textures", false);

    parent = fr;
    theScene = sc;

    Panel buttons = new Panel();
    GridLayout layout = new GridLayout(6, 1);
    
    add(buttons, "West");
    buttons.setLayout(layout);
    layout.setVgap(5);
    b = new Button [6];
    buttons.add(b[0] = new Button("New..."));
    b[0].addActionListener(this);
    b[0].setActionCommand("new");
    buttons.add(b[1] = new Button("Copy..."));
    b[1].addActionListener(this);
    b[1].setActionCommand("copy");
    buttons.add(b[2] = new Button("Delete..."));
    b[2].addActionListener(this);
    b[2].setActionCommand("delete");
    buttons.add(b[3] = new Button("Import..."));
    b[3].addActionListener(this);
    b[3].setActionCommand("import");
    buttons.add(b[4] = new Button("Edit..."));
    b[4].addActionListener(this);
    b[4].setActionCommand("edit");
    buttons.add(b[5] = new Button("Done"));
    b[5].addActionListener(this);
    b[5].setActionCommand("done");
    add(texList = new List(), "Center");
    for (int i = 0; i < theScene.getNumTextures(); i++)
      texList.add(theScene.getTexture(i).getName());
    texList.addItemListener(this);
    texList.addMouseListener(new MouseAdapter() {
      public void mouseClicked(MouseEvent e)
      {
	if (e.getClickCount() == 2 && texList.getSelectedIndex() != -1)
	  actionPerformed(new ActionEvent(texList, ActionEvent.ACTION_PERFORMED, "edit"));
      } });
    sc.addTextureListener(this);
    hilightButtons();
    pack();
    setResizable(false);
    ModellingApp.centerWindow(this);
  }

  public void dispose()
  {
    theScene.removeTextureListener(this);
    super.dispose();
  }

  void hilightButtons()
  {
    boolean selection = (texList.getSelectedIndex() != -1);
    
    b[1].setEnabled(selection);
    b[2].setEnabled(selection);
    b[4].setEnabled(selection);
  }

  public void actionPerformed(ActionEvent e)
  {
    String command = e.getActionCommand();
    int i = texList.getSelectedIndex();
    Texture tex;

    if (command.equals("new"))
      {
        showNewTextureWindow(this, theScene);
      }
    else if (command.equals("copy"))
      {
	EntriesDialog dlg = new EntriesDialog(this, "Enter name for new texture:",
		new String[] {Translate.text("Name")}, new String [] {""});
	String val[] = dlg.getEntries();
	if (val == null)
	  return;
	tex = theScene.getTexture(i).duplicate();
	tex.setName(val[0]);
	theScene.addTexture(tex);
      }
    else if (command.equals("delete"))
      {
	tex = theScene.getTexture(i);
	MessageDialog dial = new MessageDialog(this, "Delete texture '"+tex.getName()+"'?", new String [] {Translate.text("button.ok"), Translate.text("button.cancel")});
	if (dial.getChoice() == 1)
	  return;
	theScene.removeTexture(i);
      }
    else if (command.equals("import"))
      {
	showImportWindow();
      }
    else if (command.equals("edit"))
      {
	setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
	tex = theScene.getTexture(i);
	tex.edit(parent, theScene);
        tex.assignNewID();
	theScene.changeTexture(i);
	setCursor(Cursor.getDefaultCursor());
      }
    else if (command.equals("done"))
      dispose();
  }

  public void itemStateChanged(ItemEvent e)
  {
    hilightButtons();
  }
  
  /* Add a texture to the list. */
  
  public void itemAdded(int index, Object obj)
  {
    Texture tex = (Texture) obj;

    texList.add(tex.getName(), index);
    texList.deselect(texList.getSelectedIndex());
    texList.select(index);
    hilightButtons();
    if (parent instanceof LayoutWindow)
      {
	((LayoutWindow) parent).setModified();
	((LayoutWindow) parent).updateImage();
      }
  }
  
  /* Remove a texture from the list. */
  
  public void itemRemoved(int index, Object obj)
  {
    texList.removeAll();
    for (int j = 0; j < theScene.getNumTextures(); j++)
      texList.add(theScene.getTexture(j).getName());
    hilightButtons();
    if (parent instanceof LayoutWindow)
      {
	((LayoutWindow) parent).setModified();
	((LayoutWindow) parent).updateImage();
      }
  }
  
  /* Change the name of a texture in the list. */
  
  public void itemChanged(int index, Object obj)
  {
    Texture tex = (Texture) obj;
    
    texList.replaceItem(tex.getName(), index);
    texList.select(index);
    if (parent instanceof LayoutWindow)
      {
	((LayoutWindow) parent).setModified();
	((LayoutWindow) parent).updateImage();
      }
  }
  
  /** Display a window in which the user can create a new texture. */
  
  public static void showNewTextureWindow(Window parent, Scene theScene)
  {
    parent.setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
    TextField nameField = new TextField();
    Choice typeChoice = new Choice();
    Class types[] = ModellingApp.getTextureTypes();
    java.lang.reflect.Method mtd;
    for (int j = 0; j < types.length; j++)
      {
        try
          {
            mtd = types[j].getMethod("getTypeName", null);
            typeChoice.add((String) mtd.invoke(null, null));
          }
        catch (Exception ex)
          {
            ex.printStackTrace();
          }
      }
    int j = 0, k = -1;
    String name = "";
    do
      {
        j++;
        name = "Untitled "+j;
      } while (theScene.getTexture(name) != null);
    nameField.setText(name);
    nameField.selectAll();
    ComponentsDialog dlg = new ComponentsDialog(parent, "Select name and type for new texture:",
            new Component [] {nameField, typeChoice}, new String[] {Translate.text("Name"), "Type"});
    if (dlg.clickedOk())
      {
        Component frame = parent;
        while (!(frame instanceof Frame))
          frame = frame.getParent();
        try
          {
            Texture tex = (Texture) types[typeChoice.getSelectedIndex()].newInstance();
            tex.setName(nameField.getText());
            theScene.addTexture(tex);
            tex.edit((Frame) frame, theScene);
          }
        catch (Exception ex)
          {
            ex.printStackTrace();
          }
      }
    parent.setCursor(Cursor.getDefaultCursor());
  }
  
  /* Ask the user to select textures to import from a scene. */
  
  private void showImportWindow()
  {
    // First prompt them to select a file.
    
    FileDialog fd = new FileDialog(parent, "Import Textures from Scene File", FileDialog.LOAD);
    fd.show();
    if (fd.getFile() == null)
      return;
    File f = new File(fd.getDirectory(), fd.getFile());
    DataInputStream in = null;
    Scene sc = null;
    try
      {
	in = new DataInputStream(new GZIPInputStream(new BufferedInputStream(new FileInputStream(f))));
	sc = new Scene(in, false);
      }
    catch (InvalidObjectException ex)
      {
	new MessageDialog(parent, new String [] {
		"An error occurred while reading the file.  This file may", 
		"have been created with a later version of Art of Illusion."});
	return;
      }
    catch (IOException ex)
      {
	new MessageDialog(parent, new String [] {"An error occurred while reading the file.", ex.getMessage()});
	return;
      }
    finally
      {
	try
	  {
	    in.close();
	  }
	catch (Exception ex)
	  {
	    new MessageDialog(parent, new String [] {"An error occurred while reading the file.", ex.getMessage()});
	    return;
	  }
      }
    if (sc.getNumTextures() == 0)
      {
	new MessageDialog(parent, "This file does not contain any textures.");
	return;
      }
    
    // Now create the main dialog.
    
    Panel mainPanel = new Panel(), p = new Panel();
    final List baseList = new List(), importList = new List();
    final Button addButton = new Button("Add >>"), removeButton = new Button("<< Remove");
    final Vector baseVec = new Vector(), importVec = new Vector();
    final MaterialPreviewer preview = new MaterialPreviewer(sc.getTexture(0), null, 160, 160);
    for (int i = 0; i < sc.getNumTextures(); i++)
      {
	baseVec.addElement(sc.getTexture(i));
	baseList.add(sc.getTexture(i).getName());
      }
    ItemListener itemListen = new ItemListener() {
      public void itemStateChanged(ItemEvent ev)
      {
	if (ev.getSource() == baseList && baseList.getSelectedIndex() > -1)
	  {
	    if (importList.getSelectedIndex() > -1)
	      importList.deselect(importList.getSelectedIndex());
	    addButton.setEnabled(true);
	    removeButton.setEnabled(false);
	    Texture tex = (Texture) baseVec.elementAt(baseList.getSelectedIndex());
	    preview.setTexture(tex, tex.getDefaultMapping());
	    preview.render();
	  }
	else if (importList.getSelectedIndex() > -1)
	  {
	    if (baseList.getSelectedIndex() > -1)
	      baseList.deselect(baseList.getSelectedIndex());
	    addButton.setEnabled(false);
	    removeButton.setEnabled(true);
	    Texture tex = (Texture) importVec.elementAt(importList.getSelectedIndex());
	    preview.setTexture(tex, tex.getDefaultMapping());
	    preview.render();
	  }
      }
    };
    baseList.addItemListener(itemListen);
    importList.addItemListener(itemListen);
    baseList.select(0);
    ActionListener actionListen = new ActionListener() {
      public void actionPerformed(ActionEvent ev)
      {
	if (ev.getSource() == addButton)
	  {
	    int which = baseList.getSelectedIndex();
	    if (which < 0)
	      return;
	    Texture tex = (Texture) baseVec.elementAt(which);
	    baseVec.removeElementAt(which);
	    baseList.remove(which);
	    importVec.addElement(tex);
	    importList.add(tex.getName());
	    addButton.setEnabled(false);
	  }
	if (ev.getSource() == removeButton)
	  {
	    int which = importList.getSelectedIndex();
	    if (which < 0)
	      return;
	    Texture tex = (Texture) importVec.elementAt(which);
	    importVec.removeElementAt(which);
	    importList.remove(which);
	    baseVec.addElement(tex);
	    baseList.add(tex.getName());
	    removeButton.setEnabled(false);
	  }
      }
    };
    addButton.addActionListener(actionListen);
    removeButton.addActionListener(actionListen);
    removeButton.setEnabled(false);
    p.setLayout(new GridBagLayout());
    GridBagConstraints gc = new GridBagConstraints();
    gc.gridx = 0;
    gc.fill = gc.HORIZONTAL;
    gc.insets.top = gc.insets.bottom = 5;
    p.add(addButton, gc);
    p.add(removeButton, gc);
    mainPanel.setLayout(new GridBagLayout());
    gc = new GridBagConstraints();
    gc.gridy = 0;
    gc.weightx = gc.weighty = 1.0;
    gc.fill = gc.BOTH;
    gc.ipadx = 20;
    mainPanel.add(baseList, gc);
    gc.weightx = 0.0;
    gc.ipadx = 0;
    mainPanel.add(p, gc);
    gc.weightx = 1.0;
    gc.ipadx = 50;
    mainPanel.add(importList, gc);
    gc.weightx = 0.0;
    gc.ipadx = 0;
    mainPanel.add(preview, gc);
    PanelDialog dlg = new PanelDialog(parent, "Select Textures to Import:", mainPanel);
    if (!dlg.clickedOk())
      return;
    
    // Copy the selected textures, along with any images they use.
    
    boolean imageUsed[] = new boolean [sc.getNumImages()];
    for (int i = 0; i < importVec.size(); i++)
      {
	Texture tex = (Texture) importVec.elementAt(i);
	for (int j = 0; j < imageUsed.length; j++)
	  if (!imageUsed[j] && tex.usesImage(sc.getImage(j)))
	    {
	      theScene.addImage(sc.getImage(j));
	      imageUsed[j] = true;
	    }
	theScene.addTexture(tex);
      }
  }
}
