/* Copyright (C) 1999-2003 by Peter Eastman

   This program is free software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the Free Software
   Foundation; either version 2 of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but WITHOUT ANY 
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
   PARTICULAR PURPOSE.  See the GNU General Public License for more details. */

package artofillusion;

import artofillusion.ui.*;
import java.awt.*;
import java.awt.event.*;

/** This class implements the dialog box which is used for the "Object Layout" and
    "Transform Object" commands.  It allows the user enter values for the position,
    orientation, and size of an object.  The initial values are passed to the constructor
    in values[].  If this argument is omitted, all of the fields will initially be blank.
    If transformLabels is true, the rows will be labelled "Move", "Rotate", and "Scale".
    If it is false, they will be labelled "Position", "Orientation", and "Size". */

public class TransformDialog extends Dialog implements ActionListener, FocusListener
{
  private double initialValues[], finalValues[];
  private String initialStrings[];
  private TextField fields[];
  private Checkbox childrenBox, objectCenterBox, selectionCenterBox;
  private CheckboxGroup centerGroup;
  private boolean clicked, ok;
  
  private static boolean children = true;
  private static boolean selectionCenter = true;

  public TransformDialog(Frame parent, String title, double values[], boolean transformLabels, boolean extraOptions)
  {
    super(parent, title, true);
    initialValues = values;
    initialStrings = new String [9];
    fields = new TextField [9];
    for (int i = 0; i < 9; i++)
      initialStrings[i] = numberToString(values[i]);
    layoutDialog(transformLabels, extraOptions);
    pack();
    setResizable(false);
    ModellingApp.centerWindow(this);
    fields[0].requestFocus();
    show();
  }
  
  public TransformDialog(Frame parent, String title, boolean transformLabels, boolean extraOptions)
  {
    this(parent, title, new double [] {Double.NaN, Double.NaN, Double.NaN, 
    		Double.NaN, Double.NaN, Double.NaN, Double.NaN, Double.NaN, Double.NaN},
    		transformLabels, extraOptions);
  }
  
  /** Get the array of values the user entered. */
  
  public double [] getValues()
  {
    return finalValues;
  }
  
  /** Determine whether the user clicked the OK button. */
  
  public boolean clickedOk()
  {
    return ok;
  }
  
  /** Determine whether the operation should apply to unselected children of selected objects. */

  public boolean applyToChildren()
  {
    return children;
  }
  
  /** Determine whether rotations should be around the center of each object, or the center
      of the entire selection. */

  public boolean useSelectionCenter()
  {
    return selectionCenter;
  }
  
  void layoutDialog(boolean transformLabels, boolean extraOptions)
  {
    int i;
    Panel p = new Panel();
    GridBagConstraints c = new GridBagConstraints();

    setLayout(new GridBagLayout());
    c.anchor = GridBagConstraints.EAST;
    c.ipadx = 5;
    c.gridx = 0;
    c.gridy = 1;
    c.weightx = 1.0;
    if (transformLabels)
      {
	add(Translate.label("Move"), c);
	c.gridy = GridBagConstraints.RELATIVE;
	add(Translate.label("Rotate"), c);
	add(Translate.label("Scale"), c);
      }
    else
      {
	add(Translate.label("Position"), c);
	c.gridy = GridBagConstraints.RELATIVE;
	add(Translate.label("Orientation"), c);
	add(Translate.label("Size"), c);
      }
    c.anchor = GridBagConstraints.CENTER;
    c.ipadx = 0;
    c.gridx = 1;
    c.gridy = 0;
    add(new Label("        X        "), c);
    c.gridx = GridBagConstraints.RELATIVE;
    add(new Label("        Y        "), c);
    add(new Label("        Z        "), c);
    c.fill = GridBagConstraints.HORIZONTAL;
    c.gridy = 1;
    add(fields[0] = new TextField(initialStrings[0]), c);
    add(fields[1] = new TextField(initialStrings[1]), c);
    add(fields[2] = new TextField(initialStrings[2]), c);
    c.gridy = 2;
    add(fields[3] = new TextField(initialStrings[3]), c);
    add(fields[4] = new TextField(initialStrings[4]), c);
    add(fields[5] = new TextField(initialStrings[5]), c);
    c.gridy = 3;
    add(fields[6] = new TextField(initialStrings[6]), c);
    add(fields[7] = new TextField(initialStrings[7]), c);
    add(fields[8] = new TextField(initialStrings[8]), c);
    c.gridx = 0;
    c.gridy = 4;
    c.gridwidth = 4;
    if (extraOptions)
      {
	add(childrenBox = new Checkbox("Apply to unselected children of selected objects", children), c);
	c.gridy = 5;
	add(p, c);
	p.add(new Label("Rotate/Scale Around"));
	p.add(p = new Panel());
	centerGroup = new CheckboxGroup();
	p.setLayout(new GridLayout(2, 1));
	p.add(objectCenterBox = new Checkbox("Individual Object Centers", !selectionCenter, centerGroup));
	p.add(selectionCenterBox = new Checkbox("Center of Selection", selectionCenter, centerGroup));
	p = new Panel();
	c.gridy = 6;
      }
    add(p, c);
    p.add(Translate.button("ok", this));
    p.add(Translate.button("cancel", this));
    MouseListener listen = new MouseAdapter() {
      public void mousePressed(MouseEvent ev)
      {
        clicked = true;
      }
    };
    for (i = 0; i < 9; i++)
      {
	fields[i].addActionListener(this);
	fields[i].addFocusListener(this);
	fields[i].addMouseListener(listen);
      }
    addWindowListener(new WindowAdapter() {
      public void windowClosing(WindowEvent e)
      {
        finalValues = initialValues;
        dispose();
      }
    } );
  }

  public void actionPerformed(ActionEvent e)
  {
    String command = e.getActionCommand();
    String finalStrings[] = new String [9];
    int i;

    if (command.equals("cancel"))
      {
	finalValues = initialValues;
      }
    else
      {
	finalValues = new double [9];
	for (i = 0; i < 9; i++)
	  {
	    finalStrings[i] = fields[i].getText();
	    try
	      {
		finalValues[i] = Double.valueOf(finalStrings[i]).doubleValue();
		if (finalStrings[i].equals(initialStrings[i]))
		  finalValues[i] = initialValues[i];
	      }
	    catch (NumberFormatException ex)
	      {
		finalValues[i] = Double.NaN;
	      }
	  }
	if (childrenBox != null)
	  children = childrenBox.getState();
	if (centerGroup != null)
	  selectionCenter = (centerGroup.getSelectedCheckbox() == selectionCenterBox);
	ok = true;
      }
    dispose();
  }

  public void focusGained(FocusEvent e)
  {
    TextField t = (TextField) e.getSource();

    if (!clicked)
      {
	t.setSelectionStart(0);
	t.setSelectionEnd(t.getText().length());
      }
    clicked = false;
  }

  public void focusLost(FocusEvent e)
  {
    TextField t = (TextField) e.getSource();

    t.setSelectionStart(0);
    t.setSelectionEnd(0);
  }
  
  /** Given a value, return the corresponding text representation. */

  private String numberToString(double value)
  {
    String text;
    if (Double.isNaN(value))
      text = "";
    else if (value == 0.0 || value == -0.0)
      text = "0.0";
    else
      {
        // Make sure at least three significant digits are visible.
        
        int digits = (int) Math.floor(Math.log(value)/Math.log(10.0));
        double scale = Math.pow(10.0, digits < 0 ? 2-digits : 3);
        text = Double.toString(Math.round(value*scale)/scale);
      }
    return text;
  }
}