/* The TriMeshEditorWindow class represents the window for editing TriangleMesh objects. */

/* Copyright (C) 1999-2002 by Peter Eastman

   This program is free software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the Free Software
   Foundation; either version 2 of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
   PARTICULAR PURPOSE.  See the GNU General Public License for more details. */

package artofillusion;

import artofillusion.animation.*;
import artofillusion.math.*;
import artofillusion.object.*;
import artofillusion.object.TriangleMesh.*;
import artofillusion.ui.*;
import java.awt.*;
import java.awt.event.*;
import java.util.Vector;

public class TriMeshEditorWindow extends MeshEditorWindow
{
  ToolPalette modes;
  Menu editMenu, meshMenu, skeletonMenu;
  MenuItem editMenuItem[], meshMenuItem[], skeletonMenuItem[];
  CheckboxMenuItem smoothItem[];
  boolean topology;

  public TriMeshEditorWindow(EditingWindow parent, String title, ObjectInfo obj, Callback onClose, boolean allowTopology)
  {
    super(parent, title, obj, onClose);
    topology = allowTopology;
    
    toolsPanel.setLayout(new BorderLayout());  //TODO(MB) Remove this toolsPanel creation
    if (topology)
      toolsPanel.add(tools = new ToolPalette(1, 10), "North");
    else
      toolsPanel.add(tools = new ToolPalette(1, 9), "North");
    tools.addTool(defaultTool = new ReshapeMeshTool(this));
    tools.addTool(new ScaleMeshTool(this));
    tools.addTool(new RotateMeshTool(this, false));
    tools.addTool(new SkewMeshTool(this));
    tools.addTool(new TaperMeshTool(this));
    tools.addTool(new ThickenMeshTool(this));
    tools.addTool(new SkeletonTool(this));
    if (topology)
      tools.addTool(new CreateVertexTool(this));
    tools.addTool(getMetaTool() /* MoveViewTool */);
    tools.addTool(getAltTool() /* RotateViewTool */);
    tools.selectTool(defaultTool);
    toolsPanel.add(modes = new ToolPalette(1, 3), "South");
    modes.addTool(new GenericTool(this, "point.gif", "selected/point.gif"));
    modes.addTool(new GenericTool(this, "edge.gif", "selected/edge.gif"));
    modes.addTool(new GenericTool(this, "face.gif", "selected/face.gif"));
    
    //TODO(MB) Remove this menu creation
    MenuStructure ms = new MenuStructure();
    ms.loadStructure("TriMesh", false);
    //menubar = new MenuBar();
    //menubar.add(ms.createMenu(this));
    menubar = ms.createMenuBar(this);
    setMenuBar(menubar);
    menuitemsCache = null;  // The menubar was rebuild so invalidate cache
    /*
    createEditMenu((TriangleMesh) obj.object);
    createMeshMenu((TriangleMesh) obj.object);
    createSkeletonMenu((TriangleMesh) obj.object);
    createViewMenu();
    */
    updateMenus();
   
    recursivelyAddListeners(this);
    pack();
    Dimension d1 = Toolkit.getDefaultToolkit().getScreenSize(), d2;
    d2 = new Dimension((d1.width*2)/3, (d1.height*2)/3);
    setSize(d2);
    setLocation((d1.width-d2.width)/2, (d1.height-d2.height)/2);
    tools.requestFocus();

    updateMenus();
  }

  /** List of names of menu items which are only allowed if {@link #topology} is true.
    Used by {@link #isMenuItemAllowed()}.*/

  private static final String[] MENUITEMS_ALLOWED_ON_TOPOLOGY=new String[]
    { "clear", "subdivideEdges", "simplify", "bevel", "optimize",
      "closeBoundary", "joinBoundaries", "invertNormals", "createFace"
    };

  /** List of names of menu items which should be CheckboxMenuItem.
    Used by {@link #isMenuItemNameCheckbox()}.*/

  private static final String[] MENUITEMS_CHECKBOXITEMS=new String[]
    { "tolerantSelection", "freehandSelection", "displayAsQuads", "none",
      "shading", "interpolating", "approximating", "detachSkeleton"
    };

  /** Tests if menu item item should be shown in the menus for
    this object editor. */
    public boolean isMenuItemAllowed(MenuItem item)
  {
    String itemname = item.getName();
    if (Utilities.findIndexEqual(MENUITEMS_ALLOWED_ON_TOPOLOGY, itemname) > -1)
      return topology;   // Some items are only allowed if topology is true
    else if ("renderPreview".equals(itemname))
      return ModellingApp.getPreferences().getObjectPreviewRenderer() != null;
    else
      return super.isMenuItemAllowed(item);
  }

  /** Tests if menu item named itemname should be a checkbox item. */
  protected boolean isMenuItemNameCheckbox(String itemname)
  {
    if (Utilities.findIndexEqual(MENUITEMS_CHECKBOXITEMS, itemname) > -1)
      return true;
    else
      return super.isMenuItemNameCheckbox(itemname);
  }

  /* EditingWindow methods. */

  public void setTool(EditingTool tool)
  {
    for(int i = 0; i < theView.length; ++i)
    {
      if (tool instanceof GenericTool)
      {
        ((TriMeshViewer) getCurrentView()).setSelectionMode(modes.getSelection());
        ((TriMeshViewer) getCurrentView()).informSelectionChanged(this,
            ModelEvent.CHANGEDUR_OBJECTEDITOR);
        updateMenus();
      }
      else
        {
          theView[i].setTool(tool);
          currentTool = tool;
        }
    }
  }

  /** Returns true if the view has at least one selected
    element. */
  public boolean isSomethingSelected()
  {
    int i;
    boolean selected[] = ((TriMeshViewer) getCurrentView()).getSelection();

    for (i = 0; i < selected.length && !selected[i]; i++);

    return i < selected.length;
  }

  /** Updates a single menu item
    (setting enabled/disabled, (un)checked ...). */
  protected void updateMenuItem(MenuItem item)
  {
    String itemname = item.getName();
    CheckboxMenuItem cbitem = null;
    if (item instanceof CheckboxMenuItem)
      cbitem = (CheckboxMenuItem)item;

    int selectionMode = ((TriMeshViewer) getCurrentView()).getSelectionMode();
    TriangleMesh obj = (TriangleMesh)getCurrentView().getObject().object;

    // Edit Menu

    if ("clear".equals(itemname))
      item.setEnabled(isSomethingSelected());
    else if ("selectBoundary".equals(itemname))
      item.setEnabled(selectionMode == TriMeshViewer.EDGE_MODE && !getCurrentView().getObject().object.isClosed());
    else if ("extendSelection".equals(itemname))
      item.setEnabled(isSomethingSelected());

    // Mesh Menu

    else if ("subdivideEdges".equals(itemname))
    {
      item.setEnabled(true);
      if (!isSomethingSelected())
        item.setEnabled(false);
      else if (selectionMode == TriMeshViewer.EDGE_MODE)
        item.setLabel(Translate.text("menu.subdivideEdges"));
      else if (selectionMode == TriMeshViewer.FACE_MODE)
        item.setLabel(Translate.text("menu.subdivideFaces"));
      else
        item.setEnabled(false);
    }
    else if ("simplify".equals(itemname))
    {
      item.setEnabled(true);
      if (isSomethingSelected())
        item.setLabel(Translate.text("menu.simplify"));
      else
        item.setLabel(Translate.text("menu.simplifyMesh"));
    }
    else if ("createFace".equals(itemname))
      item.setEnabled(isSomethingSelected() && selectionMode == TriMeshViewer.POINT_MODE);
    else if ("editPoints".equals(itemname))
    {
      item.setEnabled(isSomethingSelected());
    }
    else if ("transformPoints".equals(itemname))
      item.setEnabled(isSomethingSelected());
    else if ("randomize".equals(itemname))
      item.setEnabled(isSomethingSelected());
    else if ("bevel".equals(itemname))
    {
      item.setEnabled(isSomethingSelected());
    }
    else if ("parameters".equals(itemname))
    {                                 // TODO(MB) Check if this is senseful
      if (isSomethingSelected())
      {
        if (selectionMode == TriMeshViewer.EDGE_MODE)
        {
          int boundary[][] = ((TriMeshViewer) getCurrentView()).findSelectedBoundaries();
          item.setEnabled(boundary.length > 0);
        }
        else
          item.setEnabled(true);
      }
      else
        item.setEnabled(false);
    }
    else if ("closeBoundary".equals(itemname))
      item.setEnabled(isSomethingSelected() && selectionMode == TriMeshViewer.EDGE_MODE);
    else if ("joinBoundaries".equals(itemname))
    {
      if (isSomethingSelected() && selectionMode == TriMeshViewer.EDGE_MODE)
      {
        int boundary[][] = ((TriMeshViewer) getCurrentView()).findSelectedBoundaries();
        item.setEnabled(boundary.length == 2);
      }
      else
        item.setEnabled(false);
    }
    else if ("extractCurve".equals(itemname))
      item.setEnabled(isSomethingSelected() && selectionMode == TriMeshViewer.EDGE_MODE);
    else if ("smoothness".equals(itemname))
      item.setEnabled(isSomethingSelected() && selectionMode != TriMeshViewer.FACE_MODE);

    else if ("fourViews".equals(itemname))
    {
      if (numViewsShown == 4)
        item.setLabel(Translate.text("menu.fourViews"));
      else
        item.setLabel(Translate.text("menu.oneView"));
    }
    
    // Smooth Menu in Mesh Menu
    else if ("none".equals(itemname))
      cbitem.setState(obj.getSmoothingMethod() == TriangleMesh.NO_SMOOTHING);
    else if ("shading".equals(itemname))
      cbitem.setState(obj.getSmoothingMethod() == TriangleMesh.SMOOTH_SHADING);
    else if ("interpolating".equals(itemname))
      cbitem.setState(obj.getSmoothingMethod() == TriangleMesh.INTERPOLATING);
    else if ("approximating".equals(itemname))
      cbitem.setState(obj.getSmoothingMethod() == TriangleMesh.APPROXIMATING);

    // Skeleton Menu

    else if ("editBone".equals(itemname))
    {
      Skeleton s = obj.getSkeleton();
      Joint selJoint = s.getJoint(getCurrentView().getSelectedJoint());

      item.setEnabled(selJoint != null);
    }
    else if ("deleteBone".equals(itemname))
    {
      Skeleton s = obj.getSkeleton();
      Joint selJoint = s.getJoint(getCurrentView().getSelectedJoint());

      item.setEnabled(selJoint != null && selJoint.children.length == 0);
    }
    else if ("bindSkeleton".equals(itemname))
      item.setEnabled(isSomethingSelected());
    else super.updateMenuItem(item);

  }

  public void keyPressed(KeyEvent e)
  {
    if (e.getKeyCode() == e.VK_DELETE || e.getKeyCode() == e.VK_BACK_SPACE)
      {
	if (getCurrentView().getCurrentTool() instanceof SkeletonTool)
	  deleteJointCommand();
	else
	  deleteCommand();
      }
    else
      super.keyPressed(e);
  }

  public void actionPerformed(ActionEvent e)
  {
    String command = e.getActionCommand();

    setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));

    if (handleMenuAction(command)) // Try to call the command by reflection
      ;  // Do nothing
    else if (command.equals("undo"))
      undoCommand();
    else if (command.equals("clear"))
      deleteCommand();
//    else if (command.equals("selectAll"))
//      selectAllCommand();
//    else if (command.equals("selectBoundary"))
//      selectBoundaryCommand();
//    else if (command.equals("extendSelection"))
//      extendSelectionCommand();
    else if (command.equals("subdivideEdges"))
      subdivideCommand();
//    else if (command.equals("simplify"))
//      simplifyCommand();
//    else if (command.equals("optimize"))
//      optimizeCommand();
    else if (command.equals("editPoints"))
      setPointsCommand();
//    else if (command.equals("transformPoints"))
//      transformPointsCommand();
//    else if (command.equals("randomize"))
//      randomizeCommand();
//    else if (command.equals("bevel"))
//      bevelCommand();
    else if (command.equals("parameters"))
      setParametersCommand();
    else if (command.equals("centerMesh"))
      centerCommand();
//    else if (command.equals("closeBoundary"))
//      closeBoundaryCommand();
//    else if (command.equals("joinBoundaries"))
//      joinBoundariesCommand();
//    else if (command.equals("extractCurve"))
//      extractCurveCommand();
    else if (command.equals("smoothness"))
      setSmoothnessCommand();
    else if (command.equals("invertNormals"))
      reverseNormalsCommand();
    else if (command.equals("meshTension"))
      setTensionCommand();
    else if (command.equals("renderPreview"))
      getCurrentView().previewObject();
    else if (command.equals("editBone"))
      editJointCommand();
    else if (command.equals("deleteBone"))
      deleteJointCommand();
//    else if (command.equals("bindSkeleton"))
//      bindSkeletonCommand();
    else if (command.equals("grid"))
      setGridCommand();
    else if (command.equals("fourViews"))
      toggleViewsCommand();
    else if (command.equals("showTemplate"))
      {
        boolean wasShown = getCurrentView().getTemplateShown();
        for(int i = 0; i < theView.length; ++i)
          theView[i].setShowTemplate(!wasShown);
        updateMenus();
        updateImage();
      }
    else if (command.equals("setTemplate"))
      setTemplateCommand();
    setCursor(Cursor.getDefaultCursor());
  }

  public void itemStateChanged(ItemEvent e)
  {
    Object source = e.getSource();
    CheckboxMenuItem sourceitem = null;
    String itemname = "";
    int i;

    if (source instanceof MenuItem)
    {
      sourceitem = (CheckboxMenuItem)source;
      itemname = sourceitem.getName();
    }

    if ("tolerantSelection".equals(itemname))
      {
        for(i = 0; i < theView.length; ++i)
          ((TriMeshViewer) theView[i]).setTolerant(sourceitem.getState());
	return;
      }
    if ("freehandSelection".equals(itemname))
      {
        for(i = 0; i < theView.length; ++i)
          ((TriMeshViewer) theView[i]).setFreehandSelection(sourceitem.getState());
	return;
      }
    if ("displayAsQuads".equals(itemname))
      {
        for(i = 0; i < theView.length; ++i)
          ((TriMeshViewer) theView[i]).setQuadMode(sourceitem.getState());
        updateImage();
	return;
      }
    if ("detachSkeleton".equals(itemname))
      {
        for(i = 0; i < theView.length; ++i)
          ((TriMeshViewer) theView[i]).setSkeletonDetached(sourceitem.getState());
	return;
      }

    final String[] SMOOTHMETHODS= new String[]{"none", "shading", "interpolating", "approximating"};
    i = Utilities.findIndexEqual(SMOOTHMETHODS, itemname);
    if (i > -1)
    {
      setSmoothingMethod(i);
      return;
    }
    super.itemStateChanged(e);
  }

  /** Creates a new view, containing the same Object3D with a
   * new MeshEditSelection.
   */
  protected MeshEditorWindow createNewViewNewSelection()
  {
    TriMeshEditorWindow w = new TriMeshEditorWindow(parentWindow, getTitle(), objInConstructor, onClose, topology);
    // w.theView.setObject(theView.getObject().getObject3D());  TODO(MB) ??? Something with editGesture ???
    return w;
  }

  /** Create a MeshViewer which is appropriate for a particular editor window.
    Here a TriMeshViewer.*/
  
  public MeshViewer createMeshViewer(ObjectInfo obj, Panel p)
  {
    return new TriMeshViewer(obj,  p);
  }
    
  public void initBaseSelectionHolder(ObjectInfo obj)
  {
    baseSelHolder = new MeshSelectionHolder(obj);
    TriangleMesh mesh = (TriangleMesh) obj.object;
    baseSelHolder.setSelection(new boolean [mesh.getVertices().length]);
    baseSelHolder.setSelectionMode(TriMeshViewer.POINT_MODE);
    baseSelHolder.decViewerRefCount();  // Start with ref count 0.
    
    for (int i = 0; i < theView.length; ++i)
      ((TriMeshViewer) theView[i]).setSelectionHolder(baseSelHolder);
  }
  
  public void initWindowMenus(ObjectInfo obj)
  {  
    //dummy
  }
  
  
  /** Create the toolbar panel on the left of an object editor. */
  
  public Panel createToolbarPanel()
  {
//    Panel p4 = new Panel();
//    
//    p4.setLayout(new BorderLayout());
//    if (topology)
//      p4.add(tools = new ToolPalette(1, 10), "North");
//    else
//      p4.add(tools = new ToolPalette(1, 9), "North");
//    tools.addTool(defaultTool = new ReshapeMeshTool(this));
//    tools.addTool(new ScaleMeshTool(this));
//    tools.addTool(new RotateMeshTool(this, false));
//    tools.addTool(new SkewMeshTool(this));
//    tools.addTool(new TaperMeshTool(this));
//    tools.addTool(new ThickenMeshTool(this));
//    tools.addTool(new SkeletonTool(this));
//    if (topology)
//      tools.addTool(new CreateVertexTool(this));
//    tools.addTool(getMetaTool() /* MoveViewTool */);
//    tools.addTool(getAltTool() /* RotateViewTool */);
//    tools.selectTool(defaultTool);
//    p4.add(modes = new ToolPalette(1, 3), "South");
//    modes.addTool(new GenericTool(this, "point.gif", "selected/point.gif"));
//    modes.addTool(new GenericTool(this, "edge.gif", "selected/edge.gif"));
//    modes.addTool(new GenericTool(this, "face.gif", "selected/face.gif"));
//    
    return new Panel();   // Dummy
  }
  
  /* Select the entire mesh. */

  public void selectAllCommand()
  {
    TriMeshViewer tmv = (TriMeshViewer) getCurrentView();
    boolean selected[] = tmv.getSelection();

    for (int i = 0; i < selected.length; i++)
      selected[i] = true;
    if (tmv.getSelectionMode() == TriMeshViewer.EDGE_MODE)
      for (int i = 0; i < selected.length; i++)
        if (tmv.isEdgeHidden(i))
          selected[i] = false;
    
    ((TriMeshViewer) getCurrentView()).setSelection(selected);
    ((TriMeshViewer) getCurrentView()).informSelectionChanged(this,
        ModelEvent.CHANGEDUR_OBJECTEDITOR);
  }

  /* Select the edges which form the boundary of the mesh. */

  public void selectBoundaryCommand()
  {
    TriangleMesh theMesh = (TriangleMesh) getCurrentView().getObject().object;
    TriangleMesh.Edge edge[] = theMesh.getEdges();
    boolean selected[] = ((TriMeshViewer) getCurrentView()).getSelection();

    for (int i = 0; i < selected.length; i++)
    {
      selected[i] = (edge[i].f2 == -1);
    }
    
    ((TriMeshViewer) getCurrentView()).setSelection(selected);
    ((TriMeshViewer) getCurrentView()).informSelectionChanged(this,
        ModelEvent.CHANGEDUR_OBJECTEDITOR);
  }

  /* Extend the selection outward by one edge. */

  public void extendSelectionCommand()
  {
    TriMeshViewer tmv = (TriMeshViewer) getCurrentView();
    TriangleMesh theMesh = (TriangleMesh) getCurrentView().getObject().object;
    int dist[] = tmv.getSelectionDistance();
    boolean selectedVert[] = new boolean [dist.length];
    boolean selected[] = tmv.getSelection();
    TriangleMesh.Edge edge[] = theMesh.getEdges();

    for (int i = 0; i < edge.length; i++)
      if ((dist[edge[i].v1] == 0 || dist[edge[i].v2] == 0) && !tmv.isEdgeHidden(i))
	selectedVert[edge[i].v1] = selectedVert[edge[i].v2] = true;
    if (tmv.getSelectionMode() == TriMeshViewer.POINT_MODE)
    {
      ((TriMeshViewer) getCurrentView()).setSelection(selected);
      ((TriMeshViewer) getCurrentView()).informSelectionChanged(this,
          ModelEvent.CHANGEDUR_OBJECTEDITOR);
    }
    else if (tmv.getSelectionMode() == TriMeshViewer.EDGE_MODE)
      {
	for (int i = 0; i < edge.length; i++)
	  selected[i] = (selectedVert[edge[i].v1] && selectedVert[edge[i].v2]);
        
        ((TriMeshViewer) getCurrentView()).setSelection(selected);
        ((TriMeshViewer) getCurrentView()).informSelectionChanged(this,
            ModelEvent.CHANGEDUR_OBJECTEDITOR);
      }
    else
      {
	TriangleMesh.Face face[] = theMesh.getFaces();
	for (int i = 0; i < face.length; i++)
	  selected[i] = (selectedVert[face[i].v1] && selectedVert[face[i].v2] && selectedVert[face[i].v3]);

        ((TriMeshViewer) getCurrentView()).setSelection(selected);
        ((TriMeshViewer) getCurrentView()).informSelectionChanged(this,
            ModelEvent.CHANGEDUR_OBJECTEDITOR);
      }
  }

  /** Delete the selected points, edges, or faces from the mesh. */

  public void deleteCommand()
  {
    TriMeshViewer tmv = (TriMeshViewer) getCurrentView();
    TriangleMesh theMesh = (TriangleMesh) getCurrentView().getObject().object;
    TriangleMesh.Vertex vert[] = (Vertex []) theMesh.getVertices();
    TriangleMesh.Edge edge[] = theMesh.getEdges();
    TriangleMesh.Face face[] = theMesh.getFaces();
    boolean selected[] = tmv.getSelection();
    boolean deleteVert[] = new boolean [vert.length];
    boolean deleteEdge[] = new boolean [edge.length];
    boolean deleteFace[] = new boolean [face.length];

    // Determine which parts of the mesh to delete.

    if (tmv.getSelectionMode() == TriMeshViewer.POINT_MODE)
      {
	for (int i = 0; i < deleteVert.length; i++)
	  deleteVert[i] = selected[i];
	for (int i = 0; i < deleteEdge.length; i++)
	  deleteEdge[i] = (deleteVert[edge[i].v1] || deleteVert[edge[i].v2]);
	for (int i = 0; i < deleteFace.length; i++)
	  deleteFace[i] = (deleteVert[face[i].v1] || deleteVert[face[i].v2] || deleteVert[face[i].v3]);
      }
    else if (tmv.getSelectionMode() == TriMeshViewer.EDGE_MODE)
      {
	for (int i = 0; i < deleteFace.length; i++)
	  deleteFace[i] = (selected[face[i].e1] || selected[face[i].e2] || selected[face[i].e3]);
	for (int i = 0; i < deleteEdge.length; i++)
	  deleteEdge[i] = (deleteFace[edge[i].f1] && (edge[i].f2 == -1 || deleteFace[edge[i].f2]));
	for (int i = 0; i < deleteVert.length; i++)
	  deleteVert[i] = vert[i].edges > 0;  // Already "orphaned" vertices are never deleted
	for (int i = 0; i < deleteFace.length; i++)
	  if (!deleteFace[i])
	    deleteVert[face[i].v1] = deleteVert[face[i].v2] = deleteVert[face[i].v3] = false;
      }
    else
      {
	for (int i = 0; i < deleteFace.length; i++)
	  deleteFace[i] = selected[i];
	for (int i = 0; i < deleteEdge.length; i++)
	  deleteEdge[i] = (deleteFace[edge[i].f1] && (edge[i].f2 == -1 || deleteFace[edge[i].f2]));
	for (int i = 0; i < deleteVert.length; i++)
	  deleteVert[i] = vert[i].edges > 0;  // Already "orphaned" vertices are never deleted
	for (int i = 0; i < deleteFace.length; i++)
	  if (!deleteFace[i])
	    deleteVert[face[i].v1] = deleteVert[face[i].v2] = deleteVert[face[i].v3] = false;
      }

    // Make sure this will still be a valid object.
    
    int breaks = 0;
    
    for (int i = 0; i < vert.length; i++)
      {
	int e[] = theMesh.getEdgesForVertex(vert[i]);
        if (e.length == 0)
          continue;

	int f, fprev = edge[e[0]].f1;
        breaks = 0;
	for (int j = 1; j < e.length; j++)
	  {
	    f = (edge[e[j]].f1 == fprev ? edge[e[j]].f2 : edge[e[j]].f1);
	    if (f == -1)
	      break;
	    if (!deleteFace[fprev] && deleteFace[f])
	      breaks++;
	    fprev = f;
	  }
	if (!deleteFace[fprev] && (edge[e[0]].f2 == -1 || deleteFace[edge[e[0]].f1]))
	  breaks++;
	if (breaks > 1)
          break;
    }
    
    if (breaks > 1)
    {
      MessageDialog dlg = new MessageDialog(this, new String [] {"Deleting the selection will",
        "result in an illegal surface."},
        new String [] {Translate.text("button.ok"), Translate.text("button.cancel")});
      if (dlg.getChoice() == 1)
        return;
    }

    // Construct the new mesh.

    int newVertCount = 0, newFaceCount = 0;
    int newVertIndex[] = new int [vert.length];
    for (int i = 0; i < deleteVert.length; i++)
      {
	newVertIndex[i] = -1;
	if (!deleteVert[i])
	  newVertCount++;
      }
    for (int i = 0; i < deleteFace.length; i++)
      if (!deleteFace[i])
	newFaceCount++;
    Vertex v[] = new Vertex [newVertCount];
    int f[][] = new int [newFaceCount][];
    newVertCount = 0;
    for (int i = 0; i < vert.length; i++)
      if (!deleteVert[i])
	{
	  newVertIndex[i] = newVertCount;
	  v[newVertCount++] = vert[i];
	}
    newFaceCount = 0;
    for (int i = 0; i < face.length; i++)
      if (!deleteFace[i])
	f[newFaceCount++] = new int [] {newVertIndex[face[i].v1], newVertIndex[face[i].v2], newVertIndex[face[i].v3]};
    TriangleMesh newmesh = new TriangleMesh(v, f);
    Vertex newvert[] = (Vertex []) newmesh.getVertices();
    Edge newedge[] = newmesh.getEdges();
    newmesh.getSkeleton().copy(theMesh.getSkeleton());
    newmesh.setTexture(theMesh.getTexture());
    newmesh.setTextureMapping(theMesh.getTextureMapping());
    if (newmesh.isClosed() && theMesh.getMaterial() != null)
      {
	newmesh.setMaterial(theMesh.getMaterial());
	newmesh.setMaterialMapping(theMesh.getMaterialMapping());
      }
    newmesh.setSmoothingMethod(theMesh.getSmoothingMethod());

    // Copy over the smoothness values for edges.

    for (int i = 0; i < edge.length; i++)
      {
	int r1 = newVertIndex[edge[i].v1];
	int r2 = newVertIndex[edge[i].v2];
	for (int j = 0; j < newedge.length; j++)
	  if ((r1 == newedge[j].v1 && r2 == newedge[j].v2) || (r1 == newedge[j].v2 && r2 == newedge[j].v1))
	    newedge[j].smoothness = edge[i].smoothness;
      }
    setUndoRecord(new UndoRecord(this, false, UndoRecord.COPY_OBJECT, new Object [] {theMesh, theMesh.duplicate()}));
    theMesh.copyObject(newmesh);
    object3DChangedDuringEditor();
  }

  /* Subdivide selected edges or faces of the mesh. */

  public void subdivideCommand()
  {
    int i, j;
    TriMeshViewer tmv = (TriMeshViewer) getCurrentView();
    TriangleMesh theMesh = (TriangleMesh) tmv.getObject().object, newmesh;
    boolean selected[] = tmv.getSelection(), newselection[];
    Edge edges[];
    Face faces[];

    if (tmv.getSelectionMode() != TriMeshViewer.EDGE_MODE &&
        tmv.getSelectionMode() != TriMeshViewer.FACE_MODE)
      return;
    for (i = 0; !selected[i] && i < selected.length; i++);
    if (i == selected.length)
      return;

    if (tmv.getSelectionMode() == TriMeshViewer.EDGE_MODE)
      {
	// Subdivide selected edges, using the appropriate method.

	i = theMesh.getVertices().length;
        newmesh = TriangleMesh.subdivideEdges(theMesh, selected, Double.MAX_VALUE);
//	if (theMesh.getSmoothingMethod() == TriangleMesh.APPROXIMATING)
//	  newmesh = TriangleMesh.subdivideLoop(theMesh, selected, Double.MAX_VALUE);
//	else if (theMesh.getSmoothingMethod() == TriangleMesh.INTERPOLATING)
//	  newmesh = TriangleMesh.subdivideButterfly(theMesh, selected, Double.MAX_VALUE);
//	else
//	  newmesh = TriangleMesh.subdivideLinear(theMesh, selected);
	setUndoRecord(new UndoRecord(this, false, UndoRecord.COPY_OBJECT, new Object [] {theMesh, theMesh.duplicate()}));
	theMesh.copyObject(newmesh);
	object3DChangedDuringEditor();

	// Update the selection.

	edges = newmesh.getEdges();
	newselection = new boolean [edges.length];
	for (j = 0; j < edges.length; j++)
	  newselection[j] = (edges[j].v1 >= i || edges[j].v2 >= i);

        ((TriMeshViewer) getCurrentView()).setSelection(newselection);
        ((TriMeshViewer) getCurrentView()).informSelectionChanged(this,
            ModelEvent.CHANGEDUR_OBJECTEDITOR);
      }
    else
      {
	// Subdivide selected faces.

	i = theMesh.getVertices().length;
	newmesh = TriangleMesh.subdivideFaces(theMesh, selected);
	setUndoRecord(new UndoRecord(this, false, UndoRecord.COPY_OBJECT, new Object [] {theMesh, theMesh.duplicate()}));
	theMesh.copyObject(newmesh);
	object3DChangedDuringEditor();

	// Update the selection.

	faces = newmesh.getFaces();
	newselection = new boolean [faces.length];
	for (j = 0; j < faces.length; j++)
	  newselection[j] = (faces[j].v1 >= i || faces[j].v2 >= i || faces[j].v3 >= i);

        ((TriMeshViewer) getCurrentView()).setSelection(newselection);
        ((TriMeshViewer) getCurrentView()).informSelectionChanged(this,
            ModelEvent.CHANGEDUR_OBJECTEDITOR);
      }
  }

  public void simplifyCommand()
  {
    TriMeshViewer tmv = (TriMeshViewer) getCurrentView();
    TriangleMesh theMesh = (TriangleMesh) tmv.getObject().object;
    boolean selection[] = tmv.getSelection();
    int selectionMode = tmv.getSelectionMode();
    ValueField errorField = new ValueField(0.01, ValueField.NONNEGATIVE);
    int i;

    ComponentsDialog dlg = new ComponentsDialog(this, "Specify tolerance for simplified mesh:",
    	new Component[] {errorField}, new String[] {"Max Surface Error"});
    if (!dlg.clickedOk())
      return;
    setUndoRecord(new UndoRecord(this, false, UndoRecord.COPY_OBJECT, new Object [] {theMesh, theMesh.duplicate()}));

    // If we are not in Edge selection mode, convert the selection to edges.

    if (selectionMode == TriMeshViewer.POINT_MODE)
    {
      Edge e[] = theMesh.getEdges();
      boolean newSel[] = new boolean [e.length];
      for (i = 0; i < e.length; i++)
        newSel[i] = (selection[e[i].v1] && selection[e[i].v2]);
      selection = newSel;
    }
    if (selectionMode == TriMeshViewer.FACE_MODE)
    {
      Edge e[] = theMesh.getEdges();
      boolean newSel[] = new boolean [e.length];
      for (i = 0; i < e.length; i++)
        newSel[i] = (selection[e[i].f1] || selection[e[i].f2]);
      selection = newSel;
    }

    // If no edges are selected, then simplify the entire mesh.

    for (i = 0; i < selection.length && !selection[i]; i++);
    if (i == selection.length)
      {
        selection = new boolean [selection.length];
        for (i = 0; i < selection.length; i++)
          selection[i] = true;
      }

    // Generate the simplified mesh.

    new TriMeshSimplifier(theMesh, selection, errorField.getValue(), this);
    object3DChangedDuringEditor();
  }

  public void optimizeCommand()
  {
    TriMeshViewer tmv = (TriMeshViewer) getCurrentView();
    MessageDialog dlg = new MessageDialog(this, "Optimize mesh connectivity?  This will rearrange the mesh edges to produce a smoother surface.",
        new String [] {Translate.text("button.ok"), Translate.text("button.cancel")});
    if (dlg.getChoice() == 1)
      return;
    TriangleMesh theMesh = (TriangleMesh) tmv.getObject().object;
    setUndoRecord(new UndoRecord(this, false, UndoRecord.COPY_OBJECT, new Object [] {theMesh, theMesh.duplicate()}));
    theMesh.copyObject(TriangleMesh.optimizeMesh(theMesh));
    for (int i = 0; i < theView.length; ++i)
      theView[i].setMesh(theMesh);
    
    boolean selection[] = ((TriMeshViewer) getCurrentView()).getSelection();
    for (int i = 0; i < selection.length; i++)
      selection[i] = false;

    ((TriMeshViewer) getCurrentView()).setSelection(selection);
    ((TriMeshViewer) getCurrentView()).informSelectionChanged(this,
        ModelEvent.CHANGEDUR_OBJECTEDITOR);
    object3DChangedDuringEditor();
  }

  public void createFaceCommand()
  {
    boolean selection[] = ((TriMeshViewer) getCurrentView()).getSelection();
    int sel[] = Utilities.createIndexList(selection);

    if (sel.length != 3)
      return;

    TriangleMesh theMesh = (TriangleMesh) getCurrentView().getObject().object;
    setUndoRecord(new UndoRecord(this, false, UndoRecord.COPY_OBJECT, new Object [] {theMesh, theMesh.duplicate()}));
    if (theMesh.addNewFace(sel[0], sel[1], sel[2]) == -1)
    {
      MessageDialog dlg = new MessageDialog(this,
        new String [] {"An edge can't be adjacent to more than",
        "two faces. New face not created."},
        new String [] {Translate.text("button.ok")});
      return;
    }
    object3DChangedDuringEditor();
  }
  
  public void repairGeometryCommand()
  {
    TriangleMesh theMesh = (TriangleMesh) getCurrentView().getObject().object;
    setUndoRecord(new UndoRecord(this, false, UndoRecord.COPY_OBJECT, new Object [] {theMesh, theMesh.duplicate()}));
    theMesh.repairGeometry();
    object3DChangedDuringEditor();
  }
    

  public void bevelCommand()
  {
    TriMeshViewer tmv = (TriMeshViewer) getCurrentView();
    final TriangleMesh theMesh = (TriangleMesh) tmv.getObject().object;
    Face face[] = theMesh.getFaces();
    final boolean selection[] = tmv.getSelection();
    final ValueField heightField = new ValueField(0.0, ValueField.NONE), widthField = new ValueField(0.0, ValueField.NONE);
    final Choice applyChoice = new Choice();
    final ObjectPreviewCanvas preview = new ObjectPreviewCanvas(new ObjectInfo(theMesh, new CoordinateSystem(), ""));
    final int bevelMode[];
    int selectionMode = tmv.getSelectionMode();

    if (selectionMode == TriMeshViewer.FACE_MODE)
      bevelMode = new int [] {TriMeshBeveler.BEVEL_FACE_GROUPS, TriMeshBeveler.BEVEL_FACES};
    else if (selectionMode == TriMeshViewer.POINT_MODE)
      bevelMode = new int [] {TriMeshBeveler.BEVEL_VERTICES, TriMeshBeveler.BEVEL_VERTICES};
    else
      bevelMode = new int [] {TriMeshBeveler.BEVEL_EDGES, TriMeshBeveler.BEVEL_EDGES};
    applyChoice.add("Selection as a Whole");
    applyChoice.add("Individual Faces");
    applyChoice.addItemListener(new ItemListener() {
      public void itemStateChanged(ItemEvent ev)
      {
        TriMeshBeveler beveler = new TriMeshBeveler(theMesh, selection, bevelMode[applyChoice.getSelectedIndex()]);
        double height = heightField.getValue();
        double width = widthField.getValue();
        preview.setObject(beveler.bevelMesh(height, width));
        preview.updateImage();
        preview.repaint();
      }
    });
    TextListener listener = new TextListener() {
      public void textValueChanged(TextEvent ev)
      {
        TriMeshBeveler beveler = new TriMeshBeveler(theMesh, selection, bevelMode[applyChoice.getSelectedIndex()]);
        double height = heightField.getValue();
        double width = widthField.getValue();
        preview.setObject(beveler.bevelMesh(height, width));
        preview.updateImage();
        preview.repaint();
      }
    };
    heightField.addTextListener(listener);
    widthField.addTextListener(listener);
    preview.setPreferredSize(200, 200);
    ComponentsDialog dlg;
    if (selectionMode == TriMeshViewer.FACE_MODE)
      dlg = new ComponentsDialog(this, "Bevel/Extrude Selected Faces:",
    	new Component[] {heightField, widthField, applyChoice, preview},
          new String [] {"Extrude Height", "Bevel Width", "Apply To", ""});
    else if (selectionMode == TriMeshViewer.POINT_MODE)
      dlg = new ComponentsDialog(this, "Bevel/Extrude Selected Points:",
          new Component[] {heightField, widthField, preview}, 
          new String [] {"Extrude Height", "Bevel Width", ""});
    else
      dlg = new ComponentsDialog(this, "Bevel/Extrude Selected Edges:",
          new Component[] {heightField, widthField, preview}, 
          new String [] {"Extrude Height", "Bevel Width", ""});
    if (!dlg.clickedOk())
      return;
    double height = heightField.getValue();
    double width = widthField.getValue();
    setUndoRecord(new UndoRecord(this, false, UndoRecord.COPY_OBJECT, new Object [] {theMesh, theMesh.duplicate()}));

    // Generate the new mesh.

    TriMeshBeveler beveler = new TriMeshBeveler(theMesh, selection, bevelMode[applyChoice.getSelectedIndex()]);
    theMesh.copyObject(beveler.bevelMesh(height, width));

    object3DChangedDuringEditor();
    ((TriMeshViewer) getCurrentView()).setSelection(beveler.getNewSelection());
    ((TriMeshViewer) getCurrentView()).informSelectionChanged(this,
        ModelEvent.CHANGEDUR_OBJECTEDITOR);

  }

  public void closeBoundaryCommand()
  {
    TriangleMesh theMesh = (TriangleMesh) getCurrentView().getObject().object;
    Vertex vt[] = (Vertex []) theMesh.getVertices();
    Edge ed[] = theMesh.getEdges();
    Face fc[] = theMesh.getFaces();
    int boundary[][] = ((TriMeshViewer) getCurrentView()).findSelectedBoundaries();
    Vec3 newvert[] = new Vec3 [vt.length+boundary.length];
    int newface[][], count = 0;

    // First, add all of the current vertices and faces to the lists.

    for (int i = 0; i < vt.length; i++)
      newvert[i] = vt[i].r;
    for (int i = 0; i < boundary.length; i++)
      count += boundary[i].length;
    newface = new int [fc.length+count][];
    for (int i = 0; i < fc.length; i++)
      newface[i] = new int [] {fc[i].v1, fc[i].v2, fc[i].v3};
    count = fc.length;

    // For each boundary, add a single vertex in the center, and faces surrounding it.

    for (int i = 0; i < boundary.length; i++)
      {
        int j = ed[boundary[i][0]].v1;
        Vec3 center = new Vec3(vt[j].r);
        for (int k = 1; k < boundary[i].length; k++)
          {
            Edge e = ed[boundary[i][k]];
            j = (e.v1 == j ? e.v2 : e.v1);
            center.add(vt[j].r);
          }
        center.scale(1.0/boundary[i].length);
        newvert[vt.length+i] = center;
        for (int k = 0; k < boundary[i].length; k++)
          {
            Edge e = ed[boundary[i][k]];
            Face f = fc[e.f1];
            if ((f.v1 == e.v1 && f.v2 == e.v2) || (f.v2 == e.v1 && f.v3 == e.v2) || (f.v3 == e.v1 && f.v1 == e.v2))
              newface[count++] = new int [] {e.v2, e.v1, vt.length+i};
            else
              newface[count++] = new int [] {e.v1, e.v2, vt.length+i};
          }
      }

    // Create the new mesh;

    TriangleMesh newmesh = new TriangleMesh(newvert, newface);
    Vertex newvt[] = (Vertex []) newmesh.getVertices();
    Edge newed[] = newmesh.getEdges();
    newmesh.setTexture(theMesh.getTexture());
    newmesh.setTextureMapping(theMesh.getTextureMapping());
    if (newmesh.isClosed() && theMesh.getMaterial() != null)
      {
	newmesh.setMaterial(theMesh.getMaterial());
	newmesh.setMaterialMapping(theMesh.getMaterialMapping());
      }
    newmesh.setSmoothingMethod(theMesh.getSmoothingMethod());

    // Copy over the smoothness values and texture parameters.

    for (int i = 0; i < vt.length; i++)
      {
	newvt[i].smoothness = vt[i].smoothness;
	newvt[i].param = vt[i].param;
      }
    for (int i = 0; i < newed.length; i++)
      {
	if (newed[i].v1 >= vt.length || newed[i].v2 >= vt.length)
	  continue;
	for (int j = 0; j < ed.length; j++)
	  if ((newed[i].v1 == ed[j].v1 && newed[i].v2 == ed[j].v2) || (newed[i].v1 == ed[j].v2 && newed[i].v2 == ed[j].v1))
	    newed[i].smoothness = ed[j].smoothness;
      }
    setUndoRecord(new UndoRecord(this, false, UndoRecord.COPY_OBJECT, new Object [] {theMesh, theMesh.duplicate()}));
    theMesh.copyObject(newmesh);
    object3DChangedDuringEditor();
  }

  public void joinBoundariesCommand()
  {
    TriangleMesh theMesh = (TriangleMesh) getCurrentView().getObject().object;
    Vertex vt[] = (Vertex []) theMesh.getVertices();
    Edge ed[] = theMesh.getEdges();
    final int boundary[][] = ((TriMeshViewer) getCurrentView()).findSelectedBoundaries();
    final int boundaryVert[][] = new int [2][];

    if (boundary.length != 2)
      return;

    // Construct a list of the vertices on each boundary;

    boundaryVert[0] = new int [boundary[0].length];
    boundaryVert[1] = new int [boundary[1].length];
    if (ed[boundary[0][0]].v1 == ed[boundary[0][1]].v1 || ed[boundary[0][0]].v1 == ed[boundary[0][1]].v2)
      boundaryVert[0][0] = ed[boundary[0][0]].v2;
    else
      boundaryVert[0][0] = ed[boundary[0][0]].v1;
    for (int i = 0; i < boundary[0].length-1; i++)
      {
        Edge e = ed[boundary[0][i]];
        boundaryVert[0][i+1] = (e.v1 == boundaryVert[0][i] ? e.v2 : e.v1);
      }
    if (ed[boundary[1][0]].v1 == ed[boundary[1][1]].v1 || ed[boundary[1][0]].v1 == ed[boundary[1][1]].v2)
      boundaryVert[1][0] = ed[boundary[1][0]].v2;
    else
      boundaryVert[1][0] = ed[boundary[1][0]].v1;
    for (int i = 0; i < boundary[1].length-1; i++)
      {
        Edge e = ed[boundary[1][i]];
        boundaryVert[1][i+1] = (e.v1 == boundaryVert[1][i] ? e.v2 : e.v1);
      }

    // Consider all possible ways of joining the two boundaries, and select the one which
    // gives the lowest mean squared distance between connected points as the initial guess.

    double offset = 0.0;
    boolean reverse = false;
    int maxsteps = (boundary[0].length > boundary[1].length ? boundary[0].length : boundary[1].length);
    double step0 = boundary[0].length/((double) maxsteps);
    double step1 = (reverse ? -1.0 : 1.0)*boundary[1].length/((double) maxsteps);
    double mindist = Double.MAX_VALUE;
    for (int i = 0; i < maxsteps-1; i++)
      {
        double dist = 0.0;
        for (int j = 0; j < maxsteps; j++)
          {
            double p0 = j*step0, p1 = (i+j)*step1;
            int i0 = ((int) Math.round(p0)+boundary[0].length) % boundary[0].length;
            int i1 = ((int) Math.round(p1)+boundary[1].length) % boundary[1].length;
            dist += vt[boundaryVert[0][i0]].r.distance2(vt[boundaryVert[1][i1]].r);
          }
        if (dist < mindist)
          {
            mindist = dist;
            offset = i*step1;
            reverse = false;
          }
        dist = 0.0;
        for (int j = 0; j < maxsteps; j++)
          {
            double p0 = j*step0, p1 = (i-j)*step1;
            int i0 = ((int) Math.round(p0)+boundary[0].length) % boundary[0].length;
            int i1 = ((int) Math.round(p1)+boundary[1].length) % boundary[1].length;
            dist += vt[boundaryVert[0][i0]].r.distance2(vt[boundaryVert[1][i1]].r);
          }
        if (dist < mindist)
          {
            mindist = dist;
            offset = i*step1;
            reverse = true;
          }
      }

    // Create a dialog allowing the user to adjust the parameters.

    final ValueSlider offsetSlider = new ValueSlider(-0.5*boundary[1].length, 0.5*boundary[1].length, 2*maxsteps, 0.0);
    final Checkbox reverseBox = new Checkbox("Reverse Direction", false);
    final ObjectPreviewCanvas preview = new ObjectPreviewCanvas(new ObjectInfo(doJoinBoundaries(boundary,
      boundaryVert, offset, reverse), new CoordinateSystem(), ""));
    final double baseOffset = offset;
    final boolean baseReverse = reverse;
    Panel p = new Panel();
    GridBagConstraints gc = new GridBagConstraints();
    p.setLayout(new GridBagLayout());
    gc.gridy = 0;
    p.add(new Label("Offset: "), gc);
    gc.fill = gc.HORIZONTAL;
    gc.anchor = gc.WEST;
    p.add(offsetSlider, gc);
    gc.gridy = 1;
    gc.gridwidth = 2;
    gc.fill = gc.NONE;
    gc.anchor = gc.CENTER;
    p.add(reverseBox, gc);
    gc.gridy = 2;
    gc.fill = gc.BOTH;
    gc.weightx = gc.weighty = 1.0;
    p.add(preview, gc);
    preview.setPreferredSize(200, 200);
    preview.setRenderMode(preview.RENDER_FLAT);
    offsetSlider.addAdjustmentListener(new AdjustmentListener() {
      public void adjustmentValueChanged(AdjustmentEvent ev)
      {
        preview.setObject(doJoinBoundaries(boundary, boundaryVert, baseOffset+offsetSlider.getValue(), baseReverse^reverseBox.getState()));
        preview.updateImage();
        preview.repaint();
      }
    } );
    reverseBox.addItemListener(new ItemListener() {
      public void itemStateChanged(ItemEvent ev)
      {
        preview.setObject(doJoinBoundaries(boundary, boundaryVert, baseOffset+offsetSlider.getValue(), baseReverse^reverseBox.getState()));
        preview.updateImage();
        preview.repaint();
      }
    } );
    PanelDialog dlg = new PanelDialog(this, "Select parameters for joining the boundary curves:", p);
    if (!dlg.clickedOk())
      return;
    setUndoRecord(new UndoRecord(this, false, UndoRecord.COPY_OBJECT, new Object [] {theMesh, theMesh.duplicate()}));
    theMesh.copyObject(preview.getObject().object);
    object3DChangedDuringEditor();
}

  /* This method does the actual work of joining together two boundary curves. */

  TriangleMesh doJoinBoundaries(int boundary[][], int boundaryVert[][], double offset, boolean reverse)
  {
    int maxsteps = (boundary[0].length > boundary[1].length ? boundary[0].length : boundary[1].length);
    double step0 = boundary[0].length/((double) maxsteps);
    double step1 = (reverse ? -1.0 : 1.0)*boundary[1].length/((double) maxsteps);
    TriangleMesh theMesh = (TriangleMesh) getCurrentView().getObject().object;
    Vertex vt[] = (Vertex []) theMesh.getVertices();
    Edge ed[] = theMesh.getEdges();
    Face fc[] = theMesh.getFaces();
    Vec3 newvert[] = new Vec3 [vt.length];
    int newface[][] = new int [fc.length+boundary[0].length+boundary[1].length][];

    // First copy over the old vertices and faces.

    for (int i = 0; i < vt.length; i++)
      newvert[i] = vt[i].r;
    for (int i = 0; i < fc.length; i++)
      newface[i] = new int [] {fc[i].v1, fc[i].v2, fc[i].v3};

    // Go around the boundaries and add new faces.

    int count = fc.length;
    double p0 = 0.0, p1 = offset;
    int i0prev = 0, i1prev = ((int) Math.round(p1)+boundary[1].length) % boundary[1].length;
    for (int i = 1; i <= maxsteps; i++)
      {
        p0 += step0;
        p1 += step1;
        while (p1 < 0.0)
          p1 += boundary[1].length;
        int i0 = ((int) Math.round(p0)) % boundary[0].length;
        int i1 = ((int) Math.round(p1)) % boundary[1].length;
        if (i0 != i0prev)
          {
            Edge e = ed[boundary[0][step0 > 0.0 ? i0prev : i0]];
            Face f = fc[e.f1];
            int v1 = boundaryVert[0][i0prev], v2 = boundaryVert[0][i0];
            if ((f.v1 == v1 && f.v2 == v2) || (f.v2 == v1 && f.v3 == v2) || (f.v3 == v1 && f.v1 == v2))
              newface[count++] = new int [] {v2, v1, boundaryVert[1][i1prev]};
            else
              newface[count++] = new int [] {v1, v2, boundaryVert[1][i1prev]};
          }
        if (i1 != i1prev)
          {
            Edge e = ed[boundary[1][step1 > 0.0 ? i1prev : i1]];
            Face f = fc[e.f1];
            int v1 = boundaryVert[1][i1prev], v2 = boundaryVert[1][i1];
            if ((f.v1 == v1 && f.v2 == v2) || (f.v2 == v1 && f.v3 == v2) || (f.v3 == v1 && f.v1 == v2))
              newface[count++] = new int [] {v2, v1, boundaryVert[0][i0]};
            else
              newface[count++] = new int [] {v1, v2, boundaryVert[0][i0]};
          }
        i0prev = i0;
        i1prev = i1;
      }

    // Create the new mesh;

    TriangleMesh newmesh = new TriangleMesh(newvert, newface);
    Vertex newvt[] = (Vertex []) newmesh.getVertices();
    Edge newed[] = newmesh.getEdges();
    newmesh.setTexture(theMesh.getTexture());
    newmesh.setTextureMapping(theMesh.getTextureMapping());
    if (newmesh.isClosed() && theMesh.getMaterial() != null)
      {
	newmesh.setMaterial(theMesh.getMaterial());
	newmesh.setMaterialMapping(theMesh.getMaterialMapping());
      }
    newmesh.setSmoothingMethod(theMesh.getSmoothingMethod());

    // Copy over the smoothness values and texture parameters.

    for (int i = 0; i < vt.length; i++)
      {
	newvt[i].smoothness = vt[i].smoothness;
	newvt[i].param = vt[i].param;
      }
    for (int i = 0; i < newed.length; i++)
      {
	if (newed[i].v1 >= vt.length || newed[i].v2 >= vt.length)
	  continue;
	for (int j = 0; j < ed.length; j++)
	  if ((newed[i].v1 == ed[j].v1 && newed[i].v2 == ed[j].v2) || (newed[i].v1 == ed[j].v2 && newed[i].v2 == ed[j].v1))
	    newed[i].smoothness = ed[j].smoothness;
      }
    return newmesh;
  }

  public void extractCurveCommand()
  {
    TriangleMesh theMesh = (TriangleMesh) getCurrentView().getObject().object;
    Vertex vt[] = (Vertex []) theMesh.getVertices();
    Edge ed[] = theMesh.getEdges();
    boolean selected[] = ((TriMeshViewer) getCurrentView()).getSelection();
    Vector edges = new Vector();
    int i;

    if (((TriMeshViewer) getCurrentView()).getSelectionMode() != TriMeshViewer.EDGE_MODE)
      return;

    // Find the select edges, and try to chain them together.

    for (i = 0; i < selected.length; i++)
      if (selected[i])
        edges.addElement(ed[i]);
    if (edges.size() == 0)
      return;
    Edge first = (Edge) edges.elementAt(0), last = first;
    Vector ordered = new Vector();
    ordered.addElement(first);
    edges.removeElementAt(0);
    while (edges.size() > 0)
      {
        for (i = 0; i < edges.size(); i++)
          {
            Edge e = (Edge) edges.elementAt(i);
            if (e.v1 == first.v1 || e.v2 == first.v1 || e.v1 == first.v2 || e.v2 == first.v2)
              {
                ordered.insertElementAt(e, 0);
                first = e;
                break;
              }
            if (e.v1 == last.v1 || e.v2 == last.v1 || e.v1 == last.v2 || e.v2 == last.v2)
              {
                ordered.addElement(e);
                last = e;
                break;
              }
          }
        if (i == edges.size())
          {
            new MessageDialog(this, "The selected edges do not form a continuous curve.");
            return;
          }
        edges.removeElementAt(i);
      }

    // Now find the sequence of vertices.

    boolean closed = (ordered.size() > 2 && (last.v1 == first.v1 || last.v2 == first.v1 || last.v1 == first.v2 || last.v2 == first.v2));
    Vec3 v[] = new Vec3 [closed ? ordered.size() : ordered.size()+1];
    float smoothness[] = new float [v.length];
    Edge second = (ordered.size() == 1 ? first : (Edge) ordered.elementAt(1));
    int prev;
    if (first.v1 == second.v1 || first.v1 == second.v2)
      prev = first.v2;
    else
      prev = first.v1;
    for (i = 0; i < ordered.size(); i++)
      {
        Edge e = (Edge) ordered.elementAt(i);
        v[i] = new Vec3(vt[prev].r);
        smoothness[i] = vt[prev].smoothness;
        prev = (e.v1 == prev ? e.v2 : e.v1);
      }
    if (!closed)
      {
        v[i] = new Vec3(vt[prev].r);
        smoothness[i] = vt[prev].smoothness;
      }
    int smoothingMethod = theMesh.getSmoothingMethod();
    if (smoothingMethod == Mesh.SMOOTH_SHADING)
      smoothingMethod = Mesh.NO_SMOOTHING;
    Curve cv = new Curve(v, smoothness, smoothingMethod, closed);
    Component parent = (Component) parentWindow;
    while (parent != null && !(parent instanceof LayoutWindow))
      parent = parent.getParent();
    if (parent != null)
      {
        ((LayoutWindow) parent).addObject(cv, new CoordinateSystem(new Vec3(), new Vec3(0.0, 0.0, 1.0), new Vec3(0.0, 1.0, 0.0)),
            "Extracted Curve", null);
        ((LayoutWindow) parent).updateImage();
      }
  }

  public void setSmoothnessCommand()
  {
    final TriangleMesh theMesh = (TriangleMesh) getCurrentView().getObject().object;
    TriangleMesh oldMesh = (TriangleMesh) theMesh.duplicate();
    final Vertex vt[] = (Vertex []) theMesh.getVertices();
    final Edge ed[] = theMesh.getEdges();
    final boolean selected[] = ((TriMeshViewer) getCurrentView()).getSelection();
    final boolean pointmode = (((TriMeshViewer) getCurrentView()).getSelectionMode() == TriMeshViewer.POINT_MODE);
    final EventProcessor processor = new EventProcessor();
    float value;
    final ValueSlider smoothness;
    ComponentsDialog dlg;
    int i;

    for (i = 0; i < selected.length && !selected[i]; i++);
    if (i == selected.length)
      return;
    if (pointmode)
      value = vt[i].smoothness;
    else
      value = ed[i].smoothness;
    value = 0.001f * (Math.round(value*1000.0f));
    smoothness = new ValueSlider(0.0, 1.0, 100, (double) value);
    smoothness.addAdjustmentListener(new AdjustmentListener() {
      public void adjustmentValueChanged(AdjustmentEvent ev)
      {
        processor.addEvent(new Callback() {
          public void execute()
          {
            float s = (float) smoothness.getValue();
            for (int i = 0; i < selected.length; i++)
              if (selected[i])
                {
                  if (pointmode)
                    vt[i].smoothness = s;
                  else
                    ed[i].smoothness = s;
                }
            theMesh.setSmoothingMethod(theMesh.getSmoothingMethod());
            //object3DChangedDuringEditor();   // TODO(MB) During editor?
            getCurrentView().getObject().object.informChanged(this, ModelEvent.CHANGEDUR_OBJECTEDITORDRAG);
            getCurrentView().objectChanged();  // TODO(MB) Enhance message mechanism to remove these lines
            getCurrentView().updateImage();
            getCurrentView().repaint();
            
//            for (int i = 0; i < theView.length; ++i)
//            {
//              theView[i].updateImage();  // TODO(MB) Use message mechanism
//              theView[i].repaint();
//            }
          }
        } );
      }
    } );
    dlg = new ComponentsDialog(this, "Set Smoothness for Selected " +
    	(pointmode ? "Points" : "Edges"), new Component [] {smoothness},
    	new String [] {Translate.text("Smoothness")});
    processor.stopProcessing();
    if (dlg.clickedOk())
    {
      setUndoRecord(new UndoRecord(this, false, UndoRecord.COPY_OBJECT, new Object [] {theMesh, oldMesh}));
      object3DChangedDuringEditor();
    }
    else
      {
        theMesh.copyObject(oldMesh);
        object3DChangedDuringEditor();
        for (i = 0; i < theView.length; ++i)
        {
          theView[i].updateImage();
          theView[i].repaint();
        }
      }
  }

  public void reverseNormalsCommand()
  {
    TriangleMesh theMesh = (TriangleMesh) getCurrentView().getObject().object;
    setUndoRecord(new UndoRecord(this, false, UndoRecord.COPY_OBJECT, new Object [] {theMesh, theMesh.duplicate()}));
    theMesh.reverseNormals();
    object3DChangedDuringEditor();
  }

  void setSmoothingMethod(int method)
  {
    TriangleMesh theMesh = (TriangleMesh) getCurrentView().getObject().object;

    setUndoRecord(new UndoRecord(this, false, UndoRecord.COPY_OBJECT, new Object [] {theMesh, theMesh.duplicate()}));
    updateMenus();
    theMesh.setSmoothingMethod(method);
    object3DChangedDuringEditor();
  }

  /* Given a list of deltas which will be added to the selected vertices, calculate the
     corresponding deltas for the unselected vertices according to the mesh tension. */

  public void adjustDeltas(Vec3 delta[])
  {
    int dist[] = getCurrentView().getSelectionDistance(), count[] = new int [delta.length];
    TriangleMesh theMesh = (TriangleMesh) getCurrentView().getObject().object;
    TriangleMesh.Edge edge[] = theMesh.getEdges();
    int maxDistance = getTensionDistance();
    double tension = getMeshTension(), scale[] = new double [maxDistance+1];


    for (int i = 0; i < delta.length; i++)
      if (dist[i] != 0)
        delta[i].set(0.0, 0.0, 0.0);
    for (int i = 0; i < maxDistance; i++)
      {
	for (int j = 0; j < count.length; j++)
	  count[j] = 0;
	for (int j = 0; j < edge.length; j++)
	  {
	    if (dist[edge[j].v1] == i && dist[edge[j].v2] == i+1)
	      {
		count[edge[j].v2]++;
		delta[edge[j].v2].add(delta[edge[j].v1]);
	      }
	    else if (dist[edge[j].v2] == i && dist[edge[j].v1] == i+1)
	      {
		count[edge[j].v1]++;
		delta[edge[j].v1].add(delta[edge[j].v2]);
	      }
	  }
	for (int j = 0; j < count.length; j++)
	  if (count[j] > 1)
	    delta[j].scale(1.0/count[j]);
      }
    for (int i = 0; i < scale.length; i++)
      scale[i] = Math.pow((maxDistance-i+1.0)/(maxDistance+1.0), tension);
    for (int i = 0; i < delta.length; i++)
      if (dist[i] > 0)
        delta[i].scale(scale[dist[i]]);
  }


}