/* Copyright (C) 2002-2003 by Peter Eastman

   This program is free software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the Free Software
   Foundation; either version 2 of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but WITHOUT ANY 
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
   PARTICULAR PURPOSE.  See the GNU General Public License for more details. */

package artofillusion.animation;

import artofillusion.*;
import artofillusion.animation.Actor.*;
import artofillusion.math.*;
import artofillusion.object.*;
import artofillusion.ui.*;
import java.awt.*;
import java.awt.event.*;
import java.text.*;

/** The ActorEditorWindow class represents the window for editing Actor objects. */

public class ActorEditorWindow extends Dialog implements ActionListener, ItemListener, TextListener
{
  Actor theObject, oldObject;
  ObjectInfo objInfo;
  EditingWindow theWindow;
  List gestureList, currentPoseList;
  Button addButton, removeButton, saveButton, extractButton;
  Button editButton, renameButton, copyButton, deleteButton;
  Button okButton, cancelButton;
  ValueField weightField;
  ActorKeyframe currentPose, key;
  ObjectPreviewCanvas preview;
  Callback onClose;
  
  /** Display a window for editing a pose of an Actor.  If key is null, the
      Actor's current pose will be edited.  Otherwise, the specified keyframe
      will be edited. */
  
  public ActorEditorWindow(EditingWindow parent, ObjectInfo info, Actor obj, ActorKeyframe key, Callback cb)
  {
    super((Frame) parent, (key == null ? info.name : "Edit Keyframe for "+info.name), false);
    objInfo = info;
    oldObject = obj;
    theObject = (Actor) obj.duplicate();
    theWindow = parent;
    this.key = key;
    if (key != null)
      theObject.applyPoseKeyframe(key);
    currentPose = (ActorKeyframe) theObject.getPoseKeyframe();
    onClose = cb;
    for (int i = currentPose.id.length-1; i >= 0; i--)
      {
        int which = theObject.findPoseIndex(currentPose.id[i]);
        if (which == -1)
          currentPose.deletePose(i);
      }
    
    // Layout the components in the window.
    
    Panel center = new Panel(), p;
    add(center, BorderLayout.CENTER);
    center.setLayout(new GridBagLayout());
    
    // The list of gestures and buttons for editing them.
    
    GridBagConstraints gc = new GridBagConstraints();
    GridBagConstraints gc2 = new GridBagConstraints();
    gc2.gridy = 0;
    gc2.fill = gc2.BOTH;
    gc2.weightx = 1.0;
    gc2.gridheight = 2;
    center.add(p = new Panel(), gc2);
    p.setLayout(new GridBagLayout());
    gc.gridx = 0;
    gc.weightx = 1.0;
    gc.weighty = 1.0;
    gc.fill = gc.BOTH;
    p.add(gestureList = new List(10), gc);
    for (int i = 0; i < theObject.gestureName.length; i++)
      gestureList.add(theObject.gestureName[i]);
    gc.weighty = 0.0;
    gc.fill = gc.HORIZONTAL;
    gc.insets.top = gc.insets.left = gc.insets.right = 5;
    p.add(editButton = new Button("Edit"), gc);
    p.add(renameButton = new Button("Rename"), gc);
    p.add(copyButton = new Button("Duplicate"), gc);
    p.add(deleteButton = new Button("Delete"), gc);
    
    // The Add and Remove buttons.
    
    gc2.gridheight = 1;
    center.add(p = new Panel(), gc2);
    p.setLayout(new GridBagLayout());
    gc.weightx = 0.0;
    p.add(addButton = new Button("Add >>"), gc);
    p.add(removeButton = new Button("<< Remove"), gc);
    p.add(saveButton = new Button("Save..."), gc);
    p.add(extractButton = new Button("Extract..."), gc);
    
    // The current pose.
    
    center.add(p = new Panel(), gc2);
    p.setLayout(new GridBagLayout());
    gc.anchor = gc.NORTH;
    gc.weightx = 1.0;
    gc.gridwidth = 2;
    gc.insets.top = gc.insets.left = gc.insets.right = 0;
    p.add(new ShortLabel("Current Pose:"), gc);
    gc.weighty = 1.0;
    gc.fill = gc.BOTH;
    p.add(currentPoseList = new List(4), gc);
    gc.weightx = gc.weighty = 0;
    gc.gridwidth = 1;
    gc.fill = gc.NONE;
    p.add(new Label("Weight: "), gc);
    gc.gridx = 1;
    p.add(weightField = new ValueField(Double.NaN, ValueField.NONE, 5), gc);
    weightField.addTextListener(this);
    buildCurrentPoseList();
    
    // The object previewer.
    
    gc2.gridwidth = 2;
    gc2.gridy = gc2.gridx = 1;
    gc2.weighty = 1.0;
    center.add(preview = new ObjectPreviewCanvas(new ObjectInfo(theObject, new CoordinateSystem(), "")), gc2);
    
    // The buttons at the bottom.
    
    add(p = new Panel(), BorderLayout.SOUTH);
    p.add(okButton = Translate.button("ok", this));
    p.add(cancelButton = Translate.button("cancel", this));
    editButton.addActionListener(this);
    renameButton.addActionListener(this);
    copyButton.addActionListener(this);
    deleteButton.addActionListener(this);
    addButton.addActionListener(this);
    removeButton.addActionListener(this);
    saveButton.addActionListener(this);
    extractButton.addActionListener(this);
    gestureList.addItemListener(this);
    gestureList.addActionListener(this);
    currentPoseList.addItemListener(this);
    pack();
    updateComponents();
    show();
  }
  
  /** Build the list showing the current pose. */
  
  private void buildCurrentPoseList()
  {
    NumberFormat nf = NumberFormat.getNumberInstance();
    nf.setMaximumFractionDigits(4);
    currentPoseList.removeAll();
    for (int i = 0; i < currentPose.id.length; i++)
      {
        int which = theObject.findPoseIndex(currentPose.id[i]);
        currentPoseList.add(theObject.gestureName[which]+" ("+nf.format(currentPose.weight[i])+")");
      }
    int items = currentPoseList.getItemCount();
    if (items > 0)
      {
        currentPoseList.select(items-1);
        weightField.setValue(currentPose.weight[items-1]);
      }
    else
      weightField.setValue(Double.NaN);
  }
  
  /** Deal with button clicks. */
  
  public void actionPerformed(ActionEvent ev)
  {
    if (ev.getSource() == addButton)
      {
        // Add a new gesture to the current pose.
        
        currentPose.addPose(theObject.gestureID[gestureList.getSelectedIndex()], 1.0);
        buildCurrentPoseList();
        updateComponents();
        updateDisplay();
      }
    else if (ev.getSource() == removeButton)
      {
        // Remove a gesture from the current pose.
        
        currentPose.deletePose(currentPoseList.getSelectedIndex());
        buildCurrentPoseList();
        updateComponents();
        updateDisplay();
      }
    else if (ev.getSource() == saveButton)
      {
        // Save the current pose as a new gesture.
        
        EntriesDialog dlg = new EntriesDialog(this, "Save the current pose as a new gesture:",
          new String [] {Translate.text("Name")}, new String [] {"New Gesture"});
        if (dlg.getEntries() == null)
          return;
        theObject.addGesture((Gesture) theObject.theObject.getPoseKeyframe(), dlg.getEntries()[0]);
        gestureList.add(dlg.getEntries()[0]);
        gestureList.select(gestureList.getItemCount()-1);
        updateComponents();
      }
    else if (ev.getSource() == extractButton)
      {
        // Save the current pose as a new object.
        
        EntriesDialog dlg = new EntriesDialog(this, "Save the current pose as a new object:",
          new String [] {Translate.text("Name")}, new String [] {"Extracted Pose"});
        if (dlg.getEntries() == null)
          return;
        ObjectInfo info = new ObjectInfo(theObject.theObject.duplicate(), 
          new CoordinateSystem(new Vec3(), Vec3.vz(), Vec3.vy()), dlg.getEntries()[0]);
        info.addTrack(new PositionTrack(info), 0);
        info.addTrack(new RotationTrack(info), 1);
        if (theWindow instanceof LayoutWindow)
          ((LayoutWindow) theWindow).addObject(info, null);
        else
          theWindow.getScene().addObject(info, null);
        theWindow.updateImage();
      }
    else if (ev.getSource() == editButton || ev.getSource() == gestureList)
      {
        // Edit a gesture.
        editGesture(gestureList.getSelectedIndex());
      }
    else if (ev.getSource() == renameButton)
      {
        // Rename a gesture.
        
        int which = gestureList.getSelectedIndex();
        EntriesDialog dlg = new EntriesDialog(this, "Enter a new name for the gesture:",
          new String [] {Translate.text("Name")}, new String [] {theObject.gestureName[which]});
        if (dlg.getEntries() == null)
          return;
        theObject.gestureName[which] = dlg.getEntries()[0];
        gestureList.replaceItem(theObject.gestureName[which], which);
        gestureList.select(which);
      }
    else if (ev.getSource() == copyButton)
      {
        // Duplicate a gesture.
        
        int which = gestureList.getSelectedIndex();
        EntriesDialog dlg = new EntriesDialog(this, "Enter name for new gesture:",
          new String [] {Translate.text("Name")}, new String [] {"Copy of "+theObject.gestureName[which]});
        if (dlg.getEntries() == null)
          return;
        theObject.addGesture((Gesture) theObject.gesture[which].duplicate(theObject.theObject), dlg.getEntries()[0]);
        gestureList.add(dlg.getEntries()[0]);
        gestureList.select(gestureList.getItemCount()-1);
        updateComponents();
        editGesture(gestureList.getItemCount()-1);
      }
    else if (ev.getSource() == deleteButton)
      {
        // Delete a gesture.
        
        int which = gestureList.getSelectedIndex();
        MessageDialog dlg = new MessageDialog(this, "Delete '"+theObject.gestureName[which]+"'?",
          new String [] {"Yes", "No"});
        if (dlg.getChoice() == 1)
          return;
        for (int i = 0; i < currentPose.id.length; i++)
          if (currentPose.id[i] == theObject.gestureID[which])
            currentPose.deletePose(i);
        theObject.deleteGestureWithID(theObject.gestureID[which]);
        gestureList.remove(which);
        buildCurrentPoseList();
        updateComponents();
      }
    else if (ev.getSource() == okButton)
      {
        // Save the edits.
        
        oldObject.copyObject(theObject);
        if (key != null)
          {
            key.copy(currentPose);
            theWindow.getScene().applyTracksToObject(objInfo);
            theWindow.updateImage();
          }
        dispose();
        if (onClose != null)
          onClose.execute();
      }
    else
      dispose();
  }
  
  /** Display a window to let the user edit a gesture. */
  
  private void editGesture(int which)
  {
    setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
    Object3D editObj = theObject.theObject.duplicate();
    editObj.applyPoseKeyframe(theObject.gesture[which]);
    EditCallback cb = new EditCallback(editObj, theObject.gestureID[which]);
    ObjectInfo info = objInfo.duplicate(editObj);
    info.name = theObject.gestureName[which];
    editObj.editGesture(theWindow, info, cb, objInfo);
    setCursor(Cursor.getDefaultCursor());
  }
  
  /** Respond to selection changes on the two lists. */
  
  public void itemStateChanged(ItemEvent ev)
  {
    updateComponents();
  }
  
  /** When the weight is changed, update the current pose. */

  public void textValueChanged(TextEvent e)
  {
    int selected = currentPoseList.getSelectedIndex();
    
    if (selected == -1 || currentPose.weight[selected] == weightField.getValue())
      return;
    currentPose.weight[selected] = weightField.getValue();
    int which = theObject.findPoseIndex(currentPose.id[selected]);
    currentPoseList.replaceItem(theObject.gestureName[which]+" ("+currentPose.weight[selected]+")", selected);
    currentPoseList.select(selected);
    updateDisplay();
  }
  
  /** Update the status of the buttons and weight field. */
  
  private void updateComponents()
  {
    int i, id, selected = gestureList.getSelectedIndex();
    
    if (selected == -1)
      {
        addButton.setEnabled(false);
        editButton.setEnabled(false);
        renameButton.setEnabled(false);
        copyButton.setEnabled(false);
        deleteButton.setEnabled(false);
      }
    else
      {
        editButton.setEnabled(true);
        renameButton.setEnabled(selected > 0);
        copyButton.setEnabled(true);
        deleteButton.setEnabled(selected > 0);
        id = theObject.gestureID[gestureList.getSelectedIndex()];
        for (i = 0; i < currentPose.id.length && currentPose.id[i] != id; i++);
        addButton.setEnabled(selected > 0 && i == currentPose.id.length);
      }
    selected = currentPoseList.getSelectedIndex();
    removeButton.setEnabled(selected != -1);
    if (selected == -1)
      {
        weightField.setValue(Double.NaN);
        weightField.setEnabled(false);
      }
    else
      {
        weightField.setValue(currentPose.weight[currentPoseList.getSelectedIndex()]);
        weightField.setEnabled(true);
      }
  }
  
  // Reapply the current pose and update the preview.
  
  private void updateDisplay()
  {
    theObject.applyPoseKeyframe(currentPose);
    preview.objectChanged();
    preview.updateImage();
  }
  
  // This is an inner class which is called when a keyframe has been edited.
  
  private class EditCallback implements Callback
  {
    public Object3D editObject;
    public int editID;
    
    public EditCallback(Object3D obj, int id)
    {
      editObject = obj;
      editID = id;
    }
    
    public void execute()
    {
      int which = theObject.findPoseIndex(editID);
      theObject.theObject.copyObject(editObject);
      theObject.gesture[which] = (Gesture) theObject.theObject.getPoseKeyframe();
      
      // Update the skeletons for all other poses, in case a joint has been added or deleted.
      
      Skeleton s = theObject.gesture[which].getSkeleton();
      if (s != null)
        for (int i = 0; i < theObject.gesture.length; i++)
          {
            if (i == which)
              continue;
            Skeleton olds = theObject.gesture[i].getSkeleton();
            Skeleton news = s.duplicate();
            Joint joint[] = news.getJoints();
            for (int j = 0; j < joint.length; j++)
              {
                Joint j2 = olds.getJoint(joint[j].id);
                if (j2 != null)
                  joint[j].copy(j2);
              }
            theObject.gesture[i].setSkeleton(news);
          }
      updateDisplay();
    }
  }
}
