/* EditKeyframesDialog is a dialog box for doing bulk editing of keyframes.  It is used 
   for the Move, Copy, Rescale, Loop, and Delete commands */

/* Copyright (C) 2001-2002 by Peter Eastman

   This program is free software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the Free Software
   Foundation; either version 2 of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but WITHOUT ANY 
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
   PARTICULAR PURPOSE.  See the GNU General Public License for more details. */

package artofillusion.animation;

import artofillusion.*;
import artofillusion.object.*;
import artofillusion.ui.*;
import java.awt.*;
import java.awt.event.*;
import java.text.*;
import java.util.*;

public class EditKeyframesDialog extends Dialog implements ActionListener
{
  LayoutWindow window;
  Scene theScene;
  Choice tracksChoice;
  ValueField startField, endField, extraField; 
  Button okButton, cancelButton;
  int operation;
  
  public static final int MOVE = 0;
  public static final int COPY = 1;
  public static final int RESCALE = 2;
  public static final int LOOP = 3;
  public static final int DELETE = 4;

  private static final String opTitle[] = new String [] {"Move", "Copy", "Rescale", "Loop", "Delete"};

  public EditKeyframesDialog(LayoutWindow win, int operation)
  {
    super(win, opTitle[operation]+" Keyframes", true);
    window = win;
    this.operation = operation;
    theScene = window.getScene();
    
    // Layout the dialog.

    setLayout(new GridBagLayout());
    GridBagConstraints gc = new GridBagConstraints();
    gc.gridx = 0;
    Panel p = new Panel();
    add(p, gc);
    p.add(new Label("Apply to "), gc);
    p.add(tracksChoice = new Choice(), gc);
    tracksChoice.add("All Tracks");
    tracksChoice.add("All Tracks of Selected Objects");
    tracksChoice.add("Selected Tracks");
    if (window.getScore().getSelectedTracks().length > 0)
      tracksChoice.select(2);
    else if (theScene.getSelection().length > 0)
      tracksChoice.select(1);
    add(p = new Panel(), gc);
    p.add(new Label(opTitle[operation]+" keyframes between time "));
    p.add(startField = new ValueField(0.0, ValueField.NONE, 5));
    add(p = new Panel(), gc);
    p.add(new Label(" and time "));
    p.add(endField = new ValueField(0.0, ValueField.NONE, 5));
    add(p = new Panel(), gc);
    if (operation == MOVE || operation == COPY)
      p.add(new Label("to time "));
    else if (operation == RESCALE)
      p.add(new Label("by a factor of "));
    else if (operation == LOOP)
      p.add(new Label("Number of times to loop: "));
    if (operation == LOOP)
      extraField = new ValueField(2.0, ValueField.INTEGER+ValueField.POSITIVE, 5);
    else
      extraField = new ValueField(1.0, ValueField.NONE, 5);
    if (operation != DELETE)
      p.add(extraField);

    KeyListener listener = new KeyAdapter() {
      public void keyPressed(KeyEvent ev)
      {
        if (ev.getKeyCode() == KeyEvent.VK_ENTER)
          {
            if (doEdit())
              dispose();
          }
        if (ev.getKeyCode() == KeyEvent.VK_ESCAPE)
          dispose();
      }
    };
    tracksChoice.addKeyListener(listener);
    startField.addKeyListener(listener);
    endField.addKeyListener(listener);
    extraField.addKeyListener(listener);
    addWindowListener(new WindowAdapter() {
      public void windowClosing(WindowEvent e)
      {
        dispose();
      }
    } );
    
    add(p = new Panel(), gc);
    p.add(okButton = Translate.button("ok", this));
    p.add(cancelButton = Translate.button("cancel", this));
    pack();
    setResizable(false);
    ModellingApp.centerWindow(this);
    show();
  }

  public void actionPerformed(ActionEvent e)
  {
    if (e.getSource() == cancelButton || doEdit())
      dispose();
  }
  
  /* Perform the operation.  Returns true if it was successful, false if it was canceled. */
  
  private boolean doEdit()
  {
    int whichTracks = tracksChoice.getSelectedIndex();
    double fps = (double) theScene.getFramesPerSecond();
    Track track[];
    
    // Find which tracks to apply the operation to.
    
    if (whichTracks == 0)
      {
        Vector tracks = new Vector();
        for (int i = 0; i < theScene.getNumObjects(); i++)
          {
            ObjectInfo info = theScene.getObject(i);
            for (int j = 0; j < info.tracks.length; j++)
              addToVector(info.tracks[j], tracks);
          }
        track = new Track [tracks.size()];
        for (int i = 0; i < track.length; i++)
          track[i] = (Track) tracks.elementAt(i);
      }
    else if (whichTracks == 1)
      {
        int sel[] = theScene.getSelection();
        Vector tracks = new Vector();
        for (int i = 0; i < sel.length; i++)
          {
            ObjectInfo info = theScene.getObject(sel[i]);
            for (int j = 0; j < info.tracks.length; j++)
              addToVector(info.tracks[j], tracks);
          }
        track = new Track [tracks.size()];
        for (int i = 0; i < track.length; i++)
          track[i] = (Track) tracks.elementAt(i);
      }
    else
      track = window.getScore().getSelectedTracks();
    
    // Find the range of times affected by the operation.
    
    double start = startField.getValue(), end = endField.getValue();
    double extra = extraField.getValue(), scaleStart = 0.0, scaleFactor = 0.0;
    int numLoops = 0;
    
    if (start > end)
      {
        double temp = start;
        start = end;
        end = temp;
      }
    double targetStart, targetEnd;
    if (operation == MOVE || operation == COPY)
      {
        targetStart = extra;
        targetEnd = targetStart + (end-start);
        if (targetStart > start && targetStart < end)
          targetStart = end;
        if (targetEnd > start && targetEnd < end)
          targetEnd = start;
      }
    else if (operation == RESCALE)
      {
        scaleFactor = extra;
        targetStart = end;
        targetEnd = start + Math.abs(scaleFactor)*(end-start);
        if (scaleFactor < 0.0)
          scaleStart = targetEnd;
        else
          scaleStart = start;
        if (targetEnd < targetStart)
          targetEnd = targetStart;
      }
    else if (operation == LOOP)
      {
        numLoops = (int) extra;
        targetStart = end;
        targetEnd = start + numLoops*(end-start);
      }
    else
      targetStart = targetEnd = start;
    
    // Determine whether there are any existing keyframes in the target range.
    
    boolean any = false;
    for (int i = 0; !any && i < track.length; i++)
      {
        double t[] = track[i].getKeyTimes();
        for (int j = 0; !any && j < t.length; j++)
          if (t[j] > targetStart && t[j] < targetEnd)
            any = true;
      }
    if (any)
      {
        NumberFormat nf = NumberFormat.getNumberInstance();
        nf.setMaximumFractionDigits(3);
        MessageDialog dlg = new MessageDialog(window, new String [] {"There are existing keyframes in the target range for",
          "this operation (from time "+nf.format(targetStart)+" to time "+nf.format(targetEnd)+").  If you continue,",
          "they will be deleted.  Do you want to continue?"}, new String []
          {"Yes", "No"});
        if (dlg.getChoice() == 1)
          return false;
      }
    
    // Prepare an undo record.
    
    UndoRecord undo = new UndoRecord(window, false);
    for (int i = 0; i < track.length; i++)
      {
        Object parent = track[i].getParent();
        Track dup;
        if (parent instanceof ObjectInfo)
          dup = track[i].duplicate((ObjectInfo) parent);
        else
          dup = track[i].duplicate(null);
        undo.addCommand(UndoRecord.COPY_TRACK, new Object [] {track[i], dup});
      }
    
    // Perform the operation.
    
    for (int i = 0; i < track.length; i++)
      {
        // Delete existing keyframes in the target region.
        
        double t[] = track[i].getKeyTimes();
        for (int j = t.length-1; j >= 0; j--)
          if (t[j] > targetStart && t[j] < targetEnd)
            track[i].deleteKeyframe(j);
        
        // Create the new keyframes.
        
        Timecourse tc = track[i].getTimecourse();
        if (tc == null)
          continue;
        Keyframe key[] = tc.getValues();
        Smoothness s[] = tc.getSmoothness();
        t = track[i].getKeyTimes();
        if (operation == MOVE || operation == RESCALE || operation == DELETE)
          for (int j = t.length-1; j >=0 ; j--)
            if (t[j] >= start && t[j] <= end)
              track[i].deleteKeyframe(j);
        if (operation == MOVE || operation == COPY)
          for (int j = 0; j < t.length; j++)
            if (t[j] >= start && t[j] <= end)
              track[i].setKeyframe(roundTime(t[j]-start+extra, fps), key[j].duplicate(), s[j]);
        if (operation == RESCALE)
          for (int j = 0; j < t.length; j++)
            if (t[j] >= start && t[j] <= end)
              track[i].setKeyframe(roundTime(scaleStart+scaleFactor*(t[j]-start), fps), key[j].duplicate(), s[j]);
        if (operation == LOOP)
          for (int j = 0; j < t.length; j++)
            if (t[j] >= start && t[j] <= end)
              for (int k = 1; k < numLoops; k++)
                track[i].setKeyframe(roundTime(t[j]+k*(end-start), fps), key[j].duplicate(), s[j]);
      }
    window.getScore().rebuildList();
    window.setTime(window.getScene().getTime());
    window.setUndoRecord(undo);
    return true;
  }
  
  /* Given a Track, add it and all of its subtracks to the specified vector. */
  
  private void addToVector(Track tr, Vector v)
  {
    Track sub[] = tr.getSubtracks();
    
    v.addElement(tr);
    for (int i = 0; i < sub.length; i++)
      addToVector(sub[i], v);
  }
  
  /* Round a time to the nearest frame. */
  
  private double roundTime(double t, double fps)
  {
    return Math.round(t*fps)/fps;
  }
}
