/* Copyright (C) 2003 by Peter Eastman

   This program is free software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the Free Software
   Foundation; either version 2 of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but WITHOUT ANY 
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
   PARTICULAR PURPOSE.  See the GNU General Public License for more details. */

package artofillusion.animation;

import artofillusion.*;
import artofillusion.animation.distortion.*;
import artofillusion.math.*;
import artofillusion.object.*;
import artofillusion.ui.*;
import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.util.Vector;

/** This is a Track which modifies the shape of an object using inverse kinematics. */

public class IKTrack extends Track
{
  ObjectInfo info;
  Vector constraints;
  WeightTrack theWeight;
  
  public IKTrack(ObjectInfo info)
  {
    super("Inverse Kinematics");
    this.info = info;
    theWeight = new WeightTrack(this);
    constraints = new Vector();
  }

  /** This method presents a window in which the user can edit the track. */
  
  public void edit(LayoutWindow win)
  {
    if (info.getSkeleton() == null)
    {
      new MessageDialog(win, "Inverse Kinematics tracks have no effect on objects without skeletons.");
      return;
    }
    new Editor(win);
  }
  
  /** Modify the position of the object. */
  
  public void apply(double time)
  {
    double weight = theWeight.getWeight(time); // DEAL WITH WEIGHT!!!!!
    Skeleton skeleton = info.getSkeleton();
    if (skeleton == null)
      return;
    Joint joint[] = skeleton.getJoints();
    if (joint.length == 0)
      return;
    boolean locked[] = new boolean [joint.length];
    Vec3 target[] = new Vec3 [joint.length];
    Mat4 toLocal = info.coords.toLocal();
    
    for (int i = 0; i < constraints.size(); i++)
    {
      Constraint c = (Constraint) constraints.elementAt(i);
      int index = skeleton.findJointIndex(c.jointID);
      if (index == -1)
        continue;
      if (c.target == null)
        locked[index] = true;
      else
        target[index] = toLocal.times(c.target.getCoords().getOrigin());
    }
    info.addDistortion(new IKDistortion(locked, target, weight));
  }
  
  /** Create a duplicate of this track. */
  
  public Track duplicate(Object obj)
  {
    IKTrack t = new IKTrack((ObjectInfo) obj);
    
    t.constraints = new Vector();
    for (int i = 0; i < constraints.size(); i++)
    {
      Constraint c = (Constraint) constraints.elementAt(i);
      t.constraints.addElement(c.duplicate());
    }
    t.theWeight = (WeightTrack) theWeight.duplicate(t);
    return t;
  }
  
  /** Make this track identical to another one. */
  
  public void copy(Track tr)
  {
    IKTrack t = (IKTrack) tr;
    
    constraints = new Vector();
    for (int i = 0; i < t.constraints.size(); i++)
    {
      Constraint c = (Constraint) t.constraints.elementAt(i);
      constraints.addElement(c.duplicate());
    }
    theWeight = (WeightTrack) t.theWeight.duplicate(t);
  }
  
  /** Get a list of all keyframe times for this track. */
  
  public double [] getKeyTimes()
  {
    return new double [0];
  }

  /** Move a keyframe to a new time, and return its new position in the list. */
  
  public int moveKeyframe(int which, double time)
  {
    return -1;
  }
  
  /** Delete the specified keyframe. */
  
  public void deleteKeyframe(int which)
  {
  }
  
  /** This track is null if it has no targets. */
  
  public boolean isNullTrack()
  {
    for (int i = 0; i < constraints.size(); i++)
      if (((Constraint) constraints.elementAt(i)).target != null)
        return false;
    return true;
  }
  
  /** This has a single child track. */
  
  public Track [] getSubtracks()
  {
    return new Track [] {theWeight};
  }

  /** Determine whether this track can be added as a child of an object. */
  
  public boolean canAcceptAsParent(Object obj)
  {
    return (obj instanceof ObjectInfo && ((ObjectInfo) obj).object instanceof Mesh);
  }
  
  /** Get the parent object of this track. */
  
  public Object getParent()
  {
    return info;
  }
  
  /** Set the parent object of this track. */
  
  public void setParent(Object obj)
  {
    info = (ObjectInfo) obj;
  }
  
  /** Get an array of any objects which this track depends on (and which therefore must
      be updated before this track is applied). */ 
  
  public ObjectInfo [] getDependencies()
  {
    Vector v = new Vector();
    for (int i = 0; i < constraints.size(); i++)
    {
      Constraint c = (Constraint) constraints.elementAt(i);
      if (c.target != null)
        v.addElement(c.target.getObject());
    }
    ObjectInfo ref[] = new ObjectInfo [v.size()];
    v.copyInto(ref);
    return ref;
  }
  
  /** Delete all references to the specified object from this track.  This is used when an
      object is deleted from the scene. */
  
  public void deleteDependencies(ObjectInfo obj)
  {
    for (int i = constraints.size()-1; i >= 0; i--)
    {
      Constraint c = (Constraint) constraints.elementAt(i);
      if (c.target != null && c.target.getObject() == obj)
        constraints.removeElementAt(i);
    }
  }

  /** Write a serialized representation of this track to a stream. */
  
  public void writeToStream(DataOutputStream out, Scene scene) throws IOException
  {
    out.writeShort(0); // Version number
    out.writeUTF(name);
    out.writeBoolean(enabled);
    out.writeInt(constraints.size());
    for (int i = 0; i < constraints.size(); i++)
    {
      Constraint c = (Constraint) constraints.elementAt(i);
      out.writeInt(c.jointID);
      out.writeBoolean(c.target != null);
      if (c.target != null)
        c.target.writeToStream(out);
    }
    theWeight.writeToStream(out, scene);
  }
  
  /** Initialize this tracked based on its serialized representation as written by writeToStream(). */
  
  public void initFromStream(DataInputStream in, Scene scene) throws IOException, InvalidObjectException
  {
    short version = in.readShort();
    
    if (version != 0)
      throw new InvalidObjectException("");
    name = in.readUTF();
    enabled = in.readBoolean();
    int numConstraints = in.readInt();
    constraints = new Vector();
    for (int i = 0; i < numConstraints; i++)
    {
      Constraint c = new Constraint(in.readInt(), in.readBoolean() ? new ObjectRef(in, scene) : null);
      constraints.addElement(c);
    }
    theWeight.initFromStream(in, scene);
  }
  
  /** This inner class represents a single constraint. */
  
  private class Constraint
  {
    public int jointID;
    public ObjectRef target;
    
    public Constraint(int id, ObjectRef t)
    {
      jointID = id;
      target = t;
    }
    
    public Constraint duplicate()
    {
      if (target == null)
        return new Constraint(jointID, null);
      return new Constraint(jointID, target.duplicate());
    }
  }

  /** Inner class for editing constraint tracks. */
  
  private class Editor extends Dialog implements ActionListener, ItemListener
  {
    LayoutWindow window;
    List constraintList;
    TextField nameField;
    Vector tempConstraints;
    int tempJointID[];
    ObjectRef tempTarget[];
    Button editButton, deleteButton;
    
    public Editor(LayoutWindow win)
    {
      super(win, "Inverse Kinematics Track", true);
      window = win;
      tempConstraints = new Vector();
      for (int i = 0; i < constraints.size(); i++)
        tempConstraints.addElement(((Constraint) constraints.elementAt(i)).duplicate());
      
      // Layout the dialog.
      
      setLayout(new GridBagLayout());
      GridBagConstraints gc = new GridBagConstraints();
      gc.gridx = 0;
      gc.anchor = gc.EAST;
      gc.insets.right = 5;
      add(new Label("Track Name:"), gc);
      add(new Label("Constraints:"), gc);
      gc.gridx = 1;
      gc.fill = gc.HORIZONTAL;
      add(nameField = new TextField(IKTrack.this.getName(), 30), gc);
      gc.fill = gc.BOTH;
      add(constraintList = new List(), gc);
      constraintList.addItemListener(this);
      constraintList.addActionListener(this);
      buildConstraintList();
      Panel p = new Panel();
      p.add(Translate.button("add", "...", this));
      p.add(editButton = Translate.button("edit", "...", this));
      p.add(deleteButton = Translate.button("delete", this));
      p.add(Translate.button("ok", this));
      p.add(Translate.button("cancel", this));
      gc.anchor = gc.CENTER;
      gc.gridx = 0;
      gc.gridwidth = 2;
      add(p, gc);
      pack();
      ModellingApp.centerWindow(this);
      editButton.setEnabled(false);
      deleteButton.setEnabled(false);
      show();
    }
    
    /** Fill in the list of constraints. */
    
    private void buildConstraintList()
    {
      Skeleton skeleton = info.getSkeleton();
      constraintList.removeAll();
      if (tempConstraints.size() == 0)
      {
        constraintList.add("(No Constraints)");
        constraintList.setEnabled(false);
        return;
      }
      constraintList.setEnabled(true);
      for (int i = 0; i < tempConstraints.size(); i++)
      {
        Constraint c = (Constraint) tempConstraints.elementAt(i);
        Joint j = skeleton.getJoint(c.jointID);
        if (c.target == null)
          constraintList.add(j.name+" is locked");
        else
          constraintList.add(j.name+" follows "+c.target.toString());
      }
    }
    
    /** Deal with button clicks. */
    
    public void actionPerformed(ActionEvent ev)
    {
      String command = ev.getActionCommand();
      
      if ("add".equals(command))
      {
        Constraint c = new Constraint(-1, null);
        if (editConstraint(c))
        {
          tempConstraints.addElement(c);
          buildConstraintList();
        }
      }
      else if ("edit".equals(command) || ev.getSource() == constraintList)
      {
        if (constraintList.getSelectedIndex() > -1)
        {
          Constraint c = (Constraint) tempConstraints.elementAt(constraintList.getSelectedIndex());
          if (editConstraint(c))
            buildConstraintList();
        }
      }
      else if ("delete".equals(command))
      {
        if (constraintList.getSelectedIndex() > -1)
          tempConstraints.removeElementAt(constraintList.getSelectedIndex());
        buildConstraintList();
      }
      else if ("ok".equals(command))
        {
          window.setUndoRecord(new UndoRecord(window, false, UndoRecord.COPY_OBJECT_INFO, new Object [] {info, info.duplicate()}));
          IKTrack.this.setName(nameField.getText());
          constraints = tempConstraints;
          window.getScore().repaintAll();
          dispose();
        }
      else
        dispose();
    }
    
    /** Update the components when a selection changes. */
    
    public void itemStateChanged(ItemEvent ev)
    {
      editButton.setEnabled(constraintList.getSelectedIndex() > -1);
      deleteButton.setEnabled(constraintList.getSelectedIndex() > -1);
    }
    
    /** Display a window for editing a single constraint. */
    
    private boolean editConstraint(Constraint c)
    {
      Skeleton skeleton = info.getSkeleton();
      Joint joint[] = skeleton.getJoints();
      List jointList = new List();
      for (int i = 0; i < joint.length; i++)
        jointList.add(joint[i].name);
      if (skeleton.findJointIndex(c.jointID) > -1)
        jointList.select(skeleton.findJointIndex(c.jointID));
      CheckboxGroup group = new CheckboxGroup();
      final Checkbox lockedBox = new Checkbox("Locked", c.target == null, group);
      final Checkbox targetBox = new Checkbox("Follows Target", c.target != null, group);
      Panel boxPanel = new Panel(new FlowLayout(FlowLayout.LEFT));
      boxPanel.add(lockedBox);
      boxPanel.add(targetBox);
      ObjectRef tempRef = (c.target == null ? new ObjectRef() : c.target);
      final ObjectRefSelector selector = new ObjectRefSelector(tempRef, window, "Target for Joint:", info);
      ItemListener listener = new ItemListener() {
        public void itemStateChanged(ItemEvent ev)
        {
          selector.setEnabled(ev.getSource() == targetBox);
        }
      };
      lockedBox.addItemListener(listener);
      targetBox.addItemListener(listener);
      selector.setEnabled(c.target != null);
      ComponentsDialog dlg = new ComponentsDialog(window, "Edit Constraint", new Component []
          {jointList, boxPanel, selector}, new String [] {"Apply Constraint To:", "Constraint Type:",
          "Target:"});
      if (!dlg.clickedOk() || jointList.getSelectedIndex() == -1)
        return false;
      if (jointList.getSelectedIndex() > -1)
        c.jointID = joint[jointList.getSelectedIndex()].id;
      if (lockedBox.getState())
        c.target = null;
      else
        c.target = selector.getSelection();
      return true;
    }
  }
}
