
/* Copyright (C) 2002-2003 by Peter Eastman

   This program is free software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the Free Software
   Foundation; either version 2 of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
   PARTICULAR PURPOSE.  See the GNU General Public License for more details. */

package artofillusion.animation;

import artofillusion.*;
import artofillusion.ui.*;
import artofillusion.object.*;
import artofillusion.animation.Joint.*;
import java.awt.*;
import java.awt.event.*;
import java.text.*;


/** This is a dialog box for editing joints in a skeleton. */

public class JointEditorDialog extends Dialog implements ActionListener,
                 ModelListener
{
  MeshEditorWindow window;
  //MeshViewer theView;
  Mesh theMesh;
  Object3D oldMesh;
  Skeleton skeleton;
  SkeletonControlHolder skelControlHolder;
  Joint joint;
  TextField nameField;
  DOFPanel ang1Panel, ang2Panel, twistPanel, lengthPanel;
  Panel propertyPanel;
  Button okButton, cancelButton;
  boolean disableUpdating;

  static JointEditorDialog propertyDialog=null;

  public static void openPropertyDialog(MeshEditorWindow win, MeshViewer view)
  {
    if (propertyDialog != null)
      return;

    propertyDialog = new JointEditorDialog(win, view, -1);
  }

  public static void closePropertyDialog()
  {
    if (propertyDialog == null)
      return;
    propertyDialog.cancelAndClose();
  }


  public JointEditorDialog(MeshEditorWindow win, MeshViewer view, int jointID)
  {
    super(win, "Edit Bone", false);
    window = win;
    //theView = view;
    theMesh = (Mesh) view.getObject().object;
    oldMesh = ((Object3D) theMesh).duplicate();
    skeleton = theMesh.getSkeleton();
    skelControlHolder = view.getSkeletonControlHolder(); // TODO(MB) Refine when different holders per EditingWindow become possible
    skelControlHolder.getModelEvent().addListener(this);
    skeleton.getModelEvent().addListener(this);

    add(propertyPanel=createPropertiesPanel(), BorderLayout.CENTER);

    editJoint(jointID);

    // theView.modelevent.addListener(this);

    pack();
    show();
  }

  /** Builds the dialog panel. The values of the (DOFs of the) joint
    are NOT set here. */

  public Panel createPropertiesPanel()
  {
    Panel result=new Panel();
    result.setLayout(new BorderLayout());

    // Layout the dialog.
    Panel p = new Panel();
    p.add(Translate.label("Name"));
    p.add(nameField = new TextField("", 20));
    result.add(p, BorderLayout.NORTH);
    p = new Panel();
    p.setLayout(new GridBagLayout());
    GridBagConstraints gc = new GridBagConstraints();
    gc.gridy = 0;
    p.add(ang1Panel = new DOFPanel("X Bend", -180.0, 180.0, null), gc);
    p.add(ang1Panel.getGraph(), gc);
    gc.insets.left = 10;
    p.add(ang2Panel = new DOFPanel("Y Bend", -180.0, 180.0, null), gc);
    p.add(ang2Panel.getGraph(), gc);
    gc.insets.left = 0;
    gc.gridy = 1;
    gc.insets.top = 10;
    p.add(twistPanel = new DOFPanel("Twist", -180.0, 180.0, null), gc);
    p.add(twistPanel.getGraph(), gc);
    gc.insets.left = 10;
    p.add(lengthPanel = new DOFPanel("Length", 0.0, Double.MAX_VALUE, null), gc);
    result.add(p, BorderLayout.CENTER);
    p = new Panel();
    p.add(okButton = Translate.button("ok", this));
    p.add(cancelButton = Translate.button("cancel", this));
    result.add(p, BorderLayout.SOUTH);

    return result;
  }


  /** Close the dialog without applying changes*/

  public void cancelAndClose()
  {
    ((Object3D) theMesh).copyObject(oldMesh);
    ((Object3D) theMesh).informChanged(this, ModelEvent.CHANGEDUR_DONTKNOW);
    skelControlHolder.getModelEvent().removeListener(this);
    skeleton.getModelEvent().removeListener(this);
    //theView.objectChanged();
    //theView.updateImage();
    //theView.modelevent.removeListener(this);
    dispose();
  }

  /** Stores data of current joint, refreshes display.
    Prepares to either close the dialog or load another joint. */
  protected void releaseCurrentJoint()
  {
    if (joint != null && joint.id != -1)
    {
      joint.getModelEvent().removeListener(this);
      valueChanged();  //TODO(MB) ???
      joint.name = nameField.getText();
      window.setUndoRecord(new UndoRecord(window, false, UndoRecord.COPY_OBJECT, new Object [] {theMesh, oldMesh})); //TODO(MB) something ?!
    }
  }

  /** Transfers data from the selected joint to the open dialog.
   */
  protected void editJoint(int jointID)
  {
    joint = skeleton.getJoint(jointID);
    if (joint == null)
      joint = Joint.createDefaultJoint();
    joint.getModelEvent().addListener(this);

    transferToDialog();
  }

  /** Transfers data from the selected joint to the open dialog.
   */
  protected void transferToDialog()
  {
    nameField.setText(joint.name);
    ang1Panel.transferToPanel(joint.angle1);
    ang2Panel.transferToPanel(joint.angle2);
    twistPanel.transferToPanel(joint.twist);
    lengthPanel.transferToPanel(joint.length);
    lengthPanel.setEnabled(joint.parent != null);
  }
  
  public void modelEventHappened(ModelEvent event)
  {
    System.out.println("JointEditorDialog.modelEventHappened "+event.getSource());
    if (event.getSource() == skelControlHolder)
    {
      if (event.hasKey("joint deselected"))
        releaseCurrentJoint();
      else if  (event.hasKey("joint selected"))
      {
        editJoint( ((Integer)event.getValue("joint selected")).intValue() );    // TODO(MB) Add new property in the view
      }
    }
    else if (event.getSource() == skeleton)
    {
      if (event.hasKey("skeleton changed"))
      {
        if (event.getCause() != this)
        {
          System.out.println("JointEditorDialog.modelEventHappened updateComponents");
          // The IK solver creates new joints, so we may have to connect to another joint
          joint.getModelEvent().removeListener(this);
          //joint.name = nameField.getText();
          editJoint(joint.getId());
//	   ang1Panel.updateComponents();
//	   ang2Panel.updateComponents();
//	   twistPanel.updateComponents();
//	   lengthPanel.updateComponents();
	}
      }
    }
    else if (event.getSource() == joint)  // TODO(MB) Remove?
    {
      System.out.println("Some Joint Moved "+((Joint)event.getSource()).id);
      if (event.hasKey("joint dof changed"))
      {
	if (event.getCause() != this)
	{
	  System.out.println("Event joint moved");
	  ang1Panel.updateComponents();
	  ang2Panel.updateComponents();
	  twistPanel.updateComponents();
	  lengthPanel.updateComponents();
	}
      }
    }
  }

  /** Deal with button clicks. */

  public void actionPerformed(ActionEvent ev)
  {
    if (ev.getSource() == cancelButton)
      cancelAndClose();
    else
      {
        releaseCurrentJoint();
        skelControlHolder.getModelEvent().removeListener(this);
        skeleton.getModelEvent().removeListener(this);
	dispose();
      }
  }

  /** This method is called when any of the values in the ValueFields is changed. */

  public void valueChanged()
  {
    boolean ok = (ang1Panel.isValid() && ang2Panel.isValid() && twistPanel.isValid() && lengthPanel.isValid());
    okButton.setEnabled(ok);
    if (!ok/* || disableUpdating*/)
      return;
    ang1Panel.recordValues();
    ang2Panel.recordValues();
    twistPanel.recordValues();
    lengthPanel.recordValues();
    joint.recalcCoords(true);
    joint.informChanged(this, ModelEvent.CHANGEDUR_DONTKNOW);  // TODO(MB) better "during" value
    if (!skeleton.getSkeletonDetached())
    {
      Skeleton.adjustMesh((Mesh) oldMesh, theMesh);
      ((Object3D)theMesh).informChanged(this, ModelEvent.CHANGEDUR_DONTKNOW);  // TODO(MB) better "during" value
    }
    skeleton.informChanged(this, ModelEvent.CHANGEDUR_DONTKNOW);  // TODO(MB) better "during" value
//    theView.objectChanged(this);
//    theView.updateImage();
  }

  /** Inner class representing a panel for editing a single degree of freedom. */

  private class DOFPanel extends Panel implements ItemListener
  {
    public ValueField valField, minField, maxField, minComfortField, maxComfortField;
    public ValueSlider stiffnessSlider;
    public Checkbox fixedBox, rangeBox, comfortBox;
    public double min, max;
    public DOF dof;
    private DOFGraph graph;

    public DOFPanel(String name, double min, double max, DOF dof)
    {
      this.min = min;
      this.max = max;
      this.dof = dof;

      // Create the components for the panel.

      valField = new ValueField(0.0, ValueField.NONE, 5);
      minField = new ValueField(0.0, ValueField.NONE, 5);
      maxField = new ValueField(Double.NaN, ValueField.NONE, 5);
      minComfortField = new ValueField(0.0, ValueField.NONE, 5);
      maxComfortField = new ValueField(Double.NaN, ValueField.NONE, 5);
      valField.setValueChecker(new ValueChecker() {
        public boolean isValid(double val)
        {
          double lower = (rangeBox.getState() ? minField.getValue() : DOFPanel.this.min);
          double upper = (rangeBox.getState() ? maxField.getValue() : DOFPanel.this.max);
          return (val >= lower && val <= upper);
        }
      } );
      minField.setValueChecker(new ValueChecker() {
        public boolean isValid(double val)
        {
          if (!rangeBox.getState())
            return true;
          if (comfortBox.getState() && val > minComfortField.getValue())
            return false;
          return (val <= valField.getValue() && val >= DOFPanel.this.min);
        }
      } );
      maxField.setValueChecker(new ValueChecker() {
        public boolean isValid(double val)
        {
          if (!rangeBox.getState())
            return true;
          if (comfortBox.getState() && val < maxComfortField.getValue())
            return false;
          return (val >= valField.getValue() && val <= DOFPanel.this.max);
        }
      } );
      minComfortField.setValueChecker(new ValueChecker() {
        public boolean isValid(double val)
        {
          if (!comfortBox.getState() || !rangeBox.getState())
            return true;
          double lower = minField.getValue();
          double upper = Math.min(maxField.getValue(), maxComfortField.getValue());
          return (val >= lower && val <= upper);
        }
      } );
      maxComfortField.setValueChecker(new ValueChecker() {
        public boolean isValid(double val)
        {
          if (!comfortBox.getState() || !rangeBox.getState())
            return true;
          double lower = Math.max(minField.getValue(), minComfortField.getValue());
          double upper = maxField.getValue();
          return (val >= lower && val <= upper);
        }
      } );
      TextListener tl = new TextListener() {
        public void textValueChanged(TextEvent e)
        {
          valField.checkIfValid();
          minField.checkIfValid();
          maxField.checkIfValid();
          minComfortField.checkIfValid();
          maxComfortField.checkIfValid();
          valueChanged();
          if (graph != null)
            graph.repaint();
        }
      };
      valField.addTextListener(tl);
      minField.addTextListener(tl);
      maxField.addTextListener(tl);
      minComfortField.addTextListener(tl);
      maxComfortField.addTextListener(tl);
      valField.sendValidEventsOnly(false);
      minField.sendValidEventsOnly(false);
      maxField.sendValidEventsOnly(false);
      minComfortField.sendValidEventsOnly(false);
      maxComfortField.sendValidEventsOnly(false);
      stiffnessSlider = new ValueSlider(0.0, 1.0, 100, 0.0);
      fixedBox = new Checkbox("Lock");
      rangeBox = new Checkbox("Restrict Total Range");
      comfortBox = new Checkbox("Restrict Comfort Range");
      fixedBox.addItemListener(this);
      rangeBox.addItemListener(this);
      comfortBox.addItemListener(this);

      // Lay them out.

      setLayout(new GridBagLayout());
      GridBagConstraints gc = new GridBagConstraints();
      gc.gridx = 0;
      add(new Label(name), gc);
      gc.gridx = 1;
      add(valField, gc);
      gc.gridx = 2;
      add(fixedBox, gc);
      gc.gridx = 0;
      gc.gridwidth = 3;
      add(rangeBox, gc);
      gc.gridwidth = 1;
      add(minField, gc);
      gc.gridx = 1;
      add(new Label(" to "), gc);
      gc.gridx = 2;
      add(maxField, gc);
      gc.gridx = 0;
      gc.gridwidth = 3;
      add(comfortBox, gc);
      gc.gridwidth = 1;
      add(minComfortField, gc);
      gc.gridx = 1;
      add(new Label(" to "), gc);
      gc.gridx = 2;
      add(maxComfortField, gc);
      gc.gridx = 0;
      gc.gridwidth = 3;
      add(new Label("Stiffness"), gc);
      add(stiffnessSlider, gc);

      if (dof != null)
      {
        internTransferToPanel();
        updateComponents();
      }
    }

    private void internTransferToPanel()
    {
      valField.setValue(dof.pos);
      minField.setValue(dof.min);
      maxField.setValue(dof.max == Double.MAX_VALUE ? Double.NaN : dof.max);
      minComfortField.setValue(dof.minComfort);
      maxComfortField.setValue(dof.maxComfort == Double.MAX_VALUE ? Double.NaN : dof.maxComfort);

      stiffnessSlider.setValue(dof.stiffness);
      fixedBox.setState(dof.fixed);
      rangeBox.setState(dof.min > min || dof.max < max || dof.comfort);
      comfortBox.setState(dof.comfort);

      updateComponents();
    }

    /** Sets newdof as dof for this panel and shows the values. */
    public void transferToPanel(DOF newdof)
    {
      dof = newdof;
      internTransferToPanel();
    }

    /** Update the states of the various components. */

    public void updateComponents()
    {
      boolean fixed = fixedBox.getState();
      boolean range = (!fixed && rangeBox.getState());
      boolean comfort = (range && comfortBox.getState());
      valField.setEnabled(!fixed);
      minField.setEnabled(range);
      maxField.setEnabled(range);
      minComfortField.setEnabled(comfort);
      maxComfortField.setEnabled(comfort);
      valField.setEditable(!fixed);
      minField.setEditable(range);
      maxField.setEditable(range);
      minComfortField.setEditable(comfort);
      maxComfortField.setEditable(comfort);
      stiffnessSlider.setEnabled(!fixed);
      rangeBox.setEnabled(!fixed);
      comfortBox.setEnabled(range);

      valField.setValue(dof.pos);
      if (graph != null)
      {
        System.out.println("DOFPanel.updateComponents repaint "+dof.pos);
	graph.repaint();
      }
    }

    /** Respond to clicks on the checkboxes. */

    public void itemStateChanged(ItemEvent ev)
    {
      if (ev.getSource() == comfortBox && comfortBox.getState())
        {
          if (minComfortField.getValue() < minField.getValue())
            minComfortField.setValue(minField.getValue());
          if (maxComfortField.getValue() > maxField.getValue())
            maxComfortField.setValue(maxField.getValue());
        }
      updateComponents();
      valField.checkIfValid();
      minField.checkIfValid();
      maxField.checkIfValid();
      minComfortField.checkIfValid();
      maxComfortField.checkIfValid();
      if (graph != null)
        graph.repaint();
      valueChanged();
    }

    /** Determine whether the values entered in the fields are all valid. */

    public boolean isValid()
    {
      if (fixedBox.getState())
        return true;
      return (valField.isValid(valField.getValue()) && minField.isValid(minField.getValue()) && maxField.isValid(maxField.getValue()) &&
        minComfortField.isValid(minComfortField.getValue()) && maxComfortField.isValid(maxComfortField.getValue()));
    }

    /** Copy the entered values into the DOF object. */

    public void recordValues()
    {
      dof.fixed = fixedBox.getState();
      boolean range = (!dof.fixed && rangeBox.getState());
      dof.pos = valField.getValue();
      if (range)
        {
          dof.min = minField.getValue();
          dof.max = maxField.getValue();
          if (Double.isNaN(dof.max))
            dof.max = Double.MAX_VALUE;
        }
      else
        {
          dof.min = min;
          dof.max = max;
        }
      dof.minComfort = minComfortField.getValue();
      dof.maxComfort = maxComfortField.getValue();
      if (Double.isNaN(dof.maxComfort))
        dof.maxComfort = Double.MAX_VALUE;
      dof.stiffness = stiffnessSlider.getValue();
      dof.comfort = (range && comfortBox.getState());
      dof.loop = ((max-min == 360.0) && (!range || (dof.max-dof.min == 360.0)));
    }

    /** Set the enabled state of every component in this panel. */

    public void setEnabled(boolean enabled)
    {
      super.setEnabled(enabled);
      Component c[] = getComponents();
      for (int i = 0; i < c.length; i++)
        c[i].setEnabled(enabled);
    }

    /* Get a graph showing the values for this panel. */

    public DOFGraph getGraph()
    {
      if (graph == null)
        graph = new DOFGraph(this);
      return graph;
    }
  }

  /* Inner class which draws the circular diagram representing the range of motion for
     a degree of freedom. */

  private class DOFGraph extends Canvas
  {
    private DOFPanel panel;
    private int dragging;
    private double lastAngle;
    private EventProcessor process;

    private static final int SIZE = 60;
    private static final int HANDLE_SIZE = 6;
    private static final int OFFSET = HANDLE_SIZE/2;

    public DOFGraph(DOFPanel dp)
    {
      panel = dp;
      dragging = -1;
      addMouseListener(new MouseAdapter() {
        public void mousePressed(MouseEvent ev)
        {
          process = new EventProcessor();
          lastAngle = 0.0;
          if (panel.fixedBox.getState())
            return;
          Point c = ev.getPoint();
          Point v = getAnglePosition(panel.valField.getValue());
          if (clicked(c, v))
            {
              dragging = 0;
              disableUpdating = true;
              return;
            }
          if (!panel.rangeBox.getState())
            return;
          v = getAnglePosition(panel.maxField.getValue());
          if (clicked(c, v))
            dragging = 2;
          v = getAnglePosition(panel.minField.getValue());
          if (clicked(c, v))
            dragging = 1;
          if (dragging > -1)
            disableUpdating = true;
          if (!panel.comfortBox.getState())
            return;
          v = getAnglePosition(panel.maxComfortField.getValue());
          if (clicked(c, v))
            dragging = 4;
          v = getAnglePosition(panel.minComfortField.getValue());
          if (clicked(c, v))
            dragging = 3;
          if (dragging > -1)
            disableUpdating = true;
        }

        public void mouseReleased(MouseEvent ev)
        {
          if (process != null)
            process.stopProcessing();
          process = null;
          dragging = -1;
          disableUpdating = false;
        }
      } );
      addMouseMotionListener(new MouseMotionAdapter() {
        public void mouseDragged(final MouseEvent ev)
        {
          process.addEvent(new Callback() {
            public void execute()
            {
              dealWithDrag(ev);
            }
          } );
        }
      } );
    }

    /** Deal with a mouse dragged event. */
    private void dealWithDrag(MouseEvent ev)
    {
      if (dragging == -1)
        return;
      System.out.println("DOFGraph dealWithDrag");
      Point pos = ev.getPoint();
      double angle = Math.atan2(pos.x-OFFSET-SIZE/2, OFFSET+SIZE/2-pos.y)*180.0/Math.PI;
      angle = Math.rint(angle);
      if (lastAngle-angle > 270.0)
        angle += 360.0;
      else if (angle-lastAngle > 270.0)
        angle -= 360.0;
      double val = panel.dof.pos;
      boolean range = panel.rangeBox.getState();
      double minv = (range ? panel.dof.min : panel.min);
      double maxv = (range ? panel.dof.max : panel.max);
      double minComfort = panel.dof.minComfort;
      double maxComfort = panel.dof.maxComfort;
      if (dragging != 1 && angle < minv)
        angle = minv;
      if (dragging != 2 && angle > maxv)
        angle = maxv;
      if ((dragging == 1 && angle > val) || (dragging == 2 && angle < val))
        angle = val;
      if (dragging == 3 && angle > maxComfort)
        angle = maxComfort;
      if (dragging == 4 && angle < minComfort)
        angle = minComfort;
      if (panel.dof.comfort)
        {
          if (dragging == 1 && angle > minComfort)
            angle = minComfort;
          if (dragging == 2 && angle < maxComfort)
            angle = maxComfort;
        }
      if (dragging == 0)
        panel.valField.setValue(angle);
      else if (dragging == 1)
        panel.minField.setValue(angle);
      else if (dragging == 2)
        panel.maxField.setValue(angle);
      else if (dragging == 3)
        panel.minComfortField.setValue(angle);
      else if (dragging == 4)
        panel.maxComfortField.setValue(angle);
      lastAngle = angle;
      panel.recordValues();
      //joint.dofChanged(this);
      repaint();
      valueChanged();
    }

    public void paint(Graphics g)
    {
      if (panel.fixedBox.getState())
        {
          // The angle is fixed, so just draw a line where it is.

          g.setColor(Color.lightGray);
          g.fillOval(0, 0, SIZE+OFFSET, SIZE+OFFSET);
          g.setColor(Color.black);
          Point p = getAnglePosition(panel.valField.getValue());
          g.drawLine(SIZE/2+OFFSET, SIZE/2+OFFSET, p.x, p.y);
          return;
        }

      // Draw the various arcs.

      int min = (int) (panel.rangeBox.getState() ? panel.minField.getValue() : panel.min);
      int max = (int) (panel.rangeBox.getState() ? panel.maxField.getValue() : panel.max);
      g.setColor(Color.black);
      g.fillArc(OFFSET, OFFSET, SIZE, SIZE, 90-min, 360-max+min);
      if (panel.rangeBox.getState() && panel.comfortBox.getState())
        {
          g.setColor(Color.lightGray);
          int minComfort = (int) panel.minComfortField.getValue();
          int maxComfort = (int) panel.maxComfortField.getValue();
          g.fillArc(OFFSET, OFFSET, SIZE, SIZE, 90-minComfort, minComfort-min);
          g.fillArc(OFFSET, OFFSET, SIZE, SIZE, 90-max, max-maxComfort);
          min = minComfort;
          max = maxComfort;
        }
      g.setColor(Color.white);
      g.fillArc(OFFSET, OFFSET, SIZE, SIZE, 90-max, max-min);
      g.setColor(Color.black);
      g.drawOval(OFFSET, OFFSET, SIZE, SIZE);

      // Draw a line marking the current position, and handles for any draggable points.

      Point p = getAnglePosition(panel.valField.getValue());
      g.drawLine(SIZE/2+OFFSET, SIZE/2+OFFSET, p.x, p.y);
      g.fillRect(p.x-OFFSET, p.y-OFFSET, HANDLE_SIZE, HANDLE_SIZE);
      if (panel.rangeBox.getState())
        {
          p = getAnglePosition(panel.minField.getValue());
          g.fillRect(p.x-OFFSET, p.y-OFFSET, HANDLE_SIZE, HANDLE_SIZE);
          p = getAnglePosition(panel.maxField.getValue());
          g.fillRect(p.x-OFFSET, p.y-OFFSET, HANDLE_SIZE, HANDLE_SIZE);
        }
      if (panel.rangeBox.getState() && panel.comfortBox.getState())
        {
          p = getAnglePosition(panel.minComfortField.getValue());
          g.fillRect(p.x-OFFSET, p.y-OFFSET, HANDLE_SIZE, HANDLE_SIZE);
          p = getAnglePosition(panel.maxComfortField.getValue());
          g.fillRect(p.x-OFFSET, p.y-OFFSET, HANDLE_SIZE, HANDLE_SIZE);
        }
    }

    /* Find the point on the circumference corresponding to a particular angle. */

    private Point getAnglePosition(double angle)
    {
      return new Point((int) (0.5*SIZE*(1.0+Math.sin(angle*Math.PI/180.0)))+OFFSET,
        (int) (0.5*SIZE*(1.0-Math.cos(angle*Math.PI/180.0)))+OFFSET);
    }

    /* Return true if the second point is within the required distance of the first point. */

    public boolean clicked(Point clickPos, Point targetPos)
    {
      return (clickPos.x >= targetPos.x-OFFSET && clickPos.x <= targetPos.x+OFFSET &&
        clickPos.y >= targetPos.y-OFFSET && clickPos.y <= targetPos.y+OFFSET);
    }

    public Dimension getPreferredSize()
    {
      return new Dimension(SIZE+2*OFFSET, SIZE+2*OFFSET);
    }
  }
}
