/* This is a Track which uses a procedure to control the orientation of an object. */

/* Copyright (C) 2001-2003 by Peter Eastman

   This program is free software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the Free Software
   Foundation; either version 2 of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but WITHOUT ANY 
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
   PARTICULAR PURPOSE.  See the GNU General Public License for more details. */

package artofillusion.animation;

import artofillusion.*;
import artofillusion.math.*;
import artofillusion.object.*;
import artofillusion.procedural.*;
import artofillusion.ui.*;
import java.awt.*;
import java.awt.event.*;
import java.io.*;

public class ProceduralRotationTrack extends Track implements ProcedureOwner
{
  ObjectInfo info;
  Procedure proc;
  Timecourse tc;
  TextureParameter parameter[];
  int smoothingMethod, mode, relCoords, joint;
  ObjectRef relObject;
  WeightTrack theWeight;
  
  private static final int ABSOLUTE = 0;
  private static final int RELATIVE = 1;

  private static final int WORLD = 0;
  private static final int PARENT = 1;
  private static final int OBJECT = 2;
  private static final int LOCAL = 3;
  
  public ProceduralRotationTrack(ObjectInfo info)
  {
    super("Rotation (procedural)");
    this.info = info;
    proc = new Procedure(new OutputModule [] {
      new OutputModule("X", "0", 0.0, null, IOPort.NUMBER),
      new OutputModule("Y", "0", 0.0, null, IOPort.NUMBER),
      new OutputModule("Z", "0", 0.0, null, IOPort.NUMBER)});
    parameter = new TextureParameter [0];
    tc = new Timecourse(new Keyframe [0], new double [0], new Smoothness [0]);
    smoothingMethod = Timecourse.INTERPOLATING;
    mode = ABSOLUTE;
    relCoords = WORLD;
    relObject = new ObjectRef();
    theWeight = new WeightTrack(this);
    joint = -1;
  }
  
  /* Modify the position of the object. */
  
  public void apply(double time)
  {
    PointInfo point = new PointInfo();
    double angles[];
    OutputModule output[] = proc.getOutputModules();
    Joint j = (joint > -1 ? info.getSkeleton().getJoint(joint) : null);

    if (j != null && mode == ABSOLUTE)
      {
        if (info.pose != null && !info.pose.equals(info.object.getPoseKeyframe()))
          {
            info.object.applyPoseKeyframe(info.pose);
            j = info.getSkeleton().getJoint(joint);
          }
        angles = new ObjectRef(info, j).getCoords().getRotationAngles();
      }
    else
      angles = info.coords.getRotationAngles();
    point.x = angles[0];
    point.y = angles[1];
    point.z = angles[2];
    point.t = time;
    ArrayKeyframe params = (ArrayKeyframe) tc.evaluate(time, smoothingMethod);
    if (params != null)
      point.param = params.val;
    proc.initForPoint(point);
    RotationKeyframe rot = new RotationKeyframe(output[0].getAverageValue(0, 0.0), output[1].getAverageValue(0, 0.0), output[2].getAverageValue(0, 0.0));
    double weight = theWeight.getWeight(time);
    Mat4 pre = null, post = null;
    if (relCoords == PARENT && info.parent != null)
      {
        pre = info.parent.coords.toLocal();
        post = info.parent.coords.fromLocal();
      }
    else if (relCoords == OBJECT)
      {
        CoordinateSystem coords = relObject.getCoords();
        if (coords != null)
	  {
	    pre = coords.toLocal();
            post = coords.fromLocal();
	  }
      }
    else if (mode == RELATIVE && relCoords == LOCAL)
      {
	pre = info.coords.fromLocal();
	post = info.coords.toLocal();
      }
    rot.applyToCoordinates(info.coords, weight, pre, post, (mode == RELATIVE), true, true, true);
    if (j != null && mode == ABSOLUTE)
      {
        Mat4 m = info.coords.fromLocal().times(j.coords.toLocal().times(info.coords.toLocal()));
        info.coords.transformAxes(m);
      }
  }
  
  /* Create a duplicate of this track. */
  
  public Track duplicate(Object obj)
  {
    ProceduralRotationTrack t = new ProceduralRotationTrack((ObjectInfo) obj);
    
    t.proc.copy(proc);
    t.mode = mode;
    t.relCoords = relCoords;
    t.smoothingMethod = smoothingMethod;
    t.tc = tc.duplicate((ObjectInfo) obj);
    t.relObject = relObject.duplicate();
    t.theWeight = (WeightTrack) theWeight.duplicate(t);
    t.joint = joint;
    return t;
  }
  
  /* Make this track identical to another one. */
  
  public void copy(Track tr)
  {
    ProceduralRotationTrack t = (ProceduralRotationTrack) tr;
    
    proc.copy(t.proc);
    mode = t.mode;
    relCoords = t.relCoords;
    smoothingMethod = t.smoothingMethod;
    tc = t.tc.duplicate(info);
    relObject = t.relObject.duplicate();
    theWeight = (WeightTrack) t.theWeight.duplicate(this);
    joint = t.joint;
  }
  
  /* Get a list of all keyframe times for this track. */
  
  public double [] getKeyTimes()
  {
    return tc.getTimes();
  }

  /* Get the timecourse describing this track. */
  
  public Timecourse getTimecourse()
  {
    return tc;
  }
  
  /* Set a keyframe at the specified time. */
  
  public void setKeyframe(double time, Keyframe k, Smoothness s)
  {
    tc.addTimepoint(k, time, s);
  }
  
  /* Set a keyframe at the specified time, based on the current state of the Scene. */
  
  public Keyframe setKeyframe(double time, Scene sc)
  {
    if (parameter.length == 0)
      return null; // There are no parameters to keyframe.
    ArrayKeyframe params = (ArrayKeyframe) tc.evaluate(time, smoothingMethod);
    double p[];
    if (params == null)
      p = getDefaultGraphValues();
    else
      {
        p = new double [params.val.length];
        System.arraycopy(params.val, 0, p, 0, p.length);
      }
    Keyframe k = new ArrayKeyframe(p);
    tc.addTimepoint(k, time, new Smoothness());
    return k;
  }

  /* Move a keyframe to a new time, and return its new position in the list. */
  
  public int moveKeyframe(int which, double time)
  {
    return tc.moveTimepoint(which, time);
  }
  
  /* Delete the specified keyframe. */
  
  public void deleteKeyframe(int which)
  {
    tc.removeTimepoint(which);
  }
  
  /* Procedural tracks are never null. */
  
  public boolean isNullTrack()
  {
    return false;
  }
    
  /* This has a single child track. */
  
  public Track [] getSubtracks()
  {
    return new Track [] {theWeight};
  }

  /* Determine whether this track can be added as a child of an object. */
  
  public boolean canAcceptAsParent(Object obj)
  {
    return (obj instanceof ObjectInfo);
  }
  
  /* Get the parent object of this track. */
  
  public Object getParent()
  {
    return info;
  }
  
  /* Set the parent object of this track. */
  
  public void setParent(Object obj)
  {
    info = (ObjectInfo) obj;
  }
  
  /* Get the smoothing method for this track. */
  
  public int getSmoothingMethod()
  {
    return smoothingMethod;
  }
  
  /* Set the smoothing method for this track. */
  
  public void setSmoothingMethod(int method)
  {
    smoothingMethod = method;
  }
  
  /* Get the names of all graphable values for this track. */
  
  public String [] getValueNames()
  {
    String names[] = new String [parameter.length];
    for (int i = 0; i < names.length; i++)
      names[i] = parameter[i].name;
    return names;
  }

  /* Get the default list of graphable values (for a track which has no keyframes). */
  
  public double [] getDefaultGraphValues()
  {
    double val[] = new double [parameter.length];
    for (int i = 0; i < val.length; i++)
      val[i] = parameter[i].defaultVal;
    return val;
  }
  
  /* Get the allowed range for graphable values.  This returns a 2D array, where elements
     [n][0] and [n][1] are the minimum and maximum allowed values, respectively, for
     the nth graphable value. */
  
  public double[][] getValueRange()
  {
    double range[][] = new double [parameter.length][2];
    for (int i = 0; i < range.length; i++)
      {
	range[i][0] = parameter[i].minVal;
	range[i][1] = parameter[i].maxVal;
      }
    return range;
  }
  
  /* Get an array of any objects which this track depends on (and which therefore must
     be updated before this track is applied). */ 
  
  public ObjectInfo [] getDependencies()
  {
    if (relCoords == OBJECT)
      {
        ObjectInfo info = relObject.getObject();
        if (info != null)
          return new ObjectInfo [] {info};
      }
     return new ObjectInfo [0];
  }
  
  /* Delete all references to the specified object from this track.  This is used when an
     object is deleted from the scene. */
  
  public void deleteDependencies(ObjectInfo obj)
  {
    if (relObject.getObject() == obj)
      relObject = new ObjectRef();
  }

  /** Find all the parameters for the procedure. */

  private TextureParameter[] findParameters()
  {
    Module module[] = proc.getModules();
    int count = 0;
    
    for (int i = 0; i < module.length; i++)
      if (module[i] instanceof ParameterModule)
        count++;
    TextureParameter params[] = new TextureParameter [count];
    count = 0;
    for (int i = 0; i < module.length; i++)
      if (module[i] instanceof ParameterModule)
        {
          params[count] = ((ParameterModule) module[i]).getParameter(this);
          ((ParameterModule) module[i]).setIndex(count++);
        }
    return params;
  }

  /* Write a serialized representation of this track to a stream. */
  
  public void writeToStream(DataOutputStream out, Scene scene) throws IOException
  {
    double t[] = tc.getTimes();
    Smoothness s[] = tc.getSmoothness();
    Keyframe v[] = tc.getValues();

    out.writeShort(1); // Version number
    out.writeUTF(name);
    out.writeBoolean(enabled);
    proc.writeToStream(out, scene);
    out.writeInt(smoothingMethod);
    out.writeInt(mode);
    out.writeInt(relCoords);
    out.writeInt(joint);
    out.writeInt(t.length);
    for (int i = 0; i < t.length; i++)
      {
        out.writeDouble(t[i]);
        ((ArrayKeyframe) v[i]).writeToStream(out); 
        s[i].writeToStream(out); 
      }
    if (relCoords == OBJECT)
      relObject.writeToStream(out);
    theWeight.writeToStream(out, scene);
  }
  
  /** Initialize this tracked based on its serialized representation as written by writeToStream(). */
  
  public void initFromStream(DataInputStream in, Scene scene) throws IOException, InvalidObjectException
  {
    short version = in.readShort();
    if (version < 0 || version > 1)
      throw new InvalidObjectException("");
    name = in.readUTF();
    enabled = in.readBoolean();
    proc.readFromStream(in, scene);
    smoothingMethod = in.readInt();
    mode = in.readInt();
    relCoords = in.readInt();
    joint = (version == 0 ? -1 : in.readInt());
    int keys = in.readInt();
    double t[] = new double [keys];
    Smoothness s[] = new Smoothness [keys];
    Keyframe v[] = new Keyframe [keys];
    for (int i = 0; i < keys; i++)
      {
        t[i] = in.readDouble();
        v[i] = new ArrayKeyframe(in, this);
        s[i] = new Smoothness(in);
      }
    tc = new Timecourse(v, t, s);
    if (relCoords == OBJECT)
      relObject = new ObjectRef(in, scene);
    else
      relObject = new ObjectRef();
    theWeight.initFromStream(in, scene);
    parameter = findParameters();
  }

  /* Present a window in which the user can edit the specified keyframe. */
  
  public void editKeyframe(LayoutWindow win, int which)
  {
    ArrayKeyframe key = (ArrayKeyframe) tc.getValues()[which];
    Smoothness s = tc.getSmoothness()[which];
    double time = tc.getTimes()[which];
    ValueField valField[] = new ValueField [parameter.length];
    ValueSlider valSlider[] = new ValueSlider [parameter.length];
    ValueField timeField = new ValueField(time, ValueField.NONE, 5);
    ValueSlider s1Slider = new ValueSlider(0.0, 1.0, 100, s.getLeftSmoothness());
    final ValueSlider s2Slider = new ValueSlider(0.0, 1.0, 100, s.getRightSmoothness());
    final Checkbox sameBox = new Checkbox("Separate Left and Right Smoothness", !s.isForceSame());
    final Label leftLabel = new Label(" ("+Translate.text("left")+")");
    final Label rightLabel = new Label(" ("+Translate.text("right")+")");
    Panel p = new Panel();
    GridBagConstraints gc = new GridBagConstraints();
    
    for (int i = 0; i < parameter.length; i++)
      {
	if (parameter[i].minVal == -Double.MAX_VALUE || parameter[i].maxVal == Double.MAX_VALUE)
	  {
	    final TextureParameter tp = parameter[i];
	    valField[i] = new ValueField(key.val[i], ValueField.NONE, 5);
	    valField[i].setValueChecker(new ValueChecker() {
	      public boolean isValid(double v)
	      {
		return (v >= tp.minVal && v <= tp.maxVal);
	      }
	    } );
	  }
	else
	  valSlider[i] = new ValueSlider(parameter[i].minVal, parameter[i].maxVal, 50, key.val[i]);
      }
    p.setLayout(new GridBagLayout());
    gc.gridx = 0;
    gc.anchor = gc.EAST;
    gc.insets.right = 5;
    for (int i = 0; i < parameter.length; i++)
      p.add(new Label(parameter[i].name), gc);
    p.add(Translate.label("Time"), gc);
    gc.gridx = 1;
    gc.anchor = gc.WEST;
    gc.insets.right = 0;
    for (int i = 0; i < parameter.length; i++)
      {
	if (valField[i] != null)
	  p.add(valField[i], gc);
	else
	  p.add(valSlider[i], gc);
      }
    p.add(timeField, gc);
    gc.gridx = 0;
    gc.gridwidth = 2;
    gc.anchor = gc.CENTER;
    p.add(sameBox, gc);
    p.add(new Label(Translate.text("Smoothness")+':'), gc);
    gc.gridwidth = 1;
    gc.gridx = 0;
    gc.anchor = gc.EAST;
    gc.insets.right = 5;
    p.add(leftLabel, gc);
    p.add(rightLabel, gc);
    gc.gridx = 1;
    gc.anchor = gc.WEST;
    gc.insets.right = 0;
    p.add(s1Slider, gc);
    p.add(s2Slider, gc);
    leftLabel.setEnabled(sameBox.getState());
    rightLabel.setEnabled(sameBox.getState());
    s2Slider.setEnabled(sameBox.getState());
    sameBox.addItemListener(new ItemListener() {
      public void itemStateChanged(ItemEvent ev)
      {
        leftLabel.setEnabled(sameBox.getState());
        rightLabel.setEnabled(sameBox.getState());
        s2Slider.setEnabled(sameBox.getState());
      }
    } );
    PanelDialog dlg = new PanelDialog(win, "Edit Keyframe", p);
    if (!dlg.clickedOk())
      return;
    win.setUndoRecord(new UndoRecord(win, false, UndoRecord.COPY_TRACK, new Object [] {this, duplicate(info)}));
    for (int i = 0; i < valField.length; i++)
      {
	if (valField[i] != null)
	  key.val[i] = valField[i].getValue();
	else
	  key.val[i] = valSlider[i].getValue();
      }
    if (sameBox.getState())
      s.setSmoothness(s1Slider.getValue(), s2Slider.getValue());
    else
      s.setSmoothness(s1Slider.getValue());
    moveKeyframe(which, timeField.getValue());
  }

  /* This method presents a window in which the user can edit the track. */
  
  public void edit(LayoutWindow win)
  {
    ProcedureEditor editor = new ProcedureEditor(proc, this, win.getScene());
    editor.setEditingWindow(win);
  }

  /** Get the title of the procedure's editing window. */
  
  public String getWindowTitle()
  {
    return "Procedural Rotation Track";
  }
  
  /** Create an object which displays a preview of the procedure. */
  
  public Object getPreview(ProcedureEditor editor)
  {
    return null;
  }
  
  /** Update the display of the preview. */
  
  public void updatePreview(Object preview)
  {
  }
  
  /** Dispose of the preview object when the editor is closed. */
  
  public void disposePreview(Object preview)
  {
  }
  
  /** Determine whether the procedure may contain View Angle modules. */
  
  public boolean allowViewAngle()
  {
    return false;
  }
  
  /** Determine whether the procedure may contain Parameter modules. */
  
  public boolean allowParameters()
  {
    return true;
  }
  
  /** Determine whether the procedure may be renamed. */
  
  public boolean canEditName()
  {
    return true;
  }
  
  /** This is called when the user clicks OK in the procedure editor. */
  
  public void acceptEdits(ProcedureEditor editor)
  {
    EditingWindow win = editor.getEditingWindow();
    win.setUndoRecord(new UndoRecord(win, false, UndoRecord.COPY_OBJECT_INFO, new Object [] {info, info.duplicate()}));
    TextureParameter newparams[] = findParameters();
    int index[] = new int [newparams.length];
    for (int i = 0; i < newparams.length; i++)
      {
	index[i] = -1;
	for (int j = 0; j < parameter.length; j++)
	  if (parameter[j].equals(newparams[i]))
	    index[i] = j;
      }
    parameter = newparams;
    Keyframe key[] = tc.getValues();
    for (int i = 0; i < key.length; i++)
      {
	double newval[] = new double [parameter.length];
	for (int j = 0; j < newval.length; j++)
	  {
	    if (index[j] > -1)
	      newval[j] = ((ArrayKeyframe) key[i]).val[index[j]];
	    else
              newval[j] = parameter[j].defaultVal;
	  }
	((ArrayKeyframe) key[i]).val = newval;
      }
    ((LayoutWindow) win).getScore().finishEditingTrack(this);
  }
  
  /** Display the Properties dialog. */
  
  public void editProperties(ProcedureEditor editor)
  {
    final Choice smoothChoice, modeChoice, jointChoice, coordsChoice;
    final ObjectRefSelector objSelector;
    ComponentsDialog dlg;
    Skeleton s = info.getSkeleton();
    Joint j[] = (s == null ? null : s.getJoints());

    smoothChoice = new Choice();
    smoothChoice.add("Discontinuous");
    smoothChoice.add("Linear");
    smoothChoice.add("Interpolating");
    smoothChoice.add("Approximating");
    smoothChoice.select(smoothingMethod);
    modeChoice = new Choice();
    modeChoice.add("Absolute");
    modeChoice.add("Relative");
    modeChoice.select(mode);
    jointChoice = new Choice();
    jointChoice.add("Object Origin");
    if (j != null)
      {
        for (int i = 0; i < j.length; i++)
          jointChoice.add(j[i].name);
        for (int i = 0; i < j.length; i++)
          if (j[i].id == joint)
            jointChoice.select(i+1);
      }
    coordsChoice = new Choice();
    coordsChoice.add("World");
    coordsChoice.add("Parent");
    coordsChoice.add("Other Object...");
    if (mode == 1)
      coordsChoice.add("Local");
    coordsChoice.select(relCoords);
    objSelector = new ObjectRefSelector(relObject, (LayoutWindow) editor.getEditingWindow(), "Position Relative To:", info);
    objSelector.setEnabled(coordsChoice.getSelectedIndex() == 2);
    modeChoice.addItemListener(new ItemListener() {
      public void itemStateChanged(ItemEvent ev)
      {
        int sel = modeChoice.getSelectedIndex();
        if (sel == 0 && coordsChoice.getItemCount() == 4)
          coordsChoice.remove(3);
        if (sel == 1 && coordsChoice.getItemCount() == 3)
          coordsChoice.add("Local");
        objSelector.setEnabled(coordsChoice.getSelectedIndex() == 2);
      }
    } );
    coordsChoice.addItemListener(new ItemListener() {
      public void itemStateChanged(ItemEvent ev)
      {
        objSelector.setEnabled(coordsChoice.getSelectedIndex() == 2);
      }
    } );
    dlg = new ComponentsDialog(editor.getParentFrame(), "Rotation Track Properties", new Component []
        {smoothChoice, modeChoice, jointChoice, coordsChoice, objSelector}, new String []
        {"Parameter Smoothing Method:", "Track Mode:", "Apply To:", "Coordinate System:", ""});
    if (!dlg.clickedOk())
      return;
    editor.saveState(false);
    smoothingMethod = smoothChoice.getSelectedIndex();
    mode = modeChoice.getSelectedIndex();
    relCoords = coordsChoice.getSelectedIndex();
    relObject = objSelector.getSelection();
    if (jointChoice.getSelectedIndex() == 0)
      joint = -1;
    else
      joint = j[jointChoice.getSelectedIndex()-1].id;
  }
}