/* Copyright (C) 2001-2003 by Peter Eastman

   This program is free software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the Free Software
   Foundation; either version 2 of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but WITHOUT ANY 
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
   PARTICULAR PURPOSE.  See the GNU General Public License for more details. */

package artofillusion.animation;

import artofillusion.*;
import artofillusion.object.*;
import artofillusion.ui.*;
import java.awt.*;
import java.awt.event.*;
import java.lang.reflect.*;
import java.util.*;

/** This is a Component which displays all the tracks for objects in a scene, and shows
    where their keyframes are. */

public class Score extends Panel implements AdjustmentListener, EditingWindow, ItemListener, TreeListener, PopupMenuManager
{
  LayoutWindow window;
  TreeList theList;
  TimeAxis theAxis;
  Vector graphs;
  Scrollbar scroll;
  ToolPalette viewTools, modeTools;
  Label helpText;
  Panel p1, p2;
  PopupMenu popupMenu;
  MenuItem popupMenuItem[];
  Marker timeMarker;
  SelectionInfo selection[];
  int scrollPos, mode, view;
  ItemListener itemListeners;
  double startTime, timeScale;
  int yoffset;

  public static final int TRACKS_MODE = 0;
  public static final int SINGLE_GRAPH_MODE = 1;
  public static final int MULTI_GRAPH_MODE = 2;

  public static final int SELECT_AND_MOVE = 0;
  public static final int SCROLL_AND_SCALE = 1;

  private final String MODE_HELP_TEXT[] = new String [] {
      Translate.text("moveKeyframeTool.helpText"),
      Translate.text("moveScoreTool.helpText")};

  public Score(LayoutWindow win)
  {
    window = win;
    setLayout(new BorderLayout());
    theList = new TreeList(win);
    theList.setPreferredSize(130, 200);
    theList.addItemListener(this);
    theList.addTreeListener(this);
    theList.setPopupMenuManager(this);
    selection = new SelectionInfo [0];
    int fps = window.getScene().getFramesPerSecond();
    timeScale = fps*5.0;
    theAxis = new TimeAxis(fps, timeScale, this);
    graphs = new Vector();
    timeMarker = new Marker(window.getScene().getTime(), Translate.text("Time"), Color.green);
    theAxis.addMarker(timeMarker);
    scroll = new Scrollbar(Scrollbar.VERTICAL, 0, 0, 0, 0);
    scroll.addAdjustmentListener(this);
    viewTools = new ToolPalette(1, 3);
    modeTools = new ToolPalette(1, 2);
    viewTools.addTool(new GenericTool(this, "trackMode.gif", "selected/trackMode.gif"));
    viewTools.addTool(new GenericTool(this, "singleMode.gif", "selected/singleMode.gif"));
    viewTools.addTool(new GenericTool(this, "multiMode.gif", "selected/multiMode.gif"));
    modeTools.addTool(new GenericTool(this, "moveKey.gif", "selected/moveKey.gif"));
    modeTools.addTool(new GenericTool(this, "panTrack.gif", "selected/panTrack.gif"));
    p1 = new Panel();
    p2 = new Panel();
    p1.setLayout(new BorderLayout());
    p1.add(new Spacer(theList, theAxis), BorderLayout.NORTH);
    p1.add(theList, BorderLayout.CENTER);
    p2.setLayout(new GridBagLayout());
    layoutGraphs();
    DividedPanel div = new DividedPanel(p1, p2, 0.0f, 1.0f, false);
    div.setLiveValidate(true, true);
    add(div, BorderLayout.CENTER);
    Panel p = new Panel();
    p.setLayout(new GridBagLayout());
    GridBagConstraints gc = new GridBagConstraints();
    gc.gridx = 0;
    gc.gridheight = 3;
    gc.fill = gc.VERTICAL;
    p.add(scroll, gc);
    gc.gridx = 1;
    gc.gridheight = 1;
    gc.anchor = gc.NORTH;
    gc.fill = gc.NONE;
    p.add(viewTools, gc);
    gc.insets.top = 5;
    p.add(modeTools, gc);
    gc.weighty = 1.0;
    p.add(new Panel(), gc);
    add(p, BorderLayout.EAST);
    add(helpText = new Label(MODE_HELP_TEXT[mode]), BorderLayout.SOUTH);
    rebuildList();
    createPopupMenu();
  }

  /** Create the popup menu. */
  
  private void createPopupMenu()
  {
    popupMenu = new PopupMenu();
    popupMenuItem = new MenuItem [5];
    popupMenu.add(popupMenuItem[0] = Translate.menuItem("editTrack", window, null));
    popupMenu.add(popupMenuItem[1] = Translate.menuItem("duplicateTracks", window, null));
    popupMenu.add(popupMenuItem[2] = Translate.menuItem("deleteTracks", window, null));
    popupMenu.add(popupMenuItem[3] = Translate.menuItem("enableTracks", window, null));
    popupMenu.add(popupMenuItem[4] = Translate.menuItem("disableTracks", window, null));
    add(popupMenu);
  }
  
  /** Display the popup menu. */

  public void showPopupMenu(Component c, int x, int y)
  {
    Track selTrack[] = getSelectedTracks();
    boolean enable = false, disable = false;

    for (int i = 0; i < selTrack.length; i++)
      {
        if (selTrack[i].isEnabled())
          disable = true;
        else
          enable = true;
      }
    popupMenuItem[0].setEnabled(selTrack.length == 1); // Edit Track
    popupMenuItem[1].setEnabled(selTrack.length > 0); // Duplicate Tracks
    popupMenuItem[2].setEnabled(selTrack.length > 0); // Delete Tracks
    popupMenuItem[3].setEnabled(enable); // Enable Tracks
    popupMenuItem[4].setEnabled(disable); // Disable Tracks
    popupMenu.show(c, x, y);
  }
  
  /** Get the popup menu for the score. */
  
  public PopupMenu getPopupMenu()
  {
    return popupMenu;
  }

  /** Get the currently selected tracks. */
  
  public Track [] getSelectedTracks()
  {
    Object obj[] = theList.getSelectedObjects();
    Track tr[] = new Track [obj.length];
    
    for (int i = 0; i < tr.length; i++)
      tr[i] = (Track) obj[i];
    return tr;
  }
  
  /** Get the currently selected keyframes. */
  
  public SelectionInfo [] getSelectedKeyframes()
  {
    return selection;
  }
  
  /** Set the currently selected keyframes. */
  
  public void setSelectedKeyframes(SelectionInfo sel[])
  {
    selection = sel;
    for (int i = 0; i < graphs.size(); i++)
      ((Component) graphs.elementAt(i)).repaint();
    window.updateMenus();
  }
  
  /** Add a set of keyframes to the selection. */
  
  public void addSelectedKeyframes(SelectionInfo newsel[])
  {
    Vector v = new Vector();
    int i, j;
    
    for (i = 0; i < selection.length; i++)
      v.addElement(selection[i]);
    for (i = 0; i < newsel.length; i++)
      {
	for (j = 0; j < selection.length; j++)
	  if (newsel[i].key == selection[j].key)
	    {
	      for (int k = 0; k < newsel[i].selected.length; k++)
		selection[j].selected[k] |= newsel[i].selected[k];
	      break;
	    }
	if (j == selection.length)
	  v.addElement(newsel[i]);
      }
    selection = new SelectionInfo [v.size()];
    for (i = 0; i < selection.length; i++)
      selection[i] = (SelectionInfo) v.elementAt(i);
    window.updateMenus();
  }
  
  /** Remove a keyframe from the selection. */
  
  public void removeSelectedKeyframe(Keyframe key)
  {
    Vector v = new Vector();
    
    for (int i = 0; i < selection.length; i++)
      if (selection[i].key != key)
	v.addElement(selection[i]);
    selection = new SelectionInfo [v.size()];
    for (int i = 0; i < selection.length; i++)
      selection[i] = (SelectionInfo) v.elementAt(i);
    window.updateMenus();
  }
  
  /** Determine whether a particular keyframe is selected. */
  
  public boolean isKeyframeSelected(Keyframe k)
  {
    for (int i = 0; i < selection.length; i++)
      if (selection[i].key == k)
        return true;
    return false;
  }

  /** Determine whether the handle for a particular value of a keyframe is selected. */
  
  public boolean isKeyframeSelected(Keyframe k, int value)
  {
    for (int i = 0; i < selection.length; i++)
      if (selection[i].key == k)
        return (selection[i].selected.length > value && selection[i].selected[value]);
    return false;
  }

  /** Rebuild the TreeList, attempting as much as possible to preserve its current state. */
  
  public void rebuildList()
  {
    Scene theScene = window.getScene();
    TreeElement allEl[] = theList.getElements();

    theList.setUpdateEnabled(false);
    theList.removeAllElements();
    for (int i = 0; i < theScene.getNumObjects(); i++)
      {
	ObjectInfo info = theScene.getObject(i);
	if (info.selected)
	  {
	    TreeElement el = new ObjectTreeElement(info, null, theList, false);
	    theList.addElement(el);
	    el.setExpanded(true);
	  }
      }
    for (int i = 0; i < theScene.getNumObjects(); i++)
      {
	ObjectInfo info = theScene.getObject(i);
	if (!info.selected || info.tracks.length == 0)
	  continue;
	ObjectTreeElement el = (ObjectTreeElement) theList.findElement(info);
	el.addTracks();
      }
    for (int i = 0; i < allEl.length; i++)
      {
	TreeElement el = theList.findElement(allEl[i].getObject());
	if (el == null)
	  continue;
	el.setExpanded(allEl[i].isExpanded());
	el.setSelected(allEl[i].isSelected());
      }
    allEl = theList.getElements();
    for (int i = 0; i < allEl.length; i++)
      allEl[i].setSelectable(allEl[i] instanceof TrackTreeElement);
    theList.setUpdateEnabled(true);
    selectedTracksChanged();
    updateScrollbar();
    repaintGraphs();
    setScrollPosition(scrollPos);
  }
  
  /** Layout the display in the right side of the Score, based on the current view mode
      and selected tracks. */

  private void layoutGraphs()
  {
    Track tr[] = getSelectedTracks();

    p2.removeAll();
    graphs.removeAllElements();
    if (view == TRACKS_MODE)
      {
	TracksPanel theTracks = new TracksPanel(window, theList, this, window.getScene().getFramesPerSecond(), timeScale);
	theTracks.setStartTime(startTime);
	theTracks.addMarker(timeMarker);
	theTracks.setMode(mode);
	GridBagConstraints gc = new GridBagConstraints();
	gc.fill = gc.BOTH;
	gc.gridx = gc.gridy = 0;
	gc.insets.bottom = 2;
	p2.add(theAxis, gc);
	gc.weightx = gc.weighty = 1.0;
	gc.gridy = 1;
	gc.insets.bottom = 0;
	p2.add(theTracks, gc);
	graphs.addElement(theTracks);
      }
    else if (view == SINGLE_GRAPH_MODE || tr.length == 0)
      {
	TrackGraph gr = new TrackGraph(window, this, theAxis);
	gr.setSubdivisions(window.getScene().getFramesPerSecond());
	gr.setStartTime(startTime);
	gr.setScale(timeScale);
	gr.addMarker(timeMarker);
	gr.setMode(mode);
	gr.setTracks(tr);
	gr.setBackground(Color.white);
	GridBagConstraints gc = new GridBagConstraints();
	gc.fill = gc.BOTH;
	gc.gridx = 1;
	gc.gridy = 0;
	gc.insets.bottom = 2;
	p2.add(theAxis, gc);
	gc.weighty = 1.0;
	gc.gridx = 0;
	gc.gridy = 1;
	gc.insets.bottom = 0;
	p2.add(gr.getAxis(), gc);
	gc.weightx = 1.0;
	gc.gridx = 1;
	p2.add(gr, gc);
	graphs.addElement(gr);
      }
    else if (view == MULTI_GRAPH_MODE)
      {
	GridBagConstraints gc = new GridBagConstraints();
	gc.fill = gc.BOTH;
	gc.gridx = 1;
	gc.gridy = 0;
	gc.insets.bottom = 2;
	p2.add(theAxis, gc);
	gc.weighty = 1.0;
	gc.insets.bottom = 0;
	for (int i = 0; i < tr.length; i++)
	  {
	    TrackGraph gr = new TrackGraph(window, this, theAxis);
	    gr.setSubdivisions(window.getScene().getFramesPerSecond());
	    gr.setStartTime(startTime);
	    gr.setScale(timeScale);
	    gr.addMarker(timeMarker);
	    gr.setMode(mode);
	    gr.setTracks(new Track [] {tr[i]});
	    gr.setBackground(Color.white);
	    gc.gridx = 0;
	    gc.gridy = i+1;
	    gc.weightx = 0.0;
	    p2.add(gr.getAxis(), gc);
	    gc.gridx = 1;
	    gc.weightx = 1.0;
	    p2.add(gr, gc);
	    graphs.addElement(gr);
/*	    if (i < tr.length-1)
	      {
		gr.showLineAtBottom(true);
		gr.getAxis().showLineAtBottom(true);
	      }*/
	    gc.insets.top = 4;
	  }
      }
    p2.validate();
    repaintAll();
  }
  
  /** Get the starting time to display. */
  
  public double getStartTime()
  {
    return startTime;
  }
  
  /** Set the starting time to display. */
  
  public void setStartTime(double time)
  {
    theAxis.setStartTime(time);
    for (int i = 0; i < graphs.size(); i++)
      ((TrackDisplay) graphs.elementAt(i)).setStartTime(time);
    startTime = time;
    repaintGraphs();
  }
  
  /** Get the number of pixels per unit time. */
  
  public double getScale()
  {
    return timeScale;
  }
  
  /** Set the number of pixels per unit time. */
  
  public void setScale(double s)
  {
    theAxis.setScale(s);
    for (int i = 0; i < graphs.size(); i++)
      ((TrackDisplay) graphs.elementAt(i)).setScale(s);
    timeScale = s;
    repaintGraphs();
  }
  
  /** Set the current time. */
  
  public void setTime(double time)
  {
    timeMarker.position = time;
    repaintGraphs();
  }
  
  /** Update the menus when the selection changes. */
  
  public void itemStateChanged(ItemEvent ev)
  {
    selectedTracksChanged();
    window.updateMenus();
  }
  
  /** This is called whenever there is a change to which tracks are selected.  It updates
      the list of selected keyframes, removing any which are no longer appropriate. */
  
  private void selectedTracksChanged()
  {
    Vector v = new Vector();
    Track sel[] = getSelectedTracks();

    for (int i = 0; i < selection.length; i++)
      for (int j = 0; j < sel.length; j++)
	if (selection[i].track == sel[j])
	  {
	    v.addElement(selection[i]);
	    break;
	  }
    selection = new SelectionInfo [v.size()];
    for (int i = 0; i < selection.length; i++)
      selection[i] = (SelectionInfo) v.elementAt(i);
    if (view == SINGLE_GRAPH_MODE)
      {
        ((TrackGraph) graphs.elementAt(0)).setTracks(sel);
        repaintAll();
      }
    if (view == MULTI_GRAPH_MODE)
      layoutGraphs();
  }

  /** This is called whenever a track is modified.  It causes all graphs to be 
      properly updated and, optionally, updates the Scene to reflect any changes
      to selected keyframes. */
  
  public void tracksModified(boolean updateScene)
  {
    for (int i = 0; i < graphs.size(); i++)
      {
	if (graphs.elementAt(i) instanceof TrackGraph)
	  ((TrackGraph) graphs.elementAt(i)).tracksModified();
	else
	  ((Component) graphs.elementAt(i)).repaint();
      }
    if (!updateScene)
      return;
    
    // Find the list of tracks with selected keyframes.

    Vector v = new Vector();
    for (int i = 0; i < selection.length; i++)
      if (!v.contains(selection[i].track))
	v.addElement(selection[i].track);

    // Now update them.

    for (int i = 0; i < v.size(); i++)
      {
	Track tr = (Track) v.elementAt(i);
	Object parent = tr.getParent();
	while (parent != null && parent instanceof Track)
	  parent = ((Track) parent).getParent();
	if (parent instanceof ObjectInfo)
	  window.getScene().applyTracksToObject((ObjectInfo) parent);
      }
    window.updateImage();
  }
  
  /** Repaint all of the graphs. */
  
  public void repaintGraphs()
  {
    for (int i = 0; i < graphs.size(); i++)
      {
        Component gr = (Component) graphs.elementAt(i);
        gr.repaint();
        if (gr instanceof TrackGraph)
          ((TrackGraph) gr).getAxis().repaint();
      }
    theAxis.repaint();
  }
  
  /** Repaint all of the child components. */
  
  public void repaintAll()
  {
    repaintGraphs();
    theList.repaint();
  }
  
  /** TreeListener methods. */
  
  public void elementsMoved(TreeEvent ev)
  {
    repaintGraphs();
  }

  public void elementOpenedOrClosed(TreeEvent ev)
  {
    repaintGraphs();
    updateScrollbar();
  }
  
  public void elementDoubleClicked(TreeEvent ev)
  {
    TreeElement el = ev.getElement();
    if (el != null && el.getObject() instanceof Track)
      {
        Track tr = (Track) el.getObject();
        tr.edit(window);
        Object parent = tr.getParent();
        while (parent != null && parent instanceof Track)
          parent = ((Track) parent).getParent();
        if (parent instanceof ObjectInfo)
          window.getScene().applyTracksToObject((ObjectInfo) parent);
        window.updateImage();
        selectedTracksChanged();
        repaintAll();
      }
  }
  
  /** Scroll the list and the tracks together. */
  
  public void setScrollPosition(int pos)
  {
    Dimension size = theList.getSize(), prefSize = theList.getPreferredSize();

    if (pos > prefSize.height-size.height)
      pos = prefSize.height-size.height;
    if (pos < 0)
      pos = 0;
    if (pos == scrollPos)
      return;
    theList.setYOffset(-pos);
    for (int i = 0; i < graphs.size(); i++)
      ((TrackDisplay) graphs.elementAt(i)).setYOffset(-pos);
    yoffset = -pos;
    theList.repaint();
    repaintGraphs();
    scrollPos = pos;
    updateScrollbar();
  }
  
  /** Respond to changes on the scrollbar. */
  
  public void adjustmentValueChanged(AdjustmentEvent ev)
  {
    setScrollPosition(ev.getValue());
  }
  
  /** Update the bounds of the scrollbar. */
  
  private void updateScrollbar()
  {
    int height = theList.getPreferredSize().height, showing = theList.getSize().height;
    scroll.setMaximum(height);
    scroll.setVisibleAmount(showing);
    scroll.setValue(scrollPos);
    scroll.setUnitIncrement(Math.max(theList.getRowHeight(), 1));
    scroll.setBlockIncrement(Math.max(showing-theList.getRowHeight(), 1));
  }
  
  /** This is called when a time marker has been moved.  If this is an intermediate 
      position in the middle of a drag, then intermediate will be true. */
  
  public void markerMoved(Marker m, boolean intermediate)
  {
    if (/*!intermediate && */m == timeMarker && m.position != window.getScene().getTime())
      window.setTime(m.position);
    else
      repaintGraphs();
  }
  
  /** Make sure the scrollbar gets adjusted when the score is resized. */
  
  public void doLayout()
  {
    theAxis.setSize(theAxis.getSize().width, theAxis.getPreferredSize().height); // Workaround for layout manager bug.
    super.doLayout();
    updateScrollbar();
  }
  
  /** Allow the user to edit the currently selected track. */
  
  public void editSelectedTrack()
  {
    Object sel[] = theList.getSelectedObjects();
    if (sel.length == 1 && sel[0] instanceof Track)
      {
        Track tr = (Track) sel[0];
        tr.edit(window);
        finishEditingTrack(tr);
      }
  }
  
  /** This method should be called when a track is done being edited. */
  
  public void finishEditingTrack(Track tr)
  {
    Object parent = tr.getParent();
    while (parent != null && parent instanceof Track)
      parent = ((Track) parent).getParent();
    if (parent instanceof ObjectInfo)
      window.getScene().applyTracksToObject((ObjectInfo) parent);
    window.updateImage();
    selectedTracksChanged();
    repaintAll();
  }
  
  /** Enable or disable all selected tracks. */
  
  public void setTracksEnabled(boolean enable)
  {
    Object sel[] = theList.getSelectedObjects();
    UndoRecord undo = new UndoRecord(window, false);
    Vector v = new Vector();

    for (int i = 0; i < sel.length; i++)
      if (sel[i] instanceof Track)
        {
          Track tr = (Track) sel[i];
          Object parent = tr.getParent();
          while (parent instanceof Track)
            parent = ((Track) parent).getParent();
          if (parent instanceof ObjectInfo && v.indexOf(parent) == -1)
            {
              v.addElement(parent);
              undo.addCommand(UndoRecord.COPY_OBJECT_INFO, new Object [] {parent, ((ObjectInfo) parent).duplicate()});
            }
          tr.setEnabled(enable);
        }
    for (int i = 0; i < v.size(); i++)
      window.getScene().applyTracksToObject((ObjectInfo) v.elementAt(i));
    theList.repaint();
    window.updateImage();
    window.updateMenus();
  }

  /** Add a keyframe to each selected track, based on the current state of the scene. */
  
  public void keyframeSelectedTracks()
  {
    Scene theScene = window.getScene();
    Object sel[] = theList.getSelectedObjects();
    double time = theScene.getTime();
    UndoRecord undo = new UndoRecord(window, false);
    Vector newkeys = new Vector();
    
    for (int i = 0; i < sel.length; i++)
      if (sel[i] instanceof Track)
        {
          Track tr = (Track) sel[i];
          if (tr.getParent() instanceof ObjectInfo)
            {
              ObjectInfo info = (ObjectInfo) tr.getParent();
              for (int j = 0; j < info.tracks.length; j++)
                if (info.tracks[j] == tr)
                  undo.addCommand(UndoRecord.SET_TRACK, new Object [] {info, new Integer(j), tr.duplicate(info)});
            }
          Keyframe k = tr.setKeyframe(time, theScene);
          if (k != null)
            newkeys.addElement(new SelectionInfo(tr, k));
        }
    window.setUndoRecord(undo);
    if (newkeys.size() > 0)
      {
        SelectionInfo newsel[] = new SelectionInfo [newkeys.size()];
        newkeys.copyInto(newsel);
        setSelectedKeyframes(newsel);
      }
    selectedTracksChanged();
    repaintGraphs();
    window.updateMenus();
  }
  
  /** Add a keyframe to the tracks of selected objects which have been modified. */
  
  public void keyframeModifiedTracks()
  {
    Scene theScene = window.getScene();
    int sel[] = theScene.getSelection();
    double time = theScene.getTime();
    UndoRecord undo = new UndoRecord(window, false);
    Vector newkeys = new Vector();
    
    for (int i = 0; i < sel.length; i++)
      {
	ObjectInfo info = theScene.getObject(sel[i]);
	boolean posx = false, posy = false, posz = false;
	boolean rotx = false, roty = false, rotz = false;
	for (int j = 0; j < info.tracks.length; j++)
	  {
	    Track tr = info.tracks[j];
	    if (!tr.isEnabled())
	      continue;
	    if (tr instanceof PositionTrack && posx && posy && posz)
	      continue;
	    if (tr instanceof RotationTrack && rotx && roty && rotz)
	      continue;
	    undo.addCommand(UndoRecord.SET_TRACK, new Object [] {info, new Integer(j), tr.duplicate(info)});
	    Keyframe k = tr.setKeyframeIfModified(time, theScene);
	    if (k != null)
	      {
                newkeys.addElement(new SelectionInfo(tr, k));
		if (tr instanceof PositionTrack)
		  {
		    PositionTrack pt = (PositionTrack) tr;
		    posx |= pt.affectsX();
		    posy |= pt.affectsY();
		    posz |= pt.affectsZ();
		  }
		if (tr instanceof RotationTrack)
		  {
		    RotationTrack rt = (RotationTrack) tr;
		    rotx |= rt.affectsX();
		    roty |= rt.affectsY();
		    rotz |= rt.affectsZ();
		  }
	      }
	  }
      }
    window.setUndoRecord(undo);
    if (newkeys.size() > 0)
      {
        SelectionInfo newsel[] = new SelectionInfo [newkeys.size()];
        newkeys.copyInto(newsel);
        setSelectedKeyframes(newsel);
      }
    selectedTracksChanged();
    repaintGraphs();
    window.updateMenus();
  }
  
  /** Duplicate the selected tracks. */
  
  public void duplicateSelectedTracks()
  {
    Scene theScene = window.getScene();
    Object sel[] = theList.getSelectedObjects();
    UndoRecord undo = new UndoRecord(window, false);
    Vector modifiedObj = new Vector(), addedTrack = new Vector();
    
    for (int i = 0; i < sel.length; i++)
      if (sel[i] instanceof Track)
        {
          Track tr = (Track) sel[i];
          if (!(tr.getParent() instanceof ObjectInfo))
            continue;
          ObjectInfo info = (ObjectInfo) tr.getParent();
          if (modifiedObj.indexOf(info) < 0)
            {
              undo.addCommand(UndoRecord.SET_TRACK_LIST, new Object [] {info, info.tracks});
              modifiedObj.addElement(info);
            }
          for (int j = 0; j < info.tracks.length; j++)
            if (info.tracks[j] == tr)
              {
                Track newtr = tr.duplicate(info);
                newtr.setName("Copy of "+tr.getName());
                info.addTrack(newtr, j+1);
                addedTrack.addElement(newtr);
              }
        }
    window.setUndoRecord(undo);
    rebuildList();
    for (int i = 0; i < addedTrack.size(); i++)
      {
	TreeElement el = theList.findElement(addedTrack.elementAt(i));
	if (el == null)
	  continue;
	el.setSelected(true);
      }
    repaintGraphs();
  }
  
  /** Delete the selected tracks. */
  
  public void deleteSelectedTracks()
  {
    Scene theScene = window.getScene();
    Object sel[] = theList.getSelectedObjects();
    UndoRecord undo = new UndoRecord(window, false);
    Vector modifiedObj = new Vector();
    
    for (int i = 0; i < sel.length; i++)
      if (sel[i] instanceof Track)
        {
          Track tr = (Track) sel[i];
          if (!(tr.getParent() instanceof ObjectInfo))
            continue;
          ObjectInfo info = (ObjectInfo) tr.getParent();
          if (modifiedObj.indexOf(info) < 0)
            {
              undo.addCommand(UndoRecord.SET_TRACK_LIST, new Object [] {info, info.tracks});
              modifiedObj.addElement(info);
            }
          info.removeTrack(tr);
        }
    window.setUndoRecord(undo);
    rebuildList();
    repaintGraphs();
  }
  
  /** Select all tracks of selected objects. */
  
  public void selectAllTracks()
  {
    Scene theScene = window.getScene();
    int sel[] = theScene.getSelection();
    
    theList.setUpdateEnabled(false);
    for (int i = 0; i < sel.length; i++)
      {
	ObjectInfo info = theScene.getObject(sel[i]);
	TreeElement el = theList.findElement(info);
	if (el != null)
          for (int j = 0; j < el.getNumChildren(); j++)
            theList.setSelected(el.getChild(j), true);
      }
    theList.setUpdateEnabled(true);
    selectedTracksChanged();
    window.updateMenus();
  }
  
  /** Add a track to the specified objects. */
  
  public void addTrack(Object obj[], Class trackClass, Object extraArgs[], boolean deselectOthers)
  {
    Scene theScene = window.getScene();
    UndoRecord undo = new UndoRecord(window, false);
    Vector added = new Vector();
    Object args[];
    if (extraArgs == null)
      args = new Object [1];
    else
      {
        args = new Object [extraArgs.length+1];
        for (int i = 0; i < extraArgs.length; i++)
          args[i+1] = extraArgs[i];
      }
    Constructor con[] = trackClass.getConstructors();
    int which;
    for (which = 0; which < con.length && con[which].getParameterTypes().length != args.length; which++);
    try
      {
        for (int i = 0; i < obj.length; i++)
          if (obj[i] instanceof ObjectInfo)
            {
              ObjectInfo info = (ObjectInfo) obj[i];
              if (trackClass == PoseTrack.class)
        	{
		  Object3D posable = info.object.getPosableObject();
		  if (posable == null)
		    continue;
		  if (posable != info.object)
		    {
		      MessageDialog dlg = new MessageDialog(window, new String []
			{"'"+info.name+"' must be converted to an Actor before a Pose track can",
			 "be attached to it.  Once you do this, you will no longer be able",
			 "to add or delete vertices.  Are you sure you want to do this?"},
			new String [] {"Yes", "No"});
		      if (dlg.getChoice() == 1)
			continue;
		      theScene.replaceObject(info.object, posable, undo);
		    }
        	}
              undo.addCommand(UndoRecord.SET_TRACK_LIST, new Object [] {info, info.tracks});
              args[0] = info;
              Track newtrack = (Track) con[which].newInstance(args);
              info.addTrack(newtrack, 0);
              added.addElement(newtrack);
            }
      }
    catch (Exception ex)
      {
        ex.printStackTrace();
      }
    window.setUndoRecord(undo);
    if (deselectOthers)
      theList.deselectAll();
    rebuildList();
    for (int i = 0; i < added.size(); i++)
      theList.setSelected(added.elementAt(i), true);
    selectedTracksChanged();
    window.updateMenus();
  }
  
  /** Edit the selected keyframe. */
  
  public void editSelectedKeyframe()
  {
    if (selection.length != 1)
      return;
    selection[0].track.editKeyframe(window, selection[0].keyIndex);
    tracksModified(true);
  }

  /** Delete all selected keyframes. */
  
  public void deleteSelectedKeyframes()
  {
    Hashtable changedTracks = new Hashtable();
    for (int i = 0; i < selection.length; i++)
      {
	Track tr = selection[i].track;
	Keyframe keys[] = tr.getTimecourse().getValues();
	for (int j = 0; j < keys.length; j++)
	  if (keys[j] == selection[i].key)
	    {
	      if (changedTracks.get(tr) == null)
	        changedTracks.put(tr, tr.duplicate(tr.getParent()));
	      tr.deleteKeyframe(j);
	      break;
	    }
      }
    selection = new SelectionInfo [0];
    UndoRecord undo = new UndoRecord(window, false);
    Enumeration enum = changedTracks.keys();
    while (enum.hasMoreElements())
      {
	Track tr = (Track) enum.nextElement();
	Object parent = tr.getParent();
	while (parent != null && parent instanceof Track)
	  parent = ((Track) parent).getParent();
	if (parent instanceof ObjectInfo)
	  window.getScene().applyTracksToObject((ObjectInfo) parent);
	undo.addCommand(UndoRecord.COPY_TRACK, new Object [] {tr, changedTracks.get(tr)});
      }
    window.setUndoRecord(undo);
    window.updateMenus();
    tracksModified(true);
  }

  /** EditingWindow methods.  Most of these either do nothing, or simply call through to 
      the corresponding methods of the LayoutWindow the Score is in. */

  public boolean confirmClose()
  {
    return true;
  }

  public void setTool(EditingTool tool)
  {
    if (view != viewTools.getSelection())
      {
        view = viewTools.getSelection();
        layoutGraphs();
      }
    if (mode != modeTools.getSelection())
      {
        mode = modeTools.getSelection();
        for (int i = 0; i < graphs.size(); i++)
          ((TrackDisplay) graphs.elementAt(i)).setMode(mode);
        setHelpText(MODE_HELP_TEXT[mode]);
      }
  }
  
  public void setHelpText(String text)
  {
    helpText.setText(text);
  }
  
  public Frame getFrame()
  {
    return window;
  }

  public void updateImage()
  {
  }

  public void updateMenus()
  {
  }
  
  public void setUndoRecord(UndoRecord command)
  {
    window.setUndoRecord(command);
  }
  
  public Scene getScene()
  {
    return window.getScene();
  }
  
  public ViewerCanvas getView()
  {
    return null;
  }
}