/* Copyright (C) 1999-2003 by Peter Eastman

   This program is free software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the Free Software
   Foundation; either version 2 of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but WITHOUT ANY 
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
   PARTICULAR PURPOSE.  See the GNU General Public License for more details. */

package artofillusion.animation;

import artofillusion.*;
import artofillusion.math.*;
import artofillusion.object.*;
import artofillusion.ui.*;
import artofillusion.object.Mesh.*;
import java.awt.*;
import java.awt.event.*;
import java.util.Vector;

/** SkeletonTool is an EditingTool used for manipulating the skeletons of objects. */

public class SkeletonTool extends EditingTool
{
  static Image icon, selectedIcon;
  Point clickPoint;
  Vec3 clickPos;
  IKSolver ik;
  Mesh oldMesh, mesh;
  UndoRecord undo;
  EventProcessor process;
  
  private static final int CLICK_TOL = 6;
  
  public SkeletonTool(EditingWindow fr)
  {
    super(fr);
    icon = loadImage("skeleton.gif");
    selectedIcon = loadImage("selected/skeleton.gif");
  }

  public void activate()
  {
    super.activate();
    theWindow.setHelpText(Translate.text("skeletonTool.helpText"));
  }

  public int whichClicks()
  {
    return ALL_CLICKS;
  }

  public Image getIcon()
  {
    return icon;
  }

  public Image getSelectedIcon()
  {
    return selectedIcon;
  }

  public void mousePressed(MouseEvent e, ViewerCanvas view)
  {
    MeshViewer mv = (MeshViewer) view;
    mesh = (Mesh) mv.getObject().object;
    Skeleton s = mesh.getSkeleton();
    Camera cam = mv.getCamera();
    CoordinateSystem objCoords = mv.getDisplayCoordinates();
    
    clickPoint = e.getPoint();
    if (e.isControlDown())
      {
	// Create a new joint.
	
	undo = new UndoRecord(theWindow, false, UndoRecord.COPY_OBJECT, new Object [] {mesh, mesh.duplicate()});
	Joint j, parent = s.getJoint(mv.getSelectedJoint());
	if (parent == null)
	  {
	    double distToJoint = cam.getDistToScreen();
	    clickPos = cam.convertScreenToWorld(clickPoint, distToJoint);
            objCoords.toLocal().transform(clickPos);
	    j = new Joint(new CoordinateSystem(clickPos, Vec3.vz(), Vec3.vy()), null, "Root "+s.getNextJointID());
	    j.angle1.fixed = j.angle2.fixed = true;
	    s.addJoint(j, -1);
	  }
	else
	  {
	    double distToJoint = cam.getWorldToView().timesZ(parent.coords.getOrigin());
	    clickPos = cam.convertScreenToWorld(clickPoint, distToJoint);
            objCoords.toLocal().transform(clickPos);
	    Vec3 zdir = clickPos.minus(parent.coords.getOrigin());
	    zdir.normalize();
	    Vec3 ydir = cam.getCameraCoordinates().getZDirection().cross(zdir);
	    ydir.normalize(); 
	    j = new Joint(new CoordinateSystem(clickPos, zdir, ydir), parent, "Bone "+s.getNextJointID());
	    s.addJoint(j, parent.id);
	  }
	mv.setSelectedJoint(j.id);
        boolean moving[] = new boolean [s.getNumJoints()];
        moving[s.findJointIndex(j.id)] = true;
	ik = new IKSolver(s, mv.getLockedJoints(), moving);
	view.updateImage();
        view.repaint();
	theWindow.updateMenus();
	oldMesh = (Mesh) mesh.duplicate();
	process = new EventProcessor();
	return;
      }
    
    // Determine which joint was clicked on.
    
    Joint joint[] = s.getJoints();
    int i;
    Mat4 objToScreen = cam.getWorldToScreen().times(objCoords.fromLocal());
    for (i = 0; i < joint.length; i++)
      {
	Vec2 pos = objToScreen.timesXY(joint[i].coords.getOrigin());
	if ((clickPoint.x > pos.x-CLICK_TOL) && 
	    (clickPoint.x < pos.x+CLICK_TOL) &&
	    (clickPoint.y > pos.y-CLICK_TOL) && 
	    (clickPoint.y < pos.y+CLICK_TOL))
	  break;
      }
    if (i == joint.length)
      {
	// No joint was clicked on, so deselect the selected joint.
	
	mv.setSelectedJoint(-1);
	view.updateImage();
        view.repaint();
	theWindow.updateMenus();
	return;
      }
    if (e.isShiftDown())
      {
	// Toggle whether this joint is locked.
	
        if (mv.isJointLocked(joint[i].id))
          mv.unlockJoint(joint[i].id);
        else
          mv.lockJoint(joint[i].id);
	view.updateImage();
        view.repaint();
	return;
      }
    
    // Make it the selected joint, and prepare to drag it.
    
    mv.setSelectedJoint(joint[i].id);
    clickPos = joint[i].coords.getOrigin();
    view.updateImage();
    view.repaint();
    theWindow.updateMenus();
    boolean moving[] = new boolean [s.getNumJoints()];
    moving[s.findJointIndex(joint[i].id)] = true;
    ik = new IKSolver(s, mv.getLockedJoints(), moving);
    oldMesh = (Mesh) mesh.duplicate();
    process = new EventProcessor();
  }
  
  public void mouseDragged(final MouseEvent e, ViewerCanvas view)
  {
    final MeshViewer mv = (MeshViewer) view;
    final CoordinateSystem objCoords = mv.getDisplayCoordinates();

    if (ik == null)
      return;
    
    // Since this can be a somewhat CPU intensive operation, we perform it on a
    // separate thread to avoid blocking the AWT even thread.

    Callback c = new Callback() {
      public void execute()
      {
        boolean converged = false;
        Camera cam = mv.getCamera();
        Mesh mesh = (Mesh) mv.getObject().object;
        Skeleton s = mesh.getSkeleton();
        Vec3 target[] = new Vec3 [s.getNumJoints()];
        Joint selectedJoint = s.getJoint(mv.getSelectedJoint());
        int jointIndex = s.findJointIndex(selectedJoint.id);
        do
        {
          Vec3 jointPos = objCoords.fromLocal().times(selectedJoint.coords.getOrigin());
          double distToJoint = cam.getWorldToView().timesZ(jointPos);
          Vec3 goal = cam.convertScreenToWorld(e.getPoint(), distToJoint);
          objCoords.toLocal().transform(goal);
  
          if (undo == null)
            undo = new UndoRecord(SkeletonTool.this.getWindow(), false, UndoRecord.COPY_OBJECT, new Object [] {mesh, mesh.duplicate()});
          target[jointIndex] = goal;
          converged = ik.solve(target, 200);
          if (!mv.getSkeletonDetached())
            Skeleton.adjustMesh(oldMesh, mesh);
          mv.objectChanged();
          mv.updateImage();
          mv.repaint();
        } while (!converged && ! process.hasEvent());
        mesh.getSkeleton().informChanged(this, ModelEvent.CHANGEDUR_OBJECTEDITORDRAG);
        if (!mv.getSkeletonDetached())
          mesh.informChanged(this, ModelEvent.CHANGEDUR_OBJECTEDITORDRAG);
      }
    };
    process.addEvent(c);
  }

  public void mouseReleased(MouseEvent e, ViewerCanvas view)
  {
    if (process != null)
      process.stopProcessing();
    process = null;
    ik = null;
    if (undo != null)
    {
      theWindow.setUndoRecord(undo);
      mesh.getSkeleton().informChanged(this, ModelEvent.CHANGEDUR_OBJECTEDITOR);
//      if (!mv.getSkeletonDetached())    // TODO(MB) Test if necessary
//        mesh.informChanged(this, ModelEvent.CHANGEDUR_OBJECTEDITORDRAG);
    }
    mesh = oldMesh = null;
    undo = null;
  }
}