/* Copyright (C) 2000-2003 by Peter Eastman

   This program is free software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the Free Software
   Foundation; either version 2 of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but WITHOUT ANY 
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
   PARTICULAR PURPOSE.  See the GNU General Public License for more details. */

package artofillusion.image;

import artofillusion.*;
import artofillusion.math.*;
import artofillusion.ui.*;
import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.util.Vector;

/** ImageOrColor represents a color which can potentially vary with position.  It provides a 
    Panel with which the user can edit it.  They can choose either a single color, or an 
    ImageMap which is then multiplied by a user specified color. */

public class ImageOrColor
{
  private ImageMap map;
  private RGBColor color;
  
  public ImageOrColor(RGBColor theColor)
  {
    color = theColor.duplicate();
  }
  
  public ImageOrColor(RGBColor scaleColor, ImageMap theMap)
  {
    color = scaleColor.duplicate();
    map = theMap;
  }
  
  /** Get the image, or null if it is a single color. */
  
  public ImageMap getImage()
  {
    return map;
  }

  /** Get the color by which the image is multiplied. */
  
  public RGBColor getColor()
  {
    return color;
  }  
  
  /** Given a texture coordinate (x and y each between 0 and 1), return the average color of
      a region of width (xsize, ysize) centered at that location.  wrapx and wrapy specify 
      whether, for purposes of interpolation, the image should be treated as wrapping around 
      so that opposite edges touch each other. */
  
  public void getColor(RGBColor theColor, boolean wrapx, boolean wrapy, double x, double y, double xsize, double ysize)
  {
    if (map == null)
      theColor.copy(color);
    else
      {
	map.getColor(theColor, wrapx, wrapy, x, y, xsize, ysize);
	theColor.multiply(color);
      }
  }

  /** Return the average color over the entire image. */
  
  public void getAverageColor(RGBColor theColor)
  {
    if (map == null)
      theColor.copy(color);
    else
      {
	theColor.setRGB(map.getAverageComponent(0), map.getAverageComponent(1), map.getAverageComponent(2));
	theColor.multiply(color);
      }
  }
  
  /** Create a duplicate of this object. */
  
  public ImageOrColor duplicate()
  {
    ImageOrColor copy = new ImageOrColor(color);
    
    copy.map = map;
    return copy;
  }
  
  /** Make this object identical to another one. */
  
  public void copy(ImageOrColor obj)
  {
    map = obj.map;
    color.copy(obj.color);
  }

  /** The following two methods are used for reading and writing files.  The first is a
      constructor which reads the necessary data from an input stream.  The other writes
      the object's representation to an output stream. */
  
  public ImageOrColor(DataInputStream in, Scene theScene) throws IOException
  {
    int index = in.readInt();
    
    if (index > -1)
      map = theScene.getImage(index);
    color = new RGBColor(in);
  }
  
  public void writeToFile(DataOutputStream out, Scene theScene) throws IOException
  {
    if (map == null)
      out.writeInt(-1);
    else
      out.writeInt(theScene.indexOf(map));
    color.writeToFile(out);
  }

  /** Get a Panel with which the user can edit the color.  The parameters are a parent Frame
      which can be used for displaying Dialogs, the Scene from which to get images, and an
      AdjustmentListener which should be notified when the color changes.  It is alright
      to pass null for the listener if notifications are not necessary.
      
      IMPORTANT NOTE: The value specified in the AdjustmentEvents will be meaningless, 
      and adjustmentValueChanged MAY be called with a null reference for the event! */
  
  public Panel getEditingPanel(Frame parent, Scene theScene, AdjustmentListener adjust)
  {
    return new EditingPanel(parent, theScene, adjust);
  }
  
  class EditingPanel extends Panel implements MouseListener
  {
    Frame fr;
    Scene sc;
    Canvas preview;
    Panel colorPatch;
    AdjustmentListener adjust;

    public EditingPanel(Frame parent, Scene theScene, AdjustmentListener adjust)
    {
      super();
      fr = parent;
      sc = theScene;
      this.adjust = adjust;
      preview = new Canvas() {
	public Dimension getPreferredSize()
	  {
	    return new Dimension(ImageMap.PREVIEW_WIDTH, ImageMap.PREVIEW_HEIGHT);
	  }
	public void paint(Graphics g)
	  {
	    if (map != null)
	      g.drawImage(map.getPreview(), 0, 0, null);
	    else
	      g.drawRect(1, 1, ImageMap.PREVIEW_WIDTH-2, ImageMap.PREVIEW_HEIGHT-2);
	  } };
      preview.addMouseListener(this);
      colorPatch = color.getSample(ImageMap.PREVIEW_WIDTH, ImageMap.PREVIEW_HEIGHT);
      colorPatch.addMouseListener(this);
      setLayout(new GridBagLayout());
      GridBagConstraints gc = new GridBagConstraints();
      add(preview, gc);
      add(colorPatch, gc);
    }
    
    public void mousePressed(MouseEvent e)
    {
    }

    public void mouseReleased(MouseEvent e)
    {
    }
  
    public void mouseClicked(MouseEvent e)
    {
      if (e.getSource() == preview)
	{
	  ImagesDialog dlg = new ImagesDialog(fr, sc, map);
	  map = dlg.getSelection();
	  preview.repaint();
	}
      else if (e.getSource() == colorPatch)
	{
	  new ColorChooser(fr, "Color", color);
	  colorPatch.setBackground(color.getColor());
	  colorPatch.repaint();
	}
      if (adjust != null)
        adjust.adjustmentValueChanged(null);
    }

    public void mouseEntered(MouseEvent e)
    {
    }
  
    public void mouseExited(MouseEvent e)
    {
    }
  }
}