/* Copyright (C) 2000-2003 by Peter Eastman

   This program is free software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the Free Software
   Foundation; either version 2 of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but WITHOUT ANY 
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
   PARTICULAR PURPOSE.  See the GNU General Public License for more details. */

package artofillusion.image;

import artofillusion.*;
import artofillusion.math.*;
import artofillusion.ui.*;
import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.util.Vector;

/** ImageOrValue represents a float value between 0 and 1, which can potentially vary with
    position.  It provides a Panel with which the user can edit it.  They can choose either
    a single value, or a component of an ImageMap which can be multiplied by a user specified
    value. */

public class ImageOrValue
{
  private ImageMap map;
  private float value;
  private int component;
  
  public ImageOrValue(float val)
  {
    value = val;
  }
  
  public ImageOrValue(float scale, ImageMap theMap, int whichComponent)
  {
    value = scale;
    map = theMap;
    component = whichComponent;
  }
  
  /** Get the image, or null if it is a single value. */
  
  public ImageMap getImage()
  {
    return map;
  }
  
  /** Get which component of the image to use. */
  
  public int getComponent()
  {
    return component;
  }
  
  /** Get the value by which the image is scaled. */
  
  public float getValue()
  {
    return value;
  }
  
  /** Given a texture coordinate (x and y each between 0 and 1), return the average value 
      of a region of width (xsize, ysize) centered at that location.  wrapx and wrapy specify
      whether, for purposes of interpolation, the image should be treated as wrapping around 
      so that opposite edges touch each other. */
  
  public float getValue(boolean wrapx, boolean wrapy, double x, double y, double xsize, double ysize)
  {
    if (map == null)
      return value;
    else
      return map.getComponent(component, wrapx, wrapy, x, y, xsize, ysize)*value;
  }

  /** Given a texture coordinate, get the gradient of the image. */

  public void getGradient(Vec2 grad, boolean wrapx, boolean wrapy, double x, double y, double xsize, double ysize)
  {
    if (map == null)
      grad.set(0.0, 0.0);
    else
      {
	map.getGradient(grad, component, wrapx, wrapy, x, y, xsize, ysize);
	grad.scale(value);
      }
  }

  /** Return the average value over the entire image. */
  
  public float getAverageValue()
  {
    if (map == null)
      return value;
    else
      return map.getAverageComponent(component)*value;
  }
  
  /** Create a duplicate of this object. */
  
  public ImageOrValue duplicate()
  {
    ImageOrValue copy = new ImageOrValue(value);
    
    copy.map = map;
    copy.component = component;
    return copy;
  }
  
  /** Make this object identical to another one. */
  
  public void copy(ImageOrValue obj)
  {
    value = obj.value;
    map = obj.map;
    component = obj.component;
  }

  /** The following two methods are used for reading and writing files.  The first is a
      constructor which reads the necessary data from an input stream.  The other writes
      the object's representation to an output stream. */
  
  public ImageOrValue(DataInputStream in, Scene theScene) throws IOException
  {
    int index = in.readInt();
    
    if (index > -1)
      map = theScene.getImage(index);
    value = in.readFloat();
    component = in.readInt();
  }
  
  public void writeToFile(DataOutputStream out, Scene theScene) throws IOException
  {
    if (map == null)
      out.writeInt(-1);
    else
      out.writeInt(theScene.indexOf(map));
    out.writeFloat(value);
    out.writeInt(component);
  }

  /** Get a Panel with which the user can edit the value.  The parameters are a parent Frame
      which can be used for displaying Dialogs, the Scene from which to get images, an
      AdjustmentListener which should be notified when the value changes, and an ActionListener
      which should be notified when the return key is pressed in the text field.  It is alright
      to pass null for the listeners if notifications are not necessary.
      
      IMPORTANT NOTE: The value specified in the AdjustmentEvents will be meaningless, 
      and adjustmentValueChanged MAY be called with a null reference for the event! */
  
  public Panel getEditingPanel(Frame parent, Scene theScene, AdjustmentListener adjust, ActionListener action)
  {
    return new EditingPanel(parent, theScene, adjust, action);
  }
  
  class EditingPanel extends Panel implements AdjustmentListener, ItemListener
  {
    Frame fr;
    Scene sc;
    Canvas preview;
    Choice componentChoice;
    Label componentLabel, valueLabel;
    ValueSlider slider;
    AdjustmentListener adjust;

    public EditingPanel(Frame parent, Scene theScene, AdjustmentListener adjust, ActionListener action)
    {
      super();
      fr = parent;
      sc = theScene;
      this.adjust = adjust;
      preview = new Canvas() {
	public Dimension getPreferredSize()
	  {
	    return new Dimension(ImageMap.PREVIEW_WIDTH, ImageMap.PREVIEW_HEIGHT);
	  }
	public void paint(Graphics g)
	  {
	    if (map != null)
	      g.drawImage(map.getPreview(), 0, 0, null);
	    else
	      g.drawRect(1, 1, ImageMap.PREVIEW_WIDTH-2, ImageMap.PREVIEW_HEIGHT-2);
	  } };
      preview.addMouseListener(new MouseAdapter() {
	public void mouseReleased(MouseEvent e)
	  {
	    ImagesDialog dlg = new ImagesDialog(fr, sc, map);
	    map = dlg.getSelection();
	    updateComponents();
	    preview.repaint();
	  } });
      componentChoice = new Choice();
      componentChoice.add("Red");
      componentChoice.add("Green");
      componentChoice.add("Blue");
      componentChoice.addItemListener(this);
      componentLabel = new Label("Component:");
      valueLabel = new Label("");
      slider = new ValueSlider(0.0, 1.0, 100, value);
      slider.addAdjustmentListener(this);
      if (action != null)
        slider.addActionListener(action);
      setLayout(new GridBagLayout());
      GridBagConstraints gc = new GridBagConstraints();
      gc.gridheight = 2;
      add(preview, gc);
      gc.gridheight = 1;
      gc.gridx = 1;
      add(valueLabel, gc);
      add(slider, gc);
      gc.gridx = 2;
      add(componentLabel, gc);
      add(componentChoice, gc);
      updateComponents();
    }
    
    void updateComponents()
    {
      if (adjust != null)
      adjust.adjustmentValueChanged(null);
      if (map == null)
        valueLabel.setText(Translate.text("Value")+":");
      else
        valueLabel.setText(Translate.text("Scale")+":");
      if (map != null && map.getComponents() > componentChoice.getItemCount())
	componentChoice.add("Alpha");
      else if (componentChoice.getItemCount() == 4)
	componentChoice.remove(3);
      if (map == null || map.getComponents() == 1)
	{
	  componentChoice.setEnabled(false);
	  ImageOrValue.this.component = 0;
	  return;
	}
      componentChoice.setEnabled(true);
      if (ImageOrValue.this.component >= componentChoice.getItemCount())
        ImageOrValue.this.component = 0;
      componentChoice.select(ImageOrValue.this.component);
    }

    public void adjustmentValueChanged(AdjustmentEvent e)
    {
      value = (float) slider.getValue();
      if (adjust != null)
        adjust.adjustmentValueChanged(e);
    }
    
    public void itemStateChanged(ItemEvent e)
    {
      ImageOrValue.this.component = componentChoice.getSelectedIndex();
      if (adjust != null)
        adjust.adjustmentValueChanged(null);
    }
  }
}