/* This class is used to save rendered images to disk. */

/* Copyright (C) 1999-2003 by Peter Eastman

   This program is free software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the Free Software
   Foundation; either version 2 of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
   PARTICULAR PURPOSE.  See the GNU General Public License for more details. */

package artofillusion.image;

import artofillusion.*;
import artofillusion.ui.*;
import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;
import java.io.*;
import java.text.*;
import jpegencoder.*;

public class ImageSaver
{
  private int format, index;
  private String name, directory;
  private boolean ok, premultiply;
  private double quality;
  private Frame parent;
  private MovieEncoder movie;

  public static final int FORMAT_JPEG = 0;
  public static final int FORMAT_TIFF = 1;
  public static final int FORMAT_MOVIE = 2;
  public static final int FORMAT_BMP = 3;
  public static final int FORMAT_HDR = 4;
  
  private static boolean premultiplyDefault = true;
  private static double qualityDefault = 50.0;

  /** Create an ImageSaver object which will be used for saving a single images.
     The constructor displays a dialog in which the user can select the name, location,
     and format.  The saveImage() method can then be used to save individual frames of the
     animation.  Use clickedOk() to determine whether the user clicked the OK or Cancel
     button. */

  public ImageSaver(Frame parent)
  {
    init(parent, false);
  }

  /** Create an ImageSaver object which will be used for saving an animation.
     The constructor displays a dialog in which the user can select the name, location,
     and format.  The saveImage() method can then be used to save individual frames of the
     animation, and lastMovieImage() should be called once all frames have been saved.
     Use clickedOk() to determine whether the user clicked the OK or Cancel button. */

  public ImageSaver(Frame parent, int width, int height, int fps)
  {
    init(parent, true);
    if (format == FORMAT_MOVIE && this.clickedOk())
    {
      movie = new MovieEncoder(width, height, fps, directory + name);
      movie.doIt(); // will start polling for images.
    }
  }

  /** Initialize the ImageSaver and display the dialog.  parent is the parent Frame,
      and animate specifies whether this object will be used for saving a single
      image or an animation. */

  private void init(Frame parent, boolean animate)
  {
    this.parent = parent;
    index = (animate ? 1 : -1);
    CheckboxGroup formatGroup = new CheckboxGroup();
    final Checkbox jpegBox = new Checkbox("JPEG", true, formatGroup);
    final Checkbox tiffBox = new Checkbox("TIFF", false, formatGroup);
    final Checkbox bmpBox = new Checkbox("BMP", false, formatGroup);
    final Checkbox hdrBox = new Checkbox("HDR", false, formatGroup);
    final Checkbox quickBox = new Checkbox("Quicktime", false);
    if (gotJMF() && animate)
      quickBox.setCheckboxGroup(formatGroup);
    Checkbox premultBox = new Checkbox("Premultiply Transparency", premultiplyDefault);
    final ValueSlider qualitySlider = new ValueSlider(0.0, 100.0, 100, qualityDefault);
    final CardLayout card = new CardLayout();
    final Panel mainPanel = new Panel(), formatsPanel = new Panel(), optionsPanel = new Panel(), p = new Panel();
    ItemListener il = new ItemListener() {
      public void itemStateChanged(ItemEvent ev)
      {
	if (jpegBox.getState() || quickBox.getState())
	  card.show(optionsPanel, "jpeg");
	else if (tiffBox.getState())
	  card.show(optionsPanel, "tiff");
	else
	  card.show(optionsPanel, "nooption");
      } };
    GridBagConstraints gc = new GridBagConstraints();

    mainPanel.setLayout(new GridBagLayout());
    gc.gridx = 0;
    gc.insets.bottom = 5;
    formatsPanel.setLayout(new GridLayout(0, 1));
    mainPanel.add(formatsPanel, gc);
    formatsPanel.add(jpegBox, gc);
    formatsPanel.add(tiffBox, gc);
    formatsPanel.add(bmpBox, gc);
    formatsPanel.add(hdrBox, gc);
    if (quickBox.getCheckboxGroup() != null)
      formatsPanel.add(quickBox, gc);
    mainPanel.add(optionsPanel, gc);
    optionsPanel.setLayout(card);
    p.setLayout(new GridBagLayout());
    p.add(new Label("Quality"), gc);
    gc.gridx = 1;
    p.add(qualitySlider, gc);
    optionsPanel.add(p, "jpeg");
    optionsPanel.add(premultBox, "tiff");
    optionsPanel.add(new Panel(), "nooption");
    premultBox.setEnabled(true);
    jpegBox.addItemListener(il);
    tiffBox.addItemListener(il);
    bmpBox.addItemListener(il);
    hdrBox.addItemListener(il);
    if (quickBox.getCheckboxGroup() != null)
      quickBox.addItemListener(il);
    il.itemStateChanged(null);
    PanelDialog dlg = new PanelDialog(parent, "Select Format For Output File:", mainPanel);
    if (dlg.clickedOk())
      {
	premultiply = premultiplyDefault = premultBox.getState();
	quality = qualityDefault = qualitySlider.getValue();
        if (jpegBox.getState())
          format = FORMAT_JPEG;
        else if (quickBox.getState())
          format = FORMAT_MOVIE;
        else if (bmpBox.getState())
          format = FORMAT_BMP;
        else if (tiffBox.getState())
          format = FORMAT_TIFF;
        else if (hdrBox.getState())
          format = FORMAT_HDR;
	FileDialog fd = new FileDialog(parent, "Save Image", FileDialog.SAVE);
	String path, sep = System.getProperty("file.separator");
	if (jpegBox.getState())
	  fd.setFile("Untitled.jpg");
	else if (quickBox.getState())
	  fd.setFile("Untitled.mov");
	else if (bmpBox.getState())
	  fd.setFile("Untitled.bmp");
	else if (tiffBox.getState())
	  fd.setFile("Untitled.tif");
	else if (hdrBox.getState())
	  fd.setFile("Untitled.hdr");
	if (ModellingApp.currentDirectory != null)
	  fd.setDirectory(ModellingApp.currentDirectory);
	fd.show();
	name = fd.getFile();
	directory = fd.getDirectory();
	ok = (name != null);
      }
  }

  /** Determine whether the Java Media Framework is installed. */

  private boolean gotJMF()
  {
    try {
      Class c = Class.forName("javax.media.protocol.DataSource");
      return true;

    } catch (ClassNotFoundException nfx) {
      // don't have JMF.
    }
    return false;
  }

  /** Determine whether the user canceled saving the image. */

  public boolean clickedOk()
  {
    return ok;
  }

  /** Save the next image to disk.  Returns false if an error occurs.*/

  public boolean saveImage(Image im)
  {
    return saveImage(new ComplexImage(im));
  }
  
  /** Save the next image to disk.  Returns false if an error occurs.*/
  
  public boolean saveImage(ComplexImage img)
  {
    String filename = name;
    if (movie != null) // Ken : to optimize.
      {
        // Save the next frame to the movie file.
        
        Image im = img.getImage();
        im = premultiplyTransparency(im);
        java.io.ByteArrayOutputStream baos = new java.io.ByteArrayOutputStream();
        java.io.BufferedOutputStream bos = new java.io.BufferedOutputStream(baos);
        JpegEncoder jpg = new JpegEncoder(im, (int)quality, bos);
        jpg.Compress();
        movie.pushBytes(baos.toByteArray());
        try
          {
            bos.close();
          }
        catch (IOException iox)
          {
            iox.printStackTrace(); // normally we can recover.
          }
        return true;
      }
    if (index > 0) // -1 to 0?
      {
	// Insert the image number into the filename.

	NumberFormat nf = NumberFormat.getNumberInstance();
	nf.setMinimumIntegerDigits(4);
	nf.setGroupingUsed(false);
	int i = name.lastIndexOf('.');
	if (i == -1 || i == (name.length()-1))
	  filename += nf.format(index++);
	else
	  filename = name.substring(0, i)+nf.format(index++)+name.substring(i);
      }
    try
      {
        if (format == FORMAT_TIFF && premultiply)
          img = new ComplexImage(premultiplyTransparency(img.getImage()));
	return saveImage(img, new File(directory, filename), format, (int) quality);
      }
    catch (Exception ex)
      {
        ex.printStackTrace();
	new MessageDialog(parent, new String [] {"An error occurred while saving the image.", ex.getMessage()});
      }
    return false;
  }

  /** Save an image to disk in the specified format.  Returns true if the image was
     successfully saved, false if an error occurred.  For JPEG, quality should be
     between 0 and 100.  For other formats, it is ignored. */

  public static boolean saveImage(Image im, File f, int format, int quality) throws IOException, InterruptedException
  {
    return saveImage(new ComplexImage(im), f, format, quality);
  }
  
  /** Save an image to disk in the specified format.  Returns true if the image was
      successfully saved, false if an error occurred.  For JPEG, quality should be
      between 0 and 100.  For other formats, it is ignored. */
  
  public static boolean saveImage(ComplexImage img, File f, int format, int quality) throws IOException, InterruptedException
  {
    BufferedOutputStream bos = new BufferedOutputStream(new FileOutputStream(f));
    Image im = img.getImage();

    if (format == FORMAT_JPEG)
      {
        im = premultiplyTransparency(im);
	JpegEncoder jpg = new JpegEncoder(im, quality, bos);
	jpg.Compress();
	bos.close();
	return true;
      }
    if (format == FORMAT_TIFF)
      {
	DataOutputStream dos = new DataOutputStream(bos);
	TIFFEncoder tiff = new TIFFEncoder(im);
	tiff.writeImage(dos);
	dos.close();
	return true;
      }
    if (format == FORMAT_BMP)
      {
        im = premultiplyTransparency(im);
	DataOutputStream dos = new DataOutputStream(bos);
	BMPEncoder bmp = new BMPEncoder(im);
	bmp.writeImage(dos);
	dos.close();
	return true;
      }
    if (format == FORMAT_HDR)
      {
        HDREncoder.writeImage(img, bos);
	bos.close();
	return true;
      }
    bos.close();
    return false;
  }

  /** This should be called after the last frame of an animation has been saved. */

  public void lastMovieImage()
  {
    if (movie != null)
    {
      movie.signalDone();
      movie = null;
    }
    return; // do nothing.
  }

  /** Determine whether this image is partially transparent, and if so, create a new image
      in which the color components are premultiplied by the transparency. */

  private static Image premultiplyTransparency(Image im)
  {
    int i, data[];

    try
    {
      PixelGrabber pg = new PixelGrabber(im, 0, 0, -1, -1, true);
      pg.grabPixels();
      data = (int []) pg.getPixels();
    }
    catch (InterruptedException ex)
    {
      return im;
    }
    for (i = 0; i < data.length && (data[i] & 0xFF000000) == 0xFF000000; i++);
    if (i == data.length)
      return im;

    // The image is partially transparent.  Premultiply by the transparency.

    for (i = 0; i < data.length; i++)
      {
	int alpha = (data[i]>>24) & 0xFF;
	int red = (data[i]>>16) & 0xFF;
	int green = (data[i]>>8) & 0xFF;
	int blue = data[i] & 0xFF;
	red = (red*(alpha+1))>>8;
	green = (green*(alpha+1))>>8;
	blue = (blue*(alpha+1))>>8;
	data[i] = (alpha<<24) + (red<<16) + (green<<8) + blue;
      }

    // Create the new image.

    MemoryImageSource source = new MemoryImageSource(im.getWidth(null), im.getHeight(null),
      data, 0, im.getWidth(null));
    return Toolkit.getDefaultToolkit().createImage(source);
  }
}
