/* ImagesDialog is a dialog box for editing the list of ImageMaps used in a scene. */

/* Copyright (C) 2001-2002 by Peter Eastman

   This program is free software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the Free Software
   Foundation; either version 2 of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but WITHOUT ANY 
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
   PARTICULAR PURPOSE.  See the GNU General Public License for more details. */

package artofillusion.image;

import artofillusion.*;
import artofillusion.ui.*;
import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;
import java.io.*;
import java.util.Vector;

public class ImagesDialog extends Dialog implements ActionListener
{
  Scene theScene;
  Frame parent;
  int selection;
  ScrollPane sp;
  Scrollbar sb;
  ImagesCanvas ic;
  Button b[];

  public ImagesDialog(Frame fr, Scene sc, ImageMap selected)
  {
    super(fr, "Images", true);

    parent = fr;
    theScene = sc;
    for (selection = 0; selection < sc.getNumImages() && sc.getImage(selection) != selected; selection++);
    if (selection == sc.getNumImages())
      selection = -1;
    sp = new ScrollPane(ScrollPane.SCROLLBARS_NEVER);
    sb = new Scrollbar(Scrollbar.VERTICAL, 0, 0, 0, 0);
    Panel buttons = new Panel();
    
    add(sp, "Center");
    sp.add(ic = new ImagesCanvas(5));
    add(sb, "East");
    sb.addAdjustmentListener(ic);
    add(buttons, "South");
    b = new Button [4];
    buttons.add(b[0] = new Button("Load..."));
    b[0].addActionListener(this);
    b[0].setActionCommand("load");
    buttons.add(b[1] = new Button("Delete..."));
    b[1].addActionListener(this);
    b[1].setActionCommand("delete");
    buttons.add(b[2] = new Button("Select None"));
    b[2].addActionListener(this);
    b[2].setActionCommand("deselect");
    buttons.add(b[3] = new Button("Done"));
    b[3].addActionListener(this);
    b[3].setActionCommand("done");
    hilightButtons();
    sp.setSize(ic.getGridWidth()*5+4, ic.getGridHeight()*4+4);
    pack();
    setResizable(false);
    ic.scrollToSelection();
    ModellingApp.centerWindow(this);
    show();
  }

  public ImageMap getSelection()
  {
    if (selection < 0)
      return null;
    return theScene.getImage(selection);
  }

  void hilightButtons()
  {
    b[1].setEnabled(selection >= 0);
    b[2].setEnabled(selection >= 0);
  }

  public void actionPerformed(ActionEvent e)
  {
    String command = e.getActionCommand();

    if (command.equals("load"))
      {
	FileDialog fd = new FileDialog(parent, "Select Image to Load", FileDialog.LOAD);
	fd.show();
	if (fd.getFile() == null)
	  return;
        setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
	try
	  {
	    theScene.addImage(ImageMap.loadImage(new File(fd.getDirectory(), fd.getFile())));
	  }
	catch (Exception ex)
	  {
            new MessageDialog(parent, new String[] {"Error loading image file."});
            ex.printStackTrace();
	    setCursor(Cursor.getDefaultCursor());
            return;
	  }
        setCursor(Cursor.getDefaultCursor());
	selection = theScene.getNumImages()-1;
	ic.imagesChanged();
	ic.scrollToSelection();
	hilightButtons();
      }
    else if (command.equals("delete"))
      {
	MessageDialog dial = new MessageDialog(this, "Delete the selected image?", new String [] {Translate.text("button.ok"), Translate.text("button.cancel")});
	if (dial.getChoice() == 1)
	  return;
	boolean success = theScene.removeImage(selection);
	if (!success)
	  {
	    new MessageDialog(this, new String [] {"This image could not be deleted, because it is", "currently being used by a texture or material."});
	    return;
	  }
	selection = -1;
	ic.imagesChanged();
	hilightButtons();
      }
    else if (command.equals("deselect"))
      {
	selection = -1;
	ic.imagesChanged();
	hilightButtons();
      }
    else if (command.equals("done"))
      dispose();
  }

  public void itemStateChanged(ItemEvent e)
  {
    hilightButtons();
  }
  
  /* ImagesCanvas is an inner class which displays the loaded images and allows the user
     to select one by clicking on it. */
  
  class ImagesCanvas extends Canvas implements MouseListener, AdjustmentListener
  {
    int w, h, gridw, gridh;

    public ImagesCanvas(int width)
    {
      w = width;
      gridw = ImageMap.PREVIEW_WIDTH + 10;
      gridh = ImageMap.PREVIEW_HEIGHT + 10;
      sb.setUnitIncrement(gridh);
      imagesChanged();
      addMouseListener(this);
    }

    public void imagesChanged()
    {
      h = (theScene.getNumImages()-1)/w + 1;
      sb.setMaximum(h*gridh+2);
      sb.setVisibleAmount(sp.getViewportSize().height);
      sp.invalidate();
      sp.validate();
      repaint();
    }

    public int getGridWidth()
    {
      return gridw;
    }
    
    public int getGridHeight()
    {
      return gridh;
    }
    
    public void scrollToSelection()
    {
      if (selection < 0)
        return;
      sp.setScrollPosition(0, (selection/w)*gridh);
    }
    
    public void paint(Graphics g)
    {
      for (int i = 0; i < theScene.getNumImages(); i++)
	g.drawImage(theScene.getImage(i).getPreview(), (i%w)*gridw+5, (i/w)*gridh+5, this);
      if (selection >= 0)
	{
	  int x = (selection%w)*gridw, y = (selection/w)*gridh;
	  
	  g.drawRect(x+1, y+1, gridw-2, gridh-2);
	  g.drawRect(x+2, y+2, gridw-4, gridh-4);
	}
    }

    public Dimension getPreferredSize()
    {
      return new Dimension(w*gridw, h*gridh);
    }

    /* Process mouse clicks.  Because delivery of mouseClicked events is unreliable under
     Windows, check for mouseReleased events instead. */

    public void mouseReleased(MouseEvent e)
    {
      Point p = e.getPoint();
      int i, j;

      i = (p.x/gridw);
      j = (p.y/gridh);
      if (i < 5 && i+j*w < theScene.getNumImages())
	selection = i+j*w;
      else
        selection = -1;
      repaint();
      hilightButtons();
    }

    /* Empty implementations of the other MouseListener methods. */

    public void mousePressed(MouseEvent e)
    {
    }

    public void mouseClicked(MouseEvent e)
    {
    }

    public void mouseEntered(MouseEvent e)
    {
    }
  
    public void mouseExited(MouseEvent e)
    {
    }
    
    /* Scroll the ImagesCanvas. */
    
    public void adjustmentValueChanged(AdjustmentEvent e)
    {
      sp.setScrollPosition(0, sb.getValue());
      repaint();
    }
  }
}

