/* LinearMaterialMapping is a MaterialMapping which represents a linear mapping (this 
   includes rotations, translations, and scalings) of between material coordinates and 
   world coordinates. */

/* Copyright (C) 2001-2002 by Peter Eastman

   This program is free software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the Free Software
   Foundation; either version 2 of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but WITHOUT ANY 
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
   PARTICULAR PURPOSE.  See the GNU General Public License for more details. */

package artofillusion.material;

import artofillusion.*;
import artofillusion.math.*;
import artofillusion.object.*;
import artofillusion.ui.*;
import java.awt.*;
import java.awt.event.*;
import java.io.*;

public class LinearMaterialMapping extends MaterialMapping
{
  CoordinateSystem coords;
  double ax, bx, cx, dx, ay, by, cy, dy, az, bz, cz, dz;
  double xscale, yscale, zscale, matScaleX, matScaleY, matScaleZ;

  public LinearMaterialMapping(Material3D theMaterial)
  {
    super(theMaterial);
    coords = new CoordinateSystem(new Vec3(), new Vec3(0.0, 0.0, 1.0), new Vec3(0.0, 1.0, 0.0));
    xscale = yscale = zscale = 1.0;
    dx = dy = dz = 0.0;
    findCoefficients();
  }

  public static String getName()
  {
    return "Linear";
  }

  public static boolean legalMapping(Object3D obj, Material mat)
  {
    return (mat instanceof Material3D);
  }

  /* Calculate the mapping coefficients. */
  
  void findCoefficients()
  {
    Vec3 zdir = coords.getZDirection(), ydir = coords.getUpDirection();
    Vec3 xdir = ydir.cross(zdir);
    ax = xdir.x/xscale;
    bx = xdir.y/xscale;
    cx = xdir.z/xscale;
    ay = ydir.x/yscale;
    by = ydir.y/yscale;
    cy = ydir.z/yscale;
    az = zdir.x/zscale;
    bz = zdir.y/zscale;
    cz = zdir.z/zscale;
    matScaleX = 1.0/xscale;
    matScaleY = 1.0/yscale;
    matScaleZ = 1.0/zscale;
  }

  /* Get the step size to use for integrating the material. */
  
  public double getStepSize()
  {
    return material.getStepSize()*Math.min(Math.min(matScaleX, matScaleY), matScaleZ);
  }

  /* Methods from MaterialMapping. */

  public void getMaterialSpec(Vec3 pos, MaterialSpec spec, double size, double time)
  {
    double s, t;
    
    ((Material3D) material).getMaterialSpec(spec, pos.x*ax+pos.y*bx+pos.z*cx-dx, 
	pos.x*ay+pos.y*by+pos.z*cy-dy, 
	pos.x*az+pos.y*bz+pos.z*cz-dz, 
	size*matScaleX, size*matScaleY, size*matScaleZ, time);
  }

  public MaterialMapping duplicate()
  {
    return duplicate(material);
  }
  
  public MaterialMapping duplicate(Material mat)
  {
    LinearMaterialMapping map = new LinearMaterialMapping((Material3D) mat);
    
    map.coords = coords.duplicate();
    map.dx = dx;
    map.dy = dy;
    map.dz = dz;
    map.xscale = xscale;
    map.yscale = yscale;
    map.zscale = zscale;
    map.findCoefficients();
    return map;
  }
  
  public void copy(MaterialMapping mapping)
  {
    LinearMaterialMapping map = (LinearMaterialMapping) mapping; 
    
    coords = map.coords.duplicate();
    dx = map.dx;
    dy = map.dy;
    dz = map.dz;
    xscale = map.xscale;
    yscale = map.yscale;
    zscale = map.zscale;
    findCoefficients();
  }

  public Panel getEditingPanel(Object3D obj, MaterialPreviewer preview)
  {
    return new Editor(obj, preview);
  }
  
  public LinearMaterialMapping(DataInputStream in, Material theMaterial) throws IOException, InvalidObjectException
  {
    super(theMaterial);

    short version = in.readShort();
    if (version != 0)
      throw new InvalidObjectException("");
    coords = new CoordinateSystem(in);
    dx = in.readDouble();
    dy = in.readDouble();
    dz = in.readDouble();
    xscale = in.readDouble();
    yscale = in.readDouble();
    zscale = in.readDouble();
    findCoefficients();
  }
  
  public void writeToFile(DataOutputStream out) throws IOException
  {
    out.writeShort(0);
    coords.writeToFile(out);
    out.writeDouble(dx);
    out.writeDouble(dy);
    out.writeDouble(dz);
    out.writeDouble(xscale);
    out.writeDouble(yscale);
    out.writeDouble(zscale);
  }
  
  /* Editor is an inner class for editing the mapping. */

  class Editor extends Panel implements TextListener
  {
    ValueField xrotField, yrotField, zrotField, xscaleField, yscaleField, zscaleField, xtransField, ytransField, ztransField;
    Object3D theObject;
    MaterialPreviewer preview;

    public Editor(Object3D obj, MaterialPreviewer preview)
    {
      super();
      
      theObject = obj;
      this.preview = preview;
      
      // Add the various components to the Panel.
      
      GridBagConstraints gc = new GridBagConstraints();
      setLayout(new GridBagLayout());
      gc.gridwidth = 6;
      gc.insets = new Insets(0, 0, 0, 5);
      gc.gridy = 0;
      add(new Label("Scale:"), gc);
      gc.gridwidth = 1;
      gc.gridy = 1;
      add(new Label("X"), gc);
      add(xscaleField = new ValueField(Math.round(xscale*1000.0)/1000.0, ValueField.NONZERO, 5), gc);
      add(new Label("Y"), gc);
      add(yscaleField = new ValueField(Math.round(yscale*1000.0)/1000.0, ValueField.NONZERO, 5), gc);
      add(new Label("Z"), gc);
      add(zscaleField = new ValueField(Math.round(zscale*1000.0)/1000.0, ValueField.NONZERO, 5), gc);
      gc.gridwidth = 6;
      gc.gridy = 2;
      add(new Label("Center:"), gc);
      gc.gridwidth = 1;
      gc.gridy = 3;
      add(new Label("X"), gc);
      add(xtransField = new ValueField(Math.round(dx*1000.0)/1000.0, ValueField.NONE, 5), gc);
      add(new Label("Y"), gc);
      add(ytransField = new ValueField(Math.round(dy*1000.0)/1000.0, ValueField.NONE, 5), gc);
      add(new Label("Z"), gc);
      add(ztransField = new ValueField(Math.round(dz*1000.0)/1000.0, ValueField.NONE, 5), gc);
      gc.gridwidth = 6;
      gc.gridy = 4;
      add(new Label("Rotation:"), gc);
      gc.gridwidth = 1;
      gc.gridy = 5;
      double angles[] = coords.getRotationAngles();
      add(new Label("X"), gc);
      add(xrotField = new ValueField(Math.round(angles[0]*1000.0)/1000.0, ValueField.NONE, 5), gc);
      add(new Label("Y"), gc);
      add(yrotField = new ValueField(Math.round(angles[1]*1000.0)/1000.0, ValueField.NONE, 5), gc);
      add(new Label("Z"), gc);
      add(zrotField = new ValueField(Math.round(angles[2]*1000.0)/1000.0, ValueField.NONE, 5), gc);
      xscaleField.addTextListener(this);
      yscaleField.addTextListener(this);
      zscaleField.addTextListener(this);
      xtransField.addTextListener(this);
      ytransField.addTextListener(this);
      ztransField.addTextListener(this);
      xrotField.addTextListener(this);
      yrotField.addTextListener(this);
      zrotField.addTextListener(this);
    }

    public void textValueChanged(TextEvent e)
    {
      xscale = xscaleField.getValue();
      yscale = yscaleField.getValue();
      zscale = zscaleField.getValue();
      dx = xtransField.getValue();
      dy = ytransField.getValue();
      dz = ztransField.getValue();
      coords.setOrientation(xrotField.getValue(), yrotField.getValue(), zrotField.getValue());
      findCoefficients();
      preview.render();
    }
  }
}