/* UniformMaterial represents a material whose properties do not vary with position. */

/* Copyright (C) 1999-2000 by Peter Eastman

   This program is free software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the Free Software
   Foundation; either version 2 of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but WITHOUT ANY 
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
   PARTICULAR PURPOSE.  See the GNU General Public License for more details. */

package artofillusion.material;

import artofillusion.*;
import artofillusion.math.*;
import artofillusion.ui.*;
import java.awt.*;
import java.awt.event.*;
import java.io.*;

public class UniformMaterial extends Material
{
  RGBColor transparencyColor, matColor, scatteringColor;
  RGBColor trueTrans, trueScat;
  float density, scattering;
  double eccentricity;
  boolean shadows;

  public UniformMaterial()
  {
    // Generate the default color.

    transparencyColor = new RGBColor(0.5f, 0.5f, 0.5f);
    matColor = new RGBColor(0.0f, 0.0f, 0.0f);
    scatteringColor = new RGBColor(0.5f, 0.5f, 0.5f);
    trueTrans = new RGBColor(0.0f, 0.0f, 0.0f);
    trueScat = new RGBColor(0.0f, 0.0f, 0.0f);
    density = 0.5f;
    scattering = 0.0f;
    eccentricity = 0.0;
    name = "";
    refraction = 1.0;
    shadows = true;
    recalcColors();
  }

  /* Recalculate the true transparency and scattering colors. */

  private void recalcColors()
  {
    trueTrans.setRGB((float) Math.pow(transparencyColor.red, density), 
	(float) Math.pow(transparencyColor.green, density), 
	(float) Math.pow(transparencyColor.blue, density));
    trueScat.copy(scatteringColor);
    trueScat.scale(density*scattering);
  }

  public static String getTypeName()
  {
    return "Uniform";
  }

  public boolean isScattering()
  {
    return (density > 0.0 && scattering > 0.0);
  }
  
  public boolean castsShadows()
  {
    return shadows;
  }

  public void getMaterialSpec(MaterialSpec spec)
  {
    spec.transparency.copy(trueTrans);
    spec.color.copy(matColor);
    if (scattering > 0.0)
      {
	spec.scattering.copy(trueScat);
	spec.eccentricity = eccentricity;
      }
  }
  
  /* The only MaterialMapping which can be used for a UniformMaterial is a UniformMapping. */
  
  public MaterialMapping getDefaultMapping()
  {
    return new UniformMaterialMapping(this);
  }

  /* Create a duplicate of the material. */
  
  public Material duplicate()
  {
    UniformMaterial m = new UniformMaterial();
    
    m.name = name;
    m.transparencyColor.copy(transparencyColor);
    m.matColor.copy(matColor);
    m.scatteringColor.copy(scatteringColor);
    m.refraction = refraction;
    m.density = density;
    m.scattering = scattering;
    m.eccentricity = eccentricity;
    m.shadows = shadows;
    m.recalcColors();
    return m;
  }
  
  /* Allow the user to interactively edit the material. */
  
  public void edit(Frame fr, Scene sc)
  {
    new Editor(fr);
  }
  
  /* The following two methods are used for reading and writing files.  The first is a
     constructor which reads the necessary data from an input stream.  The other writes
     the object's representation to an output stream. */
  
  public UniformMaterial(DataInputStream in, Scene theScene) throws IOException, InvalidObjectException
  {
    short version = in.readShort();
    
    if (version != 0)
      throw new InvalidObjectException("");
    name = in.readUTF();
    refraction = in.readDouble();
    transparencyColor = new RGBColor(in);
    matColor = new RGBColor(in);
    scatteringColor = new RGBColor(in);
    density = in.readFloat();
    scattering = in.readFloat();
    eccentricity = in.readDouble();
    shadows = in.readBoolean();
    trueTrans = new RGBColor(0.0f, 0.0f, 0.0f);
    trueScat = new RGBColor(0.0f, 0.0f, 0.0f);
    recalcColors();
  }
  
  public void writeToFile(DataOutputStream out, Scene theScene) throws IOException
  {
    out.writeShort(0);
    out.writeUTF(name);
    out.writeDouble(refraction);
    transparencyColor.writeToFile(out);
    matColor.writeToFile(out);
    scatteringColor.writeToFile(out);
    out.writeFloat(density);
    out.writeFloat(scattering);
    out.writeDouble(eccentricity);
    out.writeBoolean(shadows);
  }
  
  class Editor extends Dialog implements MouseListener, ActionListener, AdjustmentListener, ItemListener, TextListener
  {
    TextField nameField;
    ValueField refractField;
    ValueSlider densitySlider, scatSlider, eccSlider;
    Panel transPatch, colorPatch, scatPatch;
    Checkbox shadowBox;
    MaterialPreviewer preview;
    UniformMaterial newMaterial;
    Frame parent;
    
    public Editor(Frame fr)
    {
      super(fr, true);
      
      Panel p;
      GridBagConstraints gc = new GridBagConstraints();

      parent = fr;
      newMaterial = (UniformMaterial) duplicate();

      // Add the buttons at the bottom.

      p = new Panel();
      p.add(Translate.button("ok", this));
      p.add(Translate.button("cancel", this));
      add(p, BorderLayout.SOUTH);
      
      // Add the main panel.

      p = new Panel();
      p.setLayout(new GridBagLayout());
      gc.gridwidth = 2;
      gc.anchor = GridBagConstraints.CENTER;
      p.add(preview = new MaterialPreviewer(null, newMaterial, 200, 160), gc);
      gc.gridwidth = 1;
      gc.gridx = 0;
      gc.anchor = GridBagConstraints.EAST;
      gc.insets = new Insets(0, 0, 0, 5);
      p.add(Translate.label("Name"), gc);
      p.add(Translate.label("MaterialColor"), gc);
      p.add(Translate.label("Transparency"), gc);
      p.add(Translate.label("Density"), gc);
      p.add(Translate.label("Scattering"), gc);
      p.add(Translate.label("ScatteringColor"), gc);
      p.add(Translate.label("Eccentricity"), gc);
      p.add(Translate.label("IndexOfRefraction"), gc);
      gc.gridx = 1;
      gc.gridwidth = 2;
      gc.fill = GridBagConstraints.HORIZONTAL;
      p.add(nameField = new TextField(UniformMaterial.this.name, 15), gc);
      gc.gridwidth = 1;
      p.add(colorPatch = matColor.getSample(50, 30), gc);
      p.add(transPatch = transparencyColor.getSample(50, 30), gc);
      gc.gridwidth = 2;
      p.add(densitySlider = new ValueSlider(0.0, 1.0, 100, (double) density), gc);
      densitySlider.addAdjustmentListener(this);
      p.add(scatSlider = new ValueSlider(0.0, 1.0, 100, (double) scattering), gc);
      scatSlider.addAdjustmentListener(this);
      gc.gridwidth = 1;
      p.add(scatPatch = scatteringColor.getSample(50, 30), gc);
      gc.gridwidth = 2;
      p.add(eccSlider = new ValueSlider(-1.0, 1.0, 100, eccentricity), gc);
      eccSlider.addAdjustmentListener(this);
      p.add(refractField = new ValueField(refraction, ValueField.POSITIVE), gc);
      refractField.addTextListener(this);
      p.add(shadowBox = new Checkbox(Translate.text("CastsShadows"), shadows), gc);
      shadowBox.addItemListener(this);
      add(p, BorderLayout.CENTER);
      nameField.addActionListener(this);
      densitySlider.addActionListener(this);
      scatSlider.addActionListener(this);
      refractField.addActionListener(this);
      transPatch.addMouseListener(this);
      colorPatch.addMouseListener(this);
      scatPatch.addMouseListener(this);
      pack();
      setResizable(false);
      ModellingApp.centerWindow(this);
      show();
    }

    public void actionPerformed(ActionEvent e)
    {
      if (!e.getActionCommand().equals("cancel"))
	{
	  refraction = refractField.getValue();
	  density = (float) densitySlider.getValue();
	  scattering = (float) scatSlider.getValue();
	  UniformMaterial.this.name = nameField.getText();
	  transparencyColor.copy(newMaterial.transparencyColor);
	  matColor.copy(newMaterial.matColor);
	  scatteringColor.copy(newMaterial.scatteringColor);
	  eccentricity = newMaterial.eccentricity;
	  shadows = newMaterial.shadows;
	  recalcColors();
	}
      dispose();
    }

    public void mousePressed(MouseEvent e)
    {
    }

    public void mouseReleased(MouseEvent e)
    {
    }
  
    public void mouseClicked(MouseEvent e)
    {
      if (e.getSource() == transPatch)
	{
	  new ColorChooser(parent, Translate.text("Transparency"), newMaterial.transparencyColor);
	  transPatch.setBackground(newMaterial.transparencyColor.getColor());
	  transPatch.repaint();
	  newMaterial.recalcColors();
	  preview.render();
	}
      else if (e.getSource() == colorPatch)
	{
	  new ColorChooser(parent, Translate.text("MaterialColor"), newMaterial.matColor);
	  colorPatch.setBackground(newMaterial.matColor.getColor());
	  colorPatch.repaint();
	  preview.render();
	}
      else if (e.getSource() == scatPatch)
	{
	  new ColorChooser(parent, Translate.text("ScatteringColor"), newMaterial.scatteringColor);
	  scatPatch.setBackground(newMaterial.scatteringColor.getColor());
	  scatPatch.repaint();
	  newMaterial.recalcColors();
	  preview.render();
	}
    }

    public void mouseEntered(MouseEvent e)
    {
    }
  
    public void mouseExited(MouseEvent e)
    {
    }

    public void adjustmentValueChanged(AdjustmentEvent e)
    {
      newMaterial.density = (float) densitySlider.getValue();
      newMaterial.scattering = (float) scatSlider.getValue();
      newMaterial.eccentricity = (float) eccSlider.getValue();
      newMaterial.recalcColors();
      preview.render();
    }

    public void textValueChanged(TextEvent e)
    {
      newMaterial.refraction = refractField.getValue();
      preview.render();
    }
    
    public void itemStateChanged(ItemEvent e)
    {
      newMaterial.shadows = shadowBox.getState();
      preview.render();
    }
  }
}