
/* Copyright (C) 1999-2002 by Peter Eastman

   This program is free software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the Free Software
   Foundation; either version 2 of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
   PARTICULAR PURPOSE.  See the GNU General Public License for more details. */

package artofillusion.object;

import artofillusion.*;
import artofillusion.animation.*;
import artofillusion.material.*;
import artofillusion.math.*;
import artofillusion.texture.*;
import artofillusion.ui.*;
import java.awt.*;
import java.io.*;
import java.lang.reflect.*;
import java.util.Vector;

/** Object3D is the abstract superclass of any object which can be placed into a Scene. */

public abstract class Object3D implements ObjectStateCreator
{
  protected Texture theTexture;
  protected Material theMaterial;
  protected TextureMapping texMapping;
  protected MaterialMapping matMapping;

  public static final int CANT_CONVERT = 0;
  public static final int EXACTLY = 1;
  public static final int APPROXIMATELY = 2;
  
  /** Scene the object belongs to, this information will not be serialized! */
  private transient Scene theScene = null;
  protected transient ModelEvent modelevent;
  protected transient ObjectState backupState;

  public Object3D()
  {
    modelevent = new ModelEvent();
  }

  public ModelEvent getModelEvent()
  {
    if (modelevent == null)
      modelevent = new ModelEvent();

    return modelevent;
  }

  /** Create a new object which is an exact duplicate of this one. */

  public abstract Object3D duplicate();

  /** Copy all the properties of another object, to make this one identical to it.  If the
      two objects are of different classes, this will throw a ClassCastException. */

  public abstract void copyObject(Object3D obj);

  /** Get a BoundingBox which just encloses the object. */

  public abstract BoundingBox getBounds();

  /** Resize the object.  This should be interpreted such that, if setSize() is followed
      by a call to getBounds(), the dimensions of the BoundingBox will exactly match the
      dimensions specified in setSize(). */

  public abstract void setSize(double xsize, double ysize, double zsize);

  public Scene getScene()
  {
    return theScene;
  }

  public void setScene(Scene scene)
  {
    System.out.println("Object3D setScene "+this+scene);
    theScene=scene;
  }

  /** Draw a wireframe representation of an object. */

  public static void draw(Graphics g, Camera cam, WireframeMesh mesh, BoundingBox bounds)
  {
    cam.drawWireframeMesh(g, mesh, bounds);

    /*
    int i, vis = cam.visibility(bounds), from[] = mesh.from, to[] = mesh.to, last = -1;
    Vec3 vert[] = mesh.vert;

    if (vis == Camera.NOT_VISIBLE)
      return;
    if (vis == Camera.NEEDS_CLIPPING)
      for (i = 0; i < mesh.from.length; i++)
	{
	  if (from[i] == last)
	    cam.drawClippedLineTo(g, vert[(last=to[i])]);
	  else
	    cam.drawClippedLine(g, vert[from[i]], vert[(last=to[i])]);
	}
    else
      for (i = 0; i < mesh.from.length; i++)
	{
	  if (from[i] == last)
	    cam.drawLineTo(g, vert[(last=to[i])]);
	  else
	    cam.drawLine(g, vert[from[i]], vert[(last=to[i])]);
	}*/
  }

  /** Tells whether the object is closed.  For curves, this means it has no endpoints.  For
      surface, it means the surface has no boundary. */

  public boolean isClosed()
  {
    return true;
  }
  
  /** Tells whether the object can be converted to a TriangleMesh.  It should return one
      of the following values:

      CANT_CONVERT: The object cannot be converted to a TriangleMesh.
      EXACTLY: The object can be represented exactly by a TriangleMesh.
      APPROXIMATELY: The object can be converted to a TriangleMesh.  However, the resulting
      mesh will not be exactly the same shape as the original object.

      If a class overrides this method, it must also override convertToTriangleMesh(). */

  public int canConvertToTriangleMesh()
  {
    return CANT_CONVERT;
  }

  /** Return a TriangleMesh which reproduces the shape of this object.  If
      canConvertToTriangleMesh() returned APPROXIMATELY, this method should return a
      TriangleMesh which reproduces the object to within the specified tolerance.  That is,
      no point on the mesh should be further than tol from the corresponding point on the
      original surface.  If canConvertToTriangleMesh() returned EXACTLY, then tol should
      be ignored.  If canConvertToTriangleMesh() return CANT_CONVERT, this method returns null. */

  public TriangleMesh convertToTriangleMesh(double tol)
  {
    return null;
  }

  /** This will be called whenever this object is moved, or the time changes.  Most objects
      will do nothing here, and do not need to override this.  It is available for those
      cases where an object's internal properties depend explicitly on time or on the
      object's position within the scene. */

  public void sceneChanged(ObjectInfo info, Scene scene)
  {
  }

  /** Creates a ModelEvent which can be fired by the {@link #informChanged} method.
      The standard implementation returns a "object3d changed" ModelEvent.

      @param cause      What the change caused (the caller) (may be null)
      @param during     it happened during which type of operation, see ModelEvent.CHANGEDUR_* constants
   */

  protected ModelEvent createChangedModelEvent(Object cause, String during)
  {
    return getModelEvent().createClone(this, cause).put("object3d changed", null).put("happened during", during);
  }

  /** Called to inform the object about changes made on its own data.
      The method calls {@link #createChangedModelEvent} to get a ModelEvent and dispatches
      it to its scene (if any) and all listeners.
      This method should not be overridden, override
        {@link #createChangedModelEvent} instead.

      @param cause      What the change caused (the caller) (may be null)
      @param during     it happened during which type of operation, see ModelEvent.CHANGEDUR_* constants
   */

  public void informChanged(Object cause, String during)
  {
    createChangedModelEvent(cause, during).processSend(theScene);
  }
    
  /** Called to inform the object about a dragging operation in an object editor.
      The method  dispatches a ModelEventa to all listeners.
 
      @param cause      What the change caused (the caller) (may be null)
   */
  public void informDragging(Object cause, Vec3 v[])
  {
    getModelEvent().createClone(this, cause).
        put("object3d dragging", null).
        put("vertices", v).processSend();
  }
  /** If the object can be edited by the user, editable() should be overridden to return true.
      edit() should then create a window and allow the user to edit the object. */

  public boolean editable()
  {
    return false;
  }

  public void edit(EditingWindow parent, ObjectInfo info, Callback cb)
  {
  }

  /** Edit an object which represents a gesture for an Actor object.  realObject
      specifies the object in the scene which this is a gesture for. */

  public void editGesture(EditingWindow parent, ObjectInfo info, Callback cb, ObjectInfo realObject)
  {
    edit(parent, info, cb);
  }

  /** This method tells whether textures can be assigned to the object.  Objects for which
      it makes no sense to assign a texture (curves, lights, etc.) should override this
      method to return false. */

  public boolean canSetTexture()
  {
    return true;
  }

  /** This method tells whether materials can be assigned to the object.  The default
      implementation will give the correct result for most objects, but subclasses
      can override this if necessary. */

  public boolean canSetMaterial()
  {
    return (canSetTexture() && isClosed());
  }

  /** Set this object's Texture.  In most cases, you should not use this method, and instead call setTexture() on
      its ObjectInfo. */

  public void setTexture(Texture tex)
  {
    theTexture = tex;
  }
  
  /** Get this object's Texture. */

  public Texture getTexture()
  {
    return theTexture;
  }

  /** Set this object's Material.  In most cases, you should not use this method, and instead call setMaterial() on
      its ObjectInfo. */

  public void setMaterial(Material mat)
  {
    theMaterial = mat;
  }
  
  /** Get this object's Material. */

  public Material getMaterial()
  {
    return theMaterial;
  }

  /** Set this object's TextureMapping.  In most cases, you should not use this method, and instead call setTexture() on
      its ObjectInfo. */

  public void setTextureMapping(TextureMapping mapping)
  {
    texMapping = mapping;
    if (mapping instanceof LayeredMapping)
      setTexture(((LayeredMapping) mapping).getTexture());
  }
  
  /** Get this object's TextureMapping. */

  public TextureMapping getTextureMapping()
  {
    return texMapping;
  }
  
  /** Set this object's MaterialMapping.  In most cases, you should not use this method, and instead call setMaterial() on
      its ObjectInfo. */

  public void setMaterialMapping(MaterialMapping mapping)
  {
    matMapping = mapping;
  }
  
  /** Get this object's MaterialMapping. */

  public MaterialMapping getMaterialMapping()
  {
    return matMapping;
  }

  /** Get the skeleton for this object, or null if it does not have one. */

  public Skeleton getSkeleton()
  {
    return null;
  }

  /** Objects which can be rendered as part of a scene should override this method to return
      a RenderingMesh which describes the appearance of the object.  All points on the
      RenderingMesh should be within a distance tol of the true surface.  The interactive flag
      tells whether the resulting Mesh will be rendered in interactive mode.  When interactive
      is set to true, the RenderingMesh should be cached for future use, so that it may be
      rendered repeatedly without needing to be regenerated.

      The ObjectInfo contains additional information which may affect how the object is
      rendered, such as it location in the scene, texture parameters, etc.

      Objects which cannot be rendered directly (lights, cameras, curves, etc.) do not need
      to override this method. */

  public RenderingMesh getRenderingMesh(double tol, boolean interactive, ObjectInfo info)
  {
    return null;
  }

  /** Every object should override this method to return a WireframeMesh.  This will be used
      for drawing the object in wireframe mode, and also for drawing "nonrenderable" objects
      in other rendering modes. */

  public abstract WireframeMesh getWireframeMesh();

  /** The following method writes the object's data to an output stream.  Subclasses should
      override this method, but also call super.writeToFile() to save information about
      materials, etc.  In addition to this method, every Object3D must include a constructor
      with the signature

      public Classname(DataInputStream in, Scene theScene) throws IOException, InvalidObjectException

      which reconstructs the object by reading its data from an input stream.  This
      constructor, similarly, should call the overridden constructor to read information
      about materials, etc. */

  public void writeToFile(DataOutputStream out, Scene theScene) throws IOException
  {
    out.writeShort(0);
    if (canSetTexture())
      {
	if (theMaterial == null)
	  out.writeInt(-1);
	else
	  {
	    out.writeInt(theScene.indexOf(theMaterial));
	    out.writeUTF(matMapping.getClass().getName());
	    matMapping.writeToFile(out);
	  }
	if (theTexture instanceof LayeredTexture)
	  {
	    out.writeInt(-1);
	    ((LayeredMapping) texMapping).writeToFile(out, theScene);
	  }
	else
	  {
	    out.writeInt(theScene.indexOf(theTexture));
	    out.writeUTF(texMapping.getClass().getName());
	    texMapping.writeToFile(out);
	  }
      }
  }

  public Object3D(DataInputStream in, Scene theScene) throws IOException, InvalidObjectException
  {
    short version = in.readShort();
    int i;

    if (version != 0)
      throw new InvalidObjectException("");
    if (!canSetTexture())
      return;
    i = in.readInt();
    if (i > -1)
      {
	setMaterial(theScene.getMaterial(i));
	try
	  {
	    Class mapClass = ModellingApp.getClass(in.readUTF());
	    Constructor con = mapClass.getConstructor(new Class [] {DataInputStream.class, Material.class});
	    setMaterialMapping((MaterialMapping) con.newInstance(new Object [] {in, theMaterial}));
	  }
	catch (Exception ex)
	  {
	    throw new IOException();
	  }
      }
    i = in.readInt();
    if (i > -1)
      {
	setTexture(theScene.getTexture(i));
	try
	  {
	    Class mapClass = ModellingApp.getClass(in.readUTF());
	    Constructor con = mapClass.getConstructor(new Class [] {DataInputStream.class, Texture.class});
	    setTextureMapping((TextureMapping) con.newInstance(new Object [] {in, theTexture}));
	  }
	catch (Exception ex)
	  {
	    throw new IOException();
	  }
      }
    else
      {
	// This is a layered texture.

	LayeredTexture tex = new LayeredTexture();
	LayeredMapping map = (LayeredMapping) tex.getDefaultMapping();
	map.readFromFile(in, theScene);
	setTexture(tex);
	setTextureMapping(map);
      }
  }

  /** Return a Keyframe which describes the current pose of this object. */

  public abstract Keyframe getPoseKeyframe();

  /** Modify this object based on a pose keyframe. */

  public abstract void applyPoseKeyframe(Keyframe k);

  /** This will be called whenever a new pose track is created for this object.  It allows
      the object to configure the track by setting its graphable values, subtracks, etc. */

  public void configurePoseTrack(PoseTrack track)
  {
    track.setGraphableValues(new String [0], new double [0], new double [0][2]);
  }
  
  /** Allow the user to edit a keyframe returned by getPoseKeyframe(). */

  public void editKeyframe(EditingWindow parent, Keyframe k, ObjectInfo info)
  {
    new MessageDialog((Frame) parent, "This keyframe has no editable parameters.");
  }

  /** Get a version of this object to which a pose track can be attached.  For most objects,
      this simply returns itself.  Some objects, however, need to be converted to Actors
      to have pose tracks.  If this method returns null, no pose track will be attached. */

  public Object3D getPosableObject()
  {
    return this;
  }
  
  
  /** Tells the object to create and hold an internal copy of its state
    because the object's data will be changed soon and it should allow
    undoing. This method is also called by #pushUndo to create a new
    backup.*/
  public void prepareForEditing()
  {
    backupState = createCurrentState();
  }
  
  /** Tells the object to release any internal copy of its state. This happens
   regardless if the backup copy differs from the current state.*/
  public void unprepareAfterEditing()
  {
    backupState = null;
  }
  
  /** Pushes the backup state onto the undo stack and saves the current
    state as backup. Convenience method for editing windows. */
  
  public void pushUndo()  // TODO(MB) Implement
  {
    UndoManager.getDefaultManager().pushUndo(getBackupState());
    prepareForEditing();
  }

  /** Creates a state object, containing the internal data of the
    implementing object. Standard implementation uses duplicate to
    create a copy of the whole object. Method should be overridden
    for complex objects. */
  protected ObjectState createCurrentState()
  {
    return new DefaultStateWrapper(this);
  }

  /** Returns the backup state for pushing onto the undo or redo stack.
   If later #playbackState is called with this state, it will rebuild
   the data of the object as it was when the #backupState was saved last
   time.*/
  
  public ObjectState getBackupState()
  {
    return backupState;
  }

  /** Base class for more complex implementations of ObjectState. */
  public static abstract class State implements ObjectState
  {
    protected Texture theTexture;
    protected Material theMaterial;
    protected TextureMapping texMapping;
    protected MaterialMapping matMapping;
    protected ObjectStateCreator stateCreator;

    public State(Object3D obj)
    {
      theTexture = obj.theTexture.duplicate();   // TODO(MB) duplicate?
      theMaterial = obj.theMaterial.duplicate();   // TODO(MB) duplicate?
      if (obj.getTextureMapping() != null)
        texMapping = obj.getTextureMapping().duplicate();
      if (obj.getMaterialMapping() != null)
        matMapping = obj.getMaterialMapping().duplicate();
    }
    
    public void setStateCreator(ObjectStateCreator c)
    {
      stateCreator = c;
    }

    public ObjectStateCreator getStateCreator()
    {
      return stateCreator;
    }

    public abstract ObjectState replay();

  }
  
  /** Default implementation of ObjectState for simple Object3D objects. */
  public static class DefaultStateWrapper implements ObjectState
  {
    protected Object3D data, creator;
    
    public DefaultStateWrapper(Object3D towrap)
    {
      creator = towrap;
      data = creator.duplicate();
    }
    
    /** Set creator of this state. */
    public void setStateCreator(ObjectStateCreator c)
    {
      creator = (Object3D) c;
    }
    
    /** Get creator of this state. */
    public ObjectStateCreator getStateCreator()
    {
      return creator;
    }

    public ObjectState replay()
    {
      Object3D temp = creator.duplicate();
      creator.copyObject(data);
      data = temp;
      return this;
    }
  }
}