/* This is a Module which outputs an image. */

/* Copyright (C) 2000,2003 by Peter Eastman

   This program is free software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the Free Software
   Foundation; either version 2 of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but WITHOUT ANY 
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
   PARTICULAR PURPOSE.  See the GNU General Public License for more details. */

package artofillusion.procedural;

import artofillusion.*;
import artofillusion.image.*;
import artofillusion.math.*;
import artofillusion.ui.*;
import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.util.*;

public class ImageModule extends Module
{
  ImageMap map;
  boolean tilex, tiley, mirrorx, mirrory, wrapx, wrapy;
  boolean pointOk, colorOk, valueOk[], gradOk[], outside;
  double xscale, yscale, xinv, yinv;
  double x, y, xsize, ysize, lastBlur;
  PointInfo point;
  RGBColor color;
  Vec2 tempGrad;
  Vec3 gradient[];

  public ImageModule(Point position)
  {
    super("(Image)", new IOPort [] {new IOPort(IOPort.NUMBER, IOPort.INPUT, IOPort.LEFT, new String [] {"X", "(X)"}), 
      new IOPort(IOPort.NUMBER, IOPort.INPUT, IOPort.LEFT, new String [] {"Y", "(Y)"})}, 
      new IOPort [] {new IOPort(IOPort.COLOR, IOPort.OUTPUT, IOPort.RIGHT, new String [] {"Color"}),
      new IOPort(IOPort.NUMBER, IOPort.OUTPUT, IOPort.RIGHT, new String [] {"Red"}),
      new IOPort(IOPort.NUMBER, IOPort.OUTPUT, IOPort.RIGHT, new String [] {"Green"}),
      new IOPort(IOPort.NUMBER, IOPort.OUTPUT, IOPort.RIGHT, new String [] {"Blue"})}, 
      position);
    xscale = yscale = xinv = yinv = 1.0;
    tilex = tiley = true;
    color = new RGBColor(0.0f, 0.0f, 0.0f);
    tempGrad = new Vec2();
    gradient = new Vec3 [] {new Vec3(), new Vec3(), new Vec3()};
    valueOk = new boolean [3];
    gradOk = new boolean [3];
  }

  /** New point, so the color will need to be recalculated. */

  public void init(PointInfo p)
  {
    point = p;
    pointOk = colorOk = false;
    valueOk[0] = valueOk[1] = valueOk[2] = false;
    gradOk[0] = gradOk[1] = gradOk[2] = false;
  }

  /** Find the point at which the image is being evaluated. */
  
  private void findPoint(double blur)
  {
    pointOk = true;
    colorOk = valueOk[0] = valueOk[1] = valueOk[2] = false;
    x = (linkFrom[0] == null) ? point.x : linkFrom[0].getAverageValue(linkFromIndex[0], blur);
    y = (linkFrom[1] == null) ? point.y : linkFrom[1].getAverageValue(linkFromIndex[1], blur);
    x *= xinv;
    y *= yinv;
    outside = (!tilex && (x < 0.0 || x > 1.0)) || (!tiley && (y < 0.0 || y > 1.0));
    if (outside)
      return;
    if (mirrorx)
      {
	double f = Math.floor(x);
	if ((((int) f)&1) == 0)
	  x = 1.0+f-x;
	else
	  x = x-f;
      }
    else
      x = x-Math.floor(x);
    if (mirrory)
      {
	double f = Math.floor(y);
	if ((((int) f)&1) == 0)
	  y = 1.0+f-y;
	else
	  y = y-f;
      }
    else
      y = y-Math.floor(y);
    xsize = (linkFrom[0] == null) ? 0.5*point.xsize+blur : linkFrom[0].getValueError(linkFromIndex[0], blur);
    ysize = (linkFrom[1] == null) ? 0.5*point.ysize+blur : linkFrom[1].getValueError(linkFromIndex[1], blur);
    xsize *= xinv;
    ysize *= yinv;
    wrapx = tilex && !mirrorx;
    wrapy = tiley && !mirrory;
  }

  /** Calculate the color. */
  
  public void getColor(int which, RGBColor c, double blur)
  {
    if (colorOk && blur == lastBlur)
      {
        c.copy(color);
        return;
      }
    if (map == null)
      {
        color.setRGB(0.0f, 0.0f, 0.0f);
        c.copy(color);
        return;
      }
    if (!pointOk || blur != lastBlur)
      findPoint(blur);
    colorOk = valueOk[0] = valueOk[1] = valueOk[2] = true;
    lastBlur = blur;
    if (outside)
      {
        // The point is outside the map.
        
        color.setRGB(0.0f, 0.0f, 0.0f);
        c.copy(color);
        return;
      }
    map.getColor(color, wrapx, wrapy, x, y, xsize, ysize);
    c.copy(color);
  }
  
  /** Get the value of one of the components. */
  
  public double getAverageValue(int which, double blur)
  {
    int component = which-1;
    if (valueOk[component] && blur == lastBlur)
      {
	if (which == 1)
	  return color.getRed();
	if (which == 2)
	  return color.getGreen();
	return color.getBlue();
      }
    if (!pointOk || blur != lastBlur)
      findPoint(blur);
    if (outside)
      return 0.0;
    valueOk[component] = true;
    float value = map.getComponent(which-1, wrapx, wrapy, x, y, xsize, ysize);
    if (which == 1)
      color.setRGB(value, color.getGreen(), color.getBlue());
    else if (which == 2)
      color.setRGB(color.getRed(), value, color.getBlue());
    else 
      color.setRGB(color.getRed(), color.getGreen(), value);
    return value;
  }

  /** Calculate the gradient of one of the components. */

  public void getValueGradient(int which, Vec3 grad, double blur)
  {
    int component = which-1;
    if (gradOk[component] && blur == lastBlur)
      {
	grad.set(gradient[component]);
	return;
      }
    if (map == null)
      {
	grad.set(0.0, 0.0, 0.0);
	return;
      }
    if (!pointOk || blur != lastBlur)
      findPoint(blur);
    if (outside)
      {
	grad.set(0.0, 0.0, 0.0);
	return;
      }
    map.getGradient(tempGrad, component, wrapx, wrapy, x, y, xsize, ysize);
    double dx = tempGrad.x*xinv, dy = tempGrad.y*yinv;
    Vec3 g = gradient[component];
    if (dx != 0.0)
      {
        if (linkFrom[0] == null)
          g.set(dx, 0.0, 0.0);
        else
          {
            linkFrom[0].getValueGradient(linkFromIndex[0], grad, blur);
            g.x = dx*grad.x;
            g.y = dx*grad.y;
            g.z = dx*grad.z;
          }
      }
    else
      g.set(0.0, 0.0, 0.0);
    if (dy != 0.0)
      {
        if (linkFrom[1] == null)
          g.y += dy;
        else
          {
            linkFrom[1].getValueGradient(linkFromIndex[1], grad, blur);
            g.x += dy*grad.x;
            g.y += dy*grad.y;
            g.z += dy*grad.z;
          }
      }
    gradOk[component] = true;
    grad.set(g);
  }

  public void calcSize()
  {
    bounds.width = ImageMap.PREVIEW_WIDTH+IOPort.SIZE*2;
    bounds.height = ImageMap.PREVIEW_HEIGHT+IOPort.SIZE*2;
  }

  protected void drawContents(Graphics g)
  {
    if (map == null)
      {
        super.drawContents(g);
        return;
      }
    g.drawImage(map.getPreview(), bounds.x+IOPort.SIZE, bounds.y+IOPort.SIZE, null);
  }
  
  /** Create a duplicate of this module. */
  
  public Module duplicate()
  {
    ImageModule mod = new ImageModule(new Point(bounds.x, bounds.y));
    
    mod.map = map;
    mod.color.copy(color);
    mod.tilex = tilex;
    mod.tiley = tiley;
    mod.mirrorx = mirrorx;
    mod.mirrory = mirrory;
    mod.wrapx = wrapx;
    mod.wrapy = wrapy;
    return mod;
  }
  
  /** Get the image map used by this module. */
  
  public ImageMap getMap()
  {
    return map;
  }
  
  /** Set the image map used by this module. */
  
  public void setMap(ImageMap map)
  {
    this.map = map;
  }
  
  /** Allow the user to set a new value. */
  
  public boolean edit(final Frame fr, final Scene theScene)
  {
    ImageMap oldMap = map;
    final ValueField xField = new ValueField(xscale, ValueField.NONE, 10);
    final ValueField yField = new ValueField(yscale, ValueField.NONE, 10);
    Checkbox tilexBox = new Checkbox("X", tilex);
    Checkbox tileyBox = new Checkbox("Y", tiley);
    Checkbox mirrorxBox = new Checkbox("X", mirrorx);
    Checkbox mirroryBox = new Checkbox("Y", mirrory);
    final Canvas preview = new Canvas() {
      public Dimension getPreferredSize()
	{
	  return new Dimension(ImageMap.PREVIEW_WIDTH, ImageMap.PREVIEW_HEIGHT);
	}
      public void paint(Graphics g)
	{
	  if (map != null)
	    g.drawImage(map.getPreview(), 0, 0, null);
	  else
	    g.drawRect(1, 1, ImageMap.PREVIEW_WIDTH-2, ImageMap.PREVIEW_HEIGHT-2);
	} };
    preview.addMouseListener(new MouseAdapter() {
      public void mouseClicked(MouseEvent ev)
	{
	  ImagesDialog dlg = new ImagesDialog(fr, theScene, map);
	  if (dlg.getSelection() != map && dlg.getSelection() != null)
	    {
	      int w = dlg.getSelection().getWidth();
	      int h = dlg.getSelection().getHeight();
	      if (w > h)
		{
		  xField.setValue(1.0);
		  yField.setValue(((double) h)/w);
		}
	      else
		{
		  xField.setValue(((double) w)/h);
		  yField.setValue(1.0);
		}
	    }
	  map = dlg.getSelection();
	  preview.repaint();
	} });
    ComponentsDialog dlg = new ComponentsDialog(fr, "Click to Set Image:", 
      new Component [] {preview, xField, yField, tilexBox, tileyBox, mirrorxBox, mirroryBox},
      new String [] {"", "X Size", "Y Size", "Tile", "", "Mirror", ""});
    if (!dlg.clickedOk())
      {
	map = oldMap;
	return false;
      }
    xscale = xField.getValue();
    yscale = yField.getValue();
    xinv = 1.0/xscale;
    yinv = 1.0/yscale;
    tilex = tilexBox.getState();
    tiley = tileyBox.getState();
    mirrorx = mirrorxBox.getState();
    mirrory = mirroryBox.getState();
    return true;
  }

  /** Write out the parameters. */

  public void writeToStream(DataOutputStream out, Scene theScene) throws IOException
  {
    if (map == null)
      out.writeInt(-1);
    else
      out.writeInt(theScene.indexOf(map));
    out.writeDouble(xscale);
    out.writeDouble(yscale);
    out.writeBoolean(tilex);
    out.writeBoolean(tiley);
    out.writeBoolean(mirrorx);
    out.writeBoolean(mirrory);
  }
  
  /** Read in the parameters. */
  
  public void readFromStream(DataInputStream in, Scene theScene) throws IOException
  {
    int index = in.readInt();
    
    if (index > -1)
      map = theScene.getImage(index);
    else
      map = null;
    xscale = in.readDouble();
    yscale = in.readDouble();
    xinv = 1.0/xscale;
    yinv = 1.0/yscale;
    tilex = in.readBoolean();
    tiley = in.readBoolean();
    mirrorx = in.readBoolean();
    mirrory = in.readBoolean();
  }
}
