/* This is a Module which outputs a per-vertex texture parameter. */

/* Copyright (C) 2000 by Peter Eastman

   This program is free software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the Free Software
   Foundation; either version 2 of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but WITHOUT ANY 
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
   PARTICULAR PURPOSE.  See the GNU General Public License for more details. */

package artofillusion.procedural;

import artofillusion.*;
import artofillusion.math.*;
import artofillusion.ui.*;
import java.awt.*;
import java.io.*;

public class ParameterModule extends Module
{
  double minVal, maxVal, defaultVal;
  int index;
  PointInfo point;

  public ParameterModule(Point position)
  {
    super("Parameter", new IOPort [] {}, 
      new IOPort [] {new IOPort(IOPort.NUMBER, IOPort.OUTPUT, IOPort.RIGHT, new String [] {"Value"})}, 
      position);
    minVal = 0.0f;
    maxVal = 1.0f;
    defaultVal = 0.0f;
  }

  /* Cache the PointInfo object to have access to the coordinates later on. */

  public void init(PointInfo p)
  {
    point = p;
  }

  /* This module outputs the value of the parameter. */
  
  public double getAverageValue(int which, double blur)
  {
    if (point.param == null || point.param.length <= which)
      return defaultVal;
    return point.param[index];
  }
  
  /* Set the index of this parameter. */
  
  public void setIndex(int index)
  {
    this.index = index;
  }
  
  /* Get the texture parameter corresponding to this module. */
  
  public TextureParameter getParameter(Object owner)
  {
    return new TextureParameter(owner, name, minVal, maxVal, defaultVal);
  }

  /* Allow the user to set the parameters. */
  
  public boolean edit(Frame fr, Scene theScene)
  {
    TextField nameField = new TextField(name);
    ValueField minField = new ValueField(minVal, ValueField.NONE);
    ValueField maxField = new ValueField(maxVal, ValueField.NONE);
    ValueField defaultField = new ValueField(defaultVal, ValueField.NONE);
    ComponentsDialog dlg = new ComponentsDialog(fr, "Set Parameter Properties:", 
      new Component [] {nameField, minField, maxField, defaultField},
      new String [] {Translate.text("Name"), "Minimum", "Maximum", "Default"});
    if (!dlg.clickedOk())
      return false;
    name = nameField.getText();
    minVal = minField.getValue();
    maxVal = maxField.getValue();
    defaultVal = defaultField.getValue();
    if (minVal > maxVal)
      {
        new MessageDialog(fr, "The minimum value cannot be greater than then maximum value.");
        edit(fr, theScene);
      }
    if (minVal > defaultVal || maxVal < defaultVal)
      {
        new MessageDialog(fr, "The default value must be between the minimum and maximum values.");
        edit(fr, theScene);
      }
    return true;
  }
  
  /* Create a duplicate of this module. */
  
  public Module duplicate()
  {
    ParameterModule mod = new ParameterModule(new Point(bounds.x, bounds.y));
    
    mod.name = name;
    mod.minVal = minVal;
    mod.maxVal = maxVal;
    mod.defaultVal = defaultVal;
    mod.layout();
    return mod;
  }

  /* Write out the parameters. */

  public void writeToStream(DataOutputStream out, Scene theScene) throws IOException
  {
    out.writeUTF(name);
    out.writeDouble(minVal);
    out.writeDouble(maxVal);
    out.writeDouble(defaultVal);
  }
  
  /* Read in the parameters. */
  
  public void readFromStream(DataInputStream in, Scene theScene) throws IOException
  {
    name = in.readUTF();
    minVal = in.readDouble();
    maxVal = in.readDouble();
    defaultVal = in.readDouble();
    layout();
  }
}
