/* This class presents a user interface for entering scripts to be executed. */

/* Copyright (C) 2002 by Peter Eastman

   This program is free software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the Free Software
   Foundation; either version 2 of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but WITHOUT ANY 
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
   PARTICULAR PURPOSE.  See the GNU General Public License for more details. */

package artofillusion.script;

import artofillusion.*;
import artofillusion.ui.*;
import bsh.*;
import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.util.*;

public class ExecuteScriptWindow extends Frame implements ActionListener, KeyListener
{
  private LayoutWindow window;
  private TextArea scriptText;
  private String scriptName;
  
  private static String scriptDir;
  private static String lastScript;
  
  public ExecuteScriptWindow(LayoutWindow win)
  {
    window = win;
    scriptName = "Untitled.bsh";
    if (scriptDir == null)
      scriptDir = ModellingApp.TOOL_SCRIPT_DIRECTORY;
    scriptText = new TextArea(15, 60);
    if (lastScript != null)
      scriptText.setText(lastScript);
    scriptText.setFont(ModellingApp.defaultFont);
//    scriptText.addKeyListener(this);
    add(scriptText, BorderLayout.CENTER);
    Panel p = new Panel();
    add(p, BorderLayout.SOUTH);
    Button b;
    p.add(b = new Button("Execute"));
    b.setActionCommand("execute");
    b.addActionListener(this);
    p.add(b = new Button("Load..."));
    b.setActionCommand("load");
    b.addActionListener(this);
    p.add(b = new Button("Save..."));
    b.setActionCommand("save");
    b.addActionListener(this);
    p.add(b = new Button("Close"));
    b.setActionCommand("close");
    b.addActionListener(this);
    pack();
    ModellingApp.centerWindow(this);
    show();
  }
  
  public void actionPerformed(ActionEvent ev)
  {
    String command = ev.getActionCommand();
    
    if ("execute".equals(command))
      executeScript();
    if ("load".equals(command))
      loadScript();
    if ("save".equals(command))
      saveScript();
    if ("close".equals(command))
      {
        lastScript = scriptText.getText();
        dispose();
      }
  }
  
  /** Prompt the user to load a script. */
  
  private void loadScript()
  {
    FileDialog dlg = new FileDialog(this, "Select Script to Load", FileDialog.LOAD);
    dlg.setDirectory(scriptDir);
    dlg.show();
    if (dlg.getFile() == null)
      return;
    scriptDir = dlg.getDirectory();
    setCursor(Cursor.WAIT_CURSOR);
    File f = new File(dlg.getDirectory(), dlg.getFile());
    try
      {
        BufferedReader in = new BufferedReader(new FileReader(f));
        StringBuffer buf = new StringBuffer();
        int c;
        while ((c = in.read()) != -1)
          buf.append((char) c);
        in.close();
        scriptText.setText(buf.toString());
      }
    catch (Exception ex)
      {
        new MessageDialog(this, new String [] {
          "An error occurred while reading the script:",
          ex.getMessage()});
      }
    scriptName = dlg.getFile();
    setCursor(Cursor.DEFAULT_CURSOR);
  }
  
  /** Prompt the user to save a script. */
  
  private void saveScript()
  {
    FileDialog dlg = new FileDialog(this, "Save Script to File", FileDialog.SAVE);
    dlg.setFile(scriptName);
    dlg.setDirectory(scriptDir);
    dlg.show();
    if (dlg.getFile() == null)
      return;
    scriptDir = dlg.getDirectory();
    
    // Write the script to disk.
    
    setCursor(Cursor.WAIT_CURSOR);
    File f = new File(dlg.getDirectory(), dlg.getFile());
    try
      {
        BufferedWriter out = new BufferedWriter(new FileWriter(f));
        out.write(scriptText.getText().toCharArray());
        out.close();
      }
    catch (Exception ex)
      {
        new MessageDialog(this, new String [] {
          "An error occurred while writing the script:",
          ex.getMessage()});
      }
    scriptName = dlg.getFile();
    
    // Update the Scripts menus in all windows.
    
    EditingWindow allWindows[] = ModellingApp.getWindows();
    for (int i = 0; i < allWindows.length; i++)
      if (allWindows[i] instanceof LayoutWindow)
        ((LayoutWindow) allWindows[i]).rebuildMenuBar();
    setCursor(Cursor.DEFAULT_CURSOR);
  }
  
  /** Execute the script. */
  
  private void executeScript()
  {
    try
      {
        ToolScript script = ScriptRunner.parseToolScript(scriptText.getText());
        script.execute(window);
      }
    catch (Exception e)
      {
        int line = ScriptRunner.displayError(e, this, 1);
        if (line > -1)
          {
            // Find the start of the line containing the error.
            
            String text = scriptText.getText();
            int index = 0;
            for (int i = 0; i < line-1; i++)
              {
                int next = text.indexOf('\n', index);
                if (next == -1)
                  {
                    index = -1;
                    break;
                  }
                index = next+1;
              }
            if (index > -1)
              scriptText.setCaretPosition(index);
            scriptText.requestFocus();
          }
      }
    window.updateImage();
    toFront();
  }
  
  public void keyPressed(KeyEvent e)
  {
  }
  
  public void keyReleased(KeyEvent e)
  {
  }
  
  /** Provide auto-indent behavior when the user presses Return. */
  
  public void keyTyped(KeyEvent e)
  {
    if (e.getKeyChar() != '\r' && e.getKeyChar() != '\n')
      return;
    
    // The user has just pressed Return.  Find any white space characters
    // at the beginning of the previous line, and insert them before this line
    // as well.
    
    String text = scriptText.getText();
    int pos = scriptText.getCaretPosition()+1;
    int endWhiteSpace = pos-1;
    int startLine = pos-1;
    boolean extraIndent = (startLine > 0 && text.charAt(startLine-1) == '{');
    while (startLine > 0)
      {
        char c = text.charAt(startLine-1);
        if (c == '\r' || c == '\n')
          break;
        startLine--;
        if (!Character.isWhitespace(c))
          endWhiteSpace = startLine;
      }
    if (extraIndent)
      scriptText.insert("  ", pos);
    scriptText.insert(text.substring(startLine, endWhiteSpace), pos);
  }
}
