/* This class presents a user interface for entering object scripts. */

/* Copyright (C) 2002 by Peter Eastman

   This program is free software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the Free Software
   Foundation; either version 2 of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but WITHOUT ANY 
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
   PARTICULAR PURPOSE.  See the GNU General Public License for more details. */

package artofillusion.script;

import artofillusion.*;
import artofillusion.object.*;
import artofillusion.ui.*;
import bsh.*;
import java.awt.*;
import java.awt.event.*;
import java.io.*;

public class ScriptedObjectEditorWindow extends Frame implements ActionListener, KeyListener
{
  private EditingWindow window;
  private ObjectInfo info;
  private TextArea scriptText;
  private String scriptName;
  
  private static String scriptDir;
  
  public ScriptedObjectEditorWindow(EditingWindow parent, ObjectInfo obj)
  {
    super("Script '"+obj.name+"'");
    window = parent;
    info = obj;
    scriptName = "Untitled.bsh";
    if (scriptDir == null)
      scriptDir = ModellingApp.OBJECT_SCRIPT_DIRECTORY;
    scriptText = new TextArea(((ScriptedObject) info.object).getScript(), 15, 60);
    scriptText.setFont(ModellingApp.defaultFont);
//    scriptText.addKeyListener(this);
    add(scriptText, BorderLayout.CENTER);
    Panel p = new Panel();
    add(p, BorderLayout.SOUTH);
    Button b;
    p.add(b = new Button(Translate.text("button.ok")));
    b.setActionCommand("execute");
    b.addActionListener(this);
    p.add(b = new Button("Load..."));
    b.setActionCommand("load");
    b.addActionListener(this);
    p.add(b = new Button("Save..."));
    b.setActionCommand("save");
    b.addActionListener(this);
    p.add(b = new Button("Parameters..."));
    b.setActionCommand("parameters");
    b.addActionListener(this);
    p.add(b = new Button(Translate.text("button.cancel")));
    b.setActionCommand("close");
    b.addActionListener(this);
    pack();
    ModellingApp.centerWindow(this);
    show();
  }
  
  public void actionPerformed(ActionEvent ev)
  {
    String command = ev.getActionCommand();
    
    if ("execute".equals(command))
      commitChanges();
    if ("load".equals(command))
      loadScript();
    if ("save".equals(command))
      saveScript();
    if ("parameters".equals(command))
      new ParametersDialog();
    if ("close".equals(command))
      dispose();
  }
  
  /** Prompt the user to load a script. */
  
  private void loadScript()
  {
    FileDialog dlg = new FileDialog(this, "Select Script to Load", FileDialog.LOAD);
    dlg.setDirectory(scriptDir);
    dlg.show();
    if (dlg.getFile() == null)
      return;
    scriptDir = dlg.getDirectory();
    setCursor(Cursor.WAIT_CURSOR);
    File f = new File(dlg.getDirectory(), dlg.getFile());
    try
      {
        BufferedReader in = new BufferedReader(new FileReader(f));
        StringBuffer buf = new StringBuffer();
        int c;
        while ((c = in.read()) != -1)
          buf.append((char) c);
        in.close();
        scriptText.setText(buf.toString());
      }
    catch (Exception ex)
      {
        new MessageDialog(this, new String [] {
          "An error occurred while reading the script:",
          ex.getMessage()});
      }
    scriptName = dlg.getFile();
    setCursor(Cursor.DEFAULT_CURSOR);
  }
  
  /** Prompt the user to save a script. */
  
  private void saveScript()
  {
    FileDialog dlg = new FileDialog(this, "Save Script to File", FileDialog.SAVE);
    dlg.setFile(scriptName);
    dlg.setDirectory(scriptDir);
    dlg.show();
    if (dlg.getFile() == null)
      return;
    scriptDir = dlg.getDirectory();
    
    // Write the script to disk.
    
    setCursor(Cursor.WAIT_CURSOR);
    File f = new File(dlg.getDirectory(), dlg.getFile());
    try
      {
        BufferedWriter out = new BufferedWriter(new FileWriter(f));
        out.write(scriptText.getText().toCharArray());
        out.close();
      }
    catch (Exception ex)
      {
        new MessageDialog(this, new String [] {
          "An error occurred while writing the script:",
          ex.getMessage()});
      }
    scriptName = dlg.getFile();
    setCursor(Cursor.DEFAULT_CURSOR);
  }
  
  /** Commit changes to the scripted object. */
  
  private void commitChanges()
  {
    ScriptedObject so = (ScriptedObject) info.object;
    setCursor(Cursor.WAIT_CURSOR);
    so.setScript(scriptText.getText());
    so.sceneChanged(info, window.getScene());
    window.updateImage();
    setCursor(Cursor.DEFAULT_CURSOR);
    dispose();
  }
  
  public void keyPressed(KeyEvent e)
  {
  }
  
  public void keyReleased(KeyEvent e)
  {
  }
  
  /** Provide auto-indent behavior when the user presses Return. */
  
  public void keyTyped(KeyEvent e)
  {
    if (e.getKeyChar() != '\r' && e.getKeyChar() != '\n')
      return;
    
    // The user has just pressed Return.  Find any white space characters
    // at the beginning of the previous line, and insert them before this line
    // as well.
    
    String text = scriptText.getText();
    int pos = scriptText.getCaretPosition();
    int endWhiteSpace = pos-1;
    int startLine = pos-1;
    boolean extraIndent = (startLine > 0 && text.charAt(startLine-1) == '{');
    while (startLine > 0)
      {
        char c = text.charAt(startLine-1);
        if (c == '\r' || c == '\n')
          break;
        startLine--;
        if (!Character.isWhitespace(c))
          endWhiteSpace = startLine;
      }
//    if (extraIndent)
//      scriptText.insert("  ", pos);
//    scriptText.insert(text.substring(startLine, endWhiteSpace), pos);
  }
  
  /** This is an inner class for editing the list of parameters on the object. */
  
  private class ParametersDialog extends Dialog implements ActionListener, ItemListener, TextListener
  {
    private ScriptedObject script;
    private List paramList;
    private TextField nameField;
    private ValueField valueField;
    private String name[];
    private double value[];
    private int current;
    
    public ParametersDialog()
    {
      super(ScriptedObjectEditorWindow.this, "Object Parameters", true);
      script = (ScriptedObject) info.object;
      setLayout(new GridBagLayout());
      GridBagConstraints gc = new GridBagConstraints();
      gc.gridx = 0;
      gc.gridwidth = 2;
      gc.fill = gc.BOTH;
      gc.weighty = 1.0;
      name = new String [script.getNumParameters()];
      value = new double [script.getNumParameters()];
      for (int i = 0; i < name.length; i++)
        {
          name[i] = script.getParameterName(i);
          value[i] = script.getParameterValue(i);
        }
      add(paramList = new List(), gc);
      buildParameterList();
      paramList.addItemListener(this);
      gc.fill = gc.HORIZONTAL;
      gc.weighty = 0.0;
      gc.gridwidth = 1;
      add(Translate.label("Name"), gc);
      add(Translate.label("Value"), gc);
      gc.gridx = 1;
      gc.weightx = 1.0;
      add(nameField = new TextField(), gc);
      nameField.addTextListener(this);
      nameField.addFocusListener(new FocusAdapter() {
        public void focusLost(FocusEvent ev)
        {
          paramList.replaceItem(name[current], current);
        }
      } );
      add(valueField = new ValueField(0.0, ValueField.NONE), gc);
      valueField.addTextListener(this);
      gc.gridx = 0;
      gc.gridwidth = 2;
      Panel p = new Panel();
      add(p, gc);
      Button b;
      p.add(b = new Button("Add"));
      b.setActionCommand("add");
      b.addActionListener(this);
      p.add(b = new Button("Remove"));
      b.setActionCommand("remove");
      b.addActionListener(this);
      p.add(Translate.button("ok", this));
      p.add(Translate.button("cancel", this));
      setSelectedParameter(name.length == 0 ? -1 : 0);
      pack();
      show();
    }
    
    /** Build the list of parameters. */
    
    private void buildParameterList()
    {
      paramList.removeAll();
      for (int i = 0; i < name.length; i++)
        paramList.add(name[i]);
      if (name.length == 0)
        paramList.add("(no parameters)");
    }
    
    /** Update the components to show the currently selected parameter. */
    
    private void setSelectedParameter(int which)
    {
      if (which != paramList.getSelectedIndex())
        paramList.select(which);
      current = which;
      if (which == -1)
        {
          nameField.setEnabled(false);
          valueField.setEnabled(false);
        }
      else
        {
          nameField.setEnabled(true);
          valueField.setEnabled(true);
          nameField.setText(name[which]);
          valueField.setValue(value[which]);
        }
    }
    
    /** Deal with changes to the text fields. */
    
    public void textValueChanged(TextEvent ev)
    {
      if (current < 0 || current > name.length)
        return;
      if (ev.getSource() == nameField)
        {
          name[current] = nameField.getText();
//          paramList.replaceItem(name[current], current);
        }
      else
        value[current] = valueField.getValue();
    }
    
    /** Deal with selection changes. */
    
    public void itemStateChanged(ItemEvent ev)
    {
      setSelectedParameter(paramList.getSelectedIndex());
    }
    
    /** Deal with button clicks. */
    
    public void actionPerformed(ActionEvent ev)
    {
      String command = ev.getActionCommand();
      
      if ("add".equals(command))
        {
          String newName[] = new String [name.length+1];
          double newValue[] = new double [value.length+1];
          System.arraycopy(name, 0, newName, 0, name.length);
          System.arraycopy(value, 0, newValue, 0, value.length);
          newName[name.length] = "";
          newValue[value.length] = 0.0;
          name = newName;
          value = newValue;
          buildParameterList();
          setSelectedParameter(name.length-1);
          nameField.requestFocus();
        }
      if ("remove".equals(command))
        {
          int which = paramList.getSelectedIndex();
          String newName[] = new String [name.length-1];
          double newValue[] = new double [value.length-1];
          for (int i = 0, j = 0; i < name.length; i++)
            {
              if (i == which)
                continue;
              newName[j] = name[i];
              newValue[j] = value[i];
              j++;
            }
          name = newName;
          value = newValue;
          buildParameterList();
          setSelectedParameter(-1);
        }
      if ("ok".equals(command))
        {
          script.setParameters(name, value);
          dispose();
        }
      if ("cancel".equals(command))
        dispose();
    }
  }
}
