/* CylindricalMapping is a Mapping2D which wraps the texture around a cylinder. */

/* Copyright (C) 2000-2003 by Peter Eastman

   This program is free software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the Free Software
   Foundation; either version 2 of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but WITHOUT ANY 
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
   PARTICULAR PURPOSE.  See the GNU General Public License for more details. */

package artofillusion.texture;

import artofillusion.*;
import artofillusion.animation.*;
import artofillusion.math.*;
import artofillusion.object.*;
import artofillusion.ui.*;
import java.awt.*;
import java.awt.event.*;
import java.io.*;

public class CylindricalMapping extends NonlinearMapping2D
{
  CoordinateSystem coords;
  double xscale, yscale, ax, ay, dy;
  Mat4 toLocal, fromLocal;
  boolean transform;
  TextureParameter xparam, yparam, zparam;

  public CylindricalMapping(Texture theTexture)
  {
    super(theTexture);
    coords = new CoordinateSystem(new Vec3(), new Vec3(0.0, 0.0, 1.0), new Vec3(0.0, 1.0, 0.0));
    xscale = 360.0;
    yscale = 1.0;
    dy = 0.5;
    findCoefficients();
  }

  public static String getName()
  {
    return "Cylindrical";
  }

  /* Calculate the mapping coefficients. */
  
  void findCoefficients()
  {
    ax = -180.0/(Math.PI*xscale);
    ay = 1.0/yscale;
    toLocal = coords.toLocal();
    fromLocal = coords.fromLocal();
    transform = (fromLocal.m11 != 1.0 || fromLocal.m22 != 1.0 || fromLocal.m33 != 1.0);
  }

  /* Methods from TextureMapping. */

  public RenderingTriangle mapTriangle(int v1, int v2, int v3, int n1, int n2, int n3, Vec3 vert[])
  {
    Vec3 c1 = toLocal.timesDirection(vert[v1]), c2 = toLocal.timesDirection(vert[v2]), c3 = toLocal.timesDirection(vert[v3]);
    
    c1.y *= ay;
    c2.y *= ay;
    c3.y *= ay;
    return new Nonlinear2DTriangle(v1, v2, v3, n1, n2, n3, c1, c2, c3, this);
  }
  
  public void getTextureSpec(Vec3 pos, TextureSpec spec, double angle, double size, double t, double param[])
  {
    double px, py, pz, x, y, z, r, s, theta;
    
    if (!appliesToFace(angle > 0.0))
      {
        spec.diffuse.setRGB(0.0f, 0.0f, 0.0f);
        spec.specular.setRGB(0.0f, 0.0f, 0.0f);
        spec.transparent.setRGB(1.0f, 1.0f, 1.0f);
        spec.emissive.setRGB(0.0f, 0.0f, 0.0f);
        spec.roughness = spec.cloudiness = 0.0;
        spec.bumpGrad.set(0.0, 0.0, 0.0);
        return;
      }
    if (coordsFromParams && numTextureParams < param.length && param[numTextureParams] != Double.MAX_VALUE)
      {
	px = param[numTextureParams];
	py = param[numTextureParams+1];
	pz = param[numTextureParams+2];
      }
    else
      {
	px = pos.x;
	py = pos.y;
	pz = pos.z;
      }
    x = toLocal.m11*px + toLocal.m12*py + toLocal.m13*pz;
    y = toLocal.m21*px + toLocal.m22*py + toLocal.m23*pz;
    z = toLocal.m31*px + toLocal.m32*py + toLocal.m33*pz;
    theta = Math.atan(z/x);
    if (x < 0.0)
      theta += Math.PI;
    r = Math.sqrt(x*x+z*z);
    texture.getTextureSpec(spec, theta*ax, y*ay+dy, Math.abs(size*ax/r), Math.abs(size*ay), angle, t, param);
    if (texture.hasComponent(Texture.BUMP_COMPONENT))
      {
	s = spec.bumpGrad.x*ax/r;
	spec.bumpGrad.set(-s*z, spec.bumpGrad.y*ay, s*x);
	if (transform)
	  fromLocal.transform(spec.bumpGrad);
      }
  }

  public void getTransparency(Vec3 pos, RGBColor trans, double angle, double size, double t, double param[])
  {
    double px, py, pz, x, y, z, r, s, theta;
    
    if (!appliesToFace(angle > 0.0))
      {
        trans.setRGB(1.0f, 1.0f, 1.0f);
        return;
      }
    if (coordsFromParams && numTextureParams < param.length && param[numTextureParams] != Double.MAX_VALUE)
      {
	px = param[numTextureParams];
	py = param[numTextureParams+1];
	pz = param[numTextureParams+2];
      }
    else
      {
	px = pos.x;
	py = pos.y;
	pz = pos.z;
      }
    x = toLocal.m11*px + toLocal.m12*py + toLocal.m13*pz;
    y = toLocal.m21*px + toLocal.m22*py + toLocal.m23*pz;
    z = toLocal.m31*px + toLocal.m32*py + toLocal.m33*pz;
    theta = Math.atan(z/x);
    if (x < 0.0)
      theta += Math.PI;
    r = Math.sqrt(x*x+z*z);
    texture.getTransparency(trans, theta*ax, y*ay+dy, Math.abs(size*ax/r), Math.abs(size*ay), angle, t, param);
  }

  public double getDisplacement(Vec3 pos, double size, double t, double param[])
  {
    double px, py, pz, x, y, z, r, s, theta;
    
    if (coordsFromParams && numTextureParams < param.length && param[numTextureParams] != Double.MAX_VALUE)
      {
	px = param[numTextureParams];
	py = param[numTextureParams+1];
	pz = param[numTextureParams+2];
      }
    else
      {
	px = pos.x;
	py = pos.y;
	pz = pos.z;
      }
    x = toLocal.m11*px + toLocal.m12*py + toLocal.m13*pz;
    y = toLocal.m21*px + toLocal.m22*py + toLocal.m23*pz;
    z = toLocal.m31*px + toLocal.m32*py + toLocal.m33*pz;
    theta = Math.atan(z/x);
    if (x < 0.0)
      theta += Math.PI;
    r = Math.sqrt(x*x+z*z);
    return texture.getDisplacement(theta*ax, y*ay+dy, Math.abs(size*ax/r), Math.abs(size*ay), t, param);
  }

  public Mat4 getPreTransform()
  {
    return toLocal;
  }

  public void getSpecIntermed(TextureSpec spec, double x, double y, double z, double size, double angle, double t, double param[])
  {
    double r = Math.sqrt(x*x+z*z), s, theta = Math.atan(z/x);
    if (x < 0.0)
      theta += Math.PI;
    texture.getTextureSpec(spec, theta*ax, y*ay+dy, Math.abs(size*ax/r), Math.abs(size*ay), angle, t, param);
    if (texture.hasComponent(Texture.BUMP_COMPONENT))
      {
	s = spec.bumpGrad.x*ax/r;
	spec.bumpGrad.set(-s*z, spec.bumpGrad.y*ay, s*x);
	if (transform)
	  fromLocal.transform(spec.bumpGrad);
      }
  }

  public void getTransIntermed(RGBColor trans, double x, double y, double z, double size, double angle, double t, double param[])
  {
    double r = Math.sqrt(x*x+z*z), s, theta = Math.atan(z/x);
    if (x < 0.0)
      theta += Math.PI;
    texture.getTransparency(trans, theta*ax, y*ay+dy, Math.abs(size*ax/r), Math.abs(size*ay), angle, t, param);
  }

  public double getDisplaceIntermed(double x, double y, double z, double size, double t, double param[])
  {
    double r = Math.sqrt(x*x+z*z), s, theta = Math.atan(z/x);

    if (x < 0.0)
      theta += Math.PI;
    return texture.getDisplacement(theta*ax, y*ay+dy, Math.abs(size*ax/r), Math.abs(size*ay), t, param);
  }
  
  /** Given a Mesh to which this mapping has been applied, and the full list
      of texture parameters for that mesh, return the texture coordinates at
      each vertex. */
  
  public Vec2 [] findTextureCoordinates(Mesh mesh, TextureParameter param[])
  {
    int ix = -1, iy = -1, iz = -1;
    if (coordsFromParams)
      for (int i = 0; i < param.length; i++)
        {
          if (param[i].equals(xparam))
            ix = i;
          else if (param[i].equals(yparam))
            iy = i;
          else if (param[i].equals(zparam))
            iz = i;
        }
    MeshVertex vert[] = mesh.getVertices();
    Vec2 uv[] = new Vec2 [vert.length];
    for (int i = 0; i < vert.length; i++)
      {
        double x, y, z;
        if (ix == -1)
          {
            x = vert[i].r.x;
            y = vert[i].r.y;
            z = vert[i].r.z;
          }
        else
          {
            x = vert[i].param[ix];
            y = vert[i].param[iy];
            z = vert[i].param[iz];
          }
        double theta = Math.atan(z/x);
        if (x < 0.0)
          theta += Math.PI;
        if (Double.isNaN(theta))
          theta = 0.0;
        uv[i] = new Vec2(theta*ax, y*ay+dy);
      }
    return uv;
  }

  public TextureMapping duplicate()
  {
    return duplicate(texture);
  }

  public TextureMapping duplicate(Texture tex)
  {
    CylindricalMapping map = new CylindricalMapping(tex);
    
    map.coords = coords.duplicate();
    map.dy = dy;
    map.xscale = xscale;
    map.yscale = yscale;
    map.findCoefficients();
    map.coordsFromParams = coordsFromParams;
    map.numTextureParams = numTextureParams;
    map.setAppliesTo(appliesTo());
    map.xparam = xparam;
    map.yparam = yparam;
    map.zparam = zparam;
    return map;
  }
  
  public void copy(TextureMapping mapping)
  {
    CylindricalMapping map = (CylindricalMapping) mapping; 
    
    coords = map.coords.duplicate();
    dy = map.dy;
    xscale = map.xscale;
    yscale = map.yscale;
    findCoefficients();
    coordsFromParams = map.coordsFromParams;
    numTextureParams = map.numTextureParams;
    setAppliesTo(map.appliesTo());
    xparam = map.xparam;
    yparam = map.yparam;
    zparam = map.zparam;
  }

  /* Get the list of texture parameters associated with this mapping and its texture.
     That includes the texture's parameters, and possibly parameters for the texture
     coordinates. */
  
  public TextureParameter [] getParameters()
  {
    if (!coordsFromParams)
      return getTexture().getParameters();
    TextureParameter tp[] = getTexture().getParameters();
    numTextureParams = tp.length;
    TextureParameter p[] = new TextureParameter [numTextureParams+3];
    System.arraycopy(tp, 0, p, 0, numTextureParams);
    if (xparam == null)
      {
        xparam = new TextureParameter(this, "X", -Double.MAX_VALUE, Double.MAX_VALUE, Double.MAX_VALUE);
        yparam = new TextureParameter(this, "Y", -Double.MAX_VALUE, Double.MAX_VALUE, Double.MAX_VALUE);
        zparam = new TextureParameter(this, "Z", -Double.MAX_VALUE, Double.MAX_VALUE, Double.MAX_VALUE);
        xparam.type = TextureParameter.X_COORDINATE;
        yparam.type = TextureParameter.Y_COORDINATE;
        zparam.type = TextureParameter.Z_COORDINATE;
        xparam.assignNewID();
        yparam.assignNewID();
        zparam.assignNewID();
      }
    p[numTextureParams] = xparam;
    p[numTextureParams+1] = yparam;
    p[numTextureParams+2] = zparam;
    return p;
  }

  public Panel getEditingPanel(ObjectInfo obj, MaterialPreviewer preview)
  {
    return new Editor(obj.object, preview);
  }
  
  public CylindricalMapping(DataInputStream in, Texture theTexture) throws IOException, InvalidObjectException
  {
    super(theTexture);

    short version = in.readShort();
    if (version < 0 || version > 1)
      throw new InvalidObjectException("");
    coords = new CoordinateSystem(in);
    dy = in.readDouble();
    xscale = in.readDouble();
    yscale = in.readDouble();
    findCoefficients();
    coordsFromParams = in.readBoolean();
    if (version == 1)
      setAppliesTo(in.readShort());
  }
  
  public void writeToFile(DataOutputStream out) throws IOException
  {
    out.writeShort(1);
    coords.writeToFile(out);
    out.writeDouble(dy);
    out.writeDouble(xscale);
    out.writeDouble(yscale);
    out.writeBoolean(coordsFromParams);
    out.writeShort(appliesTo());
  }

  /* Editor is an inner class for editing the mapping. */

  class Editor extends Panel implements ItemListener, TextListener
  {
    ValueField xrotField, yrotField, zrotField, xscaleField, yscaleField, ytransField;
    Checkbox coordsFromParamsBox;
    Choice applyToChoice;
    Object3D theObject;
    MaterialPreviewer preview;

    public Editor(Object3D obj, MaterialPreviewer preview)
    {
      super();
      
      theObject = obj;
      this.preview = preview;
      
      // Add the various components to the Panel.
      
      GridBagConstraints gc = new GridBagConstraints();
      Panel p = new Panel();
      setLayout(new GridBagLayout());
      gc.insets = new Insets(0, 0, 0, 5);
      gc.gridy = 0;
      p.setLayout(new GridBagLayout());
      p.add(new Label("Width"), gc);
      p.add(xscaleField = new ValueField(Math.round(xscale*1000.0)/1000.0, ValueField.NONZERO, 5), gc);
      p.add(new Label("(degrees)"), gc);
      gc.gridy = 1;
      p.add(new Label("Height"), gc);
      p.add(yscaleField = new ValueField(Math.round(yscale*1000.0)/1000.0, ValueField.NONZERO, 5), gc);
      gc.gridy = 2;
      p.add(new Label("Offset"), gc);
      p.add(ytransField = new ValueField(Math.round(dy*1000.0)/1000.0, ValueField.NONE, 5), gc);
      gc.gridwidth = 6;
      gc.gridy = 0;
      add(p, gc);
      gc.gridy = 1;
      add(new Label("Rotation:"), gc);
      gc.gridwidth = 1;
      gc.gridy = 2;
      double angles[] = coords.getRotationAngles();
      add(new Label("X"), gc);
      add(xrotField = new ValueField(Math.round(angles[0]*1000.0)/1000.0, ValueField.NONE, 5), gc);
      add(new Label("Y"), gc);
      add(yrotField = new ValueField(Math.round(angles[1]*1000.0)/1000.0, ValueField.NONE, 5), gc);
      add(new Label("Z"), gc);
      add(zrotField = new ValueField(Math.round(angles[2]*1000.0)/1000.0, ValueField.NONE, 5), gc);
      gc.gridwidth = 6;
      gc.gridy = 3;
      p = new Panel();
      add(p, gc);
      p.add(new Label("Apply To:"));
      p.add(applyToChoice = new Choice());
      applyToChoice.add("Front and Back Faces");
      applyToChoice.add("Front Faces Only");
      applyToChoice.add("Back Faces Only");
      applyToChoice.select(appliesTo());
      gc.gridy = 4;
      add(coordsFromParamsBox = new Checkbox("Bind Texture Coordinates to Surface", coordsFromParams), gc);
      coordsFromParamsBox.setEnabled(theObject instanceof TriangleMesh || theObject instanceof SplineMesh || theObject instanceof Actor);
      xscaleField.addTextListener(this);
      yscaleField.addTextListener(this);
      ytransField.addTextListener(this);
      xrotField.addTextListener(this);
      yrotField.addTextListener(this);
      zrotField.addTextListener(this);
      coordsFromParamsBox.addItemListener(this);
      applyToChoice.addItemListener(this);
    }

    public void itemStateChanged(ItemEvent e)
    {
      coordsFromParams = coordsFromParamsBox.getState();
      setAppliesTo((short) applyToChoice.getSelectedIndex());
      preview.setTexture(getTexture(), CylindricalMapping.this);
      preview.render();
    }

    public void textValueChanged(TextEvent e)
    {
      xscale = xscaleField.getValue();
      yscale = yscaleField.getValue();
      dy = ytransField.getValue();
      coords.setOrientation(xrotField.getValue(), yrotField.getValue(), zrotField.getValue());
      findCoefficients();
      preview.render();
    }
  }
}