/* Copyright (C) 2001-2003 by Peter Eastman

   This program is free software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the Free Software
   Foundation; either version 2 of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but WITHOUT ANY 
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
   PARTICULAR PURPOSE.  See the GNU General Public License for more details. */

package artofillusion.texture;

import artofillusion.*;
import artofillusion.image.*;
import artofillusion.math.*;
import artofillusion.ui.*;
import java.awt.*;
import java.awt.event.*;
import java.io.*;

/** ImageMapTexture represents a texture whose properties are defined by images. */

public class ImageMapTexture extends Texture2D
{
  public ImageOrColor diffuseColor, specularColor, transparentColor, emissiveColor;
  public ImageOrValue roughness, cloudiness, transparency, specularity, shininess, bump, displacement;
  public boolean tileX, tileY, mirrorX, mirrorY;
  Vec2 tempGrad;
  
  public ImageMapTexture()
  {
    RGBColor white = new RGBColor(1.0f, 1.0f, 1.0f);
    
    // Generate the default texture.

    diffuseColor = new ImageOrColor(white);
    specularColor = new ImageOrColor(white);
    transparentColor = new ImageOrColor(white);
    emissiveColor = new ImageOrColor(new RGBColor(0.0f, 0.0f, 0.0f));
    transparency = new ImageOrValue(0.0f);
    specularity = new ImageOrValue(0.0f);
    shininess = new ImageOrValue(0.0f);
    roughness = new ImageOrValue(0.2f);
    cloudiness = new ImageOrValue(0.0f);
    bump = new ImageOrValue(0.0f);
    displacement = new ImageOrValue(0.0f);
    tileX = tileY = true;
    name = "";
    tempGrad = new Vec2();
  }

  public static String getTypeName()
  {
    return "Image Mapped";
  }

  public void getTextureSpec(TextureSpec spec, double x, double y, double xsize, double ysize, double angle, double t, double param[])
  {
    float transVal, specVal;
    boolean wrapx, wrapy;
    double f;
    
    if ((!tileX && (x < 0.0 || x > 1.0)) || (!tileY && (y < 0.0 || y > 1.0)))
      {
	// The point falls outside the image, so set the surfaceSpec to be transparent.
	
	spec.diffuse.setRGB(0.0f, 0.0f, 0.0f);
	spec.specular.setRGB(0.0f, 0.0f, 0.0f);
	spec.hilight.setRGB(0.0f, 0.0f, 0.0f);
	spec.transparent.setRGB(1.0f, 1.0f, 1.0f);
	spec.emissive.setRGB(0.0f, 0.0f, 0.0f);
	spec.roughness = spec.cloudiness = 1.0;
	spec.bumpGrad.set(0.0, 0.0, 0.0);
	return;
      }
    if (mirrorX)
      {
	f = Math.floor(x);
	if ((((int) f)&1) == 0)
	  x = 1.0+f-x;
	else
	  x = x-f;
      }
    else
      x = x-Math.floor(x);
    if (mirrorY)
      {
	f = Math.floor(y);
	if ((((int) f)&1) == 0)
	  y = 1.0+f-y;
	else
	  y = y-f;
      }
    else
      y = y-Math.floor(y);
    wrapx = tileX && !mirrorX;
    wrapy = tileY && !mirrorY;
    transVal = transparency.getValue(wrapx, wrapy, x, y, xsize, ysize);
    specVal = specularity.getValue(wrapx, wrapy, x, y, xsize, ysize);    
    diffuseColor.getColor(spec.diffuse, wrapx, wrapy, x, y, xsize, ysize);
    spec.diffuse.scale((1.0f-transVal)*(1.0f-specVal));
    specularColor.getColor(spec.specular, wrapx, wrapy, x, y, xsize, ysize);
    spec.hilight.copy(spec.specular);
    spec.specular.scale((1.0f-transVal)*specVal);
    spec.hilight.scale((1.0f-transVal)*shininess.getValue(wrapx, wrapy, x, y, xsize, ysize));
    transparentColor.getColor(spec.transparent, wrapx, wrapy, x, y, xsize, ysize);
    spec.transparent.scale(transVal);
    emissiveColor.getColor(spec.emissive, wrapx, wrapy, x, y, xsize, ysize);
    spec.roughness = roughness.getValue(wrapx, wrapy, x, y, xsize, ysize);
    spec.cloudiness = cloudiness.getValue(wrapx, wrapy, x, y, xsize, ysize);
    bump.getGradient(tempGrad, wrapx, wrapy, x, y, xsize, ysize);
    spec.bumpGrad.set(tempGrad.x*0.04, tempGrad.y*0.04, 0.0);
  }
  
  public void getTransparency(RGBColor trans, double x, double y, double xsize, double ysize, double angle, double t, double param[])
  {
    float transVal;
    boolean wrapx, wrapy;
    double f;
    
    if ((!tileX && (x < 0.0 || x > 1.0)) || (!tileY && (y < 0.0 || y > 1.0)))
      {
	// The point falls outside the image, so the texture is completely transparent.
	
	trans.setRGB(1.0f, 1.0f, 1.0f);
	return;
      }
    if (mirrorX)
      {
	f = Math.floor(x);
	if ((((int) f)&1) == 0)
	  x = 1.0+f-x;
	else
	  x = x-f;
      }
    else
      x = x-Math.floor(x);
    if (mirrorY)
      {
	f = Math.floor(y);
	if ((((int) f)&1) == 0)
	  y = 1.0+f-y;
	else
	  y = y-f;
      }
    else
      y = y-Math.floor(y);
    wrapx = tileX && !mirrorX;
    wrapy = tileY && !mirrorY;
    transVal = transparency.getValue(wrapx, wrapy, x, y, xsize, ysize);
    transparentColor.getColor(trans, wrapx, wrapy, x, y, xsize, ysize);
    trans.scale(transVal);
  }

  public double getDisplacement(double x, double y, double xsize, double ysize, double t, double param[])
  {
    boolean wrapx, wrapy;
    double f;
    
    if ((!tileX && (x < 0.0 || x > 1.0)) || (!tileY && (y < 0.0 || y > 1.0)))
      {
	// The point falls outside the image.
	
	return 0.0;
      }
    if (mirrorX)
      {
	f = Math.floor(x);
	if ((((int) f)&1) == 0)
	  x = 1.0+f-x;
	else
	  x = x-f;
      }
    else
      x = x-Math.floor(x);
    if (mirrorY)
      {
	f = Math.floor(y);
	if ((((int) f)&1) == 0)
	  y = 1.0+f-y;
	else
	  y = y-f;
      }
    else
      y = y-Math.floor(y);
    wrapx = tileX && !mirrorX;
    wrapy = tileY && !mirrorY;
    return (double) displacement.getValue(wrapx, wrapy, x, y, xsize, ysize);
  }

  public void getAverageSpec(TextureSpec spec, double time, double param[])
  {
    float transVal = transparency.getAverageValue(), specVal = specularity.getAverageValue();
    
    diffuseColor.getAverageColor(spec.diffuse);
    spec.diffuse.scale((1.0f-transVal)*(1.0f-specVal));
    specularColor.getAverageColor(spec.specular);
    spec.hilight.copy(spec.specular);
    spec.specular.scale((1.0f-transVal)*specVal);
    spec.hilight.scale((1.0f-transVal)*shininess.getAverageValue());
    transparentColor.getAverageColor(spec.transparent);
    spec.transparent.scale(transVal);
    emissiveColor.getAverageColor(spec.emissive);
    spec.roughness = roughness.getAverageValue();
    spec.cloudiness = cloudiness.getAverageValue();
    spec.bumpGrad.set(0.0, 0.0, 0.0);
  }
  
  /** Determine whether this Texture uses the specified image. */

  public boolean usesImage(ImageMap image)
  {
    return (diffuseColor.getImage() == image || specularColor.getImage() == image || 
      transparentColor.getImage() == image || emissiveColor.getImage() == image || 
      roughness.getImage() == image || cloudiness.getImage() == image || 
      transparency.getImage() == image || specularity.getImage() == image || 
      shininess.getImage() == image || bump.getImage() == image ||
      displacement.getImage() == image);
  }

  /** Create a duplicate of the texture. */
  
  public Texture duplicate()
  {
    ImageMapTexture m = new ImageMapTexture();
    
    m.name = name;
    m.diffuseColor.copy(diffuseColor);
    m.specularColor.copy(specularColor);
    m.transparentColor.copy(transparentColor);
    m.emissiveColor.copy(emissiveColor);
    m.transparency = transparency.duplicate();
    m.specularity = specularity.duplicate();
    m.shininess = shininess.duplicate();
    m.roughness = roughness.duplicate();
    m.cloudiness = cloudiness.duplicate();
    m.bump = bump.duplicate();
    m.displacement = displacement.duplicate();
    m.tileX = tileX;
    m.tileY = tileY;
    m.mirrorX = mirrorX;
    m.mirrorY = mirrorY;
    return m;
  }

  /** Determine whether this texture has a non-zero value anywhere for a particular component.
      @param component    the texture component to check for (one of the *_COMPONENT constants)
  */
  
  public boolean hasComponent(int component)
  {
    switch (component)
      {
        case DIFFUSE_COLOR_COMPONENT:
          return ((transparency.getValue() < 1.0f || transparency.getImage() != null) &&
                  (specularity.getValue() < 1.0f || specularity.getImage() != null) &&
                  diffuseColor.getColor().getMaxComponent() > 0.0);
        case SPECULAR_COLOR_COMPONENT:
          return ((transparency.getValue() < 1.0f || transparency.getImage() != null) &&
                  specularity.getValue() > 0.0f && specularColor.getColor().getMaxComponent() > 0.0);
        case TRANSPARENT_COLOR_COMPONENT:
          if (!(tileX && tileY))
            return true;
          return (transparency.getValue() > 0.0f && transparentColor.getColor().getMaxComponent() > 0.0);
        case HILIGHT_COLOR_COMPONENT:
          return ((transparency.getValue() < 1.0f || transparency.getImage() != null) &&
                  shininess.getValue() > 0.0f && specularColor.getColor().getMaxComponent() > 0.0);
        case EMISSIVE_COLOR_COMPONENT:
          return (emissiveColor.getColor().getMaxComponent() > 0.0);
        case BUMP_COMPONENT:
          return (bump.getImage() != null);
        case DISPLACEMENT_COMPONENT:
          return (displacement.getValue() != 0.0f);
      }
    return false;
  }
  
  /** Allow the user to interactively edit the texture. */
  
  public void edit(Frame fr, Scene sc)
  {
    new Editor(fr, sc);
  }
  
  /** The following two methods are used for reading and writing files.  The first is a
      constructor which reads the necessary data from an input stream.  The other writes
      the object's representation to an output stream. */
  
  public ImageMapTexture(DataInputStream in, Scene theScene) throws IOException, InvalidObjectException
  {
    short version = in.readShort();
    
    if (version < 0 || version > 1)
      throw new InvalidObjectException("");
    name = in.readUTF();
    diffuseColor = new ImageOrColor(in, theScene);
    specularColor = new ImageOrColor(in, theScene);
    transparentColor = new ImageOrColor(in, theScene);
    emissiveColor = new ImageOrColor(in, theScene);
    roughness = new ImageOrValue(in, theScene);
    cloudiness = new ImageOrValue(in, theScene);
    transparency = new ImageOrValue(in, theScene);
    specularity = new ImageOrValue(in, theScene);
    if (version == 0)
      shininess = specularity.duplicate();
    else
      shininess = new ImageOrValue(in, theScene);
    bump = new ImageOrValue(in, theScene);
    displacement = new ImageOrValue(in, theScene);
    tileX = in.readBoolean();
    tileY = in.readBoolean();
    mirrorX = in.readBoolean();
    mirrorY = in.readBoolean();
    tempGrad = new Vec2();
  }
  
  public void writeToFile(DataOutputStream out, Scene theScene) throws IOException
  {
    out.writeShort(1);
    out.writeUTF(name);
    diffuseColor.writeToFile(out, theScene);
    specularColor.writeToFile(out, theScene);
    transparentColor.writeToFile(out, theScene);
    emissiveColor.writeToFile(out, theScene);
    roughness.writeToFile(out, theScene);
    cloudiness.writeToFile(out, theScene);
    transparency.writeToFile(out, theScene);
    specularity.writeToFile(out, theScene);
    shininess.writeToFile(out, theScene);
    bump.writeToFile(out, theScene);
    displacement.writeToFile(out, theScene);
    out.writeBoolean(tileX);
    out.writeBoolean(tileY);
    out.writeBoolean(mirrorX);
    out.writeBoolean(mirrorY);
  }
  
  /** A member class which represents a dialog box for editing the texture. */
  
  class Editor extends Dialog implements ActionListener, AdjustmentListener, ItemListener
  {
    TextField nameField;
    Panel transPanel, specPanel, shinPanel, roughPanel, cloudPanel, bumpPanel, displacePanel;
    Panel diffColorPanel, specColorPanel, transColorPanel, emissColorPanel;
    Checkbox tileXBox, tileYBox, mirrorXBox, mirrorYBox;
    MaterialPreviewer preview;
    ImageMapTexture newTexture;
    Frame parent;
    
    public Editor(Frame fr, Scene sc)
    {
      super(fr, true);
      
      Panel p, p2;
      GridBagConstraints gc = new GridBagConstraints();

      parent = fr;
      newTexture = (ImageMapTexture) duplicate();

      // Add the buttons at the bottom.

      p = new Panel();
      p.add(Translate.button("ok", this));
      p.add(Translate.button("cancel", this));
      add(p, BorderLayout.SOUTH);
      
      // Add the left panel.

      p = new Panel();
      p.setLayout(new GridBagLayout());
      gc.gridwidth = 2;
      p.add(preview = new MaterialPreviewer(newTexture, null, 200, 160), gc);
      gc.gridx = 0;
      gc.anchor = GridBagConstraints.EAST;
      gc.insets = new Insets(0, 0, 0, 5);
      gc.gridwidth = 1;
      p.add(Translate.label("Name"), gc);
      p.add(Translate.label("DiffuseColor"), gc);
      p.add(Translate.label("SpecularColor"), gc);
      p.add(Translate.label("TransparentColor"), gc);
      p.add(Translate.label("EmissiveColor"), gc);
      gc.gridx = 1;
      gc.fill = GridBagConstraints.HORIZONTAL;
      p.add(nameField = new TextField(ImageMapTexture.this.name, 15), gc);
      gc.fill = GridBagConstraints.NONE;
      gc.anchor = GridBagConstraints.WEST;
      p.add(diffColorPanel = newTexture.diffuseColor.getEditingPanel(parent, sc, this), gc);
      p.add(specColorPanel = newTexture.specularColor.getEditingPanel(parent, sc, this), gc);
      p.add(transColorPanel = newTexture.transparentColor.getEditingPanel(parent, sc, this), gc);
      p.add(emissColorPanel = newTexture.emissiveColor.getEditingPanel(parent, sc, this), gc);
      add(p, BorderLayout.WEST);

      // Add the right panel.

      p = new Panel();
      p.setLayout(new GridBagLayout());
      gc.gridwidth = 2;
      gc.gridx = 0;
      gc.anchor = GridBagConstraints.CENTER;
      p.add(p2 = new Panel(), gc);
      p2.setLayout(new GridLayout(2, 2, 20, 5));
      p2.add(tileXBox = new Checkbox("Tile Horizontally", tileX));
      p2.add(mirrorXBox = new Checkbox("Mirror Horizontally", mirrorX));
      p2.add(tileYBox = new Checkbox("Tile Vertically", tileY));
      p2.add(mirrorYBox = new Checkbox("Mirror Vertically", mirrorY));
      gc.gridwidth = 1;
      gc.anchor = GridBagConstraints.EAST;
      p.add(Translate.label("Transparency"), gc);
      p.add(Translate.label("Specularity"), gc);
      p.add(Translate.label("Shininess"), gc);
      p.add(Translate.label("Roughness"), gc);
      p.add(Translate.label("Cloudiness"), gc);
      p.add(Translate.label("BumpHeight"), gc);
      p.add(Translate.label("Displacement"), gc);
      gc.gridx = 1;
      gc.anchor = GridBagConstraints.WEST;
      p.add(transPanel = newTexture.transparency.getEditingPanel(parent, sc, this, this), gc);
      p.add(specPanel = newTexture.specularity.getEditingPanel(parent, sc, this, this), gc);
      p.add(shinPanel = newTexture.shininess.getEditingPanel(parent, sc, this, this), gc);
      p.add(roughPanel = newTexture.roughness.getEditingPanel(parent, sc, this, this), gc);
      p.add(cloudPanel = newTexture.cloudiness.getEditingPanel(parent, sc, this, this), gc);
      p.add(bumpPanel = newTexture.bump.getEditingPanel(parent, sc, this, this), gc);
      p.add(displacePanel = newTexture.displacement.getEditingPanel(parent, sc, this, this), gc);
      add(p, BorderLayout.EAST);
      nameField.addActionListener(this);
      tileXBox.addItemListener(this);
      tileYBox.addItemListener(this);
      mirrorXBox.addItemListener(this);
      mirrorYBox.addItemListener(this);
      pack();
      setResizable(false);
      ModellingApp.centerWindow(this);
      show();
    }

    public void actionPerformed(ActionEvent e)
    {
      if (!e.getActionCommand().equals("cancel"))
	{
	  transparency.copy(newTexture.transparency);
	  specularity.copy(newTexture.specularity);
	  shininess.copy(newTexture.shininess);
	  roughness.copy(newTexture.roughness);
	  cloudiness.copy(newTexture.cloudiness);
	  bump.copy(newTexture.bump);
	  displacement.copy(newTexture.displacement);
	  ImageMapTexture.this.name = nameField.getText();
	  diffuseColor.copy(newTexture.diffuseColor);
	  specularColor.copy(newTexture.specularColor);
	  transparentColor.copy(newTexture.transparentColor);
	  emissiveColor.copy(newTexture.emissiveColor);
	  tileX = newTexture.tileX;
	  tileY = newTexture.tileY;
	  mirrorX = newTexture.mirrorX;
	  mirrorY = newTexture.mirrorY;
	}
      dispose();
    }

    public void adjustmentValueChanged(AdjustmentEvent e)
    {
      preview.render();
    }

    public void itemStateChanged(ItemEvent e)
    {
      newTexture.tileX = tileXBox.getState();
      newTexture.tileY = tileYBox.getState();
      newTexture.mirrorX = mirrorXBox.getState();
      newTexture.mirrorY = mirrorYBox.getState();
      preview.render();
    }
  }
}