/* UVMapping is a Mapping2D which allows the user to specify the texture
   coordinates of each vertex by hand. */

/* Copyright (C) 2002-2003 by Peter Eastman

   This program is free software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the Free Software
   Foundation; either version 2 of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but WITHOUT ANY 
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
   PARTICULAR PURPOSE.  See the GNU General Public License for more details. */

package artofillusion.texture;

import artofillusion.*;
import artofillusion.animation.*;
import artofillusion.math.*;
import artofillusion.object.*;
import artofillusion.ui.*;
import java.awt.*;
import java.awt.event.*;
import java.io.*;

public class UVMapping extends Mapping2D
{
  int numTextureParams;
  TextureParameter uparam, vparam;

  public UVMapping(Texture theTexture)
  {
    super(theTexture);
  }

  public static String getName()
  {
    return "UV";
  }

  public static boolean legalMapping(Object3D obj, Texture tex)
  {
    if (obj instanceof Actor)
      obj = ((Actor) obj).getObject();
    return (tex instanceof Texture2D && (obj instanceof TriangleMesh || obj instanceof SplineMesh || obj instanceof Actor));
  }

  /** Create a UV mapped triangle. */

  public RenderingTriangle mapTriangle(int v1, int v2, int v3, int n1, int n2, int n3, Vec3 vert[])
  {
    return new UVMappedTriangle(v1, v2, v3, n1, n2, n3, this);
  }

  /** This method is called once the texture parameters for the vertices of a triangle
      are known. */
  
  public void setParameters(RenderingTriangle tri, double p1[], double p2[], double p3[], RenderingMesh mesh)
  {
    UVMappedTriangle uv = (UVMappedTriangle) tri;
    uv.setTextureCoordinates((float) p1[numTextureParams], (float) p1[numTextureParams+1],
        (float) p2[numTextureParams], (float) p2[numTextureParams+1],
        (float) p3[numTextureParams], (float) p3[numTextureParams+1],
        mesh.vert[uv.v1], mesh.vert[uv.v2], mesh.vert[uv.v3]);
  }
  
  /** This method should not generally be called.  The mapping is undefined without knowing
      the texture coordinates for a particular triangle. */

  public void getTextureSpec(Vec3 pos, TextureSpec spec, double angle, double size, double time, double param[])
  {
    if (!appliesToFace(angle > 0.0))
      {
        spec.diffuse.setRGB(0.0f, 0.0f, 0.0f);
        spec.specular.setRGB(0.0f, 0.0f, 0.0f);
        spec.transparent.setRGB(1.0f, 1.0f, 1.0f);
        spec.emissive.setRGB(0.0f, 0.0f, 0.0f);
        spec.roughness = spec.cloudiness = 0.0;
        spec.bumpGrad.set(0.0, 0.0, 0.0);
        return;
      }
    texture.getTextureSpec(spec, pos.x, pos.y, size, size, angle, time, param);
    if (texture.hasComponent(Texture.BUMP_COMPONENT))
      {
	double s = spec.bumpGrad.x;
	double t = spec.bumpGrad.y;
	spec.bumpGrad.set(s, t, 0.0);
      }
  }

  /** This method should not generally be called.  The mapping is undefined without knowing
      the texture coordinates for a particular triangle. */

  public void getTransparency(Vec3 pos, RGBColor trans, double angle, double size, double time, double param[])
  {
    if (!appliesToFace(angle > 0.0))
      trans.setRGB(1.0f, 1.0f, 1.0f);
    else
      texture.getTransparency(trans, pos.x, pos.y, size, size, angle, time, param);
  }

  /** This method should not generally be called.  The mapping is undefined without knowing
      the texture coordinates for a particular triangle. */

  public double getDisplacement(Vec3 pos, double size, double time, double param[])
  {
    return texture.getDisplacement(pos.x, pos.y, size, size, time, param);
  }

  /** Given a Mesh to which this mapping has been applied, and the full list
      of texture parameters for that mesh, return the texture coordinates at
      each vertex. */
  
  public Vec2 [] findTextureCoordinates(Mesh mesh, TextureParameter param[])
  {
    int iu = -1, iv = -1;
    for (int i = 0; i < param.length; i++)
      {
        if (param[i].equals(uparam))
          iu = i;
        else if (param[i].equals(vparam))
          iv = i;
      }
    MeshVertex vert[] = mesh.getVertices();
    Vec2 uv[] = new Vec2 [vert.length];
    for (int i = 0; i < vert.length; i++)
      uv[i] = new Vec2(vert[i].param[iu], vert[i].param[iv]);
    return uv;
  }

  /** Given a Mesh to which this mapping has been applied, the full list
      of texture parameters for that mesh, and the desired texture coordinates
      at each vertex, set the texture parameters accordingly */
  
  public void setTextureCoordinates(Mesh mesh, TextureParameter param[], Vec2 uv[])
  {
    int iu = -1, iv = -1;
    for (int i = 0; i < param.length; i++)
      {
        if (param[i].equals(uparam))
          iu = i;
        else if (param[i].equals(vparam))
          iv = i;
      }
    MeshVertex vert[] = mesh.getVertices();
    for (int i = 0; i < vert.length; i++)
      {
        vert[i].param[iu] = uv[i].x;
        vert[i].param[iv] = uv[i].y;
      }
  }

  /** Given an Actor to which this mapping has been applied, the full list
      of texture parameters for that Actor, and the desired texture coordinates
      at each vertex, set the texture parameters accordingly */
  
  public void setTextureCoordinates(Actor actor, TextureParameter param[], Vec2 uv[])
  {
    double value[] = new double [uv.length];
    for (int i = 0; i < value.length; i++)
      value[i] = uv[i].x;
    actor.setTextureParameter(uparam, value);
    for (int i = 0; i < value.length; i++)
      value[i] = uv[i].y;
    actor.setTextureParameter(vparam, value);
  }

  public TextureMapping duplicate()
  {
    return duplicate(texture);
  }

  public TextureMapping duplicate(Texture tex)
  {
    UVMapping map = new UVMapping(tex);
    
    map.numTextureParams = numTextureParams;
    map.uparam = uparam;
    map.vparam = vparam;
    return map;
  }
  
  public void copy(TextureMapping mapping)
  {
    UVMapping map = (UVMapping) mapping; 
    
    numTextureParams = map.numTextureParams;
    uparam = map.uparam;
    vparam = map.vparam;
  }

  /* Get the list of texture parameters associated with this mapping and its texture.
     That includes the texture's parameters, and parameters for the texture coordinates. */
  
  public TextureParameter [] getParameters()
  {
    TextureParameter tp[] = getTexture().getParameters();
    numTextureParams = tp.length;
    TextureParameter p[] = new TextureParameter [numTextureParams+2];
    System.arraycopy(tp, 0, p, 0, numTextureParams);
    if (uparam == null)
      {
        uparam = new TextureParameter(this, "U", -Double.MAX_VALUE, Double.MAX_VALUE, 0.0);
        vparam = new TextureParameter(this, "V", -Double.MAX_VALUE, Double.MAX_VALUE, 0.0);
        uparam.type = TextureParameter.X_COORDINATE;
        vparam.type = TextureParameter.Y_COORDINATE;
        uparam.assignNewID();
        vparam.assignNewID();
      }
    p[numTextureParams] = uparam;
    p[numTextureParams+1] = vparam;
    return p;
  }

  public Panel getEditingPanel(ObjectInfo obj, MaterialPreviewer preview)
  {
    return new Editor(obj, preview);
  }
  
  public UVMapping(DataInputStream in, Texture theTexture) throws IOException, InvalidObjectException
  {
    super(theTexture);

    short version = in.readShort();
    if (version != 0)
      throw new InvalidObjectException("");
    setAppliesTo(in.readShort());
  }
  
  public void writeToFile(DataOutputStream out) throws IOException
  {
    out.writeShort(0);
    out.writeShort(appliesTo());
  }
  
  /* Editor is an inner class for editing the mapping. */

  class Editor extends Panel implements ActionListener, ItemListener
  {
    Choice applyToChoice;
    ObjectInfo theObject;
    MaterialPreviewer preview;

    public Editor(ObjectInfo obj, MaterialPreviewer preview)
    {
      super();
      
      theObject = obj;
      this.preview = preview;
      
      // Add the various components to the Panel.
      
      GridBagConstraints gc = new GridBagConstraints();
      setLayout(new GridBagLayout());
      gc.gridy = 0;
      gc.gridwidth = 2;
      gc.anchor = gc.CENTER;
      Button b = new Button("Edit UV Coordinates...");
      add(b, gc);
      b.addActionListener(this);
      gc.gridy = 1;
      gc.gridwidth = 1;
      gc.insets = new Insets(0, 0, 0, 5);
      add(new Label("Apply To:"), gc);
      add(applyToChoice = new Choice(), gc);
      applyToChoice.add("Front and Back Faces");
      applyToChoice.add("Front Faces Only");
      applyToChoice.add("Back Faces Only");
      applyToChoice.select(appliesTo());
      applyToChoice.addItemListener(this);
    }

    public void actionPerformed(ActionEvent ev)
    {
      Container parent = getParent();
      while (!(parent instanceof Frame))
        parent = parent.getParent();
      new UVMappingWindow((Frame) parent, theObject, UVMapping.this);
      parent = getParent();
      while (!(parent instanceof TextureMappingDialog) && parent != null)
        parent = parent.getParent();
      if (parent != null)
        ((TextureMappingDialog) parent).setPreviewMapping(UVMapping.this);
      preview.render();
    }

    public void itemStateChanged(ItemEvent e)
    {
      setAppliesTo((short) applyToChoice.getSelectedIndex());
      preview.setTexture(getTexture(), UVMapping.this);
      preview.render();
    }
  }
}