/* UVMappingWindow is a window for editing the UV texture coordinates at
   each vertex of a mesh. */

/* Copyright (C) 2002 by Peter Eastman

   This program is free software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the Free Software
   Foundation; either version 2 of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
   PARTICULAR PURPOSE.  See the GNU General Public License for more details. */

package artofillusion.texture;

import artofillusion.*;
import artofillusion.animation.*;
import artofillusion.math.*;
import artofillusion.object.*;
import artofillusion.ui.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;

public class UVMappingWindow extends Dialog implements ActionListener,
               EditingWindow, ItemListener, TextListener, ModelListener
{
  Object3D oldObj;
  Mesh editObj;
  UVMapping map;
  Vec2 coord[];
  /** Holds the coordinates when the object was constructed.
    Needed for moving "unpinned" vertices. */
  Vec2 startCoord[];
  UVMappingViewer mapView;
  MeshViewer meshView;
  MaterialPreviewer preview;
  ValueField minuField, minvField, maxuField, maxvField, uField, vField;
  Choice componentChoice, resChoice;
  Checkbox pinnedCheck;  
  boolean selectedVertices[];
  boolean pinnedVertices[];
  
  
  public static final int PINDIST_3D = 0;
  public static final int PINDIST_2D = 1;
  public static final int PINDIST_MESH = 2;
    
  int pinningDistanceMode = PINDIST_3D;
  
  public static final int PINWEIGHT_QUADNORM = 0;
  public static final int PINWEIGHT_LINNORM = 1;
  public static final int PINWEIGHT_SQRTNORM = 2;
  
  int pinningWeightMode = PINDIST_3D;

  public UVMappingWindow(Frame parent, ObjectInfo obj, UVMapping map)
  {
    super(parent, "UV Texture Coordinates", true);
    oldObj = obj.object;
    if (oldObj instanceof Actor)
      editObj = (Mesh) ((Actor) oldObj).getObject().duplicate();
    else
      editObj = (Mesh) oldObj.duplicate();
    this.map = map;
    ObjectInfo editInfo = obj.duplicate((Object3D) editObj);
    coord = map.findTextureCoordinates(editObj, obj.object.getTextureMapping().getParameters());
    startCoord = (Vec2[])Utilities.deepArrayCopy(coord);
    selectedVertices = new boolean [0];
    setLayout(new GridBagLayout());
    GridBagConstraints gc = new GridBagConstraints();
    Panel p;

    // Find the range of coordinates displayed.

    double minu = Double.MAX_VALUE, maxu = -Double.MAX_VALUE;
    double minv = Double.MAX_VALUE, maxv = -Double.MAX_VALUE;
    for (int i = 0; i < coord.length; i++)
      {
        if (coord[i].x < minu) minu = coord[i].x;
        if (coord[i].x > maxu) maxu = coord[i].x;
        if (coord[i].y < minv) minv = coord[i].y;
        if (coord[i].y > maxv) maxv = coord[i].y;
      }
    double padu = 0.1*(maxu-minu), padv = 0.1*(maxv-minv);
    minu -= padu;
    maxu += padu;
    minv -= padv;
    maxv += padv;

    // Determine the texture.

    Texture tex = obj.object.getTexture();
    if (tex instanceof LayeredTexture)
      {
        LayeredMapping layered = (LayeredMapping) obj.object.getTextureMapping();
        for (int i = 0; i < layered.getNumLayers(); i++)
          if (layered.getLayerMapping(i) == map)
            {
              tex = layered.getLayer(i);
              break;
            }
      }

    // Record the default parameter values.

    TextureParameter param[] = map.getParameters();
    double paramVal[] = null;
    if (param != null)
      {
        paramVal = new double [param.length];
        for (int i = 0; i < param.length; i++)
          paramVal[i] = param[i].defaultVal;
      }

    // Layout the three main canvases.

    Panel mapViewPanel = new Panel();
    mapViewPanel.setLayout(new BorderLayout());
    mapViewPanel.add(mapView = new UVMappingViewer((Texture2D) tex, this, minu, maxu, minv, maxv, 0, 4, 0.0, paramVal), BorderLayout.CENTER);
    mapView.setPreferredSize(200, 200);
    ToolPalette tools = new ToolPalette(6, 1);
    EditingTool defaultTool;
    mapViewPanel.add(tools, BorderLayout.NORTH);
    tools.setBackground(getBackground());
    tools.addTool(defaultTool = new ReshapeMeshTool(this));
    tools.addTool(new ScaleMeshTool(this));
    tools.addTool(new RotateMeshTool(this, true));
    tools.addTool(new SkewMeshTool(this));
    tools.addTool(new TaperMeshTool(this));
    tools.addTool(new MoveUVViewTool(this));
    mapView.setTool(defaultTool);
    mapView.setMetaTool(new MoveUVViewTool(this));
    Panel meshViewPanel = new Panel();
    meshViewPanel.setLayout(new BorderLayout());
    if (editObj instanceof TriangleMesh)
      meshView = new TriMeshViewer(editInfo, new Panel());
    else
      meshView = new SplineMeshViewer(editInfo, new Panel());
    meshView.setPreferredSize(150, 150);
    meshView.setTool(new GenericTool(this, "movePoints.gif", "selected/movePoints.gif"));
    meshView.setMetaTool(new MoveViewTool(this));
    meshView.setAltTool(new RotateViewTool(this));
    meshViewPanel.add(meshView, BorderLayout.CENTER);
    meshViewPanel.add(preview = new MaterialPreviewer(editInfo, 150, 150), BorderLayout.SOUTH);
    gc.weightx = gc.weighty = 1.0;
    gc.fill = gc.BOTH;
    DividedPanel div = new DividedPanel(mapViewPanel, meshViewPanel, 1.0f, 1.0f, false);
    add(div, gc);
    gc.weightx = gc.weighty = 0.0;
    gc.gridx = 0;
    gc.fill = gc.HORIZONTAL;

    // Layout the text fields.

    add(p = new Panel(), gc);
    p.setLayout(new FlowLayout(FlowLayout.LEFT));
    p.add(new Label("Displayed Component:"));
    p.add(componentChoice = new Choice());
    componentChoice.add(Translate.text("Diffuse"));
    componentChoice.add(Translate.text("Specular"));
    componentChoice.add(Translate.text("Transparent"));
    componentChoice.add(Translate.text("Hilight"));
    componentChoice.add(Translate.text("Emissive"));
    componentChoice.addItemListener(this);
    p.add(new Label("  Resolution:"));
    p.add(resChoice = new Choice());
    resChoice.add("Low");
    resChoice.add("Medium");
    resChoice.add("High");
    resChoice.addItemListener(this);
    add(new Label("Displayed Coordinate Range:"), gc);
    add(p = new Panel(), gc);
    p.setLayout(new FlowLayout(FlowLayout.LEFT));
    p.add(new Label("U:"));
    p.add(minuField = new ValueField(minu, ValueField.NONE, 5));
    p.add(new Label(" to "));
    p.add(maxuField = new ValueField(maxu, ValueField.NONE, 5));
    p.add(new Label("    V:"));
    p.add(minvField = new ValueField(minv, ValueField.NONE, 5));
    p.add(new Label(" to "));
    p.add(maxvField = new ValueField(maxv, ValueField.NONE, 5));
    add(new Label("Selected Vertex Coordinates:"), gc);
    add(p = new Panel(), gc);
    p.setLayout(new FlowLayout(FlowLayout.LEFT));
    p.add(new Label("U:"));
    p.add(uField = new ValueField(Double.NaN, ValueField.NONE, 5));
    p.add(new Label(" V:"));
    p.add(vField = new ValueField(Double.NaN, ValueField.NONE, 5));
    minuField.addTextListener(this);
    minvField.addTextListener(this);
    maxuField.addTextListener(this);
    maxvField.addTextListener(this);
    uField.addTextListener(this);
    vField.addTextListener(this);
    
    add(p = new Panel(), gc);
    p.setLayout(new FlowLayout(FlowLayout.LEFT));
    // TODO(MB) Localization
    p.add(pinnedCheck = new Checkbox("Pinned", false));
    pinnedCheck.addItemListener(this);
    p.add(Translate.button("moveOptions", this));
    p.add(Translate.button("moveVertices", this));
    pinnedVertices = new boolean [meshView.getSelectionDistance().length];
    
    // Layout the buttons at the bottom.

    add(p = new Panel(), gc);
    p.add(Translate.button("ok", this));
    p.add(Translate.button("cancel", this));
    pack();
    ModellingApp.centerWindow(this);
    show();
  }

  /** Regenerate the texture image based on the current settings. */

  private void rebuildImage()
  {
    int res = 1 << (2-resChoice.getSelectedIndex());

    setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
    mapView.setParameters(minuField.getValue(), maxuField.getValue(), minvField.getValue(),
        maxvField.getValue(), componentChoice.getSelectedIndex(), res);
    setCursor(Cursor.getDefaultCursor());
  }

  /** Respond to button clicks. */

  public void actionPerformed(ActionEvent ev)
  {
    String command = ev.getActionCommand();

    if (command.equals("ok"))
      {
        if (oldObj instanceof Actor)
          map.setTextureCoordinates((Actor) oldObj, map.getParameters(), coord);
        else
          map.setTextureCoordinates((Mesh) oldObj, map.getParameters(), coord);
        dispose();
      }
    else if (command.equals("cancel"))
      dispose();
    else if (command.equals("moveOptions"))
    {
      new UVPinningDialog(this);
    }
    else if (command.equals("moveVertices"))
    {
      moveVerticesCommand();
    }
  }

  /** Respond to selections. */

  public void itemStateChanged(ItemEvent ev)
  {
    if (ev.getSource() == pinnedCheck)
    {
      boolean[] selverts = mapView.getSelection();
      boolean[] pinnedverts = getPinnedVertices();
      boolean pinned = pinnedCheck.getState();
      for(int i = 0; i < selverts.length; ++i)
        if (selverts[i])
          pinnedverts[i] = pinned;
    }
    else
      rebuildImage();
  }

  /** Respond to changes to text fields. */

  public void textValueChanged(TextEvent ev)
  {
    Object source = ev.getSource();
    boolean sel[] = mapView.getSelection();

    if (source == uField || source == vField)
      {
        for (int i = 0; i < sel.length; i++)
          if (sel[i])
            {
              if (source == uField && !Double.isNaN(uField.getValue()))
                coord[i].x = uField.getValue();
              if (source == vField && !Double.isNaN(vField.getValue()))
                coord[i].y = vField.getValue();
            }
        setTextureCoords(coord);
        mapView.updateVertexPositions(coord);
        mapView.repaint();
      }
    else
      rebuildImage();
  }
  
  public void modelEventHappened(ModelEvent event)
  {
    if (event.getSource() == mapView)
    {
      if (event.hasKey("parameter changed") && event.getCause() != this)
      {
	minuField.removeTextListener(this);
	minvField.removeTextListener(this);
	maxuField.removeTextListener(this);
	maxvField.removeTextListener(this);

	minuField.setValue(mapView.getMinU());
	maxuField.setValue(mapView.getMaxU());
	minvField.setValue(mapView.getMinV());
	maxvField.setValue(mapView.getMaxV());
	
	minuField.addTextListener(this);
	minvField.addTextListener(this);
	maxuField.addTextListener(this);
	maxvField.addTextListener(this);
      }	
      else if (event.hasKey("selection changed") /* && event.getCause() != this */ )
      {
        boolean[] pinnedVerts = getPinnedVertices();
        boolean[] selVerts = mapView.getSelection();

        boolean pinned = pinnedVerts.length > 0;
        for(int i = 0; i < pinnedVerts.length; ++i)
        {
          if (selVerts[i] && !pinnedVerts[i])
          {
            pinned = false;
            break;
          }
        }
        pinnedCheck.removeItemListener(this);
        pinnedCheck.setState(pinned);
        pinnedCheck.addItemListener(this);
      }
    }
  }


  /** Update the texture coordinates of the mesh. */

  public void setTextureCoords(Vec2 coords[])
  {
    coord = coords;
    map.setTextureCoordinates(editObj, ((Object3D) editObj).getTextureMapping().getParameters(), coords);
    map.setTextureCoordinates((Mesh) preview.getObject().object, preview.getObject().object.getTextureMapping().getParameters(), coords);
    preview.render();
  }

  /** Set the currently selected EditingTool. */

  public void setTool(EditingTool tool)
  {
    mapView.setTool(tool);
  }

  /** Set the text to display at the bottom of the window. */

  public void setHelpText(String text)
  {
  }

  /** Get the Frame for this EditingWindow: either the EditingWindow itself if it is a
      Frame, or its parent if it is a Dialog. */

  public Frame getFrame()
  {
    return (Frame) getParent();
  }

  /** Return the pinned vertices. This method should be used even inside
    this class because the location where this information is stored
    could change in the future.
   */
  
  public boolean[] getPinnedVertices()
  {
    return pinnedVertices;
  }
  
  /** Update the image displayed in this window. */

  public void updateImage()
  {
  }

  /** This will be called whenever the selection changes, so rebuild the mesh
      and update the text fields. */

  public void updateMenus()
  {
    int sel[] = meshView.getSelectionDistance();

    // If necessary, rebuild the mesh.

    boolean same = (selectedVertices.length == sel.length);
    for (int i = 0; i < sel.length && same; i++)
      if (selectedVertices[i] != (sel[i] == 0))
        same = false;
    if (!same)
      {
        selectedVertices = new boolean [sel.length];
        for (int i = 0; i < sel.length; i++)
          selectedVertices[i] = (sel[i] == 0);
        mapView.setDisplayedVertices(coord, selectedVertices);
        mapView.informSelection(this, "");
      }
    updateTextFields();
  }

  /** Update the U and V text fields to reflect the current selection. */

  public void updateTextFields()
  {
    boolean sel[] = mapView.getSelection();
    boolean any = false;
    double u = 0.0, v = 0.0;
    for (int i = 0; i < sel.length; i++)
      if (sel[i])
        {
          if (!any)
            {
              u = coord[i].x;
              v = coord[i].y;
              any = true;
            }
          else
            {
              if (u != coord[i].x)
                u = Double.NaN;
              if (v != coord[i].y)
                v = Double.NaN;
            }
        }
    uField.setValue(any ? u : Double.NaN);
    vField.setValue(any ? v : Double.NaN);
    uField.setEnabled(any);
    vField.setEnabled(any);
  }
  
  /** For each UV-vertex the 2-dimensional difference between coord and
   startCoord is calculated. */
  
  public Vec2[] calcMovementsForPinning()
  {
    Vec2[] result = new Vec2[coord.length];
    
    for(int i = 0; i < result.length; ++i)
      result[i] = coord[i].minus(startCoord[i]);
    
    return result;
  }
  
  /** Calculates distance of two vertices with indexes vi1, vi2.
    The meaning of "distance" depends on set options.
    
    @return A nonegative value for a distance or -1 for infinite distance
   */

  public double calcDistanceForPinning(int vi1, int vi2)
  {
    switch (pinningDistanceMode)
    {
      case PINDIST_2D:
      {
        Vec2 v1 = startCoord[vi1];
        Vec2 v2 = startCoord[vi2];

        return v1.distance(v2);
      }        
      case PINDIST_3D:
      {
        Vec3 v1 = editObj.getVertices()[vi1].r;
        Vec3 v2 = editObj.getVertices()[vi2].r;

        return v1.distance(v2);
      }
    }
    return -1;
  }
    
  /** Called when the moveVertices button is pressed. */
  
  public void moveVerticesCommand()
  {
    // TODO(MB) Undoing
    Vec2[] movements = calcMovementsForPinning();
    boolean[] pinnedverts = getPinnedVertices();
    boolean[] selverts = mapView.getSelection();

    // Create list of indices of pinned vertices
    int[] pinnedlist = Utilities.createIndexList(pinnedverts);
    Vec2[] pinnedstarts = new Vec2[pinnedlist.length];
    Vec2[] pinnedmovs = new Vec2[pinnedlist.length];
    int[][] pinnedMeshDists = null;
    
    double[] weights = new double[pinnedlist.length];
    double weightsum, dist;
    boolean infiniteweight;
    
    int i, vi;
    
    for(i = 0; i < pinnedlist.length; ++i)
    {
      pinnedstarts[i] = startCoord[pinnedlist[i]];
      pinnedmovs[i] = movements[pinnedlist[i]];
    }
    
    // If distance mode is mesh, we need to calculate the distances
    // for each pinned vertex
    if (pinningDistanceMode == PINDIST_MESH)
    {
      pinnedMeshDists = new int[pinnedlist.length][];
      int[] meshdists;
      
      for(i = 0; i < pinnedlist.length; ++i)
      {
        meshdists = new int[movements.length];
        for(vi = 0; vi < meshdists.length; ++vi)
          meshdists[i] = (vi == pinnedlist[i]) ? 0 : -1;
        ((TriMeshViewer)meshView).extendDistance(meshdists, -1);
        pinnedMeshDists[i] = meshdists;
      }
    }
        
    // Adjust all selected, unpinned vertices
    for(vi = 0; vi < movements.length; ++vi)
    {
      if (pinnedverts[vi] || !selverts[vi])
        continue;
      
      infiniteweight = false;
      weightsum = 0;
      
      // Calculate initial weights
      for(i = 0; i < pinnedlist.length; ++i)
      {
        if (pinningDistanceMode == PINDIST_MESH)
        { // special distance calculation
          dist = pinnedMeshDists[i][vi];
        }
        else
          dist = calcDistanceForPinning(vi, pinnedlist[i]);
        
        if (dist == 0)
        {
          infiniteweight = true;
          weights[i] = -1;
        }
        else if (dist == -1)
        {  // Infinite distance -> weight = 0
          weights[i] = 0;
        }
        else
        {
          switch (pinningWeightMode)
          {
            case PINWEIGHT_LINNORM:
              weights[i] = 1/dist;
              break;
            case PINWEIGHT_QUADNORM:
              weights[i] = 1/(dist*dist);
              break;
            case PINWEIGHT_SQRTNORM:
              weights[i] = 1/Math.sqrt(dist);
              break;
            default:
              weights[i] = 0;
          }
          weightsum += weights[i];
        }
      }
      
      // Special case of one or more infinite weights (distance == 0)
      if (infiniteweight)
      {
        weightsum = 0;
        for(i = 0; i < pinnedlist.length; ++i)
        {
          if (weights[i] == -1)
          {
            weights[i] = 1;
            weightsum += 1;
          }
          else
            weights[i] = 0;
        }
      }
      
      // Normalize weights, so they sum up to 1
      if (weightsum > 0)
        for(i = 0; i < pinnedlist.length; ++i)
          weights[i] /= weightsum;
      
      {     // Motion mode:
        // Calculate the resulting movement vector
        Vec2 finalmove = new Vec2();
        for(i = 0; i < pinnedlist.length; ++i)
          finalmove.add(pinnedmovs[i].times(weights[i]));

        // Apply it
        coord[vi] = startCoord[vi].plus(finalmove);
      }
          
      if (false)  
      {  //Position mode:
        // Calculate position
        Vec2 finalpos = new Vec2();
        for(i = 0; i < pinnedlist.length; ++i)
          finalpos.add(coord[pinnedlist[i]].times(weights[i]));
        
        // Apply it
        coord[vi] = finalpos;
      }
    }
    setTextureCoords(coord);
    mapView.updateVertexPositions(coord);
    mapView.repaint();
  }   

  /** Set the current UndoRecord for this EditingWindow. */

  public void setUndoRecord(UndoRecord command)
  {
  }

  /** Get the Scene which is being edited in this window.  If it is not a window for
      editing a scene, this should return null. */

  public Scene getScene()
  {
    return null;
  }

  /** Get the ViewerCanvas in which the UV coordinates are being edited. */

  public ViewerCanvas getView()
  {
    return mapView;
  }

  /** Confirm whether this window should be closed (possibly by displaying a message to the
      user), and then close it.  If the closing is canceled, this should return false. */

  public boolean confirmClose()
  {
    return true;
  }
    
}