/* Copyright (C) 1999-2003 by Peter Eastman

   This program is free software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the Free Software
   Foundation; either version 2 of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but WITHOUT ANY 
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
   PARTICULAR PURPOSE.  See the GNU General Public License for more details. */

package artofillusion.texture;

import artofillusion.*;
import artofillusion.math.*;
import artofillusion.ui.*;
import java.awt.*;
import java.awt.event.*;
import java.io.*;

/** UniformMaterial represents a material whose properties do not vary with position. */

public class UniformTexture extends Texture
{
  public RGBColor diffuseColor, specularColor, transparentColor, emissiveColor;
  public double roughness, cloudiness;
  public float transparency, specularity, shininess;
  
  public UniformTexture()
  {
    // Generate the default color.

    diffuseColor = new RGBColor(1.0f, 1.0f, 1.0f);
    specularColor = new RGBColor(1.0f, 1.0f, 1.0f);
    transparentColor = new RGBColor(1.0f, 1.0f, 1.0f);
    emissiveColor = new RGBColor(0.0f, 0.0f, 0.0f);
    transparency = 0.0f;
    specularity = 0.0f;
    shininess = 0.0f;
    roughness = 0.2;
    cloudiness = 0.0;
    name = "";
  }
  
  /** Create a texture which is completely invisible. */

  public static UniformTexture invisibleTexture()
  {
    UniformTexture tex = new UniformTexture();

    tex.diffuseColor.setRGB(0.0f, 0.0f, 0.0f);
    tex.specularColor.setRGB(0.0f, 0.0f, 0.0f);
    tex.transparency = 1.0f;
    return tex;
  }

  public static String getTypeName()
  {
    return "Uniform";
  }

  public void getTextureSpec(TextureSpec spec)
  {
    float frac = (1.0f-transparency)*(1.0f-specularity);
    
    spec.diffuse.setRGB(diffuseColor.red*frac, diffuseColor.green*frac, diffuseColor.blue*frac);
    frac = (1.0f-transparency)*specularity;
    spec.specular.setRGB(specularColor.red*frac, specularColor.green*frac, specularColor.blue*frac);
    frac = (1.0f-transparency)*shininess;
    spec.hilight.setRGB(specularColor.red*frac, specularColor.green*frac, specularColor.blue*frac);
    spec.transparent.setRGB(transparentColor.red*transparency, transparentColor.green*transparency, transparentColor.blue*transparency);
    spec.emissive.copy(emissiveColor);
    spec.roughness = roughness;
    spec.cloudiness = cloudiness;
    spec.bumpGrad.set(0.0, 0.0, 0.0);
  }
  
  public void getTransparency(RGBColor trans)
  {
    trans.setRGB(transparentColor.red*transparency, transparentColor.green*transparency, transparentColor.blue*transparency);
  }
  
  public void getAverageSpec(TextureSpec spec, double time, double param[])
  {
    getTextureSpec(spec);
  }
  
  /** The only TextureMapping which can be used for a UniformTexture is a UniformMapping. */
  
  public TextureMapping getDefaultMapping()
  {
    return new UniformMapping(this);
  }

  /** Create a duplicate of the texture. */
  
  public Texture duplicate()
  {
    UniformTexture m = new UniformTexture();
    
    m.name = name;
    m.diffuseColor.copy(diffuseColor);
    m.specularColor.copy(specularColor);
    m.transparentColor.copy(transparentColor);
    m.emissiveColor.copy(emissiveColor);
    m.transparency = transparency;
    m.specularity = specularity;
    m.roughness = roughness;
    m.cloudiness = cloudiness;
    m.shininess = shininess;
    return m;
  }

  /** Determine whether this texture has a non-zero value anywhere for a particular component.
      @param component    the texture component to check for (one of the *_COMPONENT constants)
  */
  
  public boolean hasComponent(int component)
  {
    switch (component)
      {
        case DIFFUSE_COLOR_COMPONENT:
          return (transparency < 1.0f && specularity < 1.0f && diffuseColor.getMaxComponent() > 0.0);
        case SPECULAR_COLOR_COMPONENT:
          return (transparency < 1.0f && specularity > 0.0f && specularColor.getMaxComponent() > 0.0);
        case TRANSPARENT_COLOR_COMPONENT:
          return (transparency > 0.0f && transparentColor.getMaxComponent() > 0.0);
        case HILIGHT_COLOR_COMPONENT:
          return (transparency < 1.0f && shininess > 0.0f && specularColor.getMaxComponent() > 0.0);
        case EMISSIVE_COLOR_COMPONENT:
          return (emissiveColor.getMaxComponent() > 0.0);
        default:
          return false;
      }
  }
  
  /** Allow the user to interactively edit the material. */
  
  public void edit(Frame fr, Scene sc)
  {
    new Editor(fr);
  }
  
  /** The following two methods are used for reading and writing files.  The first is a
      constructor which reads the necessary data from an input stream.  The other writes
      the object's representation to an output stream. */
  
  public UniformTexture(DataInputStream in, Scene theScene) throws IOException, InvalidObjectException
  {
    short version = in.readShort();
    
    if (version < 0 || version > 1)
      throw new InvalidObjectException("");
    name = in.readUTF();
    diffuseColor = new RGBColor(in);
    specularColor = new RGBColor(in);
    transparentColor = new RGBColor(in);
    emissiveColor = new RGBColor(in);
    roughness = in.readDouble();
    cloudiness = in.readDouble();
    transparency = in.readFloat();
    specularity = in.readFloat();
    if (version > 0)
      shininess = in.readFloat();
    else
      shininess = specularity;
  }
  
  public void writeToFile(DataOutputStream out, Scene theScene) throws IOException
  {
    out.writeShort(1);
    out.writeUTF(name);
    diffuseColor.writeToFile(out);
    specularColor.writeToFile(out);
    transparentColor.writeToFile(out);
    emissiveColor.writeToFile(out);
    out.writeDouble(roughness);
    out.writeDouble(cloudiness);
    out.writeFloat(transparency);
    out.writeFloat(specularity);
    out.writeFloat(shininess);
  }
  
  class Editor extends Dialog implements MouseListener, ActionListener, AdjustmentListener
  {
    TextField nameField;
    ValueSlider transSlider, specSlider, shinSlider, roughSlider, clearSlider;
    Panel diffPatch, specPatch, transPatch, emissPatch;
    MaterialPreviewer preview;
    UniformTexture newTexture;
    Frame parent;
    
    public Editor(Frame fr)
    {
      super(fr, true);
      
      Panel p;
      GridBagConstraints gc = new GridBagConstraints();

      parent = fr;
      newTexture = (UniformTexture) duplicate();

      // Add the buttons at the bottom.

      p = new Panel();
      p.add(Translate.button("ok", this));
      p.add(Translate.button("cancel", this));
      add(p, BorderLayout.SOUTH);
      
      // Add the main panel.

      p = new Panel();
      p.setLayout(new GridBagLayout());
      gc.gridwidth = 2;
      gc.anchor = GridBagConstraints.CENTER;
      p.add(preview = new MaterialPreviewer(newTexture, null, 200, 160), gc);
      gc.gridwidth = 1;
      gc.gridx = 0;
      gc.anchor = GridBagConstraints.EAST;
      gc.insets = new Insets(0, 0, 0, 5);
      p.add(Translate.label("Name"), gc);
      p.add(Translate.label("DiffuseColor"), gc);
      p.add(Translate.label("SpecularColor"), gc);
      p.add(Translate.label("TransparentColor"), gc);
      p.add(Translate.label("EmissiveColor"), gc);
      p.add(Translate.label("Transparency"), gc);
      p.add(Translate.label("Specularity"), gc);
      p.add(Translate.label("Shininess"), gc);
      p.add(Translate.label("Roughness"), gc);
      p.add(Translate.label("Cloudiness"), gc);
      gc.gridx = 1;
      gc.gridwidth = 2;
      gc.fill = GridBagConstraints.HORIZONTAL;
      p.add(nameField = new TextField(UniformTexture.this.name, 15), gc);
      gc.gridwidth = 1;
      p.add(diffPatch = diffuseColor.getSample(50, 30), gc);
      p.add(specPatch = specularColor.getSample(50, 30), gc);
      p.add(transPatch = transparentColor.getSample(50, 30), gc);
      p.add(emissPatch = emissiveColor.getSample(50, 30), gc);
      gc.gridwidth = 2;
      p.add(transSlider = new ValueSlider(0.0, 1.0, 100, (double) transparency), gc);
      transSlider.addAdjustmentListener(this);
      p.add(specSlider = new ValueSlider(0.0, 1.0, 100, (double) specularity), gc);
      specSlider.addAdjustmentListener(this);
      p.add(shinSlider = new ValueSlider(0.0, 1.0, 100, shininess), gc);
      shinSlider.addAdjustmentListener(this);
      p.add(roughSlider = new ValueSlider(0.0, 1.0, 100, roughness), gc);
      roughSlider.addAdjustmentListener(this);
      p.add(clearSlider = new ValueSlider(0.0, 1.0, 100, cloudiness), gc);
      clearSlider.addAdjustmentListener(this);
      gc.gridwidth = 1;
      add(p, BorderLayout.CENTER);
      nameField.addActionListener(this);
      transSlider.addActionListener(this);
      specSlider.addActionListener(this);
      shinSlider.addActionListener(this);
      roughSlider.addActionListener(this);
      clearSlider.addActionListener(this);
      diffPatch.addMouseListener(this);
      specPatch.addMouseListener(this);
      transPatch.addMouseListener(this);
      emissPatch.addMouseListener(this);
      pack();
      setResizable(false);
      ModellingApp.centerWindow(this);
      show();
    }

    public void actionPerformed(ActionEvent e)
    {
      if (!e.getActionCommand().equals("cancel"))
	{
	  transparency = (float) transSlider.getValue();
	  specularity = (float) specSlider.getValue();
	  shininess = (float) shinSlider.getValue();
	  roughness = roughSlider.getValue();
	  cloudiness = clearSlider.getValue();
	  UniformTexture.this.name = nameField.getText();
	  diffuseColor.copy(newTexture.diffuseColor);
	  specularColor.copy(newTexture.specularColor);
	  transparentColor.copy(newTexture.transparentColor);
	  emissiveColor.copy(newTexture.emissiveColor);
	}
      dispose();
    }

    public void mousePressed(MouseEvent e)
    {
    }

    public void mouseReleased(MouseEvent e)
    {
    }
  
    public void mouseClicked(MouseEvent e)
    {
      if (e.getSource() == diffPatch)
	{
	  new ColorChooser(parent, Translate.text("DiffuseColor"), newTexture.diffuseColor);
	  diffPatch.setBackground(newTexture.diffuseColor.getColor());
	  diffPatch.repaint();
	  preview.render();
	}
      else if (e.getSource() == specPatch)
	{
	  new ColorChooser(parent, Translate.text("SpecularColor"), newTexture.specularColor);
	  specPatch.setBackground(newTexture.specularColor.getColor());
	  specPatch.repaint();
	  preview.render();
	}
      else if (e.getSource() == transPatch)
	{
	  new ColorChooser(parent, Translate.text("TransparentColor"), newTexture.transparentColor);
	  transPatch.setBackground(newTexture.transparentColor.getColor());
	  transPatch.repaint();
	  preview.render();
	}
      else if (e.getSource() == emissPatch)
	{
	  new ColorChooser(parent, Translate.text("EmissiveColor"), newTexture.emissiveColor);
	  emissPatch.setBackground(newTexture.emissiveColor.getColor());
	  emissPatch.repaint();
	  preview.render();
	}
    }

    public void mouseEntered(MouseEvent e)
    {
    }
  
    public void mouseExited(MouseEvent e)
    {
    }

    public void adjustmentValueChanged(AdjustmentEvent e)
    {
      newTexture.transparency = (float) transSlider.getValue();
      newTexture.specularity = (float) specSlider.getValue();
      newTexture.shininess = (float) shinSlider.getValue();
      newTexture.roughness = roughSlider.getValue();
      newTexture.cloudiness = clearSlider.getValue();
      preview.render();
    }
  }
}