/* Copyright (C) 1999,2000,2003 by Peter Eastman

   This program is free software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the Free Software
   Foundation; either version 2 of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but WITHOUT ANY 
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
   PARTICULAR PURPOSE.  See the GNU General Public License for more details. */

package artofillusion.ui;

import artofillusion.*;
import artofillusion.math.*;
import java.awt.*;
import java.awt.event.*;

/** ColorChooser is a Dialog in which the user can edit an RGBColor object.  It allows the
    color to be specified using the RGB, HSV, or HLS color models. */

public class ColorChooser extends Dialog implements ActionListener, AdjustmentListener, KeyListener, ItemListener
{
  RGBColor oldColor, newColor;
  ValueSlider slider1, slider2, slider3;
  Label label1, label2, label3;
  Panel oldColorPatch, newColorPatch;
  Choice modeC;
  
  static int mode = 1;
  
  public ColorChooser(Frame parent, String title, RGBColor c)
  {
    super(parent, title, true);

    GridBagConstraints gc = new GridBagConstraints();

    oldColor = c;
    newColor = c.duplicate();

    // Add the buttons at the bottom.

    Panel p = new Panel();
    p.add(Translate.button("ok", this));
    p.add(Translate.button("cancel", this));
    add(p, BorderLayout.SOUTH);

    // Add the main panel.

    p = new Panel();
    p.setLayout(new GridBagLayout());
    gc.gridwidth = 2;
    gc.fill = GridBagConstraints.HORIZONTAL;
    gc.gridx = 0;
    p.add(label1 = Translate.label("Red"), gc);
    p.add(slider1 = new ValueSlider(0.0, 1.0, 100, (double) newColor.getRed()), gc);
    slider1.addAdjustmentListener(this);
    p.add(label2 = Translate.label("Green"), gc);
    p.add(slider2 = new ValueSlider(0.0, 1.0, 100, (double) newColor.getGreen()), gc);
    slider2.addAdjustmentListener(this);
    p.add(label3 = Translate.label("Blue"), gc);
    p.add(slider3 = new ValueSlider(0.0, 1.0, 100, (double) newColor.getBlue()), gc);
    slider3.addAdjustmentListener(this);
    gc.gridwidth = 1;
    p.add(new Label("Color Model:"), gc);
    gc.gridx = 1;
    p.add(modeC = new Choice(), gc);
    modeC.addItem("RGB");
    modeC.addItem("HSV");
    modeC.addItem("HLS");
    modeC.addItemListener(this);
    modeC.select(mode);
    modeChanged();
    gc.fill = GridBagConstraints.NONE;
    gc.gridx = 2;
    p.add(new Label("Original Color"), gc);
    p.add(oldColorPatch = oldColor.getSample(50, 30), gc);
    p.add(new Label("New Color"), gc);
    p.add(newColorPatch = newColor.getSample(50, 30), gc);
    add(p, BorderLayout.CENTER);
    addAsListener(this);

    pack();
    setResizable(false);
    ModellingApp.centerWindow(this);
    show();
  }

  /* Add this as a key listener to every component. */
  
  private void addAsListener(Component c)
  {
    c.addKeyListener(this);
    if (c instanceof Container)
      {
	Component children[] = ((Container) c).getComponents();
	for (int i = 0; i < children.length; i++)
	  addAsListener(children[i]);
      }
  }

  public void actionPerformed(ActionEvent e)
  {
    float values[];
    
    if (e.getActionCommand().equals("ok"))
      {
        oldColor.setRGB(newColor.getRed(), newColor.getGreen(), newColor.getBlue());
        dispose();
      }
    else if (e.getActionCommand().equals("cancel"))
      dispose();
  }

  public void adjustmentValueChanged(AdjustmentEvent e)
  {
    float values[], newValue = e.getValue()/100.0f;

    values = new float [] {(float) slider1.getValue(), (float) slider2.getValue(), (float) slider3.getValue()};
    if (mode == 0)
      newColor.setRGB(values[0], values[1], values[2]);
    else if (mode == 1)
      newColor.setHSV(values[0]*360.0f, values[1], values[2]);
    else
      newColor.setHLS(values[0]*360.0f, values[1], values[2]);
    newColorPatch.setBackground(newColor.getColor());
    newColorPatch.repaint();
  }

  public void itemStateChanged(ItemEvent e)
  {
    mode = modeC.getSelectedIndex();
    modeChanged();
  }
    
  /* Pressing Return and Escape are equivalent to clicking OK and Cancel. */
    
  public void keyPressed(KeyEvent ev)
  {
    int code = ev.getKeyCode();

    if (code != KeyEvent.VK_ENTER && code != KeyEvent.VK_ESCAPE)
      return;
    if (code == KeyEvent.VK_ENTER && ev.getSource() instanceof Button)
      return;
    if (code == KeyEvent.VK_ENTER)
      oldColor.setRGB(newColor.getRed(), newColor.getGreen(), newColor.getBlue());
    dispose();
  }
  
  /* Unused KeyListener methods. */

  public void keyReleased(KeyEvent ev)
  {
  }

  public void keyTyped(KeyEvent ev)
  {
  }
  
  void modeChanged()
  {
    float values[];

    if (mode == 0)
      {
	label1.setText(Translate.text("Red"));
	label2.setText(Translate.text("Green"));
	label3.setText(Translate.text("Blue"));
	values = new float [] {newColor.getRed(), newColor.getGreen(), newColor.getBlue()};
      }
    else if (mode == 1)
      {
	label1.setText(Translate.text("Hue"));
	label2.setText(Translate.text("Saturation"));
	label3.setText(Translate.text("Value"));
	values = newColor.getHSV();
	values[0] /= 360.0;
      }
    else
      {
	label1.setText(Translate.text("Hue"));
	label2.setText(Translate.text("Lightness"));
	label3.setText(Translate.text("Saturation"));
	values = newColor.getHLS();
	values[0] /= 360.0;
      }
    slider1.setValue(values[0]);
    slider2.setValue(values[1]);
    slider3.setValue(values[2]);
  }
}
