/* A ComponentsDialog is a modal dialog containing a line of text, and one or more Components
   for the user to edit.  Each Component has a label next to it.  At the bottom are two
   buttons labeled OK and Cancel. */
   
/* Copyright (C) 1999,2000,2002 by Peter Eastman

   This program is free software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the Free Software
   Foundation; either version 2 of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but WITHOUT ANY 
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
   PARTICULAR PURPOSE.  See the GNU General Public License for more details. */

package artofillusion.ui;

import artofillusion.*;
import java.awt.*;
import java.awt.event.*;

public class ComponentsDialog extends Dialog implements ActionListener, KeyListener
{
  Component comp[];
  boolean ok;
  Callback okCallback, cancelCallback;
  
  /** Create a modal dialog containing a set of labeled components.
  
      @param parent       the parent of the dialog
      @param prompt       a text string to appear at the top of the dialog
      @param components   the list of components to display
      @param labels       the list of labels for each component
  */

  public ComponentsDialog(Window parent, String prompt, Component components[], String labels[])
  {
    this(parent, prompt, components, labels, null, null);
  }

  /** Create a non-modal dialog containing a set of labeled components.
  
      @param parent       the parent of the dialog
      @param prompt       a text string to appear at the top of the dialog
      @param components   the list of components to display
      @param labels       the list of labels for each component
      @param onOK         a callback to execute when the user clicks OK
      @param onCancel     a callback to execute when the user clicks Cancel
  */

  public ComponentsDialog(Window parent, String prompt, Component components[], String labels[], Callback onOK, Callback onCancel)
  {
    super(Utilities.findFrame(parent), (onOK == null && onCancel == null));

    GridBagConstraints c = new GridBagConstraints();
    Panel p = new Panel();
    
    comp = components;
    okCallback = onOK;
    cancelCallback = onCancel;
    setLayout(new GridBagLayout());
    c.gridwidth = 2;
    add(new Label(prompt), c);
    
    c.insets.top = 5;
    for (int i = 0; i < labels.length; i++)
      {
        c.gridx = 0;
        if (labels[i] == null)
          {
            c.gridwidth = 2;
            c.weightx = 0.0;
            c.weighty = 1.0;
            c.anchor = GridBagConstraints.CENTER;
            c.fill = GridBagConstraints.NONE;
            add(comp[i], c);
          }
        else
          {
            c.gridwidth = 1;
            c.insets.right = 5;
            c.weightx = c.weighty = 0.0;
            c.anchor = GridBagConstraints.EAST;
            c.fill = GridBagConstraints.NONE;
            add(new Label(labels[i]), c);
            c.insets.right = 0;
            c.gridx = 1;
            c.weightx = c.weighty = 1.0;
            c.anchor = GridBagConstraints.WEST;
            c.fill = GridBagConstraints.BOTH;
            add(comp[i], c);
          }
      }
    addAsListener(this);
    c.gridwidth = 2;
    c.gridx = 0;
    c.gridy = labels.length+1;
    c.weightx = c.weighty = 0.0;
    add(p, c);
    p.add(Translate.button("ok", this));
    p.add(Translate.button("cancel", this));
    addWindowListener(new WindowAdapter() {
      public void windowClosing(WindowEvent e)
      {
        ok = false;
        closeWindow();
      }
    } );
    pack();
    setResizable(false);
    ModellingApp.centerDialog(this, parent);
    show();
  }
  
  /** Return true if the user clicked OK, false if they clicked Cancel. */
  
  public boolean clickedOk()
  {
    return ok;
  }

  public Insets getInsets()
  {
    Insets in = super.getInsets();

    return new Insets(in.top+10, in.left+10, in.bottom+10, in.right+10);
  }

  public void actionPerformed(ActionEvent e)
  {
    String command = e.getActionCommand();

    if (command.equals("cancel"))
      ok = false;
    else
      ok = true;
    closeWindow();
  }
  
  private void closeWindow()
  {
    if (ok && okCallback != null)
      okCallback.execute();
    if (!ok && cancelCallback != null)
      cancelCallback.execute();
    dispose();
    removeAsListener(this);
  }
    
  /* Pressing Return and Escape are equivalent to clicking OK and Cancel. */
    
  public void keyPressed(KeyEvent ev)
  {
    int code = ev.getKeyCode();

    if (code != KeyEvent.VK_ENTER && code != KeyEvent.VK_ESCAPE)
      return;
    if (code == KeyEvent.VK_ENTER && ev.getSource() instanceof Button)
      return;
    ok = (code == KeyEvent.VK_ENTER);
    closeWindow();
  }
  
  /* Unused KeyListener methods. */

  public void keyReleased(KeyEvent ev)
  {
  }

  public void keyTyped(KeyEvent ev)
  {
  }

  /* Add this as a key listener to every component. */
  
  private void addAsListener(Component c)
  {
    c.addKeyListener(this);
    if (c instanceof Container)
      {
	Component children[] = ((Container) c).getComponents();
	for (int i = 0; i < children.length; i++)
	  addAsListener(children[i]);
      }
  }
  
  /* Remove this as a key listener before returning. */
  
  private void removeAsListener(Component c)
  {
    c.removeKeyListener(this);
    if (c instanceof Container)
      {
	Component children[] = ((Container) c).getComponents();
	for (int i = 0; i < children.length; i++)
	  removeAsListener(children[i]);
      }
  }
}
