/* Copyright (C) 2001,2003 by Peter Eastman

   This program is free software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the Free Software
   Foundation; either version 2 of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but WITHOUT ANY 
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
   PARTICULAR PURPOSE.  See the GNU General Public License for more details. */

package artofillusion.ui;

import artofillusion.*;

/** This a thread which processes "discardable" events.  This is typically used for 
    mouse-moved or mouse-dragged events, where significant time is needed to process 
    each event, and they might be delivered faster than they can be processed.
    <p>
    Callback objects corresponding to particular events are passed to addEvent() as they
    occur.  They are called one at a time.  If several new events are delivered before
    the last one has been finished, all but the most recent one are discarded. */

public class EventProcessor extends Thread
{
  private Callback nextEvent;
  
  public EventProcessor()
  {
    start();
  }
  
  public void run()
  {
    try
    {
      while (true)
      {
        Callback c = getNextEvent();
        c.execute();
      }
    }
    catch (InterruptedException ex)
    {
    }
  }
  
  /** Add an event to the queue. */
  
  public synchronized void addEvent(Callback event)
  {
    nextEvent = event;
    notify();
  }
  
  /** Stop the thread. */
  
  public void stopProcessing()
  {
    interrupt();
    try
      {
	join();
      }
    catch (InterruptedException ex)
      {
      }
  }
  
  /** Get the most recent event, discarding all others.  If the queue is empty, wait for
      a new event to by added. */
  
  private synchronized Callback getNextEvent() throws InterruptedException
  {
    if (nextEvent == null)
      wait();
    Callback c = nextEvent;
    nextEvent = null;
    return c;
  }
  
  /** Determine whether there is an event waiting to be processed. */
  
  public boolean hasEvent()
  {
    return (nextEvent != null);
  }
}