/* A ToolPalette is drawn as a grid of images, one for each EditingTool that is added to
   the palette.  It allows a single tool to be selected at any time. */

/* Copyright (C) 1999-2002 by Peter Eastman

   This program is free software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the Free Software
   Foundation; either version 2 of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
   PARTICULAR PURPOSE.  See the GNU General Public License for more details. */

package artofillusion.ui;

import artofillusion.*;
import java.awt.*;
import java.awt.event.*;

public class ToolPalette extends Canvas implements MouseListener
{
  int width, height, numTools, selected;
  EditingTool tool[];
  Dimension maxsize;

  /** Create a new ToolPalette.  w and h give the width and height of the tool palette,
      measured in icons. */

  public ToolPalette(int w, int h)
  {
    width = w;
    height = h;
    tool = new EditingTool [w*h];
    numTools = 0;
    selected = 0;
    maxsize = new Dimension(0, 0);
    addMouseListener(this);
    setBackground(ModellingApp.APP_BACKGROUND_COLOR);
  }

  /** Add a new tool. */

  public void addTool(EditingTool t)
  {
    int w, h;

    tool[numTools++] = t;
    w = t.getIcon().getWidth(this);
    h = t.getIcon().getHeight(this);
    if (w > maxsize.width)
      maxsize.width = w;
    if (h > maxsize.height)
      maxsize.height = h;
    if (numTools == 1)
      t.activate();
  }

  /** Return the number of the currently selected tool. */

  public int getSelection()
  {
    return selected;
  }
  
  /** Return tool with index i, as returned by e.g. {@link #getSelection}. */
  
  public EditingTool getToolByIndex(int i)
  {
    return tool[i];
  }

  public void paint(Graphics g)
  {
    for (int i = 0; i < numTools; i++)
      if (i == selected)
        g.drawImage(tool[i].getSelectedIcon(), (i%width)*maxsize.width, (i/width)*maxsize.height, this);
      else
        g.drawImage(tool[i].getIcon(), (i%width)*maxsize.width, (i/width)*maxsize.height, this);
    g.drawLine(0, 0, width*maxsize.width, 0);
    g.drawLine(0, height*maxsize.height, width*maxsize.width, height*maxsize.height);
  }

  public Dimension getPreferredSize()
  {
    return new Dimension(width*maxsize.width, height*maxsize.height+2);
  }

  /** Process mouse clicks.  Because delivery of mouseClicked events is unreliable under
      Windows, check for mouseReleased events instead. */

  public void mouseReleased(MouseEvent e)
  {
    Point p = e.getPoint();
    int i;

    i = (p.x/maxsize.width) + ((p.y-1)/maxsize.height)*width;
    if (i < numTools)
      {
	if (e.getClickCount() == 2)
	  tool[i].iconDoubleClicked();
	else
	  {
	    if (selected < tool.length)
	      tool[selected].deactivate();
	    selected = i;
	    repaint();
	    tool[i].activate();
	  }
      }
  }

  /** Change the currently selected tool. */

  public void selectTool(EditingTool which)
  {
    int i;

    for (i = 0; i < numTools; i++)
      if (tool[i] == which)
	{
	  tool[selected].deactivate();
	  selected = i;
	  repaint();
	  tool[i].activate();
	}
  }

  /** Allow the user to change tools with the keyboard. */

  public void keyPressed(KeyEvent ev)
  {
    int code = ev.getKeyCode();
    int newtool;

    if (code == KeyEvent.VK_LEFT)
      newtool = selected-1;
    else if (code == KeyEvent.VK_RIGHT)
      newtool = selected+1;
    else if (code == KeyEvent.VK_UP)
      newtool = selected-width;
    else if (code == KeyEvent.VK_DOWN)
      newtool = selected+width;
    else
      return;
    if (newtool < 0)
      newtool += numTools;
    if (newtool >= numTools)
      newtool -= numTools;
    tool[selected].deactivate();
    selected = newtool;
    repaint();
    tool[selected].activate();
  }

  /** Empty implementations of the other MouseListener methods. */

  public void mousePressed(MouseEvent e)
  {
  }

  public void mouseClicked(MouseEvent e)
  {
  }

  public void mouseEntered(MouseEvent e)
  {
  }

  public void mouseExited(MouseEvent e)
  {
  }
}