/* A ValueSlider contains a TextField and a Scrollbar which are together used for choosing
   a value.  Editing either one causes the other to change automatically.  If an illegal
   value is entered into the TextField, the text turns red to indicate this. */

/* Copyright (C) 1999-2002 by Peter Eastman

   Orientation added by David M. Turner, 21 Aug 2001 

   This program is free software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the Free Software
   Foundation; either version 2 of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but WITHOUT ANY 
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
   PARTICULAR PURPOSE.  See the GNU General Public License for more details. */

package artofillusion.ui;

import java.awt.*;
import java.awt.event.*;
import java.text.*;

public class ValueSlider extends Panel implements AdjustmentListener, TextListener
{
  TextField field;
  Scrollbar scroll;
  double value, min, max;
  int increments;
  AdjustmentListener listener = null;
  boolean forceInt, vertical;
  
  public ValueSlider(double min, double max, int increments, double value)
  {
    super();
    this.value = value;
    this.min = min;
    this.max = max;
    this.increments = increments;
    init(false);
  }
  
 public ValueSlider(double min, double max, int increments, double value, boolean vertical) 
  { 
    super(); 
    this.value = value; 
    this.min = min; 
    this.max = max; 
    this.increments = increments; 
    init (vertical); 
  } 

  void init (boolean vertical)
  {
    setLayout(new GridBagLayout()); 
    GridBagConstraints gc = new GridBagConstraints(); 

    if (vertical)
    {
      gc.gridy = 0; 
      gc.gridx = 0; 
      add(field = new TextField("", 5), gc); 
      gc.fill = GridBagConstraints.VERTICAL; 
      gc.gridx = 0; 
      gc.gridy = 1; 
      gc.gridheight = GridBagConstraints.REMAINDER; 
      gc.weighty = 1; 
      add(scroll = new Scrollbar( Scrollbar.VERTICAL, (int) ((value-min)*increments/(max-min)), 1, 0, increments+1), gc); 
    }
    else
    {
      gc.fill = GridBagConstraints.HORIZONTAL; 
      add(field = new TextField("", 5), gc); 
      gc.gridx = 1; 
      gc.weightx = 1; 
      add(scroll = new Scrollbar(Scrollbar.HORIZONTAL, (int) ((value-min)*increments/(max-min)), 1, 0, increments+1), gc); 
    }
    setText();
    field.addTextListener(this);
    scroll.addAdjustmentListener(this);
    scroll.setEnabled(!Double.isNaN(value));
    this.vertical = vertical;
  } 

  public void setForceInteger(boolean force)
  {
    forceInt = force;
  }

  public void adjustmentValueChanged(AdjustmentEvent e)
  {
    double newvalue = e.getValue()*(max-min)/increments + min;
    if (newvalue == value)
      return;
    value = newvalue;
    setText();
    field.setForeground(Color.black);
    if (listener != null)
      listener.adjustmentValueChanged(e);
  }
  
  public void textValueChanged(TextEvent e)
  {
    double val;
    
    try
      {
	val = new Double(field.getText()).doubleValue();
      }
    catch (NumberFormatException ex)
      {
        field.setForeground(Color.red);
        return;
      }
    if (val < min || val > max)
      field.setForeground(Color.red);
    else if (forceInt && val != Math.floor(val))
      field.setForeground(Color.red);
    else
      {
	value = val;
	field.setForeground(Color.black);
	scroll.setValue((int) ((value-min)*increments/(max-min)));
	scroll.setEnabled(true);
	if (listener != null)
	  listener.adjustmentValueChanged(new AdjustmentEvent(scroll, AdjustmentEvent.ADJUSTMENT_VALUE_CHANGED, 
	  	AdjustmentEvent.TRACK, scroll.getValue()));
      }
  }
  
  public double getValue()
  {
    return value;
  }
  
  public void setValue(double val)
  {
    value = val;
    scroll.setValue((int) ((val-min)*increments/(max-min)));
    setText();
  }
  
  /** Set the contents of the text field based on the value. */
  
  private void setText()
  {
    String text;
    if (Double.isNaN(value))
      text = "";
    else if (value == 0.0 || value == -0.0)
      text = "0.0";
    else
      {
        // Make sure at least three significant digits are visible.
        
        int digits = (int) Math.floor(Math.log(Math.abs(value))/Math.log(10.0));
        double scale = Math.pow(10.0, digits < 0 ? 2-digits : 3);
        text = Double.toString(Math.round(value*scale)/scale);
      }
    if (!text.equals(field.getText()))
      field.setText(text);
  }

  public void setEnabled(boolean enabled)
  {
    field.setEnabled(enabled);
    scroll.setEnabled(enabled && !Double.isNaN(value));
    super.setEnabled(enabled);
  }
  
  public void addActionListener(ActionListener l)
  {
    field.addActionListener(l);
  }
  
  public void removeActionListener(ActionListener l)
  {
    field.removeActionListener(l);
  }
  
  public void addAdjustmentListener(AdjustmentListener l)
  {
    listener = AWTEventMulticaster.add(listener, l);
  }
  
  public void removeAdjustmentListener(AdjustmentListener l)
  {
    listener = AWTEventMulticaster.remove(listener, l);
  }

  public Dimension getPreferredSize()
  {
    Dimension dim = super.getPreferredSize();
    if (vertical)
      dim.height += increments;
    else
      dim.width += increments;
    return dim;
  }
}