/* This class represents a node in an octree, used for sorting the objects by location in
   the scene.  OctreeNodes can be terminal nodes, in which case they contain a list of 
   objects, or branch nodes, in which case they contain a list of other nodes.
   
   This class is more sophisticated than most octrees.  When it subdivides a node, it does
   not simply cut it in half along each axis.  Instead, it tries to determine the optimal place
   to subdivide along each axis, based on the bounding boxes of the objects within the node. */
   
/* Copyright (C) 1999-2001 by Peter Eastman

   This program is free software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the Free Software
   Foundation; either version 2 of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but WITHOUT ANY 
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
   PARTICULAR PURPOSE.  See the GNU General Public License for more details. */

package artofillusion.raytracer;

import artofillusion.*;
import artofillusion.math.*;

public class OctreeNode
{
  OctreeNode parent, child[];
  RTObject obj[];
  boolean terminal, divided;
  BoundingBox bounds, objBounds[];
  double midx, midy, midz;
  int depth;
  
  static final int CELLS = 32;

  static Vec3 nextPos = new Vec3();
  static int leftCount[] = new int [CELLS+2];
  static int rightCount[] = new int [CELLS+2];
  static int maxDepth;

  /* The constructor takes a bounding box, an array of objects, an array of bounding boxes
     of the objects, and a reference to its parent node.  treeDepth is the depth of this node 
     in the octree, and maxDepth is the maximum allowed depth. */
  
  public OctreeNode(BoundingBox nodeBounds, RTObject tri[], BoundingBox bb[], 
	OctreeNode parentNode, int treeDepth, int maxDepth)
  {
    boolean inside[] = new boolean [tri.length];
    int count, i;

    parent = parentNode;
    bounds = nodeBounds;
    depth = treeDepth;
    this.maxDepth = maxDepth;

    // Find which objects are actually inside this node.

    for (i = 0, count = 0; i < tri.length; i++)
      if (bb[i].intersects(bounds))
	if (tri[i].intersectsBox(bounds))
	  {
	    inside[i] = true;
	    count++;
	  }
    
    // Build lists of the objects which are inside this node, and their bounding boxes.
    
    obj = new RTObject [count];
    objBounds = new BoundingBox [count];
    count = 0;
    for (i = 0; i < tri.length; i++)
      if (inside[i])
	{
	  obj[count] = tri[i];
	  objBounds[count++] = bb[i];
	}
  }

  /* Determine whether this node should be subdivided.  If so, create the child nodes.  Otherwise, mark it
     as a terminal node. */

  void subdivide()
  {
    boolean splitx, splity, splitz;

    divided = true;
    if (obj.length > 4 && depth < maxDepth-1)
      {
	findMidpoints(objBounds);
	splitx = (midx != bounds.maxx);
	splity = (midy != bounds.maxy);
	splitz = (midz != bounds.maxz);
	if (!(splitx || splity || splitz))
	  {
	    terminal = true;  // Nothing would be gained from splitting any of the axes.
	    objBounds = null;
	    return;
	  }
	child = new OctreeNode[8];
	child[0] = new OctreeNode(new BoundingBox(bounds.minx, midx, bounds.miny, midy, bounds.minz, midz), obj, objBounds, this, depth+1, maxDepth);
	if (splitz)
	  child[1] = new OctreeNode(new BoundingBox(bounds.minx, midx, bounds.miny, midy, bounds.maxz, midz), obj, objBounds, this, depth+1, maxDepth);
	if (splity)
	  {
	    child[2] = new OctreeNode(new BoundingBox(bounds.minx, midx, bounds.maxy, midy, bounds.minz, midz), obj, objBounds, this, depth+1, maxDepth);
	    if (splitz)
	      child[3] = new OctreeNode(new BoundingBox(bounds.minx, midx, bounds.maxy, midy, bounds.maxz, midz), obj, objBounds, this, depth+1, maxDepth);
	  }
	if (splitx)
	  {
	    child[4] = new OctreeNode(new BoundingBox(bounds.maxx, midx, bounds.miny, midy, bounds.minz, midz), obj, objBounds, this, depth+1, maxDepth);
	    if (splitz)
	      child[5] = new OctreeNode(new BoundingBox(bounds.maxx, midx, bounds.miny, midy, bounds.maxz, midz), obj, objBounds, this, depth+1, maxDepth);
	    if (splity)
	      {
		child[6] = new OctreeNode(new BoundingBox(bounds.maxx, midx, bounds.maxy, midy, bounds.minz, midz), obj, objBounds, this, depth+1, maxDepth);
		if (splitz)
		  child[7] = new OctreeNode(new BoundingBox(bounds.maxx, midx, bounds.maxy, midy, bounds.maxz, midz), obj, objBounds, this, depth+1, maxDepth);
	      }
	  }
	obj = null;
      }
    else
      terminal = true;  // This is a terminal node.
    objBounds = null;
  }
  
  /* Get the list of objects in this node. */
  
  public RTObject [] getObjects()
  {
    return obj;
  }
  
  /* Get the bounding box for this node. */
  
  public BoundingBox getBounds()
  {
    return bounds;
  }
  
  /* This method should be called on the root node of the ocree.  Given a point, return the 
     terminal node which contains it.  If the point is not inside this node, return null. */
  
  public OctreeNode findNode(Vec3 pos)
  {
    OctreeNode current;
    int i;
    
    if (!bounds.contains(pos))
      return null;
    current = this;
    if (!divided)
      subdivide();
    while (!current.terminal)
      {
	if (pos.x > current.midx)
	  {
	    if (pos.y > current.midy)
	      {
		if (pos.z > current.midz)
		  current = current.child[7];
		else
		  current = current.child[6];
	      }
	    else
	      {
		if (pos.z > current.midz)
		  current = current.child[5];
		else
		  current = current.child[4];
	      }
	  }
	else
	  {
	    if (pos.y > current.midy)
	      {
		if (pos.z > current.midz)
		  current = current.child[3];
		else
		  current = current.child[2];
	      }
	    else
	      {
		if (pos.z > current.midz)
		  current = current.child[1];
		else
		  current = current.child[0];
	      }
	  }
	if (!current.divided)
	  current.subdivide();
      }
    return current;
  }
  
  /* Given a ray which passes through this node, find the next node it passes through.  If this is
     the last node it passes through, return null. */
  
  public OctreeNode findNextNode(Ray r)
  {
    double t1, t2, maxt = Double.MAX_VALUE;
    Vec3 orig = r.getOrigin(), dir = r.getDirection();
    OctreeNode current;
    
    if (parent == null)
      return null;

    // Find the last point on the ray which is inside this node.
    
    if (dir.x != 0.0)
      {
	t1 = (bounds.minx-orig.x)/dir.x;
	t2 = (bounds.maxx-orig.x)/dir.x;
	if (t1 < t2)
	  {
	    if (t2 < maxt)
	      maxt = t2;
	  }
	else
	  {
	    if (t1 < maxt)
	      maxt = t1;
	  }
      }
    if (dir.y != 0.0)
      {
	t1 = (bounds.miny-orig.y)/dir.y;
	t2 = (bounds.maxy-orig.y)/dir.y;
	if (t1 < t2)
	  {
	    if (t2 < maxt)
	      maxt = t2;
	  }
	else
	  {
	    if (t1 < maxt)
	      maxt = t1;
	  }
      }
    if (dir.z != 0.0)
      {
	t1 = (bounds.minz-orig.z)/dir.z;
	t2 = (bounds.maxz-orig.z)/dir.z;
	if (t1 < t2)
	  {
	    if (t2 < maxt)
	      maxt = t2;
	  }
	else
	  {
	    if (t1 < maxt)
	      maxt = t1;
	  }
      }

    // Push it just outside this node.

    nextPos.set(orig.x+dir.x*maxt, orig.y+dir.y*maxt, orig.z+dir.z*maxt);
    nextPos.x += (dir.x > 0.0 ? Raytracer.TOL : -Raytracer.TOL);
    nextPos.y += (dir.y > 0.0 ? Raytracer.TOL : -Raytracer.TOL);
    nextPos.z += (dir.z > 0.0 ? Raytracer.TOL : -Raytracer.TOL);

    // Move up the tree until we find a node which contains that point.

    current = parent;
    while (!current.bounds.contains(nextPos))
      {
	current = current.parent;
	if (current == null)
	  return null;
      }

    // Now move back down the tree until we reach a terminal node.

    while (!current.terminal)
      {
	if (nextPos.x > current.midx)
	  {
	    if (nextPos.y > current.midy)
	      {
		if (nextPos.z > current.midz)
		  current = current.child[7];
		else
		  current = current.child[6];
	      }
	    else
	      {
		if (nextPos.z > current.midz)
		  current = current.child[5];
		else
		  current = current.child[4];
	      }
	  }
	else
	  {
	    if (nextPos.y > current.midy)
	      {
		if (nextPos.z > current.midz)
		  current = current.child[3];
		else
		  current = current.child[2];
	      }
	    else
	      {
		if (nextPos.z > current.midz)
		  current = current.child[1];
		else
		  current = current.child[0];
	      }
	  }
	if (!current.divided)
	  current.subdivide();
      }
    return current;
  }
  
  /* This method should be called on the root node of the octree.  Given a ray whose origin is
     outside the node, find the point where it enters the node, and return the terminal node which
     contains that point.  If the ray never intersects this node, return null. */
  
  public OctreeNode findFirstNode(Ray r)
  {
    double t1, t2, mint = -Double.MAX_VALUE, maxt = Double.MAX_VALUE;
    Vec3 orig = r.getOrigin(), dir = r.getDirection();

    // Find the point where the ray enters this node (if it does at all).

    if (dir.x == 0.0)
      {
	if (orig.x < bounds.minx || orig.x > bounds.maxx)
	  return null;
      }
    else
      {
	t1 = (bounds.minx-orig.x)/dir.x;
	t2 = (bounds.maxx-orig.x)/dir.x;
	if (t1 < t2)
	  {
	    if (t1 > mint)
	      mint = t1;
	    if (t2 < maxt)
	      maxt = t2;
	  }
	else
	  {
	    if (t2 > mint)
	      mint = t2;
	    if (t1 < maxt)
	      maxt = t1;
	  }
	if (mint > maxt || maxt < 0.0)
	  return null;
      }
    if (dir.y == 0.0)
      {
	if (orig.y < bounds.miny || orig.y > bounds.maxy)
	  return null;
      }
    else
      {
	t1 = (bounds.miny-orig.y)/dir.y;
	t2 = (bounds.maxy-orig.y)/dir.y;
	if (t1 < t2)
	  {
	    if (t1 > mint)
	      mint = t1;
	    if (t2 < maxt)
	      maxt = t2;
	  }
	else
	  {
	    if (t2 > mint)
	      mint = t2;
	    if (t1 < maxt)
	      maxt = t1;
	  }
	if (mint > maxt || maxt < 0.0)
	  return null;
      }
    if (dir.z == 0.0)
      {
	if (orig.z < bounds.minz || orig.z > bounds.maxz)
	  return null;
      }
    else
      {
	t1 = (bounds.minz-orig.z)/dir.z;
	t2 = (bounds.maxz-orig.z)/dir.z;
	if (t1 < t2)
	  {
	    if (t1 > mint)
	      mint = t1;
	    if (t2 < maxt)
	      maxt = t2;
	  }
	else
	  {
	    if (t2 > mint)
	      mint = t2;
	    if (t1 < maxt)
	      maxt = t1;
	  }
	if (mint > maxt || maxt < 0.0)
	  return null;
      }

    // Push it just inside this node.

    mint += Raytracer.TOL;
    nextPos.set(orig.x+dir.x*mint, orig.y+dir.y*mint, orig.z+dir.z*mint);
    
    // Return the terminal node which contains the point.
    
    return findNode(nextPos);
  }
  
  /* Analyze the distribution of objects inside this node, and determine the best place at which
     to subdivide it along each axis. */
  
  void findMidpoints(BoundingBox objBounds[])
  {
    int i, j, numToLeft, numToRight;
    double cutoff, width, min, max, cost, minCost;
    Vec3 size = bounds.getSize();
    
    // If the box is much shorter along one axis than the other two, we don't want to subdivide 
    // along that axis, since it would slow down many more rays than it would speed up.

    if (size.x > size.y)
      {
	if (size.x > size.z)
	  cutoff = Math.max(size.y, size.z);
	else
	  cutoff = size.x;
      }
    else
      {
	if (size.y > size.z)
	  cutoff = Math.max(size.x, size.z);
	else
	  cutoff = size.y;
      }
    cutoff *= 0.1;

    if (size.x > cutoff)
      {
	// Find the boundaries of each object along the x axis.

	for (i = 0; i < CELLS+2; i++)
	  leftCount[i] = rightCount[i] = 0;
	min = bounds.minx;
	max = bounds.maxx;
	width = (max-min)/CELLS;
	for (i = 0; i < objBounds.length; i++)
	  {
	    if (objBounds[i].minx <= min)
	      leftCount[0]++;
	    else
	      leftCount[(int) ((objBounds[i].minx-min)/width)+1]++;
	    if (objBounds[i].maxx < max)
	      rightCount[(int) ((objBounds[i].maxx-min)/width)+1]++;
	  }

	// Evaluate the cost function at each point, and find the optimal point at which to divide
	// the node.

	numToLeft = 0;
	numToRight = objBounds.length;
	minCost = numToRight*(CELLS-3);
	j = -1;
	for (i = 0; i < CELLS; i++)
	  {
	    numToLeft += leftCount[i];
	    numToRight -= rightCount[i];
	    cost = numToLeft*i + numToRight*(CELLS-i);
	    if (cost < minCost)
	      {
		minCost = cost;
		j = i;
	      }
	  }
	if (j == -1)
	  midx = max;
	else
	  midx = min + j*width;
      }
    else
      midx = bounds.maxx;
    
    // Now do the same for the y axis.

    if (size.y > cutoff)
      {
	for (i = 0; i < CELLS+2; i++)
	  leftCount[i] = rightCount[i] = 0;
	min = bounds.miny;
	max = bounds.maxy;
	width = (max-min)/CELLS;
	for (i = 0; i < objBounds.length; i++)
	  {
	    if (objBounds[i].miny <= min)
	      leftCount[0]++;
	    else
	      leftCount[(int) ((objBounds[i].miny-min)/width)+1]++;
	    if (objBounds[i].maxy < max)
	      rightCount[(int) ((objBounds[i].maxy-min)/width)+1]++;
	  }
	numToLeft = 0;
	numToRight = objBounds.length;
	minCost = numToRight*(CELLS-3);
	j = -1;
	for (i = 0; i < CELLS; i++)
	  {
	    numToLeft += leftCount[i];
	    numToRight -= rightCount[i];
	    cost = numToLeft*i + numToRight*(CELLS-i);
	    if (cost < minCost)
	      {
		minCost = cost;
		j = i;
	      }
	  }
	if (j == -1)
	  midy = max;
	else
	  midy = min + j*width;
      }
    else
      midy = bounds.maxy;

    // And finally the z axis.

    if (size.z > cutoff)
      {
	for (i = 0; i < CELLS+2; i++)
	  leftCount[i] = rightCount[i] = 0;
	min = bounds.minz;
	max = bounds.maxz;
	width = (max-min)/CELLS;
	for (i = 0; i < objBounds.length; i++)
	  {
	    if (objBounds[i].minz <= min)
	      leftCount[0]++;
	    else
	      leftCount[(int) ((objBounds[i].minz-min)/width)+1]++;
	    if (objBounds[i].maxz < max)
	      rightCount[(int) ((objBounds[i].maxz-min)/width)+1]++;
	  }
	numToLeft = 0;
	numToRight = objBounds.length;
	minCost = numToRight*(CELLS-3);
	j = -1;
	for (i = 0; i < CELLS; i++)
	  {
	    numToLeft += leftCount[i];
	    numToRight -= rightCount[i];
	    cost = numToLeft*i + numToRight*(CELLS-i);
	    if (cost < minCost)
	      {
		minCost = cost;
		j = i;
	      }
	  }
	if (j == -1)
	  midz = max;
	else
	  midz = min + j*width;
      }
    else
      midz = bounds.maxz;
  }
  
  /* Determine whether this node contains a given point. */
  
  public boolean contains(Vec3 p)
  {
    return bounds.contains(p);
  }
}