/* RTTriangle represents a triangle to be raytraced.  It is defined by specifying a
   RenderingMesh, the index of the RenderingTriangle within the mesh, and a matrix which 
   gives the transformation from the mesh's local coordinates to world coordinates.  To
   save time and memory, the constructor is also passed two arrays which contain the
   vertices and normals for the mesh, transformed into world coordinates. */

/* Copyright (C) 1999-2003 by Peter Eastman

   This program is free software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the Free Software
   Foundation; either version 2 of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but WITHOUT ANY 
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
   PARTICULAR PURPOSE.  See the GNU General Public License for more details. */

package artofillusion.raytracer;

import artofillusion.*;
import artofillusion.material.*;
import artofillusion.math.*;
import artofillusion.texture.*;

public class RTTriangle implements RTObject
{
  RenderingTriangle tri;
  Vec3 vert1, vert2, vert3, trueNorm, norm1, norm2, norm3, ri;
  double d, t, u, v, w, time;
  Vec2 edge2d1, edge2d2;
  int dropAxis;
  int lastRay;
  boolean interpNormals, bumpMapped, lastRayResult;
  Mat4 toLocal, fromLocal;
  MaterialMapping material;
  
  public static final double TOL = 1e-12;

  public RTTriangle(RenderingMesh mesh, int which, Mat4 fromLocal, Mat4 toLocal, MaterialMapping material, Vec3 vert[], Vec3 norm[], double time)
  {
    tri = mesh.triangle[which];
    vert1 = vert[tri.v1];
    vert2 = vert[tri.v2];
    vert3 = vert[tri.v3];
    trueNorm = fromLocal.timesDirection(mesh.faceNorm[which]);
    this.fromLocal = fromLocal;
    this.toLocal = toLocal;
    this.time = time;
    interpNormals = (tri.n1 != tri.n2 || tri.n1 != tri.n3);
    if (interpNormals)
      {
	norm1 = norm[tri.n1];
	norm2 = norm[tri.n2];
	norm3 = norm[tri.n3];
	
	// Make sure trueNorm points in the same direction as the vertex normals.
	
	int i = 0;
	if (trueNorm.dot(norm1) < 0.0) i++;
	if (trueNorm.dot(norm2) < 0.0) i++;
	if (trueNorm.dot(norm3) < 0.0) i++;
	if (i > 1)
	  trueNorm.scale(-1.0);
      }
    d = -trueNorm.dot(vert1);
    if (Math.abs(trueNorm.y) > Math.abs(trueNorm.x))
      {
	if (Math.abs(trueNorm.z) > Math.abs(trueNorm.y))
	  dropAxis = 2;
	else
	  dropAxis = 1;
      }
    else
      {
	if (Math.abs(trueNorm.z) > Math.abs(trueNorm.x))
	  dropAxis = 2;
	else
	  dropAxis = 0;
      }
    switch (dropAxis)
      {
      case 0:
	edge2d1 = new Vec2(vert1.y-vert2.y, vert1.z-vert2.z);
	edge2d2 = new Vec2(vert1.y-vert3.y, vert1.z-vert3.z);
	break;
      case 1:
	edge2d1 = new Vec2(vert1.x-vert2.x, vert1.z-vert2.z);
	edge2d2 = new Vec2(vert1.x-vert3.x, vert1.z-vert3.z);
	break;
      default:
	edge2d1 = new Vec2(vert1.x-vert2.x, vert1.y-vert2.y);
	edge2d2 = new Vec2(vert1.x-vert3.x, vert1.y-vert3.y);
      }
    double denom = 1.0/edge2d1.cross(edge2d2);
    edge2d1.scale(denom);
    edge2d2.scale(denom);
    bumpMapped = tri.mapping.getTexture().hasComponent(Texture.BUMP_COMPONENT);
    ri = new Vec3();
    lastRay = -1;
    this.material = material;
  }

  /* Get the TextureMapping for this object. */
  
  public final TextureMapping getTextureMapping()
  {
    return tri.getTextureMapping();
  }

  /* Get the MaterialMapping for this object. */
  
  public final MaterialMapping getMaterialMapping()
  {
    return material;
  }  

  /* Determine whether the given ray intersects this triangle. */

  public boolean intersects(Ray r)
  {
    double vd, v0, vx, vy;
    Vec3 orig = r.getOrigin(), dir = r.getDirection();

    if (lastRay == r.getID())
      return lastRayResult;
    lastRay = r.getID();
    vd = trueNorm.dot(dir);
    if (vd == 0.0)
      return (lastRayResult = false);  // The ray is parallel to the plane.
    v0 = -(trueNorm.dot(orig) + d);
    t = v0/vd;
    if (t < TOL)
      return (lastRayResult = false);  // Ray points away from plane of triangle.

    // Determine whether the intersection point is inside the triangle.

    ri.set(orig.x+dir.x*t, orig.y+dir.y*t, orig.z+dir.z*t);
    switch (dropAxis)
      {
      case 0:
	vx = ri.y - vert1.y;
	vy = ri.z - vert1.z;
	break;
      case 1:
	vx = ri.x - vert1.x;
	vy = ri.z - vert1.z;
	break;
      default:
	vx = ri.x - vert1.x;
	vy = ri.y - vert1.y;
      }
    v = edge2d2.x*vy - edge2d2.y*vx;
    if (v < -TOL || v > 1.0+TOL)
      return (lastRayResult = false);
    w = vx*edge2d1.y - vy*edge2d1.x;
    if (w < -TOL || w > 1.0+TOL)
      return (lastRayResult = false);
    u = 1.0-v-w;
    if (u < -TOL || u > 1.0+TOL)
      return (lastRayResult = false);
    return lastRayResult = true;
  }

  /* The following methods should be called after intersects() has returned true.  They 
     return information about the point of intersection between the ray and the triangle. */
  

  /* A ray can only intersect a triangle once. */

  public int numIntersections()
  {
    return 1;
  }

  /* Get the point of intersection. */
  
  public final void intersectionPoint(int n, Vec3 p)
  {
    p.set(ri);
  }

  /* Get the distance from the ray origin to the point of intersection. */

  public final double intersectionDist(int n)
  {
    return t;
  }

  /* Get the true normal of the triangle. */
  
  public void trueNormal(Vec3 n)
  {
    n.set(trueNorm);
  }

  /** Get the surface properties at the point of intersection.
      @param spec        the texture properties will be stored in this
      @param n           the surface normal will be stored in this
      @param viewDir     the direction from which the surface is being viewed
      @param size        the width of the region over which the texture should be averaged
  */
  
  public void intersectionProperties(TextureSpec spec, Vec3 n, Vec3 viewDir, double size)
  {
    if (!interpNormals)
      n.set(trueNorm);
    else
      {
	n.x = u*norm1.x + v*norm2.x + w*norm3.x;
	n.y = u*norm1.y + v*norm2.y + w*norm3.y;
	n.z = u*norm1.z + v*norm2.z + w*norm3.z;
	n.normalize();
      }
    tri.getTextureSpec(spec, -n.dot(viewDir), u, v, w, size, time);
    if (bumpMapped)
      {
	fromLocal.transformDirection(spec.bumpGrad);
	n.scale(spec.bumpGrad.dot(n)+1.0);
	n.subtract(spec.bumpGrad);
	n.normalize();
      }
  }

  /* Get the surface transparency at the point of intersection. */
  
  public void intersectionTransparency(int n, RGBColor trans, double angle, double size)
  {
    tri.getTransparency(trans, angle, u, v, w, size, time);
  }

  /* Get a bounding box for this triangle. */
  
  public BoundingBox getBounds()
  {
    double minx, maxx, miny, maxy, minz, maxz;
    
    minx = maxx = vert1.x;
    miny = maxy = vert1.y;
    minz = maxz = vert1.z;
    if (vert2.x < minx)
      minx = vert2.x;
    if (vert2.x > maxx)
      maxx = vert2.x;
    if (vert2.y < miny)
      miny = vert2.y;
    if (vert2.y > maxy)
      maxy = vert2.y;
    if (vert2.z < minz)
      minz = vert2.z;
    if (vert2.z > maxz)
      maxz = vert2.z;
    if (vert3.x < minx)
      minx = vert3.x;
    if (vert3.x > maxx)
      maxx = vert3.x;
    if (vert3.y < miny)
      miny = vert3.y;
    if (vert3.y > maxy)
      maxy = vert3.y;
    if (vert3.z < minz)
      minz = vert3.z;
    if (vert3.z > maxz)
      maxz = vert3.z;
    return new BoundingBox(minx, maxx, miny, maxy, minz, maxz);
  }

  /* Determine whether any part of the triangle lies within a bounding box. */

  public boolean intersectsBox(BoundingBox bb)
  {
    // First see if any vertex is inside the box.
    
    if (bb.contains(vert1) || bb.contains(vert2) || bb.contains(vert3))
      return true;

    // If any edge intersects the box, return true.

    if (edgeIntersectsBox(vert1, vert2, bb) || edgeIntersectsBox(vert2, vert3, bb) || edgeIntersectsBox(vert3, vert1, bb))
      return true;

    // None of the vertices is inside the box.  However, it is possible that the 
    // triangle cuts through the entire box.  Check each of the diagonals of
    // the box to see if one intersects the triangle.

    Ray r = new Ray();
    Vec3 orig = r.getOrigin(), dir = r.getDirection();
    double len, dist;

    orig.set(bb.minx, bb.miny, bb.minz);
    dir.set(bb.maxx-bb.minx, bb.maxy-bb.miny, bb.maxz-bb.minz);
    len = dir.length();
    dir.scale(1.0/len);
    r.newID();
    if (intersects(r))
      if (intersectionDist(0) < len)
        return true;
    orig.set(bb.maxx, bb.miny, bb.minz);
    dir.set(bb.minx-bb.maxx, bb.maxy-bb.miny, bb.maxz-bb.minz);
    len = dir.length();
    dir.scale(1.0/len);
    r.newID();
    if (intersects(r))
      if (intersectionDist(0) < len)
        return true;
    orig.set(bb.minx, bb.maxy, bb.minz);
    dir.set(bb.maxx-bb.minx, bb.miny-bb.maxy, bb.maxz-bb.minz);
    len = dir.length();
    dir.scale(1.0/len);
    r.newID();
    if (intersects(r))
      if (intersectionDist(0) < len)
        return true;
    orig.set(bb.minx, bb.miny, bb.maxz);
    dir.set(bb.maxx-bb.minx, bb.maxy-bb.miny, bb.minz-bb.maxz);
    len = dir.length();
    dir.scale(1.0/len);
    r.newID();
    if (intersects(r))
      if (intersectionDist(0) < len)
        return true;
    return false;
  }
  
  /* Determine whether a particular edge of the triangle intersects a bounding box. */
  
  boolean edgeIntersectsBox(Vec3 p1, Vec3 p2, BoundingBox bb)
  {
    double t1, t2, mint = -Double.MAX_VALUE, maxt = Double.MAX_VALUE, len;
    Vec3 dir = p2.minus(p1);
    
    len = dir.length();
    dir.scale(1.0/len);
    if (dir.x == 0.0)
      {
	if (p1.x < bb.minx || p1.x > bb.maxx)
	  return false;
      }
    else
      {
	t1 = (bb.minx-p1.x)/dir.x;
	t2 = (bb.maxx-p1.x)/dir.x;
	if (t1 < t2)
	  {
	    if (t1 > mint)
	      mint = t1;
	    if (t2 < maxt)
	      maxt = t2;
	  }
	else
	  {
	    if (t2 > mint)
	      mint = t2;
	    if (t1 < maxt)
	      maxt = t1;
	  }
	if (mint > maxt || mint > len || maxt < 0.0)
	  return false;
      }
    if (dir.y == 0.0)
      {
	if (p1.y < bb.miny || p1.y > bb.maxy)
	  return false;
      }
    else
      {
	t1 = (bb.miny-p1.y)/dir.y;
	t2 = (bb.maxy-p1.y)/dir.y;
	if (t1 < t2)
	  {
	    if (t1 > mint)
	      mint = t1;
	    if (t2 < maxt)
	      maxt = t2;
	  }
	else
	  {
	    if (t2 > mint)
	      mint = t2;
	    if (t1 < maxt)
	      maxt = t1;
	  }
	if (mint > maxt || mint > len || maxt < 0.0)
	  return false;
      }
    if (dir.z == 0.0)
      {
	if (p1.z < bb.minz || p1.z > bb.maxz)
	  return false;
      }
    else
      {
	t1 = (bb.minz-p1.z)/dir.z;
	t2 = (bb.maxz-p1.z)/dir.z;
	if (t1 < t2)
	  {
	    if (t1 > mint)
	      mint = t1;
	    if (t2 < maxt)
	      maxt = t2;
	  }
	else
	  {
	    if (t2 > mint)
	      mint = t2;
	    if (t1 < maxt)
	      maxt = t1;
	  }
	if (mint > maxt || mint > len || maxt < 0.0)
	  return false;
      }
    return true;
  }
  
  /* Get the transformation from world coordinates to the object's local coordinates. */
  
  public Mat4 toLocal()
  {
    return toLocal;
  }
}