/* This dialog box allows the user to specify options for creating array objects. */

/* Copyright 2001 Rick van der Meiden

   This program is free software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the Free Software
   Foundation; either version 2 of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but WITHOUT ANY 
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
   PARTICULAR PURPOSE.  See the GNU General Public License for more details. 
*/


package artofillusion.tools;

import artofillusion.*;
import artofillusion.math.*;
import artofillusion.object.*;
import artofillusion.object.TriangleMesh.*;
import artofillusion.ui.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;

/**
This dialog box allows the user to specify options for creating array objects.
@author Rick van der Meiden
*/

public class ArrayDialog extends Dialog implements ActionListener, ItemListener, TextListener
{
  private LayoutWindow window;
  private ArraySpec spec;

  private Vector curvesVector;
  private Button okButton, cancelButton;
  private Label linearCopiesLabel, stepXLabel, stepYLabel, stepZLabel;
  private Checkbox curveCopiesBox, curveStepBox;
  private Checkbox linearBox, intervalXBox, intervalYBox, intervalZBox;
  private Checkbox curveBox, orientationBox, useOrientationBox, useOriginBox;
  private Checkbox duplicateBox, groupBox, liveBox, deepBox;
  private Choice curveChoice;
  private ValueField linearCopiesField, stepXField, stepYField, stepZField;
  private ValueField curveCopiesField, curveStepField;
  private CheckboxGroup methodGroup;
  private CheckboxGroup modeGroup;
  
  public ArrayDialog(LayoutWindow window)
  {
    super(window, "Array", true);
    this.window = window;

    // set defaults from scene
    spec = new ArraySpec(window);

    // get available curves
    curvesVector = new Vector(10,10);
    for (int i=0; i<window.getScene().getNumObjects();i++)
    {
        ObjectInfo info = window.getScene().getObject(i);
        if (info.object instanceof Curve)
                curvesVector.addElement(info);
    }
 
    // layout dialog
    methodGroup = new CheckboxGroup();
    modeGroup = new CheckboxGroup();
    
    setLayout(new GridBagLayout());
    GridBagConstraints gc = new GridBagConstraints();

    gc.anchor = gc.CENTER;
    gc.fill = gc.NONE;
    gc.gridx = 0;
    gc.gridy = 0;
    add (new Label("Create Array of Selected Objects:"),gc);
    gc.insets.top = 15;
    gc.gridx = 0;
    gc.gridy = 1;
    add (createLinearPanel(),gc);
    gc.gridx = 0;
    gc.gridy = 2;
    add (createCurvePanel(),gc);
    gc.gridx = 0;
    gc.gridy = 3;
    add (createOptionsPanel(),gc);
    gc.gridx = 0;
    gc.gridy = 4;
    gc.anchor = gc.CENTER;
    gc.ipady = 0;
    add (createFinishPanel(),gc);

   // don't allow user to use nil curve 
   if (curvesVector.size() <= 0)
        curveBox.setEnabled(false);

    // update spec
    updateSpec();

    pack();
    show();
  }

  private Panel createLinearPanel()
  {
    Panel panel = new Panel();
    panel.setLayout(new GridBagLayout());
    GridBagConstraints gc = new GridBagConstraints();
    gc.anchor = gc.WEST;

    gc.gridx = 0;
    gc.gridy = 0;
    gc.gridwidth = 1;
    panel.add(linearBox = new Checkbox("Linear:", (spec.method==ArraySpec.METHOD_LINEAR), methodGroup), gc);
    linearBox.addItemListener(this);

    gc.gridx = 1;
    gc.gridy = 0;
    gc.insets.left = 5;
    panel.add(linearCopiesLabel = new Label("Number of Copies"), gc);

    gc.gridx = 2;
    gc.gridy = 0;
    panel.add(linearCopiesField = new ValueField(spec.linearCopies, ValueField.POSITIVE+ValueField.INTEGER, 4), gc);
    linearCopiesField.addTextListener(this);

    gc.gridx = 1;
    gc.gridy = 1;
    panel.add(stepXLabel = new Label("Step X:"), gc);

    gc.gridx = 2;
    gc.gridy = 1;
    panel.add(stepXField = new ValueField(spec.stepX, ValueField.NONE, 4), gc);
    stepXField.addTextListener(this);

    gc.gridx = 3;
    gc.gridy = 1;
    panel.add(intervalXBox = new Checkbox("Times X Size", spec.intervalX), gc);
    intervalXBox.addItemListener(this);

    gc.gridx = 1;
    gc.gridy = 2;
    panel.add(stepYLabel = new Label("Step Y:"), gc);

    gc.gridx = 2;
    gc.gridy = 2;
    panel.add(stepYField = new ValueField(spec.stepY, ValueField.NONE,4), gc);
    stepYField.addTextListener(this);

    gc.gridx = 3;
    gc.gridy = 2;
    panel.add(intervalYBox = new Checkbox("Times Y Size", spec.intervalY), gc);
    intervalYBox.addItemListener(this);

    gc.gridx = 1;
    gc.gridy = 3;
    panel.add(stepZLabel = new Label("Step Z:"), gc);

    gc.gridx = 2;
    gc.gridy = 3;
    panel.add(stepZField = new ValueField(spec.stepZ, ValueField.NONE,4), gc);
    stepZField.addTextListener(this);

    gc.gridx = 3;
    gc.gridy = 3;
    panel.add(intervalZBox = new Checkbox("Times Z Size", spec.intervalZ), gc);
    intervalZBox.addItemListener(this);

    return panel;
  }

  private Panel createCurvePanel()
  {
    Panel panel = new Panel();
    panel.setLayout(new GridBagLayout());
    GridBagConstraints gc = new GridBagConstraints();
    gc.anchor = gc.WEST;

    gc.gridx = 0;
    gc.gridy = 0;
    gc.gridwidth = 1;
    panel.add(curveBox = new Checkbox("From Curve", (spec.method==ArraySpec.METHOD_CURVE), methodGroup), gc);
    curveBox.addItemListener(this);

    gc.gridx = 1;
    gc.gridy = 0;
    gc.insets.left = 5;
    panel.add(curveChoice = new Choice(), gc);
    // put names of possible curves in choice
    
    for (int k=0; k<curvesVector.size(); k++)
    {
        ObjectInfo info = (ObjectInfo)(curvesVector.elementAt(k));
        curveChoice.add(info.name);
    }
    
    if (spec.curve != null)
        curveChoice.select(spec.curve.name);
    curveChoice.addItemListener(this);

    gc.gridx = 1;
    gc.gridy = 1;
    panel.add(curveCopiesBox = new Checkbox("Number Of Copies", spec.curveMode == spec.MODE_COPIES, modeGroup), gc);
    curveCopiesBox.addItemListener(this);
    
    gc.gridx = 2;
    gc.gridy = 1;
    panel.add(curveCopiesField = new ValueField(spec.curveCopies, ValueField.POSITIVE+ValueField.INTEGER,4), gc);
    curveCopiesField.addTextListener(this);

    gc.gridx = 1;
    gc.gridy = 2;
    panel.add(curveStepBox = new Checkbox("Step Size:", spec.curveMode == spec.MODE_STEP, modeGroup), gc);
    curveStepBox.addItemListener(this);

    gc.gridx = 2;
    gc.gridy = 2;
    panel.add(curveStepField = new ValueField(spec.curveStep, ValueField.POSITIVE,4), gc);
    curveStepField.addTextListener(this);

    gc.gridx = 1;
    gc.gridy = 3;
    panel.add(orientationBox = new Checkbox("Orientation Follows Curve", spec.orientation), gc);
    orientationBox.addItemListener(this);

    gc.gridx = 1;
    gc.gridy = 4;
    panel.add(useOriginBox = new Checkbox("Use Original Position", !spec.ignoreOrigin), gc);
    useOriginBox.addItemListener(this);

    gc.gridx = 1;
    gc.gridy = 5;
    panel.add(useOrientationBox = new Checkbox("Use Original Orientation", !spec.ignoreOrientation), gc);
    useOrientationBox.addItemListener(this);

    return panel;
  }

  private Panel createOptionsPanel()
  {
    Panel panel = new Panel();
    panel.setLayout(new GridBagLayout());
    GridBagConstraints gc = new GridBagConstraints();
    gc.anchor = gc.WEST;

    gc.gridx = 0;
    gc.gridy = 0;
    gc.gridwidth = 1;
    panel.add(deepBox = new Checkbox("Include Children", spec.deep), gc);
    deepBox.addItemListener(this);


    gc.gridx = 0;
    gc.gridy = 1;
    gc.gridwidth = 1;
    panel.add(groupBox = new Checkbox("Group", spec.group), gc);
    groupBox.addItemListener(this);

    gc.gridx = 1;
    gc.gridy = 0;
    gc.gridwidth = 1;
    gc.insets.left = 5;
    panel.add(duplicateBox = new Checkbox("Skip First Copy", !spec.dupFirst), gc);
    duplicateBox.addItemListener(this);


    gc.gridx = 1;
    gc.gridy = 1;
    gc.gridwidth = 1;
    panel.add(liveBox = new Checkbox("Live Duplicates", spec.live), gc);
    liveBox.addItemListener(this);


    return panel;
  }

  private Panel createFinishPanel()
  {
    Panel panel = new Panel();

    panel.add(okButton = Translate.button("ok", this));
    panel.add(cancelButton = Translate.button("cancel", this));
    return panel;
  }

  /** callback for actions */
  public void actionPerformed(ActionEvent e)
  {
        if ("ok".equals(e.getActionCommand()))
        {
                updateSpec();
                spec.createArray();
                window.rebuildItemList();
                window.updateImage();
        }
        dispose();
  }

  /** callback for state items */
  public void itemStateChanged(ItemEvent e)
  {
        updateSpec();
  }
  
  /** callback for changes to textfields */
  public void textValueChanged(TextEvent e)
  {
        updateSpec();    
  }

  // Update ArraySpec data
  private void updateSpec()
  {
        // get values

        if (linearBox.getState() == true)
                spec.method = spec.METHOD_LINEAR;
        if (curveBox.getState() == true)
                spec.method = spec.METHOD_CURVE;
        spec.linearCopies = (int)linearCopiesField.getValue();
        spec.stepX = stepXField.getValue();
        spec.stepY = stepYField.getValue();
        spec.stepZ = stepZField.getValue();
        spec.intervalX = intervalXBox.getState();
        spec.intervalY = intervalYBox.getState();
        spec.intervalZ = intervalZBox.getState();
        if (curvesVector.size() > 0)
                spec.curve = (ObjectInfo)(curvesVector.elementAt(curveChoice.getSelectedIndex()));

        if (curveCopiesBox.getState() == true)
                spec.curveMode = spec.MODE_COPIES;
        if (curveStepBox.getState() == true)
                spec.curveMode = spec.MODE_STEP;
        spec.curveStep = curveStepField.getValue();
        spec.curveCopies = (int)curveCopiesField.getValue();
        spec.orientation = orientationBox.getState();
        spec.ignoreOrientation = !useOrientationBox.getState();
        spec.ignoreOrigin = !useOriginBox.getState();
        spec.dupFirst = !duplicateBox.getState();
        spec.group = groupBox.getState();
        spec.live = liveBox.getState();
        spec.deep = deepBox.getState();


        // update enabled/disabled status
        linearCopiesField.setEnabled(spec.method == spec.METHOD_LINEAR);
        stepXField.setEnabled(spec.method == spec.METHOD_LINEAR);
        stepYField.setEnabled(spec.method == spec.METHOD_LINEAR);
        stepZField.setEnabled(spec.method == spec.METHOD_LINEAR);
        linearCopiesLabel.setEnabled(spec.method == spec.METHOD_LINEAR);
        stepXLabel.setEnabled(spec.method == spec.METHOD_LINEAR);
        stepYLabel.setEnabled(spec.method == spec.METHOD_LINEAR);
        stepZLabel.setEnabled(spec.method == spec.METHOD_LINEAR);
        intervalXBox.setEnabled(spec.method == spec.METHOD_LINEAR);
        intervalYBox.setEnabled(spec.method == spec.METHOD_LINEAR);
        intervalZBox.setEnabled(spec.method == spec.METHOD_LINEAR);

        curveChoice.setEnabled(spec.method == spec.METHOD_CURVE);
        curveCopiesField.setEnabled(spec.method == spec.METHOD_CURVE & spec.curveMode == spec.MODE_COPIES);
        curveCopiesBox.setEnabled(spec.method == spec.METHOD_CURVE);
        curveStepField.setEnabled(spec.method == spec.METHOD_CURVE & spec.curveMode == spec.MODE_STEP);
        curveStepBox.setEnabled(spec.method == spec.METHOD_CURVE);
        orientationBox.setEnabled(spec.method == spec.METHOD_CURVE);
        useOriginBox.setEnabled(spec.method == spec.METHOD_CURVE);
        useOrientationBox.setEnabled(spec.method == spec.METHOD_CURVE);

        // duplicateBox.setEnabled(true);
        // groupBox.setEnabled(true);
        // liveBox.setEnabled(true);

  }


}
