/* This dialog box allows the user to specify options for creating lathed objects. */

/* Copyright (C) 2001-2002 by Peter Eastman

   This program is free software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the Free Software
   Foundation; either version 2 of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but WITHOUT ANY 
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
   PARTICULAR PURPOSE.  See the GNU General Public License for more details. */

package artofillusion.tools;

import artofillusion.*;
import artofillusion.math.*;
import artofillusion.object.*;
import artofillusion.object.TriangleMesh.*;
import artofillusion.ui.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;

public class LatheDialog extends Dialog implements ActionListener, AdjustmentListener, ItemListener, TextListener
{
  LayoutWindow window;
  Curve theCurve;
  ObjectInfo theCurveInfo;
  CheckboxGroup axisGroup;
  Checkbox xBox, yBox, zBox, endsBox;
  ValueField radiusField, segmentsField;
  ValueSlider angleSlider;
  Button okButton, cancelButton;
  ObjectPreviewCanvas preview;
  
  private static int counter = 1;

  public LatheDialog(LayoutWindow window, ObjectInfo curveinfo)
  {
    super(window, "Lathe", true);
    this.window = window;
    theCurveInfo = curveinfo;
    theCurve = (Curve)curveinfo.object;

    // Layout the window.
    
    setLayout(new GridBagLayout());
    GridBagConstraints gc = new GridBagConstraints();
    gc.anchor = gc.WEST;
    gc.gridx = 0;
    gc.gridwidth = 2;
    add(new Label("Select Lathe Axis:"), gc);
    axisGroup = new CheckboxGroup();
    add(xBox = new Checkbox("X axis", false, axisGroup), gc);
    add(yBox = new Checkbox("Y axis", true, axisGroup), gc);
    add(zBox = new Checkbox("Z axis", false, axisGroup), gc);
    add(endsBox = new Checkbox("Line through endpoints", false, axisGroup), gc);
    xBox.addItemListener(this);
    yBox.addItemListener(this);
    zBox.addItemListener(this);
    endsBox.addItemListener(this);
    add(new Label("Total Rotation Angle:"), gc);
    gc.fill = gc.HORIZONTAL;
    add(angleSlider = new ValueSlider(0.0, 360.0, 180, 360.0), gc);
    angleSlider.addAdjustmentListener(this);
    gc.gridwidth = 1;
    add(new Label("Radius:"), gc);
    add(new Label("Segments:"), gc);
    gc.gridx = 1;
    add(radiusField = new ValueField(0.0, ValueField.NONE), gc);
    add(segmentsField = new ValueField(8.0, ValueField.POSITIVE+ValueField.INTEGER), gc);
    radiusField.addTextListener(this);
    segmentsField.addTextListener(this);

    // Add the preview canvas.
    
    gc.gridx = 2;
    gc.gridwidth = 1;
    gc.gridheight = 9;
    gc.weightx = gc.weighty = 1.0;
    gc.fill = gc.BOTH;
    add(preview = new ObjectPreviewCanvas(null), gc);
    preview.setPreferredSize(150, 150);
    
    // Add the buttons at the bottom.
    
    Panel p = new Panel();
    p.add(okButton = Translate.button("ok", this));
    p.add(cancelButton = Translate.button("cancel", this));
    gc.gridx = 0;
    gc.gridheight = 1;
    gc.gridwidth = gc.REMAINDER;
    gc.anchor = gc.CENTER;
    gc.fill = gc.NONE;
    add(p, gc);
    selectDefaults();
    makeObject();
    pack();
    show();
  }
  
  // Deal with changes to the checkboxes and choices.
  
  public void itemStateChanged(ItemEvent e)
  {
    makeObject();
  }
  
  // Deal with changes to the text fields.
  
  public void textValueChanged(TextEvent e)
  {
    makeObject();
  }
  
  // Deal with changes to the slider.
  
  public void adjustmentValueChanged(AdjustmentEvent e)
  {
    makeObject();
  }
  
  // Deal with clicks on the buttons.
  
  public void actionPerformed(ActionEvent e)
  {
    if ("ok".equals(e.getActionCommand()))
      {
	window.addObject(preview.getObject().object,
            new CoordinateSystem(
                new Vec3(theCurveInfo.coords.getOrigin()),
                Vec3.vz(), Vec3.vy()),
            "Lathed Object "+(counter++), null);

	window.setSelection(window.getScene().getNumObjects()-1);
	window.setUndoRecord(new UndoRecord(window, false, UndoRecord.DELETE_OBJECT, new Object [] {new Integer(window.getScene().getNumObjects()-1)}));
	window.updateImage();
      }
    dispose();
  }

  // Select default values for the various options.
  
  private void selectDefaults()
  {
    MeshVertex vert[] = theCurve.getVertices();

    if (!theCurve.isClosed() && vert[0].r.distance(vert[vert.length-1].r) > 0.0)
      {
        axisGroup.setSelectedCheckbox(endsBox);
        return;
      }
    endsBox.setEnabled(false);
    double minx = Double.MAX_VALUE;
    for (int i = 0; i < vert.length; i++)
      if (vert[i].r.x < minx)
        minx = vert[i].r.x;
    minx = Math.max(-minx, 0.0);
    radiusField.setValue(Math.ceil(minx));
  }
  
  // Create the extruded object.
  
  private void makeObject()
  {
    MeshVertex vert[] = theCurve.getVertices();
    Vec3 axis, radius, center = new Vec3();
    int segments = (int) segmentsField.getValue();
    double angle = angleSlider.getValue(), angleStep = angle*Math.PI/(segments*180.0);
    boolean closed = false;

    if (angle == 360.0)
      closed = true;
    else
      segments++;

    // Determine the rotation axis.
    
    if (axisGroup.getSelectedCheckbox() == xBox)
      {
        axis = Vec3.vx();
        radius = Vec3.vy();
      }
    else if (axisGroup.getSelectedCheckbox() == yBox)
      {
        axis = Vec3.vy();
        radius = Vec3.vx();
      }
    else if (axisGroup.getSelectedCheckbox() == zBox)
      {
        axis = Vec3.vz();
        radius = Vec3.vx();
      }
    else
      {
        axis = vert[0].r.minus(vert[vert.length-1].r);
        axis.normalize();
        center.set(vert[0].r);
        radius = new Vec3();
        for (int i = 0; i < vert.length; i++)
          radius.add(vert[i].r);
        radius.scale(1.0/vert.length);
        radius.subtract(center);
        radius.subtract(axis.times(axis.dot(center)));
        radius.normalize();
      }
    center.add(radius.times(-radiusField.getValue()));

    // Calculate the vertices of the lathed surface.
    
    Vec3 v[][] = new Vec3 [segments][vert.length], cm = new Vec3();
    CoordinateSystem coords = new CoordinateSystem(center, axis, radius);
    for (int i = 0; i < segments; i++)
      {
	Mat4 m = coords.fromLocal().times(Mat4.zrotation(i*angleStep)).times(coords.toLocal());
	for (int j = 0; j < vert.length; j++)
	  {
	    v[i][j] = m.times(vert[j].r);
	    cm.add(v[i][j]);
	  }
      }
    
    // Create the arrays of smoothness values.
    
    float usmooth[] = new float [segments], vsmooth[] = new float [vert.length];
    float s[] = theCurve.getSmoothness();
    for (int i = 0; i < segments; i++)
      usmooth[i] = 1.0f;
    for (int i = 0; i < s.length; i++)
      vsmooth[i] = s[i];
    int smoothMethod = theCurve.getSmoothingMethod();
    if (smoothMethod == Mesh.NO_SMOOTHING)
      {
        for (int i = 0; i < s.length; i++)
          vsmooth[i] = 0.0f;
        smoothMethod = Mesh.APPROXIMATING;
      }
    else
      for (int i = 0; i < s.length; i++)
        vsmooth[i] = s[i];

    // Center it.
    
    cm.scale(1.0/(segments*vert.length));
    for (int i = 0; i < v.length; i++)
      for (int j = 0; j < v[i].length; j++)
	v[i][j].subtract(cm);
    SplineMesh mesh = new SplineMesh(v, usmooth, vsmooth, smoothMethod, closed, theCurve.isClosed());
    mesh.setTexture(window.getScene().getDefaultTexture());
    mesh.setTextureMapping(mesh.getTexture().getDefaultMapping());
    mesh.makeRightSideOut();
    preview.setObject(mesh);
    preview.updateImage();
    preview.repaint();
  }
}