/* This dialog box allows the user to specify options for creating skinned objects. */

/* Copyright (C) 2001-2002 by Peter Eastman

   This program is free software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the Free Software
   Foundation; either version 2 of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but WITHOUT ANY 
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
   PARTICULAR PURPOSE.  See the GNU General Public License for more details. */

package artofillusion.tools;

import artofillusion.*;
import artofillusion.math.*;
import artofillusion.object.*;
import artofillusion.object.TriangleMesh.*;
import artofillusion.ui.*;
import java.awt.*;
import java.awt.event.*;
import java.util.Vector;

public class SkinDialog extends Dialog implements ActionListener, ItemListener
{
  LayoutWindow window;
  List curveList;
  Checkbox reverseBox;
  Button upButton, downButton, okButton, cancelButton;
  ObjectPreviewCanvas preview;
  ObjectInfo curve[];
  boolean reverse[];
  
  private static int counter = 1;

  public SkinDialog(LayoutWindow window, Vector curves)
  {
    super(window, "Skin", true);
    this.window = window;
    curve = new ObjectInfo [curves.size()];
    reverse = new boolean [curves.size()];
    Scene scene = window.getScene();
    
    // Layout the window.
    
    setLayout(new GridBagLayout());
    GridBagConstraints gc = new GridBagConstraints();
    gc.fill = gc.BOTH;
    gc.gridx = 0;
    gc.gridheight = 4;
    add(curveList = new List(), gc);
    for (int i = 0; i < curves.size(); i++)
      {
	curve[i] = (ObjectInfo) curves.elementAt(i);
	curveList.add(curve[i].name);
      }
    curveList.select(0);
    curveList.addItemListener(this);
    gc.gridx = 1;
    gc.gridheight = 1;
    add(upButton = new Button("Move Up"), gc);
    upButton.addActionListener(this);
    add(downButton = new Button("Move Down"), gc);
    downButton.addActionListener(this);
    add(reverseBox = new Checkbox("Reverse Direction"), gc);
    reverseBox.addItemListener(this);

    // Add the preview canvas.
    
    gc.gridx = 2;
    gc.gridheight = 4;
    gc.fill = gc.BOTH;
    gc.weightx = gc.weighty = 1.0;
    add(preview = new ObjectPreviewCanvas(null), gc);
    preview.setPreferredSize(150, 150);
    
    // Add the buttons at the bottom.
    
    gc.gridx = 0;
    gc.gridwidth = 3;
    gc.weightx = gc.weighty = 0.0;
    Panel p = new Panel();
    p.add(okButton = Translate.button("ok", this));
    p.add(cancelButton = Translate.button("cancel", this));
    add(p, gc);
    makeObject();
    pack();
    updateComponents();
    show();
  }
  
  /** Find the center of the curves the skin should consist of.*/
  
  protected Vec3 findCurvesCenter()
  {
    Vec3 center = new Vec3();
    ObjectInfo info;
            
    for (int i = 0; i < curve.length; i++)
      center.add(curve[i].coords.getOrigin());

    center.scale(1.0/curve.length);
    
    return center;
  }

  
  // Deal with changes to the checkboxes and choices.
  
  public void itemStateChanged(ItemEvent e)
  {
    if (e.getSource() == reverseBox)
      reverse[curveList.getSelectedIndex()] = reverseBox.getState();
    makeObject();
    updateComponents();
  }
  
  // Enable or disable components, based on the current selections.
  
  private void updateComponents()
  {
    int which = curveList.getSelectedIndex();
    
    upButton.setEnabled(which > 0);
    downButton.setEnabled(which < curve.length-1);
    reverseBox.setState(reverse[which]);
  }
  
  // Deal with clicks on the buttons.
  
  public void actionPerformed(ActionEvent e)
  {
    if (okButton == e.getSource())
      {
	window.addObject(preview.getObject().object, new CoordinateSystem(findCurvesCenter(), Vec3.vz(), Vec3.vy()), "Skinned Object "+(counter++), null);
	window.setSelection(window.getScene().getNumObjects()-1);
	window.setUndoRecord(new UndoRecord(window, false, UndoRecord.DELETE_OBJECT, new Object [] {new Integer(window.getScene().getNumObjects()-1)}));
	window.updateImage();
	dispose();
      }
    else if (upButton == e.getSource())
      {
	int which = curveList.getSelectedIndex();
	ObjectInfo swap1 = curve[which];
	curve[which] = curve[which-1];
	curve[which-1] = swap1;
	boolean swap2 = reverse[which];
	reverse[which] = reverse[which-1];
	reverse[which-1] = swap2;
	String swap3 = curveList.getItem(which);
	curveList.remove(which);
	curveList.add(swap3, which-1);
	curveList.select(which-1);
	makeObject();
	updateComponents();
      }
    else if (downButton == e.getSource())
      {
	int which = curveList.getSelectedIndex();
	ObjectInfo swap1 = curve[which];
	curve[which] = curve[which+1];
	curve[which+1] = swap1;
	boolean swap2 = reverse[which];
	reverse[which] = reverse[which+1];
	reverse[which+1] = swap2;
	String swap3 = curveList.getItem(which);
	curveList.remove(which);
	curveList.add(swap3, which+1);
	curveList.select(which+1);
	makeObject();
	updateComponents();
      }
    else
      dispose();
  }
  
  // Create the skinned object.
  
  private void makeObject()
  {
    Vec3 v[][] = new Vec3 [curve.length][], center = new Vec3();
    float us[] = new float [curve.length], vs[] = new float [((Curve) curve[0].object).getVertices().length];
    int smoothMethod = Mesh.INTERPOLATING;
    boolean closed = false;
    
    for (int i = 0; i < curve.length; i++)
      {
	Curve cv = (Curve) curve[i].object;
	MeshVertex vert[] = cv.getVertices();
	v[i] = new Vec3 [vert.length];
	float smooth[] = cv.getSmoothness();
	if (cv.getSmoothingMethod() > smoothMethod)
	  smoothMethod = cv.getSmoothingMethod();
	closed |= cv.isClosed();
	for (int j = 0; j < vert.length; j++)
	  {
	    int k = (reverse[i] ? vert.length-j-1 : j);
	    v[i][j] = curve[i].coords.fromLocal().times(vert[k].r);
	    center.add(v[i][j]);
	    if (cv.getSmoothingMethod() != Mesh.NO_SMOOTHING)
	      vs[j] += smooth[k];
	  }
	us[i] = 1.0f;
      }
    for (int i = 0; i < vs.length; i++)
      vs[i] /= curve.length;
    
    // Center it.
    
    center.scale(1.0/(v.length*v[0].length));
    for (int i = 0; i < v.length; i++)
      for (int j = 0; j < v[i].length; j++)
	v[i][j].subtract(center);
    SplineMesh mesh = new SplineMesh(v, us, vs, smoothMethod, false, closed);
    mesh.setTexture(window.getScene().getDefaultTexture());
    mesh.setTextureMapping(mesh.getTexture().getDefaultMapping());
    mesh.makeRightSideOut();
    preview.setObject(mesh);
    preview.updateImage();
    preview.repaint();
  }
}